!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fileio_trj_mod
!> @brief   read/write trajectory file
!! @authors Norio Takase (NT), Chigusa Kobayashi (CK)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

module fileio_trj_mod

  use trajectory_str_mod
  use molecules_str_mod
  use select_atoms_str_mod
  use fileio_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: open_trj
  public  :: close_trj
  public  :: read_trj
  public  :: write_trj
  public  :: seek_trj

  private :: open_pdb_trj
  private :: close_pdb_trj
  private :: read_pdb_trj
  private :: write_pdb_trj_sel
  private :: write_pdb_trj

  private :: open_dcd_trj
  private :: close_dcd_trj
  private :: read_dcd_trj
  private :: read_dcd_trj_header
  private :: write_dcd_trj_header
  private :: write_dcd_trj_sel
  private :: write_dcd_trj
  private :: seek_dcd_trj

  private :: to_pbcbox
  private :: to_symmat
  private :: to_deg

  ! parameters
  integer,    private            :: GRXMagic = 1993

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_trj
  !> @brief        Open trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    filename   : trajectory file name
  !! @param[in]    trj_format : trajectory file format
  !! @param[in]    trj_type   : trajectory file type
  !! @param[in]    in_out     : information for input or output
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_trj(file, filename, trj_format, trj_type, in_out)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    character(*),            intent(in)    :: filename
    integer,                 intent(in)    :: trj_format
    integer,                 intent(in)    :: trj_type
    integer,                 intent(in)    :: in_out


    file%name       = filename
    file%trj_format = trj_format
    file%trj_type   = trj_type
    file%in_out     = in_out

    select case (file%trj_format)

    case (TrjFormatPDB)

      call open_pdb_trj(file)

    case (TrjFormatDCD)

      call open_dcd_trj(file)

    end select

    file%step_no = 1

    return

  end subroutine open_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_trj
  !> @brief        Close trajectory file
  !! @authors      NT
  !! @param[inout] file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_trj(file)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file


    select case (file%trj_format)

    case (TrjFormatPDB)

      call close_pdb_trj(file)

    case (TrjFormatDCD)

      call close_dcd_trj(file)
      call update_dcd_trj_header(file)

    end select

    return

  end subroutine close_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_trj
  !> @brief        Read trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[inout] trajectory : trajectory information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(inout) :: trajectory


    select case (file%trj_format)

    case (TrjFormatPDB)

      call read_pdb_trj(file, trajectory)

    case (TrjFormatDCD)

      call read_dcd_trj_header(file)
      call read_dcd_trj(file, trajectory)

    end select

    file%step_no = file%step_no + 1

    return

  end subroutine read_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_trj
  !> @brief        Write trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    selatoms   : atom selection information
  !! @param[in]    molecule   : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_trj(file, trajectory, selatoms, molecule)

    ! formal arguments
    type(s_trj_file),             intent(inout) :: file
    type(s_trajectory),           intent(in)    :: trajectory
    type(s_selatoms),   optional, intent(in)    :: selatoms
    type(s_molecule),   optional, intent(in)    :: molecule


    select case (file%trj_format)

    case (TrjFormatPDB)

      if (.not. present(molecule)) then
        call error_msg('Write_Trj> Molecular information is necessary '//&
                       'for the output in the PDB format')
      end if

      if (present(selatoms)) then
        call write_pdb_trj_sel(file, trajectory, selatoms, molecule)
      else
        call write_pdb_trj(file, trajectory, molecule)
      end if

    case (TrjFormatDCD)

      if (present(selatoms)) then
        call write_dcd_trj_header(file, size(selatoms%idx))
        call write_dcd_trj_sel(file, trajectory, selatoms)
      else
        call write_dcd_trj_header(file, size(trajectory%coord(1,:)))
        call write_dcd_trj(file, trajectory)
      end if

    case (TrjFormatCharmmRst)

      call error_msg('Write_Trj> Charmm restart is not supported.')

    case (TrjFormatNamdRst)

      call error_msg('Write_Trj> NAMD restart is not supported.')

    end select

    file%step_no = file%step_no + 1

    return

  end subroutine write_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    seek_trj
  !> @brief        Seek trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    step_no    : step number 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine seek_trj(file, trajectory, step_no)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(inout) :: trajectory
    integer,                 intent(in)    :: step_no


    select case (file%trj_format)

    case (TrjFormatPDB)

      call error_msg('Seek_Trj> Seek is not supported. [PDB]')

    case (TrjFormatDCD)

      call read_dcd_trj_header(file)
      call seek_dcd_trj(file, trajectory, step_no)

    end select
    return

  end subroutine seek_trj

  !*********************************************************************
  ! pdb

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_pdb_trj
  !> @brief        Open pdb trajectory file
  !! @authors      NT
  !! @param[inout] file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_pdb_trj(file)

    ! formal arugments
    type(s_trj_file),        intent(inout) :: file
    

    call open_file(file%unit_no, file%name, file%in_out)

    return 

  end subroutine open_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_pdb_trj
  !> @brief        Close pdb trajectory file
  !! @authors      NT
  !! @param[in]    file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_pdb_trj(file)

    ! formal arugments
    type(s_trj_file),        intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_pdb_trj
  !> @brief        Read pdb trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[inout] trajectory : trajectory information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_pdb_trj(file, trajectory)
    
    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(inout) :: trajectory

    ! local variables
    character(80)               :: line
    integer                     :: i, j, natm, unit

    real(wp),           pointer :: crd(:,:)


    unit =  file%unit_no
    crd  => trajectory%coord
    natm =  size(crd(1,:))

    i = 0

    do while (.true.)

      read(unit, '(a80)', end=900, err=900) line

      if ((line(1:4) == 'ATOM') .or. &
          (line(1:4) == 'HETA')) then

        i = i + 1 

        ! atom coordinates
        read(line, '(30x,3f8.3)', err=900) (crd(j,i), j=1,3)
        if (i >= natm) &
          exit

      else if (line(1:4) == 'CRYS') then

        ! crystal
        read(line, '(6x,3f9.3)', err=900) (trajectory%pbc_box(j,j), j=1,3)

      end if

    end do

    return

900 call error_msg('Read_Pdb_Trj> read error')

  end subroutine read_pdb_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_pdb_trj_sel
  !> @brief        Write pdb trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    selatoms   : atom selection information
  !! @param[in]    molecule   : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_pdb_trj_sel(file, trajectory, selatoms, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_selatoms),   target, intent(in)    :: selatoms
    type(s_molecule),   target, intent(in)    :: molecule

    ! local varialbles
    integer                     :: i, j, iatm, nsel, unit, molec, molep

    real(wp),           pointer :: crd(:,:)
    integer,            pointer :: numr(:), numm(:), aidx(:)
    character(4),       pointer :: nama(:), namr(:)


    unit =  file%unit_no
    nama => molecule%atom_name
    namr => molecule%residue_name
    numr => molecule%residue_no
    numm => molecule%molecule_no
    crd  => trajectory%coord


    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, '(a6,3(f9.3),a37)') 'CRYST1', &
           (trajectory%pbc_box(j,j), j=1,3), &
           '  90.00  90.00  90.00 P 1           1'
    end if

    write(unit, '(a5,5x,i4)') 'MODEL', file%step_no


    aidx  => selatoms%idx
    nsel  = size(selatoms%idx)

    molep = -1

    do i = 1, nsel
      iatm = aidx(i)

      molec = numm(iatm)
      if (molep /= -1 .and. molep /= molec) then
        write(unit,'(a3)') 'TER'
      end if
      molep = molec

      if (numr(iatm) <= 9999) then

        write(unit, '(a4,i7,1x,a4,1x,a4,1x,i4,4x,3f8.3)') &
          'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      else if (numr(iatm) > 999999) then 

        write(unit, '(a4,i7,1x,a4,1x,a4,i6,3x,3f8.3)') &
          'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      else

        write(unit, '(a4,i7,1x,a4,1x,a4,i5,4x,3f8.3)') &
          'ATOM', iatm, nama(iatm), namr(iatm), numr(iatm), (crd(j,iatm),j=1,3)

      end if
    end do

    write(unit,'(a3)') 'TER'
    write(unit,'(a6)') 'ENDMDL'

    return

  end subroutine write_pdb_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_pdb_trj
  !> @brief        Write pdb trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    molecule   : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_pdb_trj(file, trajectory, molecule)

    ! formal arguments
    type(s_trj_file),           intent(inout) :: file
    type(s_trajectory), target, intent(in)    :: trajectory
    type(s_molecule),   target, intent(in)    :: molecule

    ! local varialbles
    integer                     :: i, j, natm, unit, molec, molep

    real(wp),           pointer :: crd(:,:)
    integer,            pointer :: numr(:), numm(:)
    character(4),       pointer :: nama(:), namr(:)


    unit =  file%unit_no
    nama => molecule%atom_name
    namr => molecule%residue_name
    numr => molecule%residue_no
    crd  => trajectory%coord


    if (file%trj_type == TrjTypeCoorBox) then
      write(unit, '(a6,3(f9.3),a37)') 'CRYST1', &
           (trajectory%pbc_box(j,j), j=1,3),    &
           '  90.00  90.00  90.00 P 1           1'
    end if

    write(unit, '(a5,5x,i4)') 'MODEL', file%step_no


    natm = size(crd(1,:))

    molep = -1

    do i = 1, natm

      molec = numm(i)
      if (molep /= -1 .and. molep /= molec) then
        write(unit,'(a3)') 'TER'
      end if
      molep = molec

      if (numr(i) <= 9999) then

        write(unit, '(a4,i7,1x,a4,1x,a4,1x,i4,4x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      else if (numr(i) > 999999) then 

        write(unit, '(a4,i7,1x,a4,1x,a4,i6,3x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      else

        write(unit, '(a4,i7,1x,a4,1x,a4,i5,4x,3f8.3)') &
             'ATOM', i, nama(i), namr(i), numr(i), (crd(j,i),j=1,3)

      end if
    end do

    write(unit,'(a3)') 'TER'
    write(unit,'(a6)') 'ENDMDL'

    return

  end subroutine write_pdb_trj

  !*********************************************************************
  ! dcd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_dcd_trj
  !> @brief        Open DCD trajectory file
  !! @authors      NT
  !! @param[inout] file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_dcd_trj(file)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file

    ! local variable 
    integer                  :: unit, endian
    integer(4)               :: i4


    ! check endianess
    !
    endian = IOFileNativeEndian

    ! GCC gfortan do not support raw binary format extension (form = "binary")
#ifndef __GFORTRAN__

    if (file%in_out == IOFileInput) then

      unit = get_unit_no()
      open(unit, &
           file    = file%name, &
           status  = 'old',     &
           form    = 'binary',  &
           convert = 'little_endian', &
           err     = 900)

      read(unit) i4
      if (i4 == 84) then
        endian = IOFileLittleEndian
      else
        endian = IOFileBigEndian
      end if

      close(unit)
      call free_unit_no(unit)

    end if

#endif

    ! open file
    !
    call open_binary_file(file%unit_no, file%name, file%in_out, endian)

    file%rw_header = .true.
    file%byte_swap = .false.

    return

900 call error_msg('Open_Dcd_Trj> I/O Error.'//trim(file%name))

    return

  end subroutine open_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_dcd_trj
  !> @brief        Close DCD trajectory file
  !! @authors      NT
  !! @param[in]    file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_dcd_trj(file)

    ! formal arguments
    type(s_trj_file),        intent(in)    :: file


    call close_file(file%unit_no)

    return

  end subroutine close_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_dcd_trj_header
  !> @brief        Read DCD trajectory file header
  !! @authors      NT
  !! @param[inout] file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_dcd_trj_header(file)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file

    ! local variables
    integer                  :: i, unit
    integer(4)               :: icntrl(20), ntitle, natom
    character(4)             :: hdr
    character(80)            :: title(10)


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    unit = file%unit_no


    ! read header,cntrl
    !
    read(unit) hdr, icntrl(1:20)

    if (hdr /= 'CORD') &
      call error_msg('Read_Dcd_Trj_Header> Bad dcd file format.')
      
    if (icntrl(11) >= 1) then
      file%trj_type = TrjTypeCoorBox
    else
      file%trj_type = TrjTypeCoor
    end if


    ! read title
    !
    read(unit) ntitle,(title(i),i=1,ntitle)


    ! read natom
    !
    read(unit) natom
      
    return

  end subroutine read_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj_header
  !> @brief        Write DCD trajectory file header
  !! @authors      NT
  !! @param[inout] file      : trajectory file handle
  !! @param[in]    num_atoms : number of atom
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj_header(file, num_atoms)

    ! parameters
    character(4),            parameter     :: hdr = 'CORD'
    integer(4),              parameter     :: ntitle = 4

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    integer,                 intent(in)    :: num_atoms

    ! local variables
    integer                  :: i, unit
    integer(4)               :: icntrl(20)
    character(80)            :: title(4)
    character(24)            :: name, date


    if (.not. file%rw_header) &
      return
    file%rw_header = .false.

    unit = file%unit_no


    ! write header,cntrl
    !
    icntrl(:)  = 0
    icntrl(2)  = file%step_no                 ! first time step     (NPRIV)
    icntrl(3)  = 1                            ! output period       (NSAVC)
    icntrl(4)  = 0                            ! number of time step (NSTEP)
    icntrl(10) = transfer(1.0_sp, icntrl(10)) ! length of time step (DELTA)
    if (file%trj_type == TrjTypeCoorBox) &
    icntrl(11) = 1                            ! flag for with unit-cell
    icntrl(20) = 24                           ! PRETEND TO BE CHARMM24 -JCP

    write(unit) hdr, icntrl(1:20)


    ! write title
    !
    call fdate (date)
    call getlog(name)

    title(1) = 'REMARKS CREATED BY GENESIS ANALYSIS TOOL'
    title(2) = 'REMARKS DATE: ' // date // ' CREATED BY USER: ' // name
    title(3) = ''
    title(4) = ''

    write(unit) ntitle, (title(i),i=1,ntitle)


    ! write natom
    !
    write(unit) num_atoms


    return

  end subroutine write_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_dcd_trj_header
  !> @brief        Update DCD trajectory file header
  !! @authors      NT
  !! @param[inout] file : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_dcd_trj_header(file)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file


    ! GCC gfortan do not support raw binary format extension (form = "binary")
#ifndef __GFORTRAN__

    ! local variables
    integer                  :: unit


    if (file%in_out /= IOFileInput) then

      unit = get_unit_no()
      open(unit,               &
           file   = file%name, &
           status = 'old',     &
           form   = 'binary',  &
           access = 'direct',  &
           recl   = 4)

      write(unit,rec=3) file%step_no - 1
      write(unit,rec=6) file%step_no - 1

      close(unit)
      call free_unit_no(unit)

    end if

#else
    call error_msg('This feature is not available!')
#endif

    return

  end subroutine update_dcd_trj_header

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_dcd_trj
  !> @brief        Read DCD trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[inout] trajectory : trajectory information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_dcd_trj(file, trajectory)
    
    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(inout) :: trajectory

    ! local variables
    real(dp)                 :: symmat(6), pbcbox(6)
    integer                  :: i, natm, unit

    real(sp), allocatable    :: crd4f(:)


    unit = file%unit_no


    ! read periodic boundary condition
    !
    if (file%trj_type == TrjTypeCoorBox) then

      read(unit) symmat(1:6)

      call to_pbcbox(symmat, pbcbox)

      trajectory%pbc_box(1,1) = real(pbcbox(1),wp)
      trajectory%pbc_box(2,2) = real(pbcbox(2),wp)
      trajectory%pbc_box(3,3) = real(pbcbox(3),wp)

    end if


    ! read coordinates
    !

    natm = size(trajectory%coord(1,:))

    allocate(crd4f(natm))

    read(unit) crd4f(1:natm)
    trajectory%coord(1,1:natm) = real(crd4f(1:natm),wp)
    read(unit) crd4f(1:natm)
    trajectory%coord(2,1:natm) = real(crd4f(1:natm),wp)
    read(unit) crd4f(1:natm)
    trajectory%coord(3,1:natm) = real(crd4f(1:natm),wp)

    deallocate(crd4f)

    return

  end subroutine read_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj_sel
  !> @brief        Write DCD trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    selatoms   : atom selection information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj_sel(file, trajectory, selatoms)
    
    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(in)    :: trajectory
    type(s_selatoms),        intent(in)    :: selatoms

    ! local variables
    real(dp)                 :: pbcbox(6), symmat(6)
    integer                  :: natm, unit


    unit = file%unit_no


    ! write periodic boundary condition
    !
    if (file%trj_type == TrjTypeCoorBox) then

      pbcbox(1) = real(trajectory%pbc_box(1,1),dp)
      pbcbox(2) = real(trajectory%pbc_box(2,2),dp)
      pbcbox(3) = real(trajectory%pbc_box(3,3),dp)
      pbcbox(4) = 90.0_dp
      pbcbox(5) = 90.0_dp
      pbcbox(6) = 90.0_dp

      call to_symmat(pbcbox, symmat)

      write(unit) symmat(1:6)

    end if


    ! write atom coordinates
    !
    natm = size(selatoms%idx)

    write(unit) real(trajectory%coord(1,selatoms%idx(1:natm)),sp)
    write(unit) real(trajectory%coord(2,selatoms%idx(1:natm)),sp)
    write(unit) real(trajectory%coord(3,selatoms%idx(1:natm)),sp)

    return

  end subroutine write_dcd_trj_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_dcd_trj
  !> @brief        Write DCD trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[in]    trajectory : trajectory information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_dcd_trj(file, trajectory)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(in)    :: trajectory

    ! local variables
    real(dp)                 :: pbcbox(6), symmat(6)
    integer                  :: natm, unit


    unit = file%unit_no


    ! write periodic boundary condition
    !
    if (file%trj_type == TrjTypeCoorBox) then

      pbcbox(1) = real(trajectory%pbc_box(1,1),dp)
      pbcbox(2) = real(trajectory%pbc_box(2,2),dp)
      pbcbox(3) = real(trajectory%pbc_box(3,3),dp)
      pbcbox(4) = 90.0_dp
      pbcbox(5) = 90.0_dp
      pbcbox(6) = 90.0_dp

      call to_symmat(pbcbox, symmat)

      write(unit) symmat(1:6)

    end if


    ! write atom coordinates
    !
    natm = size(trajectory%coord(1,:))

    write(unit) real(trajectory%coord(1,1:natm),sp)
    write(unit) real(trajectory%coord(2,1:natm),sp)
    write(unit) real(trajectory%coord(3,1:natm),sp)

    return

  end subroutine write_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    seek_dcd_trj
  !> @brief        Seek DCD trajectory file
  !! @authors      NT
  !! @param[inout] file       : trajectory file handle
  !! @param[inout] trajectory : trajectory information
  !! @param[in]    step_no    : step number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine seek_dcd_trj(file, trajectory, step_no)

    ! formal arguments
    type(s_trj_file),        intent(inout) :: file
    type(s_trajectory),      intent(inout) :: trajectory
    integer,                 intent(in)    :: step_no

    ! local variables
    real(dp)                 :: symmat(6)
    integer                  :: unit, natm

    real(4),   allocatable   :: crd4f(:)


    if (file%step_no > step_no) then
      call error_msg('Seek_Dcd_Trj> Seeking to the back is not supported.')
    else if (file%step_no == step_no) then
      return 
    end if

    unit = file%unit_no
    natm = size(trajectory%coord(1,:))

    allocate(crd4f(natm))

    do while (file%step_no /= step_no)

      if (file%trj_type == TrjTypeCoorBox) then
        read(unit) symmat(1:6)
      end if

      read(unit) crd4f(1:natm)
      read(unit) crd4f(1:natm)
      read(unit) crd4f(1:natm)

      file%step_no = file%step_no + 1

    end do
    
    !write(MsgOut,'(A,A,A,I7)') 'Seek_Dcd_Trj> file:', trim(file%name), &
    !                           ' current:', file%step_no

    deallocate(crd4f)

    return

  end subroutine seek_dcd_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    to_pbcbox
  !> @brief        to_pbcbox
  !! @authors      NT
  !! @param[in]    symmat : symmetric shape matrix
  !! @param[inout] pbcbox : periodic boundary condition box
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_pbcbox(symmat, pbcbox)

    ! formal arguments 
    real(dp),                intent(in)    :: symmat(:)
    real(dp),                intent(inout) :: pbcbox(:)

    ! local variables
    real(dp)                 :: ab, bc, ca


    pbcbox(1)= sqrt(symmat(1)*symmat(1)+symmat(2)*symmat(2)+symmat(4)*symmat(4))
    pbcbox(2)= sqrt(symmat(2)*symmat(2)+symmat(3)*symmat(3)+symmat(5)*symmat(5))
    pbcbox(3)= sqrt(symmat(4)*symmat(4)+symmat(5)*symmat(5)+symmat(6)*symmat(6))

    ab = symmat(2)*(symmat(1)+symmat(3))+symmat(4)*symmat(5)
    bc = symmat(5)*(symmat(3)+symmat(6))+symmat(2)*symmat(4)
    ca = symmat(4)*(symmat(1)+symmat(6))+symmat(2)*symmat(5)

    pbcbox(4) = to_deg(acos(bc/(pbcbox(2)*pbcbox(3))))
    pbcbox(5) = to_deg(acos(ca/(pbcbox(3)*pbcbox(1))))
    pbcbox(6) = to_deg(acos(ab/(pbcbox(1)*pbcbox(2))))

    return

  end subroutine to_pbcbox

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    to_symmat
  !> @brief        to_symmat
  !! @authors      NT
  !! @param[in]    pbcbox : periodic boundary condition box
  !! @param[inout] symmat : symmetric shape matrix
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_symmat(pbcbox, symmat)

    ! formal arguments 
    real(dp),                intent(in)    :: pbcbox(:)
    real(dp),                intent(inout) :: symmat(:)


    if (pbcbox(4) - EPS > 90.0_wp .or. &
        pbcbox(4) + EPS < 90.0_wp .or. &
        pbcbox(5) - EPS > 90.0_wp .or. &
        pbcbox(5) + EPS < 90.0_wp .or. &
        pbcbox(6) - EPS > 90.0_wp .or. &
        pbcbox(6) + EPS < 90.0_wp) &
      write(MsgOut,*) &
      'To_Symmat> WARNING: Non rectangular unit cell is not supported.'

    symmat(1) = pbcbox(1)
    symmat(2) = 0.0_wp   ! angle A and B = 90.0 deg.
    symmat(3) = pbcbox(2)
    symmat(4) = 0.0_wp   ! angle A and C = 90.0 deg.
    symmat(5) = 0.0_wp   ! angle B and C = 90.0 deg.
    symmat(6) = pbcbox(3)

    return

  end subroutine to_symmat

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      to_deg
  !> @brief        radian to degree
  !! @authors      NT, CK
  !! @param[in]    radian : radian
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function to_deg(radian)

    ! return value
    real(dp)                  :: to_deg

    ! formal arguments
    real(dp),                 intent(in)    :: radian


    to_deg = radian*RAD/PI

    return

  end function to_deg

end module fileio_trj_mod
