!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   output_mod
!> @brief   output data for analysis
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

module output_mod

  use output_str_mod
  use fileio_pdb_mod
  use fileio_control_mod
  use fileio_mod
  use string_mod
  use messages_mod

  implicit none
  private

  ! structures
  type, public :: s_out_info
    character(MaxFilename) :: angfile  = ''
    character(MaxFilename) :: disfile  = ''
    character(MaxFilename) :: enefile  = ''
    character(MaxFilename) :: pdbfile  = ''
    character(MaxFilename) :: rmsfile  = ''
    character(MaxFilename) :: rstfile  = ''
    character(MaxFilename) :: torfile  = ''
    character(MaxFilename) :: trjfile  = ''
    character(MaxFilename) :: trrfile  = ''
    character(MaxFilename) :: gprfile  = ''
    character(MaxFilename) :: qntfile  = ''
    character(MaxFilename) :: coorfile = ''
    character(MaxFilename) :: velfile  = ''
    character(MaxFilename) :: xscfile  = ''
  end type s_out_info

  ! subroutines
  public  :: show_ctrl_output
  public  :: read_ctrl_output
  public  :: setup_output

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_output
  !> @brief        show control parameters in OUTPUT section
  !! @authors      NT
  !! @param[in]    keys : comma separated keys for show
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_output(keys)

    ! formal arguments
    character(*),            intent(in)    :: keys


    if (len_trim(keys) == 0) &
      return

    write(MsgOut,'(A)') '[OUTPUT]'

    if (index(keys,'ang') /= 0) &
      write(MsgOut,'(A)') 'angfile        = output.ang      # angle file'

    if (index(keys,'dis') /= 0) &
      write(MsgOut,'(A)') 'disfile        = output.dis      # distance file'

    if (index(keys,'ene') /= 0) &
      write(MsgOut,'(A)') 'enefile        = output.ene      # REMD energy file'

    if (index(keys,'pdb') /= 0) &
      write(MsgOut,'(A)') 'pdbfile        = output.pdb      # PDB file'

    if (index(keys,'rst') /= 0) &
      write(MsgOut,'(A)') 'rstfile        = output.rst      # restart file'

    if (index(keys,'tor') /= 0) &
      write(MsgOut,'(A)') 'torfile        = output.tor      # torsion file'
    
    if (index(keys,'TRJ') /= 0) &
      write(MsgOut,'(A)') 'trjfile        = output{}.trj    # trajectory file'
    
    if (index(keys,'trj') /= 0) &
      write(MsgOut,'(A)') 'trjfile        = output.trj      # trajectory file'
    
    if (index(keys,'rms') /= 0) &
      write(MsgOut,'(A)') 'rmsfile        = output.rms      # RMSD file'

    if (index(keys,'trr') /= 0) &
      write(MsgOut,'(A)') 'trrfile        = output.trr      # TRROT file'
    
!    if (index(keys,'gpr') /= 0) &
!      write(MsgOut,'(A)') 'gprfile        = output.gpr      # go model parameter file'
    
    if (index(keys,'qnt') /= 0) &
      write(MsgOut,'(A)') 'qntfile        = output.qnt      # fraction of native contact file'

    if (index(keys,'coor') /= 0) &
      write(MsgOut,'(A)') 'coorfile       = output.coor     # coordinate restart file'

    if (index(keys,'vel') /= 0) &
      write(MsgOut,'(A)') 'velfile        = output.vel      # velocity restart file'

    if (index(keys,'xsc') /= 0) &
      write(MsgOut,'(A)') 'xscfile        = output.xsc      # extended system file'
    
    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_output

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_output
  !> @brief        read control parameters in OUTPUT section
  !! @authors      NT
  !! @param[in]    handle   : unit number of control file
  !! @param[inout] out_info : OUTPUT section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_output(handle, out_info)

    ! parameters
    character(*),            parameter :: Section = 'Output'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_out_info),        intent(inout) :: out_info


    ! read parameters from control file
    !
    call read_ctrlfile_string(handle, Section, 'Angfile', out_info%angfile)
    call read_ctrlfile_string(handle, Section, 'Disfile', out_info%disfile)
    call read_ctrlfile_string(handle, Section, 'Enefile', out_info%enefile)
    call read_ctrlfile_string(handle, Section, 'Pdbfile', out_info%pdbfile)
    call read_ctrlfile_string(handle, Section, 'Rmsfile', out_info%rmsfile)
    call read_ctrlfile_string(handle, Section, 'Rstfile', out_info%rstfile)
    call read_ctrlfile_string(handle, Section, 'Torfile', out_info%torfile)
    call read_ctrlfile_string(handle, Section, 'Trjfile', out_info%trjfile)
    call read_ctrlfile_string(handle, Section, 'Trrfile', out_info%trrfile)
    call read_ctrlfile_string(handle, Section, 'Gprfile', out_info%gprfile)
    call read_ctrlfile_string(handle, Section, 'Qntfile', out_info%qntfile)
    call read_ctrlfile_string(handle, Section, 'Coorfile',out_info%coorfile)
    call read_ctrlfile_string(handle, Section, 'Velfile', out_info%velfile)
    call read_ctrlfile_string(handle, Section, 'Xscfile', out_info%xscfile)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Output> Output Files'
    if (len_trim(out_info%angfile) > 0) &
      write(MsgOut,'(A20,A)') '  angfile         = ', trim(out_info%angfile)
    if (len_trim(out_info%disfile) > 0) &
      write(MsgOut,'(A20,A)') '  disfile         = ', trim(out_info%disfile)
    if (len_trim(out_info%enefile) > 0) &
      write(MsgOut,'(A20,A)') '  enefile         = ', trim(out_info%enefile)
    if (len_trim(out_info%pdbfile) > 0) &
      write(MsgOut,'(A20,A)') '  pdbfile         = ', trim(out_info%pdbfile)
    if (len_trim(out_info%rmsfile) > 0) &
      write(MsgOut,'(A20,A)') '  rmsfile         = ', trim(out_info%rmsfile)
    if (len_trim(out_info%rstfile) > 0) &
      write(MsgOut,'(A20,A)') '  rstfile         = ', trim(out_info%rstfile)
    if (len_trim(out_info%torfile) > 0) &
      write(MsgOut,'(A20,A)') '  torfile         = ', trim(out_info%torfile)
    if (len_trim(out_info%trjfile) > 0) &
      write(MsgOut,'(A20,A)') '  trjfile         = ', trim(out_info%trjfile)
    if (len_trim(out_info%trrfile) > 0) &
      write(MsgOut,'(A20,A)') '  trrfile         = ', trim(out_info%trrfile)
    if (len_trim(out_info%gprfile) > 0) &
      write(MsgOut,'(A20,A)') '  gprfile         = ', trim(out_info%gprfile)
    if (len_trim(out_info%qntfile) > 0) &
      write(MsgOut,'(A20,A)') '  qntfile         = ', trim(out_info%qntfile)
    if (len_trim(out_info%coorfile) > 0) &
      write(MsgOut,'(A20,A)') '  coorfile        = ', trim(out_info%coorfile)
    if (len_trim(out_info%velfile) > 0) &
      write(MsgOut,'(A20,A)') '  velfile         = ', trim(out_info%velfile)
    if (len_trim(out_info%xscfile) > 0) &
      write(MsgOut,'(A20,A)') '  xscfile         = ', trim(out_info%xscfile)
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_output

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_output
  !> @brief        setup output files
  !! @authors      NT
  !! @param[in]    out_info : OUTPUT section control parameters information
  !! @param[out]   output   : output information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_output(out_info, output)

    ! formal argments
    type(s_out_info),        intent(in)    :: out_info
    type(s_output),          intent(inout) :: output


    output%angfile  = out_info%angfile
    output%disfile  = out_info%disfile
    output%enefile  = out_info%enefile
    output%pdbfile  = out_info%pdbfile
    output%rmsfile  = out_info%rmsfile
    output%rstfile  = out_info%rstfile
    output%torfile  = out_info%torfile
    output%trjfile  = out_info%trjfile
    output%trrfile  = out_info%trrfile
    output%gprfile  = out_info%gprfile
    output%qntfile  = out_info%qntfile
    output%coorfile = out_info%coorfile
    output%velfile  = out_info%velfile
    output%xscfile  = out_info%xscfile

    return

  end subroutine setup_output

end module output_mod
