!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ta_analyze_mod
!> @brief   run analyzing trajectories
!! @authors Norio Takase (NT), Takaharu Mori (TM)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module ta_analyze_mod

  use ta_option_str_mod
  use fileio_trj_mod
  use trajectory_str_mod
  use output_str_mod
  use molecules_str_mod
  use fileio_mod
  use messages_mod
  use constants_mod
 
  implicit none
  private

  ! subroutines
  public  :: analyze
  private :: analyze_dis
  private :: analyze_ang
  private :: analyze_tor
  private :: analyze_qnt
  private :: out_result

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze
  !> @brief        run analyzing trajectories
  !! @authors      NT, TM
  !! @param[in]    trj_list   : trajectory file list information
  !! @param[in]    output     : output information
  !! @param[inout] option     : option information
  !! @param[inout] trajectory : trajectory information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze(trj_list, output, option, trajectory)

    ! formal arguments
    type(s_trj_list),        intent(in)    :: trj_list
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(inout) :: option
    type(s_trajectory),      intent(inout) :: trajectory

    ! local variables
    type(s_trj_file)         :: trj_in
    integer                  :: nstru, ifile, istep, num_trjfiles
    integer                  :: dis_unit, ang_unit, tor_unit, qnt_unit


    if (option%check_only) &
      return

    ! open output file
    !
    if (option%out_dis) &
      call open_file(dis_unit, output%disfile, IOFileOutputNew)

    if (option%out_ang) &
      call open_file(ang_unit, output%angfile, IOFileOutputNew)

    if (option%out_tor) &
      call open_file(tor_unit, output%torfile, IOFileOutputNew)

    if (option%out_qnt) &
      call open_file(qnt_unit, output%qntfile, IOFileOutputNew)


    ! analysis loop
    !
    nstru = 0
    num_trjfiles = size(trj_list%md_steps)

    do ifile = 1, num_trjfiles

      ! open trajectory file
      !
      call open_trj(trj_in, trj_list%filenames(ifile), &
                            trj_list%trj_format,       &
                            trj_list%trj_type, IOFileInput)

      do istep = 1, trj_list%md_steps(ifile)

        ! read trajectory
        !   coordinates of one MD snapshot are saved in trajectory%coord)
        !
        call read_trj(trj_in, trajectory)

        if (mod(istep, trj_list%ana_periods(ifile)) == 0) then

          nstru = nstru + 1
          write(MsgOut,*) '      number of structures = ', nstru

          if (option%out_dis) then
            call analyze_dis(trajectory, option)
            call out_result (nstru, dis_unit, option%distance)
          end if

          if (option%out_ang) then
            call analyze_ang(trajectory, option)
            call out_result (nstru, ang_unit, option%angle)
          end if

          if (option%out_tor) then
            call analyze_tor(trajectory, option)
            call out_result (nstru, tor_unit, option%torsion)
          end if

          if (option%out_qnt) then
            call analyze_qnt(trajectory, option)
            call out_result (nstru, qnt_unit, (/option%qnative/))
          end if

        end if

      end do

      ! close trajectory file
      !
      call close_trj(trj_in)

    end do


    ! close output file
    !
    if (option%out_qnt) call close_file(qnt_unit)
    if (option%out_tor) call close_file(tor_unit)
    if (option%out_ang) call close_file(ang_unit)
    if (option%out_dis) call close_file(dis_unit)

    return

  end subroutine analyze

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze_dis
  !> @brief        analyze distances
  !! @authors      NT, TM
  !! @param[in]    trajectory : trajectory information
  !! @param[inout] option     : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze_dis(trajectory, option)

    ! formal argments
    type(s_trajectory),      intent(in)    :: trajectory
    type(s_option),          intent(inout) :: option

    ! local variables
    real(wp)                 :: x1, y1, z1, x2, y2, z2, dis2
    integer                  :: i, idx1, idx2


    do i = 1, size(option%distance)

      idx1 = option%dist_list(1,i)
      idx2 = option%dist_list(2,i)

      x1 = trajectory%coord(1,idx1)
      y1 = trajectory%coord(2,idx1)
      z1 = trajectory%coord(3,idx1)
      x2 = trajectory%coord(1,idx2)
      y2 = trajectory%coord(2,idx2)
      z2 = trajectory%coord(3,idx2)
      dis2 = (x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2
      option%distance(i) = sqrt(dis2)

    end do

    return

  end subroutine analyze_dis

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze_ang
  !> @brief        analize angles
  !! @authors      NT, TM
  !! @param[in]    trajectory : trajectory information
  !! @param[inout] option     : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze_ang(trajectory, option)

    ! formal argments
    type(s_trajectory),      intent(in)    :: trajectory
    type(s_option),          intent(inout) :: option

    ! local variables
    real(wp)                 :: dx12, dy12, dz12, dx32, dy32, dz32, d12, d32
    real(wp)                 :: cosp, ang
    integer                  :: i, idx1, idx2, idx3


    do i = 1, size(option%angle)
      
      idx1 = option%angl_list(1, i)
      idx2 = option%angl_list(2, i)
      idx3 = option%angl_list(3, i)

      dx12 = trajectory%coord(1, idx1) - trajectory%coord(1, idx2)
      dy12 = trajectory%coord(2, idx1) - trajectory%coord(2, idx2)
      dz12 = trajectory%coord(3, idx1) - trajectory%coord(3, idx2)

      dx32 = trajectory%coord(1, idx3) - trajectory%coord(1, idx2)
      dy32 = trajectory%coord(2, idx3) - trajectory%coord(2, idx2)
      dz32 = trajectory%coord(3, idx3) - trajectory%coord(3, idx2)

      d12 = sqrt(dx12*dx12 + dy12*dy12 + dz12*dz12)
      d32 = sqrt(dx32*dx32 + dy32*dy32 + dz32*dz32)

      if (d12  >  0.0_wp .and. d32  >  0.0_wp) then

        dx12 = dx12 / d12
        dy12 = dy12 / d12
        dz12 = dz12 / d12

        dx32 = dx32 / d32
        dy32 = dy32 / d32
        dz32 = dz32 / d32

        cosp = dx12*dx32 + dy12*dy32 + dz12*dz32
        if (cosp  >  1.0_wp) then
          cosp = 1.0_wp
        else if (cosp < -1.0_wp) then
          cosp = -1.0_wp
        end if

        ang = acos(cosp)
      end if

      ang = ang * (180.0_wp / PI)

      option%angle(i) = ang

    end do

    return

  end subroutine analyze_ang

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze_tor
  !> @brief        analize torsions
  !! @authors      NT, TM
  !! @param[in]    trajectory : trajectory informatin
  !! @param[inout] option     : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze_tor(trajectory, option)

    ! formal argments
    type(s_trajectory),      intent(in)    :: trajectory
    type(s_option),          intent(inout) :: option

    ! local variables
    real(wp)                 :: dx21,dy21,dz21, dx32,dy32,dz32, dx43,dy43,dz43
    real(wp)                 :: px12,py12,pz12, px23,py23,pz23, r12, r23, r12r23
    real(wp)                 :: cosp, s1223, px123,py123,pz123, s32123, tor
    integer                  :: i, idx1, idx2, idx3, idx4


    do i = 1, size(option%torsion)
      
      idx1 = option%tors_list(1, i)
      idx2 = option%tors_list(2, i)
      idx3 = option%tors_list(3, i)
      idx4 = option%tors_list(4, i)

      dx21 = trajectory%coord(1, idx2) - trajectory%coord(1, idx1)
      dy21 = trajectory%coord(2, idx2) - trajectory%coord(2, idx1)
      dz21 = trajectory%coord(3, idx2) - trajectory%coord(3, idx1)
      dx32 = trajectory%coord(1, idx3) - trajectory%coord(1, idx2)
      dy32 = trajectory%coord(2, idx3) - trajectory%coord(2, idx2)
      dz32 = trajectory%coord(3, idx3) - trajectory%coord(3, idx2)
      dx43 = trajectory%coord(1, idx4) - trajectory%coord(1, idx3)
      dy43 = trajectory%coord(2, idx4) - trajectory%coord(2, idx3)
      dz43 = trajectory%coord(3, idx4) - trajectory%coord(3, idx3)

      px12 = dy21*dz32 - dy32*dz21
      py12 = dz21*dx32 - dz32*dx21
      pz12 = dx21*dy32 - dx32*dy21
      px23 = dy32*dz43 - dy43*dz32
      py23 = dz32*dx43 - dz43*dx32
      pz23 = dx32*dy43 - dx43*dy32
      r12  = px12*px12 + py12*py12 + pz12*pz12
      r23  = px23*px23 + py23*py23 + pz23*pz23
      r12r23 = sqrt( r12 * r23 )

      if (r12r23 > 0.0_wp) then

        ! in the case of bond angle = 0 or pi can not calculate
        ! torsional angle   then do not calculate torsional energy

        r12r23 = 1.0_wp / r12r23
        s1223  = px12 * px23 + py12 * py23 + pz12 * pz23
        cosp   = s1223 * r12r23

        if (cosp >  1.0_wp) cosp =  1.0_wp
        if (cosp < -1.0_wp) cosp = -1.0_wp

        px123  = py12 * pz23 - py23 * pz12
        py123  = pz12 * px23 - pz23 * px12
        pz123  = px12 * py23 - px23 * py12 
        s32123 = dx32 * px123 + dy32 * py123 + dz32 * pz123

        tor = acos(cosp)                                
        if (s32123 < 0.0_wp) then                        
          tor = (2.0_wp * pi) - tor                        
        end if

      end if

      tor = tor * (180.0_wp / PI)
      if (tor .gt. 180.0_wp) tor = tor - 360.0_wp

      option%torsion(i) = tor

    end do

    return

  end subroutine analyze_tor

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze_qnt
  !> @brief        analysis qnative
  !! @authors      TM
  !! @param[in]    trajectory : trajectory informatin
  !! @param[inout] option     : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze_qnt(trajectory, option)

    ! formal argments
    type(s_trajectory),      intent(in)    :: trajectory
    type(s_option),          intent(inout) :: option

    ! local variables
    real(wp)                 :: x1, y1, z1, x2, y2, z2, dist
    integer                  :: i, ncon, idx1, idx2, found


    ncon = size(option%contact_list(1,:))

    found = 0
    do i = 1, ncon

      idx1 = option%contact_list(1,i)
      idx2 = option%contact_list(2,i)

      x1 = trajectory%coord(1,idx1)
      y1 = trajectory%coord(2,idx1)
      z1 = trajectory%coord(3,idx1)
      x2 = trajectory%coord(1,idx2)
      y2 = trajectory%coord(2,idx2)
      z2 = trajectory%coord(3,idx2)
      dist = sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)

      if (dist <= option%cutoff_qnative) &
        found = found + 1

    end do

    option%qnative = real(found,wp) / real(ncon,wp)


    return

  end subroutine analyze_qnt

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    out_result
  !> @brief        output analysis results
  !! @authors      NT
  !! @param[in]    sturct_no : sturcture number
  !! @param[in]    unit_no   : file unit number
  !! @param[in]    results   : analysis results
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine out_result(struct_no, unit_no, results)

    ! formal arguments
    integer,                 intent(in)    :: struct_no
    integer,                 intent(in)    :: unit_no
    real(wp),                intent(in)    :: results(:)

    ! local variables
    integer                  :: i


    write(unit_no, '(i10,1x,100(f8.3,1x))') &
         struct_no, (results(i), i=1, size(results))

    return

  end subroutine out_result

end module ta_analyze_mod
