!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   ta_option_mod
!> @brief   module for analysis options
!! @authors Norio Takase (NT), Takaharu Mori (TM)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module ta_option_mod

  use ta_option_str_mod
  use molecules_str_mod
  use fileio_gpr_mod
  use fileio_control_mod
  use string_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_opt_info
    logical                     :: check_only     = .true.
    character(Maxline), allocatable :: dis_atoms(:)
    character(Maxline), allocatable :: ang_atoms(:)
    character(Maxline), allocatable :: tor_atoms(:)
    real(wp)                    :: cutoff_qnative = 5.0_wp
  end type s_opt_info

  ! subroutines
  public  :: show_ctrl_option
  public  :: read_ctrl_option
  public  :: setup_option
  private :: parse_atom_defs
  private :: get_atom_index

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_option
  !> @brief        show control parameters in OPTION section
  !! @authors      NT
  !! @param[in]    mode : show mode
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_option(mode)

    ! formal arguments
    character(*),            intent(in)    :: mode


    write(MsgOut,'(A)') '[OPTION]'
    write(MsgOut,'(A)') 'check_only     = YES             # (YES/NO)'

    select case (mode)
      
    case ('distance')
      write(MsgOut,'(A)') '# distance1      = P1:306:ALA:O   P1:768:ASN:HD21'
      write(MsgOut,'(A)') '# distance2      = P1:771:GLU:OE1 P1:768:ASN:ND2'
    case ('angle')
      write(MsgOut,'(A)') '# angle1         = P1:306:ALA:O   P1:768:ASN:HD21 P1:277:TIP3:H1'
      write(MsgOut,'(A)') '# angle2         = P1:771:GLU:OE1 P1:768:ASN:ND2  P1:272:TIP3:H1'
    case ('torsion')
      write(MsgOut,'(A)') '# torsion1       = P1:306:ALA:O   P1:768:ASN:HD21 P1:277:TIP3:H1 P1:271:TIP3:H2'
      write(MsgOut,'(A)') '# torsion2       = P1:771:GLU:OE1 P1:768:ASN:ND2  P1:272:TIP3:H1 P1:204:TIP3:OH2'
    case ('qnative')
      write(MsgOut,'(A)') 'cutoff_qnative = 5.0             # cutoff distance of native contact'
    case default
      write(MsgOut,'(A)') '# distance1      = P1:306:ALA:O   P1:768:ASN:HD21'
      write(MsgOut,'(A)') '# distance2      = P1:771:GLU:OE1 P1:768:ASN:ND2'
      write(MsgOut,'(A)') '# angle1         = P1:306:ALA:O   P1:768:ASN:HD21 P1:277:TIP3:H1'
      write(MsgOut,'(A)') '# angle2         = P1:771:GLU:OE1 P1:768:ASN:ND2  P1:272:TIP3:H1'
      write(MsgOut,'(A)') '# torsion1       = P1:306:ALA:O   P1:768:ASN:HD21 P1:277:TIP3:H1 P1:271:TIP3:H2'
      write(MsgOut,'(A)') '# torsion2       = P1:771:GLU:OE1 P1:768:ASN:ND2  P1:272:TIP3:H1 P1:204:TIP3:OH2'
      write(MsgOut,'(A)') '# cutoff_qnative = 5.0             # cutoff distance of native contact'
    end select

    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_option
  !> @brief        read control parameters in OPTION section
  !! @authors      NT, TM
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   opt_info : OPTION section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_option(handle, opt_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Option'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_opt_info),        intent(inout) :: opt_info

    ! local variables
    integer                  :: i, ndis, nang, ntor
    character(MaxLine)       :: value
    character(20)            :: dis_name, ang_name, tor_name


    ! read parameters
    !
    call read_ctrlfile_logical(handle, Section, &
                              'check_only', opt_info%check_only)

    call read_ctrlfile_real   (handle, Section, & 
                              'cutoff_qnative', opt_info%cutoff_qnative)

    ! check readable distance atom definitions
    !
    ndis = 0
    do while (.true.)
      value = ''
      write(dis_name,'(A8,I0)') 'distance', ndis + 1
      call read_ctrlfile_string(handle, Section, dis_name, value)
      if (value == '') &
        exit
      ndis = ndis + 1
    end do
    allocate(opt_info%dis_atoms(ndis))

    ! check readable angle atom definitions
    !
    nang = 0
    do while (.true.)
      value = ''
      write(ang_name,'(A5,I0)') 'angle', nang + 1
      call read_ctrlfile_string(handle, Section, ang_name, value)
      if (value == '') &
        exit
      nang = nang + 1
    end do
    allocate(opt_info%ang_atoms(nang))

    ! check readable torsion atom definitions
    !
    ntor = 0
    do while (.true.)
      value = ''
      write(tor_name,'(A7,I0)') 'torsion', ntor + 1
      call read_ctrlfile_string(handle, Section, tor_name, value)
      if (value == '') &
        exit
      ntor = ntor + 1
    end do
    allocate(opt_info%tor_atoms(ntor))


    ! read distance atom definitions
    !
    do i = 1, ndis
      write(dis_name,'(A8,I0)') 'distance', i
      call read_ctrlfile_string(handle, Section, dis_name, &
                                opt_info%dis_atoms(i))
    end do

    ! read angle atom definitions
    !
    do i = 1, nang
      write(ang_name,'(A5,I0)') 'angle', i
      call read_ctrlfile_string(handle, Section, ang_name, &
                                opt_info%ang_atoms(i))
    end do

    ! read torsion atom definitions
    !
    do i = 1, ntor
      write(tor_name,'(A7,I0)') 'torsion', i
      call read_ctrlfile_string(handle, Section, tor_name, &
                                opt_info%tor_atoms(i))
    end do


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Option> Parameters of Options'
    if (opt_info%check_only) then
      write(MsgOut,'(A20,A3)') '  check only      = ', 'yes'
    else
      write(MsgOut,'(A20,A2)') '  check only      = ', 'no'
    end if

    if (ndis > 0) then
      write(MsgOut,'(A20,I10)') '   # of distance  = ', ndis
      do i = 1, ndis
        write(MsgOut,'(A14,I0,A3,A)') &
             '     distance ', i, ' = ', trim(opt_info%dis_atoms(i))
      end do
    end if

    if (nang > 0) then
      write(MsgOut,'(A20,I10)') '   # of angle     = ', nang
      do i = 1, nang
        write(MsgOut,'(A14,I0,A3,A)') &
             '        angle ', i, ' = ', trim(opt_info%ang_atoms(i))
      end do
    end if

    if (ntor > 0) then
      write(MsgOut,'(A20,I10)') '   # of torsion   = ', ntor
      do i = 1, ntor
        write(MsgOut,'(A14,I0,A3,A)') &
             '      torsion ', i, ' = ', trim(opt_info%tor_atoms(i))
      end do
    end if

    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_option
  !> @brief        setup option information
  !! @authors      NT
  !! @param[in]    opt_info : OPTION section control parameters information
  !! @param[in]    molecule : molecule information
  !! @param[in]    gpr      : GPR information
  !! @param[inout] option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_option(opt_info, molecule, gpr, option)

    ! formal arguments
    type(s_opt_info),        intent(in)    :: opt_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_gpr),             intent(in)    :: gpr
    type(s_option),          intent(inout) :: option
    
    ! local variables
    integer                  :: i, ndis, nang, ntor, ncon


    ! distance
    !
    ndis = size(opt_info%dis_atoms)
    option%out_dis = ndis > 0

    if (option%out_dis) then

      allocate(option%distance(ndis), option%dist_list(2,ndis))

      write(MsgOut,'(A)') 'Setup_Option> distance atom indices: '
      do i = 1, ndis
        call parse_atom_defs( &
             molecule, opt_info%dis_atoms(i), option%dist_list(:,i))
        write(MsgOut,'(I5,A,I7,I7)') i, ' ) ', option%dist_list(:,i)
      end do

      write(MsgOut, '(A)') ' '

    end if

    ! angle
    !
    nang = size(opt_info%ang_atoms)
    option%out_ang = nang > 0

    if (option%out_ang) then

      allocate(option%angle(nang), option%angl_list(3,nang))

      write(MsgOut,'(A)') 'Setup_Option> angle atom indices: '
      do i = 1, nang
        call parse_atom_defs( &
             molecule, opt_info%ang_atoms(i), option%angl_list(:,i))
        write(MsgOut,'(I5,A,I7,I7,I7)') i, ' ) ', option%angl_list(:,i)
      end do
      write(MsgOut, '(A)') ' '

    end if

    ! torsion
    !
    ntor = size(opt_info%tor_atoms)
    option%out_tor = ntor > 0

    if (option%out_tor) then

      allocate(option%torsion(ntor), option%tors_list(4,ntor))

      write(MsgOut,'(A)') 'Setup_Option> torsion atom indices: '
      do i = 1, ntor
        call parse_atom_defs( &
             molecule, opt_info%tor_atoms(i), option%tors_list(:,i))
        write(MsgOut,'(I5,A,I7,I7,I7,I7)') i, ' ) ', option%tors_list(:,i)
      end do
      write(MsgOut, '(A)') ' '

    end if

    ! qnative
    !
    ncon = gpr%num_pairs
    option%out_qnt = ncon > 0

    if (option%out_qnt) then

      allocate(option%contact_list(2,ncon))

      do i = 1, ncon
        option%contact_list(1,i) = gpr%pair_list(1,i)
        option%contact_list(2,i) = gpr%pair_list(2,i)
      end do

      option%cutoff_qnative = opt_info%cutoff_qnative

      write(MsgOut,'(A)') 'Setup_Option> :'
      write(MsgOut,'(A20,F10.3)') '  cutoff_qnative  = ', option%cutoff_qnative
      write(MsgOut,'(A20,I10)')   '  num_contacts    = ', ncon
      do i = 1, ncon
        write(MsgOut,'(I10,2I12)') i, option%contact_list(1,i), &
                                      option%contact_list(2,i)
      end do
      write(MsgOut,'(A)') ' '

    end if

    return

  end subroutine setup_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    parse_atom_defs
  !> @brief        parse distance/angle/torsion atom definition line
  !! @authors      NT
  !! @param[in]    molecule : molecule information
  !! @param[in]    defs     : atom definitions
  !! @param[inout] atoms    : atom indices
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine parse_atom_defs(molecule, defs, atoms)

    ! formal argments
    type(s_molecule),        intent(in)    :: molecule
    character(*),            intent(in)    :: defs
    integer,                 intent(inout) :: atoms(:)

    ! local variables
    integer                  :: idx, i
    character(Maxline)       :: s(4)


    atoms(:) = -1

    s(:) = ''
    read(defs, *, err=900, end=100) s(1), s(2), s(3), s(4)

100 do i = 1, 4
      if (s(i) /= '') then
        idx = get_atom_index(molecule, s(i))
        if (idx == -1) &
          call error_msg('Parse_atom_defs> read error. atom not found : '//&
                         trim(s(i)))
        if (size(atoms) < i) &
          call error_msg('Parse_atom_defs> read error. bad format : '//&
                         trim(defs))
        atoms(i) = idx
      else
        exit
      end if
    end do

    do i = 1, size(atoms)
      if (atoms(i) == -1) &
        call error_msg('Parse_atom_defs> read error. bad format : '//&
                       trim(defs))
    end do

    return

900 call error_msg('Parse_atom_defs> read error.')

  end subroutine parse_atom_defs

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      get_atom_index
  !> @brief        get atom index
  !! @authors      NT
  !! @return       atom index
  !! @param[in]    molecule : molecule information
  !! @param[in]    def      : atom definition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  function get_atom_index(molecule, def)

    ! return value
    integer                  :: get_atom_index

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    character(*),            intent(in)    :: def

    ! local variables
    integer                  :: i, res_no, ntok
    character(Maxline)       :: cstr
    character(80)            :: an, sn, seg_name, res_name, atom_name


    get_atom_index = -1

    cstr = def
    ntok = 1
    do i = 1, len(cstr)
      if (cstr(i:i) == ':') then
        cstr(i:i) = ' '
        ntok = ntok + 1
      end if
    end do

    if (ntok < 3) then
      goto 900

    else if (ntok == 3) then
      
      read(cstr,*,err=900,end=900) res_no, res_name, atom_name
      do i = 1, molecule%num_atoms
        read(molecule%atom_name(i),*,err=900) an
        if (res_no    == molecule%residue_no(i)   .and. &
            res_name  == molecule%residue_name(i) .and. &
            atom_name == an) then
          get_atom_index = i
          return 
        end if
      end do

    else

      read(cstr,*,err=900,end=900) seg_name, res_no, res_name, atom_name
      do i = 1, molecule%num_atoms
        read(molecule%segment_name(i),*,err=900) sn
        read(molecule%atom_name(i),*,err=900) an
        if (seg_name  == sn                       .and. &
            res_no    == molecule%residue_no(i)   .and. &
            res_name  == molecule%residue_name(i) .and. &
            atom_name == an) then
          get_atom_index = i
          return 
        end if
      end do

    end if

    return 

900 call error_msg('Get_Atom_Index> read error '//def)

  end function get_atom_index

end module ta_option_mod
