!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_enefunc_table_mod
!> @brief   define lookup table for energy calculation
!! @authors Jaewoon Jung (JJ), Yuji Sugita (YS), Takaharu Mori (TM), 
!!          Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_enefunc_table_mod

  use at_mathlib_mod
  use at_energy_mod
  use at_energy_str_mod
  use at_enefunc_str_mod
  use molecules_str_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! constants
  integer,          parameter :: MAX_EXCL       = 16 ! = 4(1-2) + 4x3 (1-3)
  integer,          parameter :: MAX_NB_14      = 36 ! = 3x4x3 (1-4) 
  real(wp),         parameter :: FACT_NUM_NB15  = 1.5_wp
  real(wp),         parameter :: FACT_THRESHOLD = 0.95_wp

  ! subroutines
  public  :: setup_enefunc_table
  private :: setup_table_general_pme_cubic
  private :: setup_table_general_pme_linear
  private :: setup_table_general_cutoff_cubic
  private :: setup_table_water_pme_cubic
  private :: setup_table_water_pme_linear
  private :: setup_table_water_cutoff_cubic
  private :: setup_solute_water_lists
  private :: count_nonb_excl_water

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_table
  !> @brief        define lookup table of potential energy function
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[in]    molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @authors      JJ, TM
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_table(ene_info, molecule, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: cutoff, cutoff2_water, cutoff2
    real(wp)                 :: el_fact, alpha, alpha2m, alpha2sp
    integer                  :: cutoff_int
    integer                  :: cutoff_int2


    if (ene_info%table) then
      enefunc%table%table = .true.
    else
      enefunc%table%table = .false.
      return
    end if

    ! set variables
    !
    enefunc%table%table_order = ene_info%table_order
    enefunc%table%water_model = ene_info%water_model
    enefunc%table%density     = ene_info%table_density

    ! set common parameters
    !
    el_fact  = ELECOEF/ene_info%dielec_const
    alpha    = ene_info%pme_alpha
    alpha2m  = -alpha**2
    alpha2sp = 2.0_wp*alpha/sqrt(PI)

    cutoff        = ene_info%cutoffdist
    cutoff2       = cutoff * cutoff
    cutoff2_water = (cutoff + 4.5_wp)*(cutoff + 4.5_wp)        !TODO 4.5?

    if (enefunc%table%table_order == 3) then
      cutoff_int  = int(enefunc%table%density*cutoff2       )
      cutoff_int2 = int(enefunc%table%density*cutoff2_water )
    else
      cutoff_int  = int(enefunc%table%density*cutoff2*2.1_wp)  !TODO 2.1?
      cutoff_int2 = cutoff_int + 1
    end if

    call alloc_enefunc(enefunc, EneFuncTable, cutoff_int2, 1)

    ! make water and solute lists
    !
    call setup_solute_water_lists(molecule, enefunc)

    ! define lookup table of potential energy function
    !
    if (ene_info%electrostatic == ElectrostaticPME) then

      if (enefunc%table%table_order == 3) then
        call setup_table_general_pme_cubic (ene_info, alpha, alpha2m, alpha2sp,&
                                            cutoff2, cutoff_int, cutoff_int2,  &
                                            enefunc)
        call setup_table_water_pme_cubic   (ene_info, alpha, alpha2m, alpha2sp,&
                                            cutoff2, cutoff_int, cutoff_int2,  &
                                            molecule, enefunc)
      else 
        call setup_table_general_pme_linear(ene_info, alpha, alpha2m, alpha2sp,&
                                            cutoff2, cutoff_int, cutoff_int2,  &
                                            enefunc)
        call setup_table_water_pme_linear  (ene_info, alpha, alpha2m, alpha2sp,&
                                            cutoff2, cutoff_int, cutoff_int2,  &
                                            molecule, enefunc)
      end if

    else

      if (enefunc%table%table_order == 3) then
        call setup_table_general_cutoff_cubic(ene_info, cutoff2, cutoff_int,   &
                                              cutoff_int2, enefunc)
        call setup_table_water_cutoff_cubic  (ene_info, cutoff2, cutoff_int,   &
                                              cutoff_int2, molecule, enefunc)
      else
        call error_msg &
          ('Setup_Enefunc_Table> table_order = 1 is not available with CUTOFF')
      end if

    end if

    ! exclude 1-2, 1-3 interactions for solute-solute interaction
    !
    call count_nonb_excl_water(molecule, enefunc)

    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Setup_Enefunc_Table> Setup Variables for LookupTable'
      write(MsgOut,'(A20,I10,A20,I10)')                       &
            '  num_solutes     = ', enefunc%table%num_solute, &
            '  num_waters      = ', enefunc%table%num_water
      write(MsgOut,'(A)') ''
    end if

    return

  end subroutine setup_enefunc_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general_pme_cubic
  !> @brief        define lookup table of potential energy function
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general_pme_cubic(ene_info, alpha, alpha2m, alpha2sp, &
                                           cutoff2, cutoff_int, cutoff_int2,   &
                                           enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: alpha
    real(wp),                intent(in)    :: alpha2m
    real(wp),                intent(in)    :: alpha2sp
    real(wp),                intent(in)    :: cutoff2
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int2
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_rij, inv_r2, inv_r3, inv_r6, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2
    real(wp)                 :: lj1, lj2, lj4, lj12, lj6
    real(wp)                 :: density, el_fact
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:), table_grad(:)
    real(wp),        pointer :: table_ecor(:), table_decor(:)
    real(wp),        pointer :: dtable_ecor(:), dtable_decor(:)

    real(wp),    allocatable :: elec_array(:), delec_array(:)
    real(wp),    allocatable :: ecor_array(:), decor_array(:)
    real(wp),    allocatable :: lj12_array(:), dlj12_array(:)
    real(wp),    allocatable :: lj6_array(:), dlj6_array(:)


    el_fact      = ELECOEF / enefunc%dielec_const
    density      =  enefunc%table%density
    table_ene    => enefunc%table%table_ene
    table_ecor   => enefunc%table%table_ecor
    table_decor  => enefunc%table%table_decor
    dtable_ecor  => enefunc%table%dtable_ecor
    dtable_decor => enefunc%table%dtable_decor
    table_grad   => enefunc%table%table_grad

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)

    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_array(cutoff_int), delec_array(cutoff_int))
    allocate(ecor_array(cutoff_int), decor_array(cutoff_int))
    allocate(lj12_array(cutoff_int), dlj12_array(cutoff_int))
    allocate(lj6_array (cutoff_int), dlj6_array (cutoff_int))

    table_ene(1:6*cutoff_int2)  = 0.0_wp
    table_ecor(1:cutoff_int2)   = 0.0_wp
    table_decor(1:cutoff_int2)  = 0.0_wp
    dtable_ecor(1:cutoff_int2)  = 0.0_wp
    dtable_decor(1:cutoff_int2) = 0.0_wp
    table_grad(1:6*cutoff_int2) = 0.0_wp

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(6*i-5)  = inv_r12 - lj1
        table_ene(6*i-3)  = inv_r6 - lj2
        table_grad(6*i-5) = -12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = -6.0_wp*inv_r6*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r3  = sqrt(inv_r6)

        table_ene(6*i-5)  = lj12*(inv_r6-lj1)**2
        table_ene(6*i-3)  = lj6*(inv_r3-lj2)**2
        table_grad(6*i-5) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(6*i-3) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)
      end do

    else

      do i = 2, switch_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        switch   = 1.0_wp
        dswitch  = 0.0_wp
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij - switch*6.0_wp*inv_r6*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij  *(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)
      end do

    end if

    table_ecor(cutoff_int)   = 0.0_wp
    table_decor(cutoff_int)  = 0.0_wp
    dtable_ecor(cutoff_int)  = 0.0_wp
    dtable_decor(cutoff_int) = 0.0_wp
    table_ene(6*cutoff_int-5:6*cutoff_int)  = 0.0_wp
    table_grad(6*cutoff_int-5:6*cutoff_int) = 0.0_wp

    do i = 2, cutoff_int-1
      table_ene(6*i-4)  = (lj12_array(i-1) + lj12_array(i))   / 2.0_wp
      table_grad(6*i-4) = (dlj12_array(i-1) + dlj12_array(i)) / 2.0_wp
      table_ene(6*i-2)  = (lj6_array(i-1) + lj6_array(i))     / 2.0_wp
      table_grad(6*i-2) = (dlj6_array(i-1) + dlj6_array(i))   / 2.0_wp
      table_ene(6*i)    = (elec_array(i-1) + elec_array(i))   / 2.0_wp
      table_grad(6*i)   = (delec_array(i-1) + delec_array(i)) / 2.0_wp
      dtable_ecor(i)    = (ecor_array(i-1) + ecor_array(i))   / 2.0_wp
      dtable_decor(i)   = (decor_array(i-1) + decor_array(i)) / 2.0_wp
    end do

    table_ene(2)               = lj12_array(1)
    table_ene(4)               = lj6_array(1)
    table_ene(6)               = elec_array(1)
    table_ene(6*cutoff_int-4)  = lj12_array(cutoff_int-1)
    table_ene(6*cutoff_int-2)  = lj6_array(cutoff_int-1)
    table_ene(6*cutoff_int)    = elec_array(cutoff_int-1)
    table_grad(2)              = dlj12_array(1)
    table_grad(4)              = dlj6_array(1)
    table_grad(6)              = delec_array(1)
    table_grad(6*cutoff_int-4) = dlj12_array(cutoff_int-1)
    table_grad(6*cutoff_int-2) = dlj6_array(cutoff_int-1)
    table_grad(6*cutoff_int)   = delec_array(cutoff_int-1)

    deallocate(elec_array, delec_array)
    deallocate(ecor_array, decor_array)
    deallocate(lj12_array, dlj12_array)
    deallocate(lj6_array, dlj6_array)

    return

  end subroutine setup_table_general_pme_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general_pme_linear
  !> @brief        define lookup table of potential energy function
  !!               (linear)
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general_pme_linear(ene_info, alpha, alpha2m, alpha2sp,&
                                            cutoff2, cutoff_int, cutoff_int2,  &
                                            enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: alpha
    real(wp),                intent(in)    :: alpha2m
    real(wp),                intent(in)    :: alpha2sp
    real(wp),                intent(in)    :: cutoff2
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int2
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_r2, inv_rij, inv_r6, inv_r3, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2
    real(wp)                 :: lj1, lj2, lj4, dist, lj12, lj6
    real(wp)                 :: density, el_fact
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:), table_grad(:)
    real(wp),        pointer :: table_ecor(:), table_decor(:)


    el_fact     = ELECOEF / enefunc%dielec_const
    density     =  enefunc%table%density
    table_ene   => enefunc%table%table_ene
    table_ecor  => enefunc%table%table_ecor
    table_decor => enefunc%table%table_decor
    table_grad  => enefunc%table%table_grad

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(1.0_wp/switchdist2*cutoff2*density)

    table_ene  (1:3*cutoff_int2) = 0.0_wp
    table_ecor (1:cutoff_int2)   = 0.0_wp
    table_decor(1:cutoff_int2)   = 0.0_wp
    table_grad (1:3*cutoff_int2) = 0.0_wp

    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = switch_int+1, cutoff_int
        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(3*i-2)  = inv_r12 - lj1
        table_ene(3*i-1)  = inv_r6 - lj2
        table_grad(3*i-2) = -12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = -6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2
      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = density, switch_int

        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r3  = sqrt(inv_r6)

        table_ene(3*i-2)  = lj12*(inv_r6-lj1)**2
        table_ene(3*i-1)  = lj6*(inv_r3-lj2)**2
        table_grad(3*i-2) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(3*i-1) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
      end do

    else

      do i = switch_int+1, cutoff_int
        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(3*i-2)  = switch*inv_r12
        table_ene(3*i-1)  = switch*inv_r6
        table_grad(3*i-2) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2
      end do

      do i = density, switch_int

        rij2     = cutoff2*density/real(i,wp)
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij  *(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(3*i-2)  = switch*inv_r12
        table_ene(3*i-1)  = switch*inv_r6
        table_grad(3*i-2) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = dswitch*inv_r6*inv_rij - switch*6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
      end do

    end if

    return

  end subroutine setup_table_general_pme_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general_cutoff_cubic
  !> @brief        define lookup table for cutoff
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general_cutoff_cubic(ene_info, cutoff2, cutoff_int, &
                                              cutoff_int2, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: cutoff2
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int2
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2, cutoff2_inv, shift, dshift
    real(wp)                 :: lj1, lj2, lj4, el_fact, lj12, lj6
    real(wp)                 :: density
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:)
    real(wp),        pointer :: table_grad(:)

    real(wp),    allocatable :: elec_array(:), delec_array(:)
    real(wp),    allocatable :: lj12_array(:), dlj12_array(:)
    real(wp),    allocatable :: lj6_array(:), dlj6_array(:)


    density    =  enefunc%table%density
    table_ene  => enefunc%table%table_ene
    table_grad => enefunc%table%table_grad

    cutoff2_inv = 1.0_wp/cutoff2
    el_fact = ELECOEF/enefunc%dielec_const

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)

    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_array(cutoff_int2), delec_array(cutoff_int2))
    allocate(lj12_array(cutoff_int2), dlj12_array(cutoff_int2))
    allocate(lj6_array (cutoff_int2), dlj6_array (cutoff_int2))

    table_ene(1:6*cutoff_int2)  = 0.0_wp
    table_grad(1:6*cutoff_int2) = 0.0_wp

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int
        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(6*i-5)  = inv_r12 - lj1
        table_ene(6*i-3)  = inv_r6 - lj2
        table_grad(6*i-5) = -12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = -6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int
        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r3   = sqrt(inv_r6)

        table_ene(6*i-5)  = lj12*(inv_r6-lj1)**2
        table_ene(6*i-3)  = lj6*(inv_r3-lj2)**2
        table_grad(6*i-5) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(6*i-3) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

    else

      do i = 2, switch_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        switch   = 1.0_wp
        dswitch  = 0.0_wp
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

    end if

    table_ene (6*cutoff_int-5:6*cutoff_int) = 0.0_wp
    table_grad(6*cutoff_int-5:6*cutoff_int) = 0.0_wp

    do i = 2, cutoff_int-1
      table_ene(6*i-4)  = (lj12_array(i-1) + lj12_array(i))   / 2.0_wp
      table_grad(6*i-4) = (dlj12_array(i-1) + dlj12_array(i)) / 2.0_wp
      table_ene(6*i-2)  = (lj6_array(i-1) + lj6_array(i))     / 2.0_wp
      table_grad(6*i-2) = (dlj6_array(i-1) + dlj6_array(i))   / 2.0_wp
      table_ene(6*i)    = (elec_array(i-1) + elec_array(i))   / 2.0_wp
      table_grad(6*i)   = (delec_array(i-1) + delec_array(i)) / 2.0_wp
    end do

    table_ene(2)               = lj12_array(1)
    table_ene(4)               = lj6_array(1)
    table_ene(6)               = elec_array(1)
    table_ene(6*cutoff_int-4)  = lj12_array(cutoff_int-1)
    table_ene(6*cutoff_int-2)  = lj6_array(cutoff_int-1)
    table_ene(6*cutoff_int)    = elec_array(cutoff_int-1)
    table_grad(2)              = dlj12_array(1)
    table_grad(4)              = dlj6_array(1)
    table_grad(6)              = delec_array(1)
    table_grad(6*cutoff_int-4) = dlj12_array(cutoff_int-1)
    table_grad(6*cutoff_int-2) = dlj6_array(cutoff_int-1)
    table_grad(6*cutoff_int)   = delec_array(cutoff_int-1)

    deallocate(elec_array, delec_array)
    deallocate(lj12_array, dlj12_array)
    deallocate(lj6_array, dlj6_array)

    return

  end subroutine setup_table_general_cutoff_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water_pme_cubic
  !> @brief        define lookup table of water
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water_pme_cubic(ene_info, alpha, alpha2m, alpha2sp, &
                                         cutoff2, cutoff_int, cutoff_int2,   &
                                         molecule, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: alpha
    real(wp),                 intent(in)    :: alpha2m
    real(wp),                 intent(in)    :: alpha2sp
    real(wp),                 intent(in)    :: cutoff2
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int2
    type(s_molecule),         intent(in)    :: molecule
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: co6_OO, co6_OH, co6_HH
    real(wp)                  :: co12_OO, co12_OH, co12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_rij, inv_r2, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2
    real(wp)                  :: lj1, lj2, lj4, lj12, lj6
    real(wp)                  :: term_lj6, term_lj12
    real(wp)                  :: density, el_fact
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)

    real(wp),         allocatable :: elec_WW(:,:), evdw_WW(:,:), de_WW(:,:)

    el_fact      = ELECOEF / enefunc%dielec_const
    density      =  enefunc%table%density
    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    ! set switch distance etc.
    !
    switchdist   = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    lj1 = 1.0_wp/(cutoff2-switchdist2)
    lj1 = lj1*lj1*lj1
    switch_int = int(switchdist2*density)

    allocate(elec_WW(cutoff_int2,3), &
             evdw_WW(cutoff_int2,3), &
             de_WW  (cutoff_int2,3))

    table_ene_WW(1:6*cutoff_int2,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int2,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = molecule%atom_cls_no(enefunc%table%water_list(1,1))
    atmcls_H = molecule%atom_cls_no(enefunc%table%water_list(2,1))
    charge_O = molecule%charge(enefunc%table%water_list(1,1))
    charge_H = molecule%charge(enefunc%table%water_list(2,1))

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int

        rij2   = real(i,wp)/density
        rij    = sqrt(rij2)

        co6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * (inv_r12-lj1)
        term_lj12_OH = co12_OH * (inv_r12-lj1)
        term_lj12_HH = co12_HH * (inv_r12-lj1)
        term_lj6_OO   = co6_OO * (inv_r6-lj2)
        term_lj6_OH   = co6_OH * (inv_r6-lj2)
        term_lj6_HH   = co6_HH * (inv_r6-lj2)

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO   = co6_OO * inv_r6
        term_lj6_OH   = co6_OH * inv_r6
        term_lj6_HH   = co6_HH * inv_r6

        table_de_WW(2*i-1,1) = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(2*i-1,2) = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(2*i-1,3) = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2      = real(i,wp)/density
        rij       = sqrt(rij2)
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO   = co6_OO * term_lj6
        term_lj6_OH   = co6_OH * term_lj6
        term_lj6_HH   = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO   = co6_OO * term_lj6
        term_lj6_OH   = co6_OH * term_lj6
        term_lj6_HH   = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_de_WW(2*i-1,1) = term_lj_OO
        table_de_WW(2*i-1,2) = term_lj_OH
        table_de_WW(2*i-1,3) = term_lj_HH

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do

    else

      do i = 2, switch_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp

        co6_OO  = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH  = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH  = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO   = co6_OO * inv_r6
        term_lj6_OH   = co6_OH * inv_r6
        term_lj6_HH   = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij  &
                             - switch*(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij  &
                             - switch*(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij  &
                             - switch*(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch*(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch*(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch*(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do

    end if

    table_ene_WW(6*cutoff_int-1,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-3,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-5,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int-1,1:3)  = 0.0_wp
    table_ene_WW(6*cutoff_int,1:3)   = 0.0_wp
    table_ene_WW(6*cutoff_int-2,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-4,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int,1:3)    = 0.0_wp

    do i = 2, cutoff_int-1

      table_ene_WW(6*i-2,1:3) = (elec_WW(i-1,1:3) + elec_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i-4,1:3) = (evdw_WW(i-1,1:3) + evdw_WW(i,1:3)) / 2.0_wp
      table_de_WW(2*i,1:3)    = (de_WW(i-1,1:3) + de_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i,1:3)   = table_de_WW(2*i,1:3)

    end do

    table_ene_WW(4,1:3) = elec_WW(1,1:3)
    table_ene_WW(2,1:3) = evdw_WW(1,1:3)
    table_de_WW(2,1:3)  = de_WW(1,1:3)
    table_ene_WW(6,1:3) = de_WW(1,1:3)

    table_ene_WW(6*cutoff_int-2,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int-4,1:3) = evdw_WW(cutoff_int-1,1:3)
    table_de_WW(2*cutoff_int,1:3)    = de_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)

    deallocate(elec_WW, evdw_WW, de_WW)

    return

  end subroutine setup_table_water_pme_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water_pme_linear
  !> @brief        define lookup table of water (linear interpolation)
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water_pme_linear(ene_info, alpha, alpha2m, alpha2sp, &
                                          cutoff2, cutoff_int, cutoff_int2,   &
                                          molecule, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: alpha
    real(wp),                 intent(in)    :: alpha2m
    real(wp),                 intent(in)    :: alpha2sp
    real(wp),                 intent(in)    :: cutoff2
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int2
    type(s_molecule),         intent(in)    :: molecule
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: co6_OO, co6_OH, co6_HH
    real(wp)                  :: co12_OO, co12_OH, co12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2
    real(wp)                  :: lj1, lj2, lj4, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6
    real(wp)                  :: density, el_fact
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)

    el_fact      = ELECOEF / enefunc%dielec_const
    density      =  enefunc%table%density
    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    lj1 = 1.0_wp/(cutoff2-switchdist2)
    lj1 = lj1*lj1*lj1
    switch_int = int(1.0_wp/switchdist2*cutoff2*density)

    table_ene_WW(1:6*cutoff_int2,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int2,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = molecule%atom_cls_no(enefunc%table%water_list(1,1))
    atmcls_H = molecule%atom_cls_no(enefunc%table%water_list(2,1))
    charge_O = molecule%charge(enefunc%table%water_list(1,1))
    charge_H = molecule%charge(enefunc%table%water_list(2,1))

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = switch_int+1, cutoff_int

        rij2     = cutoff2*density/real(i,wp)
        rij    = sqrt(rij2)

        co6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * (inv_r12-lj1)
        term_lj12_OH = co12_OH * (inv_r12-lj1)
        term_lj12_HH = co12_HH * (inv_r12-lj1)
        term_lj6_OO  = co6_OO * (inv_r6-lj2)
        term_lj6_OH  = co6_OH * (inv_r6-lj2)
        term_lj6_HH  = co6_HH * (inv_r6-lj2)

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        table_de_WW(i,1)   = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(i,2)   = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(i,3)   = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1)   = table_de_WW(i,1)                      &
                           - (table_ene_WW(6*i-4,1)                &
                           + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2)                      &
                           - (table_ene_WW(6*i-4,2)                &
                           + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3)                      &
                           - (table_ene_WW(6*i-4,3)                &
                           +alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = density, switch_int
        rij2      = cutoff2*density/real(i,wp)
        rij       = sqrt(rij2)
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO   = co6_OO * term_lj6
        term_lj6_OH   = co6_OH * term_lj6
        term_lj6_HH   = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO  = co6_OO * term_lj6
        term_lj6_OH  = co6_OH * term_lj6
        term_lj6_HH  = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_de_WW(i,1)   = term_lj_OO
        table_de_WW(i,2)   = term_lj_OH
        table_de_WW(i,3)   = term_lj_HH

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1) = table_de_WW(i,1) &
            -(table_ene_WW(6*i-4,1)+alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2) = table_de_WW(i,2) &
            -(table_ene_WW(6*i-4,2)+alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3) = table_de_WW(i,3) &
            -(table_ene_WW(6*i-4,3)+alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, cutoff_int
        table_ene_WW(6*i-2,1:3) = table_ene_WW(6*(i+1)-5,1:3) - &
                                  table_ene_WW(6*i-5,1:3)
        table_ene_WW(6*i-1,1:3) = table_ene_WW(6*(i+1)-4,1:3) - &
                                  table_ene_WW(6*i-4,1:3)
        table_ene_WW(6*i,1:3)   = table_ene_WW(6*(i+1)-3,1:3) - &
                                  table_ene_WW(6*i-3,1:3)
      end do

    else

      do i = switch_int+1, cutoff_int

        rij2     = cutoff2*density/real(i,wp)
        rij    = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp

        co6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(i,1)   = (  dswitch * term_lj_OO * rij  &
                           - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(i,2)   = (  dswitch * term_lj_OH * rij  &
                           - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(i,3)   = (  dswitch * term_lj_HH * rij  &
                           - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1)   = table_de_WW(i,1)       &
                           - (table_ene_WW(6*i-4,1) &
                           + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2)       &
                           - (table_ene_WW(6*i-4,2) &
                           + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3)       &
                           - (table_ene_WW(6*i-4,3) &
                           +alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, switch_int
        rij2     = cutoff2*density/real(i,wp)
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp/rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        term_lj12_OO  = co12_OO * inv_r12
        term_lj12_OH  = co12_OH * inv_r12
        term_lj12_HH  = co12_HH * inv_r12
        term_lj6_OO   = co6_OO * inv_r6
        term_lj6_OH   = co6_OH * inv_r6
        term_lj6_HH   = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(i,1)   = (  dswitch * term_lj_OO * rij     &
                           - switch*(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(i,2)   = (  dswitch * term_lj_OH * rij     &
                           - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(i,3)   = (  dswitch * term_lj_HH * rij     &
                           - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1)   = table_de_WW(i,1) &
                -(table_ene_WW(6*i-4,1)+alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2) &
                -(table_ene_WW(6*i-4,2)+alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3) &
                -(table_ene_WW(6*i-4,3)+alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, cutoff_int
        table_ene_WW(6*i-2,1:3) = table_ene_WW(6*(i+1)-5,1:3) - &
                                  table_ene_WW(6*i-5,1:3)
        table_ene_WW(6*i-1,1:3) = table_ene_WW(6*(i+1)-4,1:3) - &
                                  table_ene_WW(6*i-4,1:3)
        table_ene_WW(6*i,1:3)   = table_ene_WW(6*(i+1)-3,1:3) - &
                                  table_ene_WW(6*i-3,1:3)
      end do

    end if

    return

  end subroutine setup_table_water_pme_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water_cutoff_cubic
  !> @brief        define lookup table of water for cutoff
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water_cutoff_cubic(ene_info, cutoff2, cutoff_int, &
                                            cutoff_int2, molecule, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: cutoff2
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int2
    type(s_molecule),         intent(in)    :: molecule
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: co6_OO, co6_OH, co6_HH
    real(wp)                  :: co12_OO, co12_OH, co12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2, cutoff2_inv, shift, dshift
    real(wp)                  :: term_lj12, term_lj6
    real(wp)                  :: lj1, lj2, lj4, el_fact, lj12, lj6
    real(wp)                  :: density
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)

    real(wp),         allocatable :: elec_WW(:,:), evdw_WW(:,:), de_WW(:,:)


    density      =  enefunc%table%density
    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    cutoff2_inv = 1.0_wp / cutoff2
    el_fact = ELECOEF / enefunc%dielec_const

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)

    if (.not.ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_WW(cutoff_int2,3), &
             evdw_WW(cutoff_int2,3), &
             de_WW  (cutoff_int2,3))

    table_ene_WW(1:6*cutoff_int2,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int2,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = molecule%atom_cls_no(enefunc%table%water_list(1,1))
    atmcls_H = molecule%atom_cls_no(enefunc%table%water_list(2,1))
    charge_O = molecule%charge(enefunc%table%water_list(1,1))
    charge_H = molecule%charge(enefunc%table%water_list(2,1))

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift

        co6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * (inv_r12-lj1)
        term_lj12_OH = co12_OH * (inv_r12-lj1)
        term_lj12_HH = co12_HH * (inv_r12-lj1)
        term_lj6_OO  = co6_OO * (inv_r6-lj2)
        term_lj6_OH  = co6_OH * (inv_r6-lj2)
        term_lj6_HH  = co6_HH * (inv_r6-lj2)

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        table_de_WW(2*i-1,1)   = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(2*i-1,2)   = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(2*i-1,3)   = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)   - table_de_WW(2*i-3,1:3)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2      = real(i,wp)/density
        rij       = sqrt(rij2)
        shift     = (1.0_wp - (rij2*cutoff2_inv))
        dshift    = -4.0_wp*shift*rij*cutoff2_inv
        shift     = shift * shift
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO  = co6_OO * term_lj6
        term_lj6_OH  = co6_OH * term_lj6
        term_lj6_HH  = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = co12_OO * term_lj12
        term_lj12_OH = co12_OH * term_lj12
        term_lj12_HH = co12_HH * term_lj12
        term_lj6_OO  = co6_OO * term_lj6
        term_lj6_OH  = co6_OH * term_lj6
        term_lj6_HH  = co6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_de_WW(2*i-1,1) = term_lj_OO
        table_de_WW(2*i-1,2) = term_lj_OH
        table_de_WW(2*i-1,3) = term_lj_HH

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW  (i-1,1:3) = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)

      end do

    else

      do i = 2, switch_int

        rij2     = real(i,wp)/density
        rij    = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift

        co6_OO  = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        co6_OH  = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        co6_HH  = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        co12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        co12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        co12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch*(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch*(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch*(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)  - table_de_WW(2*i-3,1:3)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = co12_OO * inv_r12
        term_lj12_OH = co12_OH * inv_r12
        term_lj12_HH = co12_HH * inv_r12
        term_lj6_OO  = co6_OO * inv_r6
        term_lj6_OH  = co6_OH * inv_r6
        term_lj6_HH  = co6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch*(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO))*inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch*(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH))*inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch*(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH))*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW  (i-1,1:3) = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)

      end do

    end if

    table_ene_WW(6*cutoff_int-5,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-3,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-1,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int-1,1:3)  = 0.0_wp
    table_ene_WW(6*cutoff_int-4,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-2,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int,1:3)   = 0.0_wp
    table_de_WW(2*cutoff_int,1:3)    = 0.0_wp

    do i = 2, cutoff_int-1

      table_ene_WW(6*i-2,1:3) = (elec_WW(i-1,1:3) + elec_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i-4,1:3) = (evdw_WW(i-1,1:3) + evdw_WW(i,1:3)) / 2.0_wp
      table_de_WW(2*i,1:3)    = (de_WW(i-1,1:3) + de_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i,1:3)   = table_de_WW(2*i,1:3)

    end do

    table_ene_WW(4,1:3) = elec_WW(1,1:3)
    table_ene_WW(2,1:3) = evdw_WW(1,1:3)
    table_de_WW (2,1:3) = de_WW(1,1:3)
    table_ene_WW(6,1:3) = evdw_WW(1,1:3)

    table_ene_WW(6*cutoff_int-2,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int-4,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)
    table_de_WW (2*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)

    deallocate(elec_WW, evdw_WW, de_WW)

    return

  end subroutine setup_table_water_cutoff_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_solute_water_lists
  !> @brief        setup solute and water lists
  !! @authors      CK, TM, JJ
  !! @param[in]    molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_solute_water_lists(molecule, enefunc)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: i, nbond, nwater
    integer                  :: nangle
    integer                  :: i1, i2, i3, natom
    integer                  :: mole_no, tmp_mole_no, wat_found, solute_found
    integer                  :: len_wat, len_ri1, len_ri2, len_ri3
    character(4)             :: ri1, ri2, ri3, ci1, ci2, ci3

    logical,     allocatable :: solute(:)


    nbond  = enefunc%num_bonds
    natom  = molecule%num_atoms
    nangle = enefunc%num_angles

    ! count the number of water molecules
    !
    wat_found    = 0
    solute_found = 0
    tmp_mole_no  = -1
    len_wat      = len_trim(enefunc%table%water_model)

    ! allocate local array
    !
    allocate(solute(natom))

    solute(1:natom) = .true.

    if (enefunc%table%num_water == 0) then

      do i = 1, nbond

        ri1 = molecule%residue_name(enefunc%bond_list(1,i))
        ri2 = molecule%residue_name(enefunc%bond_list(2,i))
        len_ri1 = len_trim(ri1)
        len_ri2 = len_trim(ri2)
    
        if (ri1(1:len_ri1) == enefunc%table%water_model(1:len_wat) .and. &
            ri2(1:len_ri2) == enefunc%table%water_model(1:len_wat)) then

          mole_no = molecule%molecule_no(enefunc%bond_list(1,i))
          if (mole_no /= tmp_mole_no) then
            wat_found   = wat_found + 1
            tmp_mole_no = mole_no
          end if

        end if
      end do
    
      nwater = wat_found
      if (nwater == 0) then
        call error_msg('Setup_Solute_Water_Lists> Water was not found. Table in [ENERGY] must be NO')
      end if

      enefunc%table%num_water = nwater
      call alloc_enefunc(enefunc, EneFuncTableWater, nwater, 1)
    
      ! search water molecules by angles
      !
      wat_found   =  0
      tmp_mole_no = -1

      do i = 1, nangle

        ri1 = molecule%residue_name(enefunc%angl_list(1,i))
        ri2 = molecule%residue_name(enefunc%angl_list(2,i))
        ri3 = molecule%residue_name(enefunc%angl_list(3,i))
        len_ri1 = len_trim(ri1)
        len_ri2 = len_trim(ri2)
        len_ri3 = len_trim(ri3)

        if (ri1(1:len_ri1) == enefunc%table%water_model(1:len_wat) .and. &
            ri2(1:len_ri2) == enefunc%table%water_model(1:len_wat) .and. &
            ri3(1:len_ri3) == enefunc%table%water_model(1:len_wat)) then

          mole_no = molecule%molecule_no(enefunc%angl_list(1,i))
          ci1 = molecule%atom_name(enefunc%angl_list(1,i))
          ci2 = molecule%atom_name(enefunc%angl_list(2,i))
          ci3 = molecule%atom_name(enefunc%angl_list(3,i))
      
          if (mole_no /= tmp_mole_no) then

            wat_found   = wat_found + 1
            tmp_mole_no = mole_no

            if (ci1(1:1) == 'H' .and. ci2(1:1) == 'O' .and. ci3(1:1) == 'H')then

              enefunc%table%water_list(1,wat_found) = enefunc%angl_list(2,i)

              if (enefunc%angl_list(1,i) > enefunc%angl_list(3,i)) then
                enefunc%table%water_list(2,wat_found) = enefunc%angl_list(3,i)
                enefunc%table%water_list(3,wat_found) = enefunc%angl_list(1,i)
              else
                enefunc%table%water_list(2,wat_found) = enefunc%angl_list(1,i)
                enefunc%table%water_list(3,wat_found) = enefunc%angl_list(3,i)
              end if

            end if
          end if
        end if
      end do

      if (wat_found /= enefunc%table%num_water) then
        call error_msg('Setup_Water> number of water is incorrect')
      end if

    endif

    ! check solute list
    !
    do i = 1, enefunc%table%num_water
      i1 = enefunc%table%water_list(1,i)
      i2 = enefunc%table%water_list(2,i)
      i3 = enefunc%table%water_list(3,i)
      solute(i1) = .false. 
      solute(i2) = .false. 
      solute(i3) = .false. 
    end do

    do i = 1, natom
      if (solute(i)) solute_found = solute_found + 1
    end do
    enefunc%table%num_solute = solute_found

    call alloc_enefunc(enefunc, EneFuncTableSolute, solute_found, natom)

    solute_found = 0             
    enefunc%table%solute_list_inv(1:natom) = 0

    do i = 1, natom
      if (solute(i)) then
        solute_found = solute_found + 1
        enefunc%table%solute_list(solute_found) = i
        enefunc%table%solute_list_inv(i) = solute_found
      end if
    end do
 
    ! deallocate local array
    !
    deallocate(solute)

    return

  end subroutine setup_solute_water_lists

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_water
  !> @brief        exclude 1-2, 1-3 interactions for solute-solute interaction
  !! @authors      JJ
  !! @param[in]    molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_water(molecule, enefunc)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, k, k1, k2, k3, k4, nsolute
    integer                  :: num_excl, num_excl_total, num_excl_total1
    integer                  :: num_nb14, num_nb14_total
    integer                  :: alloc_stat, dealloc_stat
    logical                  :: duplicate

    integer,         pointer :: solute(:)

    integer,     allocatable :: nonb_excl_list(:,:), nonb_excl_list1(:,:)
    integer,     allocatable :: nb14_calc_list(:,:)


    solute  => enefunc%table%solute_list_inv
    nsolute = enefunc%table%num_solute

    ! allocate nonbonded exclusion list (solute only and all) and 
    ! 1-4 interaction list
    !
    allocate(enefunc%table%num_nonb_excl(nsolute), stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc
    allocate(nonb_excl_list(MAX_EXCL,nsolute),stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    allocate(enefunc%num_nonb_excl(molecule%num_atoms), stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc
    allocate(nonb_excl_list1(MAX_EXCL,molecule%num_atoms),stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    allocate(enefunc%num_nb14_calc(nsolute), stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc
    allocate(nb14_calc_list(MAX_NB_14,nsolute),stat=alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    enefunc%table%num_nonb_excl(1:nsolute) = 0
    enefunc%num_nb14_calc(1:nsolute) = 0
    enefunc%num_nonb_excl(1:molecule%num_atoms) = 0

    ! exclude 1-2 interaction
    !
    num_excl_total  = 0
    num_excl_total1 = 0

    do k = 1, molecule%num_bonds

      k1 = solute(molecule%bond_list(1,k))
      k2 = solute(molecule%bond_list(2,k))

      if (k1 /= 0 .and. k2 /= 0) then

        if (k1 < k2) then
          num_excl = enefunc%table%num_nonb_excl(k1) + 1
          enefunc%table%num_nonb_excl(k1) = num_excl
          nonb_excl_list(num_excl,k1) = k2
          num_excl_total = num_excl_total + 1
        else
          num_excl = enefunc%table%num_nonb_excl(k2) + 1
          enefunc%table%num_nonb_excl(k2) = num_excl
          nonb_excl_list(num_excl,k2) = k1
          num_excl_total = num_excl_total + 1
        end if

      end if

      k1 = molecule%bond_list(1,k)
      k2 = molecule%bond_list(2,k)

      if (k1 < k2) then
        num_excl = enefunc%num_nonb_excl(k1) + 1
        enefunc%num_nonb_excl(k1) = num_excl
        nonb_excl_list1(num_excl,k1) = k2
        num_excl_total1 = num_excl_total1 + 1
      else
        num_excl = enefunc%num_nonb_excl(k2) + 1
        enefunc%num_nonb_excl(k2) = num_excl
        nonb_excl_list1(num_excl,k2) = k1
        num_excl_total1 = num_excl_total1 + 1
      end if

    end do

    ! exclude 1-3 interaction
    !
    do k = 1, molecule%num_angles

      k1 = solute(molecule%angl_list(1,k))
      k3 = solute(molecule%angl_list(3,k))

      if (k1 /= 0 .and. k3 /= 0) then 

        if (k1 < k3) then

          num_excl = enefunc%table%num_nonb_excl(k1)
          duplicate = .false.

          do i = 1, enefunc%table%num_nonb_excl(k1)
            if (k3 == nonb_excl_list(i,k1)) duplicate = .true.
          end do

          if (.not.duplicate) then
            num_excl = num_excl + 1
            enefunc%table%num_nonb_excl(k1) = num_excl
            nonb_excl_list(num_excl,k1) = k3
            num_excl_total = num_excl_total + 1
          end if

        else

          num_excl = enefunc%table%num_nonb_excl(k3)
          duplicate = .false.

          do i = 1, enefunc%table%num_nonb_excl(k3)
            if (k1 == nonb_excl_list(i,k3)) duplicate = .true.
          end do

          if (.not.duplicate) then
            num_excl = num_excl + 1
            enefunc%table%num_nonb_excl(k3) = num_excl
            nonb_excl_list(num_excl,k3) = k1
            num_excl_total = num_excl_total + 1
          end if

        end if
      end if

      k1 = molecule%angl_list(1,k)
      k3 = molecule%angl_list(3,k)

      if (k1 < k3) then

        num_excl = enefunc%num_nonb_excl(k1)
        duplicate = .false.

        do i = 1, enefunc%num_nonb_excl(k1)
          if (k3 == nonb_excl_list1(i,k1)) duplicate = .true.
        end do

        if (.not.duplicate) then
           num_excl = num_excl + 1
           enefunc%num_nonb_excl(k1) = num_excl
           nonb_excl_list1(num_excl,k1) = k3
           num_excl_total1 = num_excl_total1 + 1
        end if

      else

        num_excl = enefunc%num_nonb_excl(k3)
        duplicate = .false.

        do i = 1, enefunc%num_nonb_excl(k3)
          if (k1 == nonb_excl_list1(i,k3)) duplicate = .true.
        end do

        if (.not.duplicate) then
          num_excl = num_excl + 1
          enefunc%num_nonb_excl(k3) = num_excl
          nonb_excl_list1(num_excl,k3) = k1
          num_excl_total1 = num_excl_total1 + 1
        end if
      end if

    end do

    ! count 1-4 interaction
    !
    num_nb14_total = 0
    do k = 1, molecule%num_dihedrals

      k1 = solute(molecule%dihe_list(1,k))
      k4 = solute(molecule%dihe_list(4,k))

      if (k1 /= 0 .and. k4 /= 0) then

        if (k1 < k4) then

          num_nb14 = enefunc%num_nb14_calc(k1)
          duplicate = .false.

          do i = 1, enefunc%table%num_nonb_excl(k1)
            if (k4 == nonb_excl_list(i,k1)) duplicate = .true.
          end do

          do i = 1, num_nb14
            if (k4 == nb14_calc_list(i,k1)) duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            enefunc%num_nb14_calc(k1) = num_nb14
            nb14_calc_list(num_nb14,k1) = k4
            num_nb14_total = num_nb14_total + 1
          end if

        else

          num_nb14 = enefunc%num_nb14_calc(k4)
          duplicate = .false.

          do i = 1, enefunc%table%num_nonb_excl(k4)
            if (k1 == nonb_excl_list(i,k4)) duplicate = .true.
          end do

          do i = 1, num_nb14
            if (k1 == nb14_calc_list(i,k4)) duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            enefunc%num_nb14_calc(k4) = num_nb14
            nb14_calc_list(num_nb14,k4) = k1
            num_nb14_total = num_nb14_total + 1
          end if

        end if
      end if
    end do

    ! pack 2D-array into 1D-array
    !
    allocate(enefunc%nonb_excl_list(num_excl_total1), stat = alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    call pack_array(molecule%num_atoms, enefunc%num_nonb_excl,  &
                    nonb_excl_list1, enefunc%nonb_excl_list)
    deallocate(nonb_excl_list1, stat = dealloc_stat)

    allocate(enefunc%table%nonb_excl_list(num_excl_total), stat = alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    call pack_array(nsolute, enefunc%table%num_nonb_excl,  &
                    nonb_excl_list, enefunc%table%nonb_excl_list)
    deallocate(nonb_excl_list, stat = dealloc_stat)

    allocate(enefunc%nb14_calc_list(num_nb14_total), stat = alloc_stat)
    if (alloc_stat /=0) call error_msg_alloc

    call pack_array(nsolute, enefunc%num_nb14_calc,  &
                    nb14_calc_list, enefunc%nb14_calc_list)
    deallocate(nb14_calc_list, stat = dealloc_stat)

    return

  end subroutine count_nonb_excl_water

end module at_enefunc_table_mod
