!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_energy_nonbonds_mod
!> @brief   calculate nonbonded energy
!! @authors Takaharu Mori (TM), Yuji Sugita (YS), Jaewoon Jung (JJ), 
!!          Takashi Imai (TI), 
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_energy_nonbonds_mod

  use at_energy_table_mod
  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use molecules_str_mod
  use timers_mod
  use mpi_parallel_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: compute_energy_nonbond_cutoff 
  public  :: compute_energy_nonbond_list
  public  :: compute_energy_nonbond14 
  private :: compute_energy_nonbond_pbc

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_cutoff
  !> @brief        Calculate nonbond energy by cutoff
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    boundary : boundary conditions information
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_cutoff(enefunc, molecule, pairlist, &
                                           boundary, nonb_ene, coord,   &
                                           force, virial, eelec, evdw)

    ! formal arguments
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_molecule),        intent(in)    :: molecule
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(in)    :: coord(:,:)
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: eelec
    real(wp),                intent(inout) :: evdw


    call timer(TimerPmeReal, TimerOn)

    if (enefunc%table%table) then

      call timer(TimerNonBond, TimerOn)

      ! use lookup table
      !
      if (enefunc%table%table_order == 3) then
        call compute_energy_nonbond_table(enefunc, molecule, pairlist,       &
                                          boundary, nonb_ene, coord,         &
                                          force, virial, eelec, evdw)
      else
        call error_msg &
          ('Compute_Energy_Nonbond_Cutoff> table_order = 1 is disabled')
      end if

      call timer(TimerNonBond, TimerOff)

    else

      ! use traditional scheme
      !
      call compute_energy_nonbond14(enefunc, molecule, coord, force, virial, &
                                    eelec, evdw)

      call compute_energy_nonbond_pbc(enefunc, molecule, pairlist, boundary, &
                                      coord, force, virial,                  &
                                      eelec, evdw)

    end if

    call timer(TimerPmeReal, TimerOff)

    return

  end subroutine compute_energy_nonbond_cutoff

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond14
  !> @brief        calculate nonbonded14 energy
  !  @authors      YS, JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond14(enefunc, molecule, coord, force, virial, &
                                      eelec, evdw)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: dij(1:3), rij, rij2
    real(wp)                  :: inv_rij, inv_r3, inv_r6, inv_r10, inv_r12
    real(wp)                  :: lj6, lj10, lj12, CCOEF
    real(wp)                  :: term_elec, shift, dshift, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2
    real(wp)                  :: c1_switch, c2_switch, c4_switch, coef
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown
    real(wp)                  :: c12, c6, term_lj12, term_lj6
    real(wp)                  :: work(1:3)
    integer                   :: i, j, k, l, natom
    integer                   :: num_nb14, ini_nb14, fin_nb14
    integer                   :: id, my_id, nthread
    integer                   :: omp_get_thread_num, omp_get_num_threads

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: dielec_const
    real(wp),         pointer :: cutoff, switchdist
    integer,          pointer :: atmcls(:)


    call timer(TimerNonBond, TimerOn)

    charge       => molecule%charge
    atmcls       => molecule%atom_cls_no

    dielec_const => enefunc%dielec_const
    cutoff       => enefunc%cutoffdist
    switchdist   => enefunc%switchdist

    natom        =  molecule%num_atoms
    CCOEF        =  ELECOEF/dielec_const
    cutoff2      =  cutoff * cutoff
    switchdist2  =  switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup   = cutoff2*cutoff2*cutoff2
      c2_switchup   = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    ! calculate energy and gradient
    !
    num_nb14 = 0

    !$omp parallel                                                            &
    !$omp private(i, k, j, ini_nb14, fin_nb14, dij, rij2, rij, work,          &
    !$omp         c2_switch, c4_switch, switch, dswitch, shift, dshift,       &
    !$omp         lj6, lj10, lj12, inv_rij, inv_r3, inv_r6, inv_r10, inv_r12, &
    !$omp         coef, term_elec, term_lj12, term_lj6, id, my_id)            &
    !$omp firstprivate(num_nb14)                                              &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)           &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
    nthread = omp_get_num_threads()
#else
    id      = 0
    nthread = 1
#endif
    my_id = my_city_rank * nthread + id

    do i = 1, natom-1

      ini_nb14 = num_nb14 + 1
      fin_nb14 = num_nb14 + enefunc%num_nb14_calc(i)
      num_nb14 = fin_nb14

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb14 = fin_nb14 + 1

      do k = ini_nb14, fin_nb14

        j = enefunc%nb14_calc_list(k)

        ! compute distance
        !
        dij(1:3) = coord(1:3,i) - coord(1:3,j)
        rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
        rij      = sqrt(rij2)
        inv_rij  = 1.0_wp / rij 

        ! electrostatic energy 
        !
        shift  = (1.0_wp - (rij2/cutoff2))
        dshift = -4.0_wp*shift*rij/cutoff2
        shift  = shift * shift
        if (enefunc%forcefield == ForcefieldCHARMM19) &
          shift = shift * enefunc%nb14_qq_scale
        if (enefunc%forcefield == ForcefieldCHARMM19) &
          dshift = dshift * enefunc%nb14_qq_scale
        term_elec = charge(i) * charge(j) * CCOEF
        eelec  = eelec + shift * ( term_elec * inv_rij )

        ! lennard-jones energy and gradient
        !
        lj6  = enefunc%nb14_lj6 (atmcls(i),atmcls(j))
        lj10 = enefunc%nb14_lj10(atmcls(i),atmcls(j))
        lj12 = enefunc%nb14_lj12(atmcls(i),atmcls(j))

        if (enefunc%forcefield == ForcefieldKBGO) then

          if (rij2 > switchdist2) then
            c2_switch = cutoff2 - rij2
            c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
            switch    = c2_switch*c4_switch*c1_switch
            dswitch   = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
          else
            switch    = 1.0_wp
            dswitch   = 0.0_wp
          end if

          rij2 = 1.0_wp / rij2
          inv_r6  = rij2 * rij2 * rij2
          inv_r10 = inv_r6 * rij2 * rij2
          inv_r12 = inv_r6 * inv_r6
          evdw = evdw + switch*(lj12 * inv_r12 - lj10 * inv_r10 + lj6 * inv_r6)
          coef = (dswitch * (lj12*inv_r12 - lj10*inv_r10 + lj6*inv_r6) * rij  &
                 - switch * (12.0_wp*lj12*inv_r12 - &
                             10.0_wp*lj10*inv_r10 + &
                             6.0_wp*lj6*inv_r6)) * rij2
        else if (enefunc%force_switch) then

          if (rij2 > switchdist2) then
            rij2 = 1.0_wp / rij2
            inv_r3 = rij2 * inv_rij
            inv_r6 = inv_r3 * inv_r3
            term_lj12 = lj12 * c12 * (inv_r6-c1_switchup)**2
            term_lj6 = lj6 * c6 * (inv_r3-c2_switchup)**2
            evdw = evdw + term_lj12 - term_lj6
            coef = -12.0_wp * lj12 * c12 * (inv_r6-c1_switchup) * inv_r6 * rij2
            coef = coef + 6.0_wp * lj6 * c6 * (inv_r3-c2_switchup) * inv_r3*rij2
          else
            rij2 = 1.0_wp / rij2
            inv_r3 = rij2 * inv_rij
            inv_r6 = inv_r3 * inv_r3 
            inv_r12 = inv_r6 * inv_r6
            coef = -12.0_wp * lj12 * inv_r12 * rij2
            coef = coef + 6.0_wp * lj6 * inv_r6 * rij2
            inv_r12 = inv_r12 - c1_switchdown
            inv_r6 = inv_r6 - c2_switchdown
            term_lj12 = lj12 * inv_r12
            term_lj6 = lj6 * inv_r6
            evdw = evdw + term_lj12 - term_lj6
          end if

        else 

          if (rij2 > switchdist2) then
            c2_switch = cutoff2 - rij2
            c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
            switch  = c2_switch*c4_switch*c1_switch
            dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
          else
            switch  = 1.0_wp
            dswitch = 0.0_wp
          end if

          rij2    = 1.0_wp / rij2
          inv_r6  = rij2 * rij2 * rij2
          inv_r12 = inv_r6 * inv_r6
          evdw    = evdw + switch*(lj12 * inv_r12 - lj6 * inv_r6)
          coef    = ( dswitch * (lj12*inv_r12 - lj6*inv_r6) * rij  &
                     - switch * (12.0_wp*lj12*inv_r12-6.0_wp*lj6*inv_r6) )*rij2
        end if

        ! electrostatic gradient
        !
        coef = coef + (dshift*(term_elec*inv_rij) - &
                        shift*(term_elec*rij2))*inv_rij

        work(1:3) = coef*dij(1:3)

        ! store force
        !
        force(1:3,i) = force(1:3,i) - work(1:3)
        force(1:3,j) = force(1:3,j) + work(1:3)

        ! virial
        !
        do l = 1, 3
          virial(1:3,l) = virial(1:3,l) - dij(1:3)*work(l)
        end do

      end do
    end do
    !$omp end parallel
 
    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond14

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_list
  !> @brief        calculate nonbonded energy with pairlist (NOBC)
  !! @authors      YS, JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_list(enefunc, molecule, pairlist, coord, &
                                         force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: dij(1:3), rij, rij2
    real(wp)                  :: inv_rij, inv_r6, inv_r12, inv_r10, dr126
    real(wp)                  :: lj6, lj10, lj12, CCOEF
    real(wp)                  :: term_elec, shift, dshift, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2
    real(wp)                  :: c1_switch, c2_switch, c4_switch, coef
    real(wp)                  :: work(1:3)
    integer                   :: i, j, k, l, natom
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, nthread, my_id
    integer                   :: omp_get_num_threads, omp_get_thread_num

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: dielec_const
    real(wp),         pointer :: cutoff, switchdist
    integer,          pointer :: atmcls(:)


    call timer(TimerNonBond, TimerOn)

    charge       => molecule%charge
    atmcls       => molecule%atom_cls_no

    dielec_const => enefunc%dielec_const
    cutoff       => enefunc%cutoffdist
    switchdist   => enefunc%switchdist

    natom        =  molecule%num_atoms
    CCOEF        =  ELECOEF/dielec_const
    cutoff2      =  cutoff * cutoff
    switchdist2  =  switchdist * switchdist

    c1_switch    =  1.0_wp/(cutoff2-switchdist2)
    c1_switch    =  c1_switch*c1_switch*c1_switch

    ! calculate energy and gradient
    !
    num_nb15 = 0

    !$omp parallel                                                            &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, k, j, dij, rij, rij2,     &
    !$omp         work, c2_switch, c4_switch, switch, dswitch, shift, dshift, &
    !$omp         lj6, lj10, lj12, inv_rij, inv_r6, inv_r10, inv_r12,         &
    !$omp         dr126, coef, term_elec)                                     &
    !$omp firstprivate(num_nb15)                                              &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)           &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
    nthread = omp_get_num_threads()
#else
    id      = 0
    nthread = 1
#endif
    my_id   = my_city_rank * nthread + id

    do i = 1, natom-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + pairlist%num_nb15_calc(i,id+1)
      num_nb15 = fin_nb15

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15
        j = pairlist%nb15_calc_list(k,id+1)

        ! compute distance
        !
        dij(1:3) = coord(1:3,i) - coord(1:3,j)
        rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        ! cutoff
        !
        if (rij2 < cutoff2) then

          rij = sqrt(rij2)        

          if (rij2 > switchdist2) then
            c2_switch = cutoff2 - rij2
            c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
            switch  = c2_switch*c4_switch*c1_switch
            dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
          else
            switch  = 1.0_wp
            dswitch = 0.0_wp
          end if

          shift  = (1.0_wp - (rij2/cutoff2))
          dshift = -4.0_wp*shift*rij/cutoff2
          shift  = shift * shift

          lj6  = enefunc%nonb_lj6 (atmcls(i),atmcls(j))
          lj10 = enefunc%nonb_lj10(atmcls(i),atmcls(j))
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))

          ! lj energy and gradient
          !
          if (enefunc%forcefield == ForcefieldKBGO) then

            rij2    = 1.0_wp / rij2
            inv_rij = 1.0_wp / rij
            inv_r6  = rij2 * rij2 * rij2
            inv_r10 = inv_r6 * rij2 * rij2
            inv_r12 = inv_r6 * inv_r6
            inv_r6  = lj6 * inv_r6
            inv_r10 = lj10 * inv_r10
            inv_r12 = lj12 * inv_r12
            dr126   = inv_r12 - inv_r10 + inv_r6
            evdw    = evdw + switch*dr126 
            
            ! gradient
            !
            coef = (  dswitch * dr126 * rij &
                     - switch * (12.0_wp*inv_r12 - &
                                 10.0_wp*inv_r10 + &
                                 6.0_wp*inv_r6) ) * rij2
          else

            rij2    = 1.0_wp / rij2
            inv_rij = 1.0_wp / rij
            inv_r6  = rij2 * rij2 * rij2
            inv_r12 = inv_r6 * inv_r6
            inv_r6  = lj6 * inv_r6
            inv_r12 = lj12 * inv_r12
            dr126   = inv_r12 - inv_r6
            evdw    = evdw + switch*dr126 
            
            ! gradient
            !
            coef = (  dswitch * dr126 * rij &
                     - switch * (12.0_wp*inv_r12-6.0_wp*inv_r6) ) * rij2
          end if

          ! electrostatic energy and gradient
          !
          term_elec = charge(i) * charge(j) * CCOEF
          eelec = eelec + shift * ( term_elec * inv_rij )
          coef  = coef + (dshift*(term_elec*inv_rij) &
                         - shift*(term_elec*rij2))*inv_rij

          work(1:3) = coef*dij(1:3)

          ! store force
          !
          force(1:3,i) = force(1:3,i) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do l = 1, 3
            virial(1:3,l) = virial(1:3,l) - dij(1:3)*work(l)
          end do

        end if

      end do

    end do
    !$omp end parallel 

    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond_list

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pbc
  !> @brief        calculate nonbonded energy with pairlist (PBC)
  !! @authors      TI, JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    boundary : boundary conditions information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         trans_vec is not used
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pbc(enefunc, molecule, pairlist, boundary, &
                                        coord, force, virial, eelec, evdw)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    type(s_pairlist), target, intent(in)    :: pairlist
    type(s_boundary), target, intent(in)    :: boundary
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: bsize(1:3), bsize_inv(1:3)
    real(wp)                  :: dij(1:3), rij, rij2
    real(wp)                  :: inv_rij, inv_r3, inv_r6, inv_r10, inv_r12
    real(wp)                  :: dr126, lj6, lj10, lj12, CCOEF
    real(wp)                  :: term_elec, shift, dshift, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2, cutoff2_inv
    real(wp)                  :: c1_switch, c2_switch, c4_switch, coef
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown
    real(wp)                  :: c12, c6, term_lj12, term_lj6
    real(wp)                  :: work(1:3)
    integer                   :: i, j, k, l, natom
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, nthread, my_id
    integer                   :: omp_get_num_threads, omp_get_thread_num

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: dielec_const
    real(wp),         pointer :: cutoff, switchdist
    integer,          pointer :: atmcls(:)


    call timer(TimerNonBond, TimerOn)

    charge        => molecule%charge
    atmcls        => molecule%atom_cls_no

    dielec_const  => enefunc%dielec_const
    cutoff        => enefunc%cutoffdist
    switchdist    => enefunc%switchdist

    bsize(1)       = boundary%box_size_x
    bsize(2)       = boundary%box_size_y
    bsize(3)       = boundary%box_size_z
    bsize_inv(1:3) = 1.0_wp/bsize(1:3)

    natom          = molecule%num_atoms
    CCOEF          = ELECOEF/dielec_const
    cutoff2        = cutoff * cutoff
    cutoff2_inv    = 1.0_wp/cutoff2
    switchdist2    = switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup = cutoff2*cutoff2*cutoff2
      c2_switchup = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    ! calculate energy and gradient
    !
    num_nb15 = 0

    !$omp parallel                                                            &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, k, j, dij, rij, rij2,     &
    !$omp         work, c2_switch, c4_switch, switch, dswitch, shift, dshift, &
    !$omp         lj6, lj10, lj12, inv_rij, inv_r3, inv_r6, inv_r10, inv_r12, &
    !$omp         dr126, coef, term_elec, term_lj12, term_lj6)                &
    !$omp firstprivate(num_nb15)                                              &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)           &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
    nthread = omp_get_num_threads()
#else
    id      = 0
    nthread = 1
#endif
    my_id   = my_city_rank * nthread + id
    id      = id + 1
    
    do i = 1, natom - 1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + pairlist%num_nb15_calc(i,id)
      num_nb15 = fin_nb15
    
      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15
        j = pairlist%nb15_calc_list(k,id)

        ! compute distance
        !
        dij(1:3) = coord(1:3,i) - coord(1:3,j)

        ! consider the periodic boundary
        !
        dij(1:3) = dij(1:3) - bsize(1:3) * anint(dij(1:3)*bsize_inv(1:3))
        rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij  = sqrt(rij2)
          lj6  = enefunc%nonb_lj6 (atmcls(i),atmcls(j))
          lj10 = enefunc%nonb_lj10(atmcls(i),atmcls(j))
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))
          inv_rij = 1.0_wp / rij

          ! electrostatic energy 
          !
          shift  = (1.0_wp - (rij2*cutoff2_inv))
          dshift = -4.0_wp*shift*rij*cutoff2_inv
          shift  = shift * shift
          term_elec = charge(i) * charge(j) * CCOEF
          eelec  = eelec + shift * ( term_elec * inv_rij )

          ! lj energy and gradient
          !
          if (enefunc%forcefield == ForcefieldKBGO) then

            if (rij2 > switchdist2) then
              c2_switch = cutoff2 - rij2
              c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
              switch  = c2_switch*c4_switch*c1_switch
              dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
            else
              switch  = 1.0_wp
              dswitch = 0.0_wp
            end if

            rij2    = 1.0_wp / rij2
            inv_rij = 1.0_wp / rij
            inv_r6  = rij2 * rij2 * rij2
            inv_r10 = inv_r6 * rij2 * rij2
            inv_r12 = inv_r6 * inv_r6
            inv_r12 = lj12 * inv_r12
            inv_r10 = lj10 * inv_r10
            inv_r6  = lj6 * inv_r6
            dr126   = inv_r12 - inv_r10 + inv_r6
            evdw    = evdw + switch*dr126
            coef    = (  dswitch * dr126 * rij &
                        - switch * (12.0_wp*inv_r12 - &
                                    10.0_wp*inv_r10 + &
                                    6.0_wp*inv_r6) ) * rij2

          else if (enefunc%force_switch) then

            if (rij2 > switchdist2) then
              rij2      = 1.0_wp / rij2
              inv_rij   = sqrt(rij2)
              inv_r3    = rij2 * inv_rij
              inv_r6    = rij2 * rij2 * rij2
              term_lj12 = lj12 * c12 * (inv_r6-c1_switchup)**2
              term_lj6  = lj6 * c6 * (inv_r3-c2_switchup)**2
              evdw      = evdw + term_lj12 - term_lj6
              coef      = -12.0_wp * lj12 * c12 * (inv_r6-c1_switchup) * &
                                                   inv_r6 * rij2
              coef = coef + 6.0_wp * lj6 * c6 * (inv_r3-c2_switchup) * &
                                                 inv_r3 * rij2
            else
              rij2      = 1.0_wp / rij2
              inv_rij   = sqrt(rij2)
              inv_r3    = rij2 * inv_rij
              inv_r6    = rij2 * rij2 * rij2
              inv_r12   = inv_r6 * inv_r6
              coef      = -12.0_wp * lj12 * inv_r12 * rij2
              coef      = coef + 6.0_wp * lj6 * inv_r6 * rij2
              inv_r12   = inv_r12 - c1_switchdown
              inv_r6    = inv_r6 - c2_switchdown
              term_lj12 = lj12 * inv_r12
              term_lj6  = lj6 * inv_r6
              evdw      = evdw + term_lj12 - term_lj6
            end if

          else

            if (rij2 > switchdist2) then
              c2_switch = cutoff2 - rij2
              c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
              switch  = c2_switch*c4_switch*c1_switch
              dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
            else
              switch  = 1.0_wp
              dswitch = 0.0_wp
            end if

            rij2    = 1.0_wp / rij2
            inv_rij = 1.0_wp / rij
            inv_r6  = rij2 * rij2 * rij2
            inv_r12 = inv_r6 * inv_r6
            inv_r12 = lj12 * inv_r12
            inv_r6  = lj6 * inv_r6
            dr126   = inv_r12 - inv_r6
            evdw    = evdw + switch*dr126
            coef    = (  dswitch * dr126 * rij &
                       - switch * (12.0_wp*inv_r12-6.0_wp*inv_r6) ) * rij2
          end if

          ! electrostatic gradient
          !
          coef  = coef + (dshift*(term_elec*inv_rij) - &
                           shift*(term_elec*rij2))*inv_rij

          work(1:3) = coef*dij(1:3)

          ! store force
          !
          force(1:3,i) = force(1:3,i) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do l = 1, 3
            virial(1:3,l) = virial(1:3,l) - dij(1:3)*work(l)
          end do

        end if  

      end do
    end do
    !$omp end parallel   

    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond_pbc

end module at_energy_nonbonds_mod
