!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_energy_pme_mod
!> @brief   Smooth particle mesh ewald method
!! @authors Takashi Imai (TI), Jaewoon Jung (JJ), Takaharu Mori (TM), 
!!          Chigusa Kobayashi (CK), Yuji Sugita (YS)
!! @note    TI modified for NPT (2010/11/23)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_energy_pme_mod

  use at_mathlib_mod
  use at_energy_table_mod
  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use molecules_str_mod
  use at_boundary_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! variables
  integer, public, parameter :: PmeMaxNspline = 10

  ! subroutines
  public  :: compute_energy_nonbond_pme
  private :: pme_real_nb14
  private :: pme_real
  private :: pme_bond_corr
  private :: pme_pre
  private :: pme_recip_ffte

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pme
  !> @brief        Calculate nonbond energy by PME
  !! @authors      TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    boundary : boundary conditions information
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pme(enefunc, molecule, pairlist, boundary, &
                                        nonb_ene, coord, force, virial,        &
                                        eelec, evdw)

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_molecule),        intent(in)    :: molecule
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(in)    :: coord(:,:)
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: eelec
    real(wp),                intent(inout) :: evdw


    call timer(TimerNonBond, TimerOn)

    ! Calculate PME real part
    !
    if (real_calc) then
      call timer(TimerPmeReal, TimerOn)

      if (enefunc%table%table) then

        ! use lookup table
        !
        call compute_energy_nonbond_table(enefunc, molecule, pairlist,    &
                                          boundary, nonb_ene, coord,      &
                                          force, virial, eelec, evdw)
        call pme_bond_corr_table(enefunc, molecule, coord,                &
                                          force, virial, eelec)

      else

        ! use traditional scheme
        !
        call pme_real_nb14(enefunc, molecule, coord, force, virial,       &
                           eelec, evdw)
        call pme_real(enefunc, molecule, pairlist, boundary, coord,       &
                      force, virial, eelec, evdw)
        call pme_bond_corr(enefunc, molecule, coord, force, virial, eelec)

      end if

      call timer(TimerPmeReal, TimerOff)
    end if

    ! Calculate PME reciprocal part
    !
    if (reciprocal_calc) then
      call timer(TimerPmeRecip, TimerOn)

      call pme_pre(boundary, enefunc)
      call pme_recip_ffte(enefunc, boundary, molecule, &
                          coord, force, virial, eelec)

      call timer(TimerPmeRecip, TimerOff)
    end if

    ! Add self energy
    !
    eelec = eelec + enefunc%pme%u_self

    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond_pme

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_real_nb14
  !> @brief        Calculate PME nonbond 1-4 term real part + Lennard-Jones 
  !! @authors      TI, JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_real_nb14(enefunc, molecule, coord, force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: dij(1:3), rij, rij2, inv_r2, inv_rij
    real(wp)                  :: inv_r3, inv_r6, inv_r10, inv_r12
    real(wp)                  :: lj6, lj10, lj12, el_fact
    real(wp)                  :: term_elec1, term_elec2, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown
    real(wp)                  :: c12, c6, term_lj12, term_lj6
    real(wp)                  :: c1_switch, c2_switch, c4_switch, grad_coef
    real(wp)                  :: work(1:3), alpha, alpha2m, alpha2sp
    integer                   :: i, j, k, l
    integer                   :: num_nb14, ini_nb14, fin_nb14
    integer                   :: id, my_id, nthread
    integer                   :: omp_get_thread_num, omp_get_num_threads

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: cutoff, switchdist
    real(wp),         pointer :: nb14_lj6(:,:), nb14_lj10(:,:), nb14_lj12(:,:)
    integer,          pointer :: natom
    integer,          pointer :: atmcls(:)
    integer,          pointer :: num_nb14_calc(:), nb14_calc_list(:)


    ! use pointers
    !
    natom          => molecule%num_atoms
    charge         => molecule%charge
    atmcls         => molecule%atom_cls_no
    cutoff         => enefunc%cutoffdist
    switchdist     => enefunc%switchdist
    nb14_lj6       => enefunc%nb14_lj6
    nb14_lj10      => enefunc%nb14_lj10
    nb14_lj12      => enefunc%nb14_lj12
    num_nb14_calc  => enefunc%num_nb14_calc
    nb14_calc_list => enefunc%nb14_calc_list
    el_fact        =  ELECOEF/enefunc%dielec_const
    alpha          =  enefunc%pme%alpha
    alpha2m        =  enefunc%pme%alpha2m
    alpha2sp       =  enefunc%pme%alpha2sp


    cutoff2     = cutoff * cutoff
    switchdist2 = switchdist * switchdist
    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup   = cutoff2*cutoff2*cutoff2
      c2_switchup   = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup - switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup - switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    num_nb14 = 0

    ! calculate energy and gradient
    !
    !$omp parallel default(shared)                                           &
    !$omp firstprivate(num_nb14)                                             &
    !$omp private(i, k, j, ini_nb14, fin_nb14, dij, rij2, inv_r2, rij,       &
    !$omp         work, c2_switch, c4_switch, switch, dswitch, lj6,          &
    !$omp         lj10, lj12, inv_rij, inv_r3, inv_r6, inv_r10, inv_r12,     &
    !$omp         grad_coef, term_elec1, term_elec2, term_lj12, term_lj6,    &
    !$omp         id, my_id)                                                 &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)          &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
    nthread = omp_get_num_threads()
#else
    id      = 0
    nthread = 1
#endif
    my_id = my_city_rank * nthread + id

    do i = 1, natom-1

      ini_nb14 = num_nb14 + 1
      fin_nb14 = num_nb14 + num_nb14_calc(i)
      num_nb14 = fin_nb14

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb14 = fin_nb14 + 1

      do k = ini_nb14, fin_nb14

        j = nb14_calc_list(k)

        ! compute distance
        !
        dij(1:3) = coord(1:3,i) - coord(1:3,j)
        rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
        rij  = sqrt(rij2)

        lj6  = nb14_lj6 (atmcls(i),atmcls(j))
        lj10 = nb14_lj10(atmcls(i),atmcls(j))
        lj12 = nb14_lj12(atmcls(i),atmcls(j))

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r3  = inv_r2 * inv_rij
        inv_r6  = inv_r3 * inv_r3

        ! lennard-jones energy and gradient
        !
        if (enefunc%forcefield == ForcefieldKBGO) then

          if (rij2 > switchdist2) then
            c2_switch = cutoff2 - rij2
            c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
            switch    = c2_switch*c4_switch*c1_switch
            dswitch   = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
          else
            switch  = 1.0_wp
            dswitch = 0.0_wp
          end if
          inv_r10 = inv_r6*inv_r2*inv_r2
          inv_r12 = inv_r6*inv_r6
          evdw    = evdw + switch*(lj12*inv_r12 - lj10*inv_r10 + lj6*inv_r6)
          grad_coef = (dswitch*(lj12*inv_r12 - lj10*inv_r10 + lj6*inv_r6)*rij &
                    - switch*(12.0_wp*lj12*inv_r12 - 10.0_wp*lj10*inv_r10     &
                    + 6.0_wp*lj6*inv_r6))*inv_r2

        else if (enefunc%force_switch) then

          if (rij2 > switchdist2) then
            term_lj12 = lj12*c12*(inv_r6 - c1_switchup)**2
            term_lj6  = lj6 *c6 *(inv_r3 - c2_switchup)**2
            evdw      = evdw + term_lj12 - term_lj6
            grad_coef = -12.0_wp*lj12*c12*(inv_r6 - c1_switchup)*inv_r6*inv_r2
            grad_coef = grad_coef &
                      + 6.0_wp*lj6*c6*(inv_r3 - c2_switchup)*inv_r3*inv_r2
          else
            inv_r12   = inv_r6*inv_r6
            grad_coef = -12.0_wp*lj12*inv_r12*inv_r2
            grad_coef = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
            inv_r12   = inv_r12 - c1_switchdown
            inv_r6    = inv_r6  - c2_switchdown
            term_lj12 = lj12*inv_r12
            term_lj6  = lj6 *inv_r6
            evdw      = evdw + term_lj12 - term_lj6
          end if
        else

          if (rij2 > switchdist2) then
            c2_switch = cutoff2 - rij2
            c4_switch = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
            switch    = c2_switch*c4_switch*c1_switch
            dswitch   = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4_switch)
          else
            switch    = 1.0_wp
            dswitch   = 0.0_wp
          end if
          inv_r12 = inv_r6*inv_r6
          evdw    = evdw + switch*(lj12*inv_r12 - lj6*inv_r6)
          grad_coef = (dswitch*(lj12*inv_r12 - lj6*inv_r6)*rij  &
                    - switch*(12.0_wp*lj12*inv_r12 - 6.0_wp*lj6*inv_r6))*inv_r2

        end if

        ! Electrostatic
        !
        term_elec1 = charge(i)*charge(j)*el_fact
        term_elec2 = term_elec1*erfc04(alpha*rij)*inv_rij
        eelec      = eelec + term_elec2
        grad_coef  = grad_coef      &
                   - (term_elec2 + alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2

        work(1:3) = grad_coef*dij(1:3)

        ! store force
        !
        force(1:3,i) = force(1:3,i) - work(1:3)
        force(1:3,j) = force(1:3,j) + work(1:3)

        ! virial
        !
        do l = 1, 3
          virial(1:3,l) = virial(1:3,l) - dij(1:3)*work(l)
        end do

      end do

    end do
    !$omp end parallel

    return

  end subroutine pme_real_nb14

  !======1=========2=========3=========4=========5=========6=========7=========8
  ! 
  !  Subroutine    pme_real
  !> @brief        Calculate PME real part + Lennard-Jones 
  !! @authors      TI, JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    boundary : boundary conditions information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_real(enefunc, molecule, pairlist, boundary, coord, force, &
                      virial, eelec, evdw)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    type(s_pairlist), target, intent(in)    :: pairlist
    type(s_boundary), target, intent(in)    :: boundary
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: rtmp(1:3)
    real(wp)                  :: dij(1:3), rij, rij2, inv_rij
    real(wp)                  :: inv_r2, inv_r3, inv_r6, inv_r10, inv_r12
    real(wp)                  :: lj6, lj10, lj12
    real(wp)                  :: term_lj12, term_lj10, term_lj6, term_lj
    real(wp)                  :: term_elec1, term_elec2, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2, el_fact
    real(wp)                  :: c1_switch, c2_switch, c4, grad_coef
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown
    real(wp)                  :: c12, c6
    real(wp)                  :: work(1:3), alpha, alpha2m, alpha2sp
    real(wp)                  :: vxx, vyx, vzx, vyy, vzy, vzz
    real(wp)                  :: box_size(1:3)
    integer                   :: i, j, k, l
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: omp_get_num_threads, omp_get_thread_num
    integer                   :: id, nthread, my_id

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: cutoff, switchdist
    real(wp),         pointer :: nonb_lj6(:,:), nonb_lj10(:,:), nonb_lj12(:,:)
    integer,          pointer :: natom
    integer,          pointer :: atmcls(:)
    integer,          pointer :: num_nb15_calc(:,:), nb15_calc_list(:,:)

  
    ! use pointers
    natom          => molecule%num_atoms
    charge         => molecule%charge
    atmcls         => molecule%atom_cls_no
    cutoff         => enefunc%cutoffdist
    switchdist     => enefunc%switchdist
    nonb_lj6       => enefunc%nonb_lj6
    nonb_lj10      => enefunc%nonb_lj10
    nonb_lj12      => enefunc%nonb_lj12
    el_fact        =  ELECOEF/enefunc%dielec_const
    alpha          =  enefunc%pme%alpha
    alpha2m        =  enefunc%pme%alpha2m
    alpha2sp       =  enefunc%pme%alpha2sp

    num_nb15_calc  => pairlist%num_nb15_calc
    nb15_calc_list => pairlist%nb15_calc_list
    
    box_size(1)    =  boundary%box_size_x
    box_size(2)    =  boundary%box_size_y
    box_size(3)    =  boundary%box_size_z

    cutoff2     = cutoff * cutoff 
    switchdist2 = switchdist * switchdist
    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup   = cutoff2*cutoff2*cutoff2
      c2_switchup   = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    num_nb15 = 0
    
    ! calculate energy and gradient
    !
    !$omp parallel default(shared)                                              & 
    !$omp firstprivate(num_nb15)                                                &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, k, j, dij,                  &
    !$omp         rij2, rij, inv_r2, work, c2_switch, c4, switch, dswitch,      &
    !$omp         lj6, lj10, lj12, inv_rij, inv_r3, inv_r6, inv_r10, inv_r12,   &
    !$omp         grad_coef, term_elec1, term_elec2, term_lj12, term_lj10,      &
    !$omp         term_lj6, term_lj, rtmp)                                      &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)             &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
    nthread = omp_get_num_threads()
#else
    id      = 0
    nthread = 1
#endif
    my_id   = my_city_rank * nthread + id
    
    do i = 1, natom-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calc(i,id+1)
      num_nb15 = fin_nb15
  
      rtmp(1:3) = coord(1:3,i) 

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_list(k,id+1)

        ! compute distance
        !
        dij = rtmp(1:3) - coord(1:3,j)

        !consider the periodic boundary
        !
        dij(1:3) = dij(1:3) - box_size(1:3)*anint(dij(1:3)/box_size(1:3))
        rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij = sqrt(rij2)

          lj6  = nonb_lj6 (atmcls(i),atmcls(j))
          lj10 = nonb_lj10(atmcls(i),atmcls(j))
          lj12 = nonb_lj12(atmcls(i),atmcls(j))

          inv_r2  = 1.0_wp / rij2
          inv_rij = 1.0_wp / rij
          inv_r3  = inv_r2 * inv_rij
          inv_r6  = inv_r3 * inv_r3

          ! Lennard-Jones energy/gradient
          !
          if (enefunc%forcefield == ForcefieldKBGO) then

            if (rij2 > switchdist2) then
              c2_switch = cutoff2 - rij2
              c4 = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
              switch  = c2_switch*c4*c1_switch
              dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4)
            else
              switch  = 1.0_wp
              dswitch = 0.0_wp
            end if
            inv_r10   = inv_r6*inv_r2* inv_r2
            inv_r12   = inv_r6*inv_r6
            term_lj12 = lj12*inv_r12
            term_lj10 = lj10*inv_r10
            term_lj6  = lj6*inv_r6
            term_lj   = term_lj12 - term_lj10 + term_lj6
            evdw      = evdw + switch * term_lj
            grad_coef = (dswitch*term_lj*rij - switch*(12.0_wp*term_lj12 &
                         - 10.0_wp*term_lj10 + 6.0_wp*term_lj6))*inv_r2

          else if (enefunc%force_switch) then

            if (rij2 > switchdist2) then
              term_lj12 = lj12*c12*(inv_r6 - c1_switchup)**2
              term_lj6  = lj6 *c6 *(inv_r3 - c2_switchup)**2
              evdw      = evdw + term_lj12 - term_lj6
              grad_coef = -12.0_wp*lj12*c12 &
                        * (inv_r6 - c1_switchup)*inv_r6*inv_r2
              grad_coef = grad_coef &
                        + 6.0_wp*lj6*c6*(inv_r3 - c2_switchup)*inv_r3*inv_r2
            else
              inv_r12   = inv_r6*inv_r6
              grad_coef = -12.0_wp*lj12*inv_r12*inv_r2
              grad_coef = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12   = inv_r12 - c1_switchdown
              inv_r6    = inv_r6  - c2_switchdown
              term_lj12 = lj12*inv_r12
              term_lj6  = lj6 *inv_r6
              evdw      = evdw + term_lj12 - term_lj6
            end if

          else

            if (rij2 > switchdist2) then
              c2_switch = cutoff2 - rij2
              c4 = c2_switch*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
              switch  = c2_switch*c4*c1_switch
              dswitch = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch - c4)
            else
              switch  = 1.0_wp
              dswitch = 0.0_wp
            end if
            inv_r12   = inv_r6*inv_r6
            term_lj12 = lj12*inv_r12
            term_lj6  = lj6*inv_r6
            term_lj   = term_lj12 - term_lj6
            evdw      = evdw + switch*term_lj
            grad_coef = (dswitch*term_lj*rij - switch*(12.0_wp*term_lj12 &
                      - 6.0_wp*term_lj6))*inv_r2

          end if

          ! Electrostatic energy and gradient (PME version)
          !
          term_elec1 = charge(i)*charge(j)*el_fact
          term_elec2 = term_elec1*erfc04(alpha*rij)*inv_rij
          eelec      = eelec + term_elec2

          grad_coef  = grad_coef - (term_elec2 &
                     + alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2
          work(1:3)  = grad_coef*dij(1:3)

          ! store force
          !
          force(1:3,i) = force(1:3,i) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do l = 1, 3
            virial(1:3,l) = virial(1:3,l) - dij(1:3)*work(l)
          end do

        end if

      end do

    end do
    !$omp end parallel

    return

  end subroutine pme_real

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr
  !> @brief        Calculate PME bonded energy correction
  !! @authors      TI, JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    r        : coordinates 
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr(enefunc, molecule, r, force, virial, eelec)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: r(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec

    ! local variables
    real(wp)                  :: u, el_fact, alpha, alpha2m, alpha2sp
    real(wp)                  :: rij(3), rij1, rij2, qij, uij, fij(3)
    real(wp)                  :: vxx,vyx,vzx,vyy,vzy,vzz
    integer                   :: i, j, k, jatm
    integer                   :: num_excl,ini_excl,fin_excl
    integer                   :: iatm

    real(wp),         pointer :: q(:), force_wk(:,:)
    integer,          pointer :: natm, num_nonb_excl(:), nonb_excl_list(:)
    integer,          pointer :: i_list(:)

    ! use pointers
    !
    natm           => molecule%num_atoms
    q              => molecule%charge
    num_nonb_excl  => enefunc%num_nonb_excl
    nonb_excl_list => enefunc%nonb_excl_list
    force_wk       => enefunc%pme%force_wk
    i_list         => enefunc%pme%i_list
    el_fact        =  ELECOEF/enefunc%dielec_const
    alpha          =  enefunc%pme%alpha
    alpha2m        =  enefunc%pme%alpha2m
    alpha2sp       =  enefunc%pme%alpha2sp

    ! Initializing the energy and force 
    !
    u             = 0.0_wp
    num_excl      = 0
    i_list(:)     = 0
    force_wk(:,:) = 0.0_wp


    ! Running only the 1-2 and 1-3 connected atoms
    !
    !$omp parallel default (none)                                         &
    !$omp firstprivate(num_excl)                                          &
    !$omp private(i, ini_excl, fin_excl, j, jatm, rij2, k, rij, qij,      &
    !$omp         rij1, uij, fij, vxx, vyx, vzx, vyy, vzy, vzz)           &
    !$omp shared(natm, alpha, alpha2sp, alpha2m, r, q, el_fact,           &
    !$omp        num_nonb_excl, nonb_excl_list, nproc_city, my_city_rank, &
    !$omp        i_list, force_wk) &
    !$omp reduction(+:eelec) reduction(+:virial)
    !
    do i = 1, natm-1

      ini_excl = num_excl + 1
      fin_excl = num_excl + num_nonb_excl(i)
      num_excl = fin_excl

      if (mod(i-1,nproc_city) == my_city_rank) then

        !$omp do
        do j = ini_excl, fin_excl

          jatm      = nonb_excl_list(j)
          i_list(j) = i

          ! Calculating R_ij
          rij2 = 0.0_wp

          do k = 1, 3
            rij(k) = r(k,i) - r(k,jatm)
            rij2 = rij2 + rij(k)**2
          end do

          qij  = q(i) * q(jatm)
          rij1 = sqrt(rij2)

          ! Calculating U_ij and adding it to U
          !
          uij = qij * erf_n(alpha * rij1) / rij1
          eelec = eelec - uij * el_fact

          ! Calculating F_ij and adding it to F_i (symmetry considered)
          !
          do k = 1, 3
            fij(k) = (uij - alpha2sp*qij*exp(alpha2m*rij2))*rij(k)/rij2
            force_wk(k,j) = fij(k)*el_fact
          end do

          ! virial
          !
          vxx = rij(1) * fij(1) * el_fact
          vyx = rij(2) * fij(1) * el_fact
          vzx = rij(3) * fij(1) * el_fact
          vyy = rij(2) * fij(2) * el_fact
          vzy = rij(3) * fij(2) * el_fact
          vzz = rij(3) * fij(3) * el_fact
          virial(1,1) = virial(1,1) - vxx
          virial(2,1) = virial(2,1) - vyx
          virial(3,1) = virial(3,1) - vzx
          virial(1,2) = virial(1,2) - vyx
          virial(2,2) = virial(2,2) - vyy
          virial(3,2) = virial(3,2) - vzy
          virial(1,3) = virial(1,3) - vzx
          virial(2,3) = virial(2,3) - vzy
          virial(3,3) = virial(3,3) - vzz

        end do
        !$omp end do

      end if

    end do
    !$omp end parallel


    do j = 1, size(enefunc%nonb_excl_list(:))

      iatm = i_list(j)
      jatm = nonb_excl_list(j)

      if (iatm > 0) then
        do k = 1, 3
          force(k,iatm) = force(k,iatm) - force_wk(k,j)
          force(k,jatm) = force(k,jatm) + force_wk(k,j)
        end do
      end if

    end do

    return

  end subroutine pme_bond_corr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_pre
  !> @brief        Prepare functions for PME calculation
  !! @authors      TI, JJ, TM
  !! @param[in]    boundary : boundary conditions information
  !! @note         Extracted from setup_pme for NPT calculation
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_pre(boundary, enefunc)

    ! formal arguments
    type(s_boundary),         intent(in)    :: boundary
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: fact, gfact(3), g2, alpha2m
    integer                  :: i, j, k, is, js, ks
    integer                  :: ngrid(3)
    real(wp), pointer        :: gx(:), gy(:), gz(:), b2(:,:)
    real(wp), pointer        :: vir_fact(:,:,:), theta(:,:,:)

    ! use pointers
    !
    gx         => enefunc%pme%gx
    gy         => enefunc%pme%gy
    gz         => enefunc%pme%gz
    vir_fact   => enefunc%pme%vir_fact
    theta      => enefunc%pme%theta
    b2         => enefunc%pme%b2
    ngrid(1:3) =  enefunc%pme%ngrid(1:3)
    alpha2m    =  enefunc%pme%alpha2m

    ! Prepareing theta = F^-1[theta](h), h shifted
    ! Gx, Gy, Gz, and vir_fact=-2(1+G^2/4a)/G^2 are also prepared 
    ! for virial calculation
    !
    gfact(1) = 2.0_wp*PI/boundary%box_size_x
    gfact(2) = 2.0_wp*PI/boundary%box_size_y
    gfact(3) = 2.0_wp*PI/boundary%box_size_z

    fact = 0.25_wp / alpha2m

    !$omp parallel do default(none)                                   &
    !$omp private(i, j, k, g2, ks, is, js)                            &
    !$omp shared(ngrid, gfact, gx, gy, gz, vir_fact, theta, fact, b2)
    !
    do k = 1, ngrid(3)/2 + 1
      ks    = k - 1
      gz(k) = gfact(3) * real(ks,wp)
      do j = 1, ngrid(2)/2 + 1
        js    = j - 1
        gy(j) = gfact(2) * real(js,wp)
        do i =1, ngrid(1)/2 + 1
          is    = i - 1
          gx(i) = gfact(1) * real(is,wp)
          g2    = gx(i)**2 + gy(j)**2 + gz(k)**2
          vir_fact(i,j,k) = -2.0_wp * (1.0_wp - g2 * fact)/g2
          theta(i,j,k)    = b2(i,1) * b2(j,2) * b2(k,3) * exp(g2 * fact)/g2
        end do
      end do

      do j = ngrid(2)/2 + 2, ngrid(2)
        js    = j - 1 - ngrid(2)
        gy(j) = gfact(2) * real(js,wp)
        do i = 1, ngrid(1)/2 + 1
          is    = i - 1
          gx(i) = gfact(1) * real(is,wp)
          g2    = gx(i)**2 + gy(j)**2 + gz(k)**2
          vir_fact(i,j,k) = -2.0_wp * (1.0_wp - g2 * fact)/g2
          theta(i,j,k)    = b2(i,1) * b2(j,2) * b2(k,3) * exp(g2 * fact)/g2
        end do
      end do
    end do
    !$omp end parallel do

    !$omp parallel do default(none)                                   &
    !$omp private(i, j, k, g2, ks, is, js)                            &
    !$omp shared(ngrid, gfact, gx, gy, gz, vir_fact, theta, fact, b2)
    !
    do k = ngrid(3)/2 + 2, ngrid(3)
      ks    = k - 1 - ngrid(3)
      gz(k) = gfact(3) * real(ks,wp)
      do j = 1, ngrid(2)/2 + 1
        js    = j - 1
        gy(j) = gfact(2) * real(js,wp)
        do i = 1, ngrid(1)/2 + 1
          is    = i - 1
          gx(i) = gfact(1) * real(is,wp)
          g2    = gx(i)**2 + gy(j)**2 + gz(k)**2
          vir_fact(i,j,k) = -2.0_wp * (1.0_wp - g2 * fact)/g2
          theta(i,j,k)    = b2(i,1) * b2(j,2) * b2(k,3) * exp(g2 * fact)/g2
        end do
      end do

      do j = ngrid(2)/2 + 2, ngrid(2)
        js    = j - 1 - ngrid(2)
        gy(j) = gfact(2) * real(js,wp)
        do i = 1, ngrid(1)/2 + 1
          is    = i - 1
          gx(i) = gfact(1) * real(is,wp)
          g2    = gx(i)**2 + gy(j)**2 + gz(k)**2
          vir_fact(i,j,k) = -2.0_wp * (1.0_wp - g2 * fact)/g2
          theta(i,j,k)    = b2(i,1) * b2(j,2) * b2(k,3) * exp(g2 * fact)/g2
        end do
      end do
    end do
    !$omp end parallel do

    do k = 1, ngrid(3)
      do j = 1, ngrid(2)
        do i = ngrid(1)/2+2, ngrid(1)
          theta(i,j,k)    = theta(ngrid(1)-i+2,j,k)
          vir_fact(i,j,k) = vir_fact(ngrid(1)-i+2,j,k)
        end do
      end do
    end do

    do i = ngrid(1)/2 + 2, ngrid(1)
      gx(i) = gx(ngrid(1)-i+2)
    end do

    vir_fact(1,1,1) = 0.0_wp
    theta(1,1,1)    = 0.0_wp

    return

  end subroutine pme_pre

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_recip_ffte
  !> @brief        Calculate PME reciprocal part
  !! @authors      TI, JJ, CK, TM
  !! @param[in]    molecule : molecule information
  !! @param[in]    r        : coordinates 
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_recip_ffte(enefunc, boundary, molecule, r, force, virial, eelec)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_boundary),         intent(in)    :: boundary
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: r(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec

    ! local variables
    real(wp)                  :: u, el_fact, origin(3), box_inv(3), r_scale(3)
    real(wp)                  :: dv(3), vol_fact2, vol_fact4
    real(wp)                  :: vxx,vyx,vzx,vyy,vzy,vzz
    real(wp)                  :: tqq
    real(wp)                  :: bsc(PmeMaxNspline,3), bscd(PmeMaxNspline,3)
    real(wp)                  :: bs_fact3, bs_fact3d
    integer                   :: i, k
    integer                   :: ix, iy, iz, ixs, iys, izs, ixx, iyy, izz, ii(3)
    integer                   :: iproc, jproc
    integer                   :: istart_atom, iend_atom
    integer                   :: ngrid(3), n_bs

    real(wp),         pointer :: q(:), f(:,:), v(:,:), gx(:), gy(:), gz(:)
    real(wp),         pointer :: vir_fact(:,:,:), theta(:,:,:)
    real(wp),         pointer :: qdf(:,:,:), qdf_recv(:,:,:,:), qdf_send(:,:,:,:)
    integer,          pointer :: natm
    integer,          pointer :: x_start1(:), x_end1(:)
    integer,          pointer :: y_start(:), y_end(:)
    integer,          pointer :: y_local(:), y_start1(:), y_end1(:)
    integer,          pointer :: z_start(:), z_end(:), z_local(:)
    complex(wp),      pointer :: ftqdf_localA(:,:,:), ftqdf_localB(:,:,:)

    ! use pointers
    !
    natm         => molecule%num_atoms
    q            => molecule%charge
    f            => enefunc%pme%f
    v            => enefunc%pme%v
    gx           => enefunc%pme%gx
    gy           => enefunc%pme%gy
    gz           => enefunc%pme%gz
    qdf          => enefunc%pme%qdf
    x_start1     => enefunc%pme%x_start1
    x_end1       => enefunc%pme%x_end1
    y_start      => enefunc%pme%y_start
    y_end        => enefunc%pme%y_end
    y_local      => enefunc%pme%y_local
    y_start1     => enefunc%pme%y_start1
    y_end1       => enefunc%pme%y_end1
    z_start      => enefunc%pme%z_start
    z_end        => enefunc%pme%z_end
    z_local      => enefunc%pme%z_local
    qdf_recv     => enefunc%pme%qdf_recv
    qdf_send     => enefunc%pme%qdf_send
    ftqdf_localA => enefunc%pme%ftqdf_localA
    ftqdf_localB => enefunc%pme%ftqdf_localB
    theta        => enefunc%pme%theta
    vir_fact     => enefunc%pme%vir_fact

    istart_atom  =  enefunc%pme%istart_atom
    iend_atom    =  enefunc%pme%iend_atom
    ngrid(1:3)   =  enefunc%pme%ngrid(1:3)
    bs_fact3     =  enefunc%pme%bs_fact3
    bs_fact3d    =  enefunc%pme%bs_fact3d
    n_bs         =  enefunc%pme%n_bs
    el_fact      =  ELECOEF/enefunc%dielec_const

    origin(1)    = boundary%origin_x
    origin(2)    = boundary%origin_y
    origin(3)    = boundary%origin_z
    box_inv(1)   = 1.0_wp/boundary%box_size_x
    box_inv(2)   = 1.0_wp/boundary%box_size_y
    box_inv(3)   = 1.0_wp/boundary%box_size_z
    r_scale(1)   = real(ngrid(1),wp)*box_inv(1)
    r_scale(2)   = real(ngrid(2),wp)*box_inv(2)
    r_scale(3)   = real(ngrid(3),wp)*box_inv(3)

    vol_fact2    = 2.0_wp * PI * el_fact * box_inv(1)*box_inv(2)*box_inv(3) 
    vol_fact4    = 2.0_wp * vol_fact2


    ! Initializing the energy and force
    !
    u = 0.0_wp
    f(1:3,1:natm) = 0.0_wp

    ! Translation to scaled fractional coordinates (v(x,y,z))
    !
    do i = istart_atom, iend_atom
      do k = 1, 3
        v(k,i) = (r(k,i) - origin(k)) * r_scale(k)
        v(k,i) = v(k,i) + real(ngrid(k) / 2,wp) &
                - real(ngrid(k),wp) * anint(v(k,i) / real(ngrid(k),wp))
      end do
    end do

    ! Initialization of Q-fanction
    !
    do iz = 1, ngrid(3)
      do iy = 1, ngrid(2)
        do ix = 1, ngrid(1)
          qdf(ix,iy,iz) = 0.0_wp
        end do
      end do
    end do

    ! Calculateing Q-fanction
    !
    do i = istart_atom, iend_atom

      do k = 1, 3
        ii(k) = int(v(k,i))
        dv(k) = v(k,i) - real(ii(k),wp)
        call b_spline_coef(n_bs, dv(k), bsc(1:n_bs,k))
      end do

      do iz = 1, n_bs

        izz = ii(3) - iz + 1
        if (izz < 0) then
          izs = izz + ngrid(3) + 1
        else if (izz == ngrid(3)) then
          izs = izz
        else
          izs = izz + 1
        end if

        do iy = 1, n_bs

          iyy = ii(2) - iy + 1
          if (iyy < 0) then
            iys = iyy + ngrid(2) + 1
          else if (iyy == ngrid(2)) then
            iys = iyy
          else
            iys = iyy + 1
          end if

          do ix = 1, n_bs

            ixx = ii(1) - ix + 1
            if (ixx < 0) then
              ixs = ixx + ngrid(1) + 1
            else if (ixx == ngrid(1)) then
              ixs = ixx
            else
              ixs = ixx + 1
            end if
            qdf(ixs,iys,izs) = qdf(ixs,iys,izs) &
                             + q(i) * bsc(ix,1)*bsc(iy,2)*bsc(iz,3)*bs_fact3
          end do
        end do
      end do
    end do

    ! Q-fanction is saved to qdf_send for communication
    !
    do jproc = 0, my_city_rank - 1
      do iz = z_start(jproc), z_end(jproc)
        izs = iz - z_start(jproc) + 1
        do iy = y_start(jproc), y_end(jproc)
          iys = iy - y_start(jproc) + 1
          do ix = 1, ngrid(1)
            qdf_send(ix,iys,izs,jproc+1) = qdf(ix,iy,iz)
          end do
        end do
      end do
    end do

    do jproc = my_city_rank + 1, nproc_city - 1
      do iz = z_start(jproc), z_end(jproc)
        izs = iz - z_start(jproc) + 1
        do iy = y_start(jproc), y_end(jproc)
          iys = iy - y_start(jproc) + 1
          do ix = 1, ngrid(1)
            qdf_send(ix,iys,izs,jproc+1) = qdf(ix,iy,iz)
          end do
        end do
      end do
    end do


    do iz = z_start(my_city_rank), z_end(my_city_rank)
      izs = iz - z_start(my_city_rank) + 1
      do iy = y_start(my_city_rank), y_end(my_city_rank)
        iys = iy - y_start(my_city_rank) + 1
        do ix = 1, ngrid(1)
          ftqdf_localA(ix,iys,izs) = qdf(ix,iy,iz)
        end do
      end do
    end do

#ifdef MPI
    call mpi_alltoall &
         (qdf_send, ngrid(1)*y_local(0)*z_local(0), mpi_wp_real, &
          qdf_recv, ngrid(1)*y_local(0)*z_local(0), mpi_wp_real, &
          mpi_comm_city, ierror)
#endif

    ! Q-fanction is accumulated to qdf_send for each processor
    !
    do iproc = 0, my_city_rank - 1
      do iz = 1, z_local(my_city_rank)
        do iy = 1, y_local(my_city_rank)
          do ix = 1, ngrid(1)
            ftqdf_localA(ix,iy,iz) = ftqdf_localA(ix,iy,iz) + &
                 qdf_recv(ix,iy,iz,iproc+1)
          end do
        end do
      end do
    end do

    do iproc = my_city_rank + 1, nproc_city - 1
      do iz = 1, z_local(my_city_rank)
        do iy = 1, y_local(my_city_rank)
          do ix = 1, ngrid(1)
            ftqdf_localA(ix,iy,iz) = ftqdf_localA(ix,iy,iz) + &
                 qdf_recv(ix,iy,iz,iproc+1)
          end do
        end do
      end do
    end do

    ! Q -> F^-1[Q]
    !
    call pzfft3dv(ftqdf_localA, ftqdf_localB, ngrid(1), ngrid(2), ngrid(3), &
                  grid_commy, grid_commz, nprocy, nprocz, 0)
    call pzfft3dv(ftqdf_localA, ftqdf_localB, ngrid(1), ngrid(2), ngrid(3), &
                  grid_commy, grid_commz, nprocy, nprocz, -2)

    ! Calculating U = Sum{Th*(F(Q)*F-1(Q))} and Virial
    !
    do iz = 1, ngrid(3)
      do iy = y_start1(my_city_rank), y_end1(my_city_rank)

        iys = iy - y_start1(my_city_rank) + 1

        do ix = x_start1(my_city_rank), x_end1(my_city_rank)
          ixs = ix - x_start1(my_city_rank) + 1
          tqq = real(ftqdf_localB(ixs,iys,iz),wp)**2 + &
                imag(ftqdf_localB(ixs,iys,iz))**2
          tqq = tqq * theta(ix,iy,iz) * vol_fact2
          u   = u + tqq

          ! virial
          !
          vxx = tqq * (1.0_wp + vir_fact(ix,iy,iz) * gx(ix) * gx(ix))
          vyx = tqq * vir_fact(ix,iy,iz) * gy(iy) * gx(ix)
          vzx = tqq * vir_fact(ix,iy,iz) * gz(iz) * gx(ix)
          vyy = tqq * (1.0_wp + vir_fact(ix,iy,iz) * gy(iy) * gy(iy))
          vzy = tqq * vir_fact(ix,iy,iz) * gz(iz) * gy(iy)
          vzz = tqq * (1.0_wp + vir_fact(ix,iy,iz) * gz(iz) * gz(iz))

          virial(1,1) = virial(1,1) + vxx
          virial(2,1) = virial(2,1) + vyx
          virial(3,1) = virial(3,1) + vzx
          virial(1,2) = virial(1,2) + vyx
          virial(2,2) = virial(2,2) + vyy
          virial(3,2) = virial(3,2) + vzy
          virial(1,3) = virial(1,3) + vzx
          virial(2,3) = virial(2,3) + vzy
          virial(3,3) = virial(3,3) + vzz
        end do

      end do
    end do

    ! F^-1[Th]*F^-1[Q] (=X)
    !
    do iz = 1, ngrid(3)
      do iy = y_start1(my_city_rank), y_end1(my_city_rank)
        iys = iy - y_start1(my_city_rank) + 1
        do ix = x_start1(my_city_rank), x_end1(my_city_rank)
          ixs = ix - x_start1(my_city_rank) + 1
          ftqdf_localB(ixs,iys,iz) = cmplx(theta(ix,iy,iz)) * &
               ftqdf_localB(ixs,iys,iz)
        end do
      end do
    end do

    ! X -> F[X] (=Conv[Th*Q])
    !
    call pzfft3dv(ftqdf_localB, ftqdf_localA, ngrid(1), ngrid(2), ngrid(3), &
                  grid_commy, grid_commz, nprocy, nprocz, 2)

    do iz = 1, z_local(my_city_rank)
      do iy = 1, y_local(my_city_rank)
        do ix = 1, ngrid(1)
          qdf_send(ix,iy,iz,my_city_rank+1) = real(ftqdf_localA(ix,iy,iz),wp)
        end do
      end do
    end do

    ! X is saved on qdf_send and distributed
    !
#ifdef MPI
    call mpi_allgather(qdf_send(1,1,1,my_city_rank+1), &
                       z_local(0)*y_local(0)*ngrid(1), &
                       mpi_wp_real, qdf_recv,          &
                       z_local(0)*y_local(0)*ngrid(1), &
                       mpi_wp_real, mpi_comm_city, ierror)
#else
    do iz = 1, z_local(my_city_rank)
      do iy = 1, y_local(my_city_rank)
        do ix = 1, ngrid(1)
          qdf_recv(ix,iy,iz,my_city_rank+1) = &
               qdf_send(ix,iy,iz,my_city_rank+1)
        end do
      end do
    end do
#endif

    ! X is saved on qdf
    !
    do iproc = 0, nproc_city - 1
      do iz = z_start(iproc), z_end(iproc)
        izs = iz - z_start(iproc) + 1
        do iy = y_start(iproc), y_end(iproc)
          iys = iy - y_start(iproc) + 1
          do ix = 1, ngrid(1)
            qdf(ix,iy,iz) = qdf_recv(ix,iys,izs,iproc+1)
          end do
        end do
      end do
    end do

    ! Calculating Fi=Sum{dQ/dri*Conv[Th*Q]}
    !
    !$omp parallel do default(none)                           &
    !$omp private(i, k, ii, dv, bsc, bscd,                    & 
    !$omp         iz, izz, izs, iy, iyy, iys, ix, ixx, ixs)   &
    !$omp shared(istart_atom, iend_atom, n_bs, v, f, q, qdf,  &
    !$omp        ngrid, r_scale, vol_fact4, bs_fact3d)
    !
    do i = istart_atom, iend_atom

      do k = 1, 3
        ii(k) = int(v(k,i))
        dv(k) = v(k,i) - real(ii(k),wp)
        call b_spline_dev_coef(n_bs, dv(k), bsc(1:n_bs,k), bscd(1:n_bs,k))
      end do

      do iz = 1, n_bs

        izz = ii(3) - iz + 1
        if (izz < 0) then
          izs = izz + ngrid(3) + 1
        else if (izz == ngrid(3)) then
          izs = izz
        else
          izs = izz + 1
        end if

        do iy = 1, n_bs

          iyy = ii(2) - iy + 1
          if (iyy < 0) then
            iys = iyy + ngrid(2) + 1
          else if (iyy == ngrid(2)) then
            iys = iyy
          else
            iys = iyy + 1
          end if

          do ix = 1, n_bs

            ixx = ii(1) - ix + 1
            if (ixx < 0) then
              ixs = ixx + ngrid(1) + 1
            else if (ixx == ngrid(1)) then
              ixs = ixx
            else
              ixs = ixx + 1
            end if

            f(1,i) = f(1,i) + bscd(ix,1)*bsc (iy,2)*bsc (iz,3)*qdf(ixs,iys,izs)
            f(2,i) = f(2,i) + bsc (ix,1)*bscd(iy,2)*bsc (iz,3)*qdf(ixs,iys,izs)
            f(3,i) = f(3,i) + bsc (ix,1)*bsc (iy,2)*bscd(iz,3)*qdf(ixs,iys,izs)

          end do
        end do
      end do

      do k = 1, 3
        f(k,i) = -f(k,i) * q(i) * r_scale(k) * vol_fact4 * bs_fact3d
      end do

    end do
    !$omp end parallel do

    eelec = eelec + u
    force(1:3,1:natm) = force(1:3,1:natm) + f(1:3,1:natm)

    return

  end subroutine pme_recip_ffte

end module at_energy_pme_mod
