!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_setup_mpi_mod
!> @brief   setup MPI
!! @authors Jaewoon Jung (JJ), Takaharu Mori (TM)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_setup_mpi_mod

  use at_remd_mod
  use at_energy_mod
  use at_energy_str_mod
  use fileio_control_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: setup_mpi_md
  public  :: setup_mpi_remd

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mpi_md
  !> @brief        setup mpi in MD
  !> @authors      JJ, TM
  !! @param[in]    ene_info : ENERGY section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mpi_md(ene_info)
  
    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info

    ! local variables
    real(wp)                 :: ratio, recip
    integer                  :: i, alloc_stat, dealloc_stat
    integer                  :: omp_get_num_threads
    integer                  :: color, key, nthread
    integer                  :: node_recip, node_real, expon

    integer,     allocatable :: globrank(:), locrank(:)
    logical,     allocatable :: nonbreal(:), nonbrecip(:)


#ifdef MPI
      ! Equalize communicators (comm_country & comm_city) between MD and REMD
      !
      mpi_comm_country = mpi_comm_world
      nproc_country    = nproc_world
      my_country_rank  = my_world_rank
#endif

    select case (ene_info%electrostatic)

    case (ElectrostaticPME)

#ifdef MPI

      if (ene_info%pme_multiple) then

        if (nproc_country == 1) &
          call error_msg('Setup_Mpi_Md> "PME_multiple" must be "NO", '//&
                         'if only 1 MPI is used for one replica.')

        ratio = ene_info%pme_mul_ratio
        recip = real(nproc_country,wp) / (1.0_wp + ratio)
        expon = int(log(recip)/log(real(2,wp)))
        node_recip = 2**expon
        node_real  = nproc_country - node_recip

        if (my_country_rank < node_real) then
          real_calc = .true.
          reciprocal_calc = .false.
          color = 1
          key   = my_country_rank
        else
          real_calc = .false.
          reciprocal_calc = .true.
          color = 2
          key   = my_country_rank
        end if

        call mpi_comm_split(mpi_comm_country, color, key, mpi_comm_city, ierror)
        call mpi_comm_size (mpi_comm_city, nproc_city,   ierror)
        call mpi_comm_rank (mpi_comm_city, my_city_rank, ierror)

      else

        real_calc = .true.
        reciprocal_calc = .true.
        mpi_comm_city = mpi_comm_country
        nproc_city    = nproc_country
        my_city_rank  = my_country_rank

      end if
#else
        real_calc = .true.
        reciprocal_calc = .true.
        nproc_city   = nproc_country
        my_city_rank = my_country_rank
#endif

    case (ElectrostaticCUTOFF)

#ifdef MPI
      real_calc = .true.
      reciprocal_calc = .false.
      mpi_comm_city = mpi_comm_country
      nproc_city    = nproc_country
      my_city_rank  = my_country_rank
#else
      real_calc = .true.
      reciprocal_calc = .false.
      nproc_city   = nproc_country
      my_city_rank = my_country_rank
#endif

    end select

#ifdef MPI
    ! Write the summary of setup MP
    !
    allocate(globrank (nproc_world), &
             locrank  (nproc_world), &
             nonbreal (nproc_world), &
             nonbrecip(nproc_world), stat=alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    call mpi_gather(my_world_rank,   1, mpi_integer,                     &
                    globrank,  1, mpi_integer, 0, mpi_comm_world, ierror)
    call mpi_gather(my_city_rank, 1, mpi_integer,                        &
                    locrank,   1, mpi_integer, 0, mpi_comm_world, ierror)
    call mpi_gather(real_calc, 1, mpi_logical,                           &
                    nonbreal,  1, mpi_logical, 0, mpi_comm_world, ierror)
    call mpi_gather(reciprocal_calc, 1, mpi_logical,                     &
                    nonbrecip, 1, mpi_logical, 0, mpi_comm_world, ierror)

    if (main_rank) then

      write(MsgOut,'(a)') 'Setup_Mpi_Md> Summary of Setup MPI'
      write(MsgOut,'(a,i10)') '  number of MPI processes   = ', &
                              nproc_world
#ifdef OMP

      !$omp parallel shared(nthread)
      nthread = omp_get_num_threads()
      !$omp end parallel
      write(MsgOut,'(a,i10)') '  number of OpenMP threads  = ', &
                              nthread
#else

      nthread = 1

#endif

      write(MsgOut,'(a,i10)') '  total number of CPU cores = ', &
                              nproc_world*nthread
      write(MsgOut,'(a)') ''

    end if


    ! deallocate local array
    !
    deallocate(globrank, locrank, nonbreal, nonbrecip, stat=dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

#else

#ifdef OMP

    !$omp parallel shared(nthread)
    nthread = omp_get_num_threads()
    !$omp end parallel
    write(MsgOut,'(a,i10)') '  number of OpenMP threads  = ', &
                            nthread
#else

    nthread = 1

#endif

    write(MsgOut,'(a)') 'Setup_Mpi_Md> Single CPU is used.'
    write(MsgOut,'(a)') ''

#endif

    return

  end subroutine setup_mpi_md

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mpi_remd
  !> @brief        setup mpi in REMD
  !> @authors      TM, JJ
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[in]    rep_info : REMD section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mpi_remd(ene_info, rep_info)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_rep_info),        intent(in)    :: rep_info

    ! local variables
    real(wp)                 :: ratio, recip
    integer                  :: i, alloc_stat, dealloc_stat
    integer                  :: omp_get_num_threads, nthread
    integer                  :: nreplicas, nproc_per_rep
    integer                  :: color, key, temp
    integer                  :: node_recip, node_real, expon
    character(6)             :: mast, nonb1, nonb2

    integer,     allocatable :: globrank(:), repno(:), reprank(:)
    logical,     allocatable :: repmast(:), nonbreal(:), nonbrecip(:)


#ifdef MPI

    nreplicas  = product(rep_info%nreplicas(1:rep_info%dimension))

    ! Check relationship between nreplicas and num_procs
    !
    if (nreplicas <= nproc_world) then
      if (mod(nproc_world,nreplicas) /= 0)  &
        call error_msg('Setup_Mpi_Remd> mod(total_MPI_procs,nreplicas) is not ZERO')
      nproc_per_rep = int(nproc_world/nreplicas)
    else
      if (mod(nreplicas,nproc_world) /= 0)  &
        call error_msg('Setup_Mpi_Remd> mod(nreplicas,total_MPI_procs) is not ZERO')
      nproc_per_rep = 1
    end if


    ! Communicator within country
    !
    my_country_no = int(my_world_rank / nproc_per_rep)

    call mpi_comm_split(mpi_comm_world, my_country_no, my_world_rank, &
                        mpi_comm_country, ierror)
    call mpi_comm_size (mpi_comm_country, nproc_country,   ierror)
    call mpi_comm_rank (mpi_comm_country, my_country_rank, ierror)

    if (my_country_rank == 0) then
      replica_main_rank = .true.
    else
      replica_main_rank = .false.
    end if


    ! Communicator between replicas (mpi_comm_airplane)
    !
    temp  = my_country_no
    color = my_country_rank
    key   = temp

    call mpi_comm_split(mpi_comm_world, color, key, mpi_comm_airplane, ierror)
    call mpi_comm_size (mpi_comm_airplane, nproc_airplane,   ierror)
    call mpi_comm_rank (mpi_comm_airplane, my_airplane_rank, ierror)


    ! Equalize communicator (mpi_comm_city) between MD and REMD
    !
    select case (ene_info%electrostatic)

    case (ElectrostaticPME)

      if (ene_info%pme_multiple) then

        if (nproc_country == 1) &
          call error_msg('Setup_Mpi_Remd> "PME_multiple" must be "NO", '//&
                         'if only 1 MPI is used for one replica.')

        ratio = ene_info%pme_mul_ratio
        recip = real(nproc_country,wp) / (1.0_wp + ratio)
        expon = int(log(recip)/log(real(2,wp)))
        node_recip = 2**expon
        node_real  = nproc_country - node_recip

        if (my_country_rank < node_real) then
          real_calc = .true.
          reciprocal_calc = .false.
          color = 1
          key   = my_country_rank
        else
          real_calc = .false.
          reciprocal_calc = .true.
          color = 2
          key   = my_country_rank
        end if

        call mpi_comm_split(mpi_comm_country, color, key, mpi_comm_city, ierror)
        call mpi_comm_size (mpi_comm_city, nproc_city,   ierror)
        call mpi_comm_rank (mpi_comm_city, my_city_rank, ierror)

      else

        real_calc = .true.
        reciprocal_calc = .true.
        mpi_comm_city = mpi_comm_country
        nproc_city    = nproc_country
        my_city_rank  = my_country_rank

      end if

    case (ElectrostaticCUTOFF)

      real_calc = .true.
      reciprocal_calc = .false.
      mpi_comm_city = mpi_comm_country
      nproc_city    = nproc_country
      my_city_rank  = my_country_rank

    end select


    ! Write the summary of setup MPI
    !
    allocate(globrank (nproc_world), &
             repno    (nproc_world), &
             reprank  (nproc_world), &
             repmast  (nproc_world), &
             nonbreal (nproc_world), &
             nonbrecip(nproc_world), &
             stat=alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    call mpi_gather(my_world_rank, 1, mpi_integer,                      &
                    globrank, 1, mpi_integer, 0, mpi_comm_world, ierror)
    call mpi_gather(my_country_no, 1, mpi_integer,                      &
                    repno, 1, mpi_integer, 0, mpi_comm_world, ierror)
    call mpi_gather(my_city_rank, 1, mpi_integer,                       &
                    reprank, 1, mpi_integer, 0, mpi_comm_world, ierror)
    call mpi_gather(replica_main_rank, 1, mpi_logical,                  &
                    repmast, 1, mpi_logical, 0, mpi_comm_world, ierror)
    call mpi_gather(real_calc, 1, mpi_logical,                          &
                    nonbreal, 1, mpi_logical, 0, mpi_comm_world, ierror)
    call mpi_gather(reciprocal_calc, 1, mpi_logical,                    &
                    nonbrecip, 1, mpi_logical, 0, mpi_comm_world, ierror)

    if (main_rank) then

      write(MsgOut,'(a)') 'Setup_Mpi_Remd> Summary of Setup MPI'
      write(MsgOut,'(a,i10)') '  number of MPI processes                = ', &
                              nproc_world
      write(MsgOut,'(a,i10)') '  number of MPI processes in one replica = ', &
                              nproc_country
#ifdef OMP

      !$omp parallel shared(nthread)
      nthread = omp_get_num_threads()
      !$omp end parallel
      write(MsgOut,'(a,i10)') '  number of OpenMP threads               = ', &
                              nthread
#else

      nthread = 1

#endif
      write(MsgOut,'(a,i10)') '  total number of CPU cores              = ', &
                              nproc_world*nthread
      write(MsgOut,'(a)') '     world_rank     country_no   country_rank'

      do i = 1, nproc_world

        mast  = '      '
        nonb1 = '      '
        nonb2 = '      '

        if (repmast(i))   mast  = 'MASTER'
        if (nonbreal(i))  nonb1 = 'REAL  '
        if (nonbrecip(i)) nonb2 = 'RECP  '

        write(MsgOut,'(3I15,2X,3A6)') &
              globrank(i), repno(i), reprank(i), NONB1, NONB2, mast
      end do

      write(MsgOut,'(a)') ''

    end if


    ! deallocate local array
    !
    deallocate(globrank, repno, reprank, repmast,      &
               nonbreal, nonbrecip, stat=dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

#else

    replica_main_rank = .true.

    select case (ene_info%electrostatic)

    case (ElectrostaticPME)

      real_calc = .true.
      reciprocal_calc = .true.
      nproc_city   = nproc_country
      my_city_rank = my_country_rank

    case (ElectrostaticCUTOFF)

      real_calc = .true.
      reciprocal_calc = .false.
      nproc_city   = nproc_country
      my_city_rank = my_country_rank
    
    end select

#ifdef OMP

    !$omp parallel shared(nthread)
    nthread = omp_get_num_threads()
    !$omp end parallel
    write(MsgOut,'(a,i10)') '  number of OpenMP threads  = ', &
                            nthread
#else

    nthread = 1

#endif

    write(MsgOut,'(a)') 'Setup_Mpi_Remd> Single CPU is used.'
    write(MsgOut,'(a)') ''

#endif

    return

  end subroutine setup_mpi_remd

end module at_setup_mpi_mod
