!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fileio_rst
!> @brief   Restart file I/O module
!! @authors Takaharu Mori (TM), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module fileio_rst_mod

  use fileio_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_rst
    integer                       :: rstfile_type    = 0
    integer                       :: num_atoms       = 0
    integer                       :: num_deg_freedom = 0
    integer                       :: iseed           = 0
    real(dp),         allocatable :: coord(:,:)
    real(dp),         allocatable :: velocity(:,:)
    real(dp)                      :: thermostat_momentum  = 0.0_wp
    real(dp)                      :: barostat_momentum(3) = 0.0_wp
    real(dp)                      :: box_size_x      = 0.0_wp
    real(dp)                      :: box_size_y      = 0.0_wp
    real(dp)                      :: box_size_z      = 0.0_wp
    ! min
    real(dp)                      :: energy          = 0.0_wp
    real(dp)                      :: delta_r         = 0.0_wp
    ! remd
    integer                       :: iseed_remd      = 0
    integer                       :: dimension       = 0
    integer                       :: nreplicas       = 0
    integer,          allocatable :: repid2parmsetid(:)
    integer,          allocatable :: num_criteria(:,:,:)
    integer,          allocatable :: num_exchanges(:,:,:)
  end type s_rst

  ! parameters for allocatable variables
  integer,      public, parameter :: RestartAtom      = 1
  integer,      public, parameter :: RestartReplica   = 2

  ! parameters
  integer,      public, parameter :: RstfileTypeUndef = 0
  integer,      public, parameter :: RstfileTypeMin   = 1
  integer,      public, parameter :: RstfileTypeMd    = 2
  integer,      public, parameter :: RstfileTypeRemd  = 3

  integer,              parameter :: HeaderLength     = 80

  ! subroutines
  public  :: input_rst
  public  :: output_rst
  public  :: alloc_rst
  public  :: dealloc_rst
  public  :: dealloc_rst_all
  private :: read_rst
  private :: write_rst

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    input_rst
  !> @brief        a driver subroutine for reading Restart file
  !! @authors      TM
  !! @param[in]    rst_filename : filename of restart file
  !! @param[out]   rst          : structure of restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_rst(rst_filename, rst)

    ! formal arguments
    character(*),            intent(in)    :: rst_filename
    type(s_rst),             intent(inout) :: rst

    ! local variables
    integer                  :: file, endian, ival


    ! check restart file endian
    !

    file = get_unit_no()

    open(file, &
         file    = rst_filename,    &
         status  = 'old',           &
         form    = 'unformatted',   &
         convert = 'little_endian', &
         access  = 'direct',        &
         recl    = 4)
    read(file,rec=1) ival
    close(file)

    if (ival /= HeaderLength) then

      open(file, &
           file    = rst_filename,  &
           status  = 'old',         &
           form    = 'unformatted', &
           convert = 'big_endian',  &
           access  = 'direct',      &
           recl    = 4)
      read(file,rec=1) ival
      close(file)

      if (ival /= HeaderLength) &
        call error_msg('Input_Rst> unknown file format.')

      endian = IOFileBigEndian

    else

      endian = IOFileLittleEndian

    end if

    call free_unit_no(file)


    ! open Restart file
    !
    call open_binary_file(file, rst_filename, IOFileInput, endian)

    ! read Restart file
    !
    call read_rst(file, rst)

    ! close Restart file
    !
    call close_file(file)

    return

  end subroutine input_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_rst
  !> @brief        a driver subroutine for writing Restart file
  !! @authors      TM
  !! @param[in]    rst_filename : filename of restart file
  !! @param[in]    rst          : structure of restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_rst(rst_filename, rst)

    ! formal arguments
    character(*),            intent(in) :: rst_filename
    type(s_rst),             intent(in) :: rst

    ! local variables
    integer                  :: file


    ! open Restart file
    !
    call open_binary_file(file, rst_filename, IOFileOutputReplace)

    ! write Restart file
    !
    call write_rst(file, rst)

    ! close Restart file
    !
    call close_file(file)

    return

  end subroutine output_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    alloc_rst
  !> @brief        allocate restart information
  !! @authors      TM, CK
  !! @param[inout] rst       : structure of restart information
  !! @param[in]    variable  : selected variable
  !! @param[in]    var_size  : size of the selected variable
  !! @param[in]    var_size2 : 2nd size of the selected variable
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine alloc_rst(rst, variable, var_size1, var_size2)

    ! formal arguments
    type(s_rst),             intent(inout) :: rst
    integer,                 intent(in)    :: variable
    integer,                 intent(in)    :: var_size1
    integer,       optional, intent(in)    :: var_size2

    ! local variables
    integer                  :: alloc_stat
    integer                  :: dealloc_stat


    ! initialize
    !
    alloc_stat   = 0
    dealloc_stat = 0

    ! allocate selected variables
    !
    select case (variable)

    case(RestartAtom)

      if (allocated(rst%coord)) then
        if (size(rst%coord(1,:)) == var_size1) return
        deallocate(rst%coord,    &
                   rst%velocity, &
                   stat = dealloc_stat)
      end if

      allocate(rst%coord(3,var_size1),    &
               rst%velocity(3,var_size1), &
               stat = alloc_stat)

      rst%coord   (1:3,1:var_size1) = 0.0_dp
      rst%velocity(1:3,1:var_size1) = 0.0_dp

    case(RestartReplica)

      if (allocated(rst%num_criteria)) then
        if (size(rst%num_criteria(:,1,1)) == var_size1 .and. &
            size(rst%num_criteria(1,:,1)) == var_size2 ) return
        deallocate(rst%repid2parmsetid, &
                   rst%num_criteria,    &
                   rst%num_exchanges,   &
                   stat = dealloc_stat)
      end if

      allocate(rst%repid2parmsetid(var_size1),           &
               rst%num_criteria (var_size1,var_size2,2), &
               rst%num_exchanges(var_size1,var_size2,2), &
               stat = alloc_stat)

      rst%repid2parmsetid(1:var_size1)               = 0
      rst%num_criteria (1:var_size1,1:var_size2,1:2) = 0
      rst%num_exchanges(1:var_size1,1:var_size2,1:2) = 0

    case default

      call error_msg('Alloc_Rst> bad variable')

    end select

    if (alloc_stat /= 0)   call error_msg_alloc
    if (dealloc_stat /= 0) call error_msg_dealloc    

    return

  end subroutine alloc_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_rst
  !> @brief        deallocate restart information
  !! @authors      TM, CK
  !! @param[inout] psf      : structure of restart information
  !! @param[in]    variable : selected variable
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_rst(rst, variable)

    ! formal arguments
    type(s_rst),             intent(inout) :: rst
    integer,                 intent(in)    :: variable

    ! local variables
    integer                  :: dealloc_stat


    ! initialize
    !
    dealloc_stat = 0

    ! deallocate selected variable
    !
    select case (variable)

    case (RestartAtom)

      if (allocated(rst%coord)) then
        deallocate(rst%coord,    &
                   rst%velocity, &
                   stat = dealloc_stat)
      end if
      
    case(RestartReplica)

      if (allocated(rst%repid2parmsetid)) then
        deallocate(rst%repid2parmsetid, &
                   rst%num_criteria,    &
                   rst%num_exchanges,   &
                   stat = dealloc_stat)
      end if

    case default
      call error_msg('Dealloc_Restart> bad variable')

    end select

    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_rst_all
  !> @brief        deallocate all restart information
  !! @authors      TM, CK
  !! @param[inout] rst : structure of restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_rst_all(rst)

    ! format arguments
    type(s_rst),             intent(inout) :: rst


    call dealloc_rst(rst, RestartAtom)
    call dealloc_rst(rst, RestartReplica)

    return

  end subroutine dealloc_rst_all

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_rst
  !> @brief        read restart data
  !! @authors      TM, CK
  !! @param[in]    file : unit number of restart file
  !! @param[out]   rst  : structure of restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_rst(file, rst)

    ! formal arguments
    integer,                 intent(in)    :: file
    type(s_rst),             intent(inout) :: rst

    ! local variables    
    integer                  :: i
    character(HeaderLength)  :: head1, head2


    ! deallocate old data
    !
    call dealloc_rst_all(rst)

    ! read header 
    !
    rewind(file)

    read(file) head1
    read(file) head2
    read(file) rst%num_atoms

    ! allocate buffer
    !
    call alloc_rst(rst, RestartAtom, rst%num_atoms)

    ! select restart file type
    !
    read(file) rst%rstfile_type

    select case(rst%rstfile_type)

    case(RstfileTypeMin)

      read(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      read(file) rst%energy, rst%delta_r
      read(file) rst%coord(1,:)
      read(file) rst%coord(2,:)
      read(file) rst%coord(3,:)

      ! write the summary
      !
      if (main_rank) then

        write(MsgOut,'(A)') 'Read_Rst> Summary of RST file'

        write(MsgOut,'(A)') '  RstfileType     =        MIN'
        write(MsgOut,'(A20,I10,A20,I10)')                &
             '  num_atoms       = ', rst%num_atoms
        write(MsgOut,'(A20,3F10.3)')                     &
             '  boxsize (x,y,z) = ', rst%box_size_x,     &
                                     rst%box_size_y,     &
                                     rst%box_size_z

        write(MsgOut,'(A)') ' '
      end if

    case(RstfileTypeMd)

      read(file) rst%iseed, rst%num_deg_freedom
      read(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      read(file) rst%thermostat_momentum
      read(file) rst%barostat_momentum(1:3)
      read(file) rst%coord(1,:)
      read(file) rst%coord(2,:)
      read(file) rst%coord(3,:)
      read(file) rst%velocity(1,:)
      read(file) rst%velocity(2,:)
      read(file) rst%velocity(3,:)

      ! write the summary
      !
      if (main_rank) then
        write(MsgOut,'(A)') 'Read_Rst> Summary of RST file'

        write(MsgOut,'(A)') '  RstfileType     =         MD'
        write(MsgOut,'(A20,I10,A20,I10)')                &
             '  num_atoms       = ', rst%num_atoms,      &
             '  iseed           = ', rst%iseed
        write(MsgOut,'(A20,3F10.3)')                     &
             '  boxsize (x,y,z) = ', rst%box_size_x,     &
                                     rst%box_size_y,     &
                                     rst%box_size_z
        write(MsgOut,'(A)') ''

      end if

    case(RstfileTypeRemd)

      read(file) rst%iseed, rst%num_deg_freedom
      read(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      read(file) rst%thermostat_momentum
      read(file) rst%barostat_momentum(1:3)
      read(file) rst%coord(1,:)
      read(file) rst%coord(2,:)
      read(file) rst%coord(3,:)
      read(file) rst%velocity(1,:)
      read(file) rst%velocity(2,:)
      read(file) rst%velocity(3,:)

      read(file) rst%iseed_remd
      read(file) rst%nreplicas
      read(file) rst%dimension

      call alloc_rst(rst, RestartReplica, rst%nreplicas, rst%dimension)

      read(file) rst%repid2parmsetid(:)
      do i = 1, rst%nreplicas
        read(file) rst%num_criteria (i,:,1)
        read(file) rst%num_criteria (i,:,2)
        read(file) rst%num_exchanges(i,:,1)
        read(file) rst%num_exchanges(i,:,2)
      end do

      if (main_rank) then
        write(MsgOut,'(A)') 'Read_Rst> Summary of RST file'

        write(MsgOut,'(A)') '  RstfileType     =       REMD'
        write(MsgOut,'(A20,I10,A20,I10)')              &
             '  num_atoms       = ', rst%num_atoms,    &
             '  iseed           = ', rst%iseed
        write(MsgOut,'(A20,3F10.3)')                   &
             '  boxsize (x,y,z) = ', rst%box_size_x,   &
                                     rst%box_size_y,   &
                                     rst%box_size_z
        write(MsgOut,'(A20,I10,A20,I10)')              &
             '  nreplicas         = ', rst%nreplicas,  &
             '  dimension         = ', rst%dimension
        write(MsgOut,'(A20,I10)') &
             '  remd iseed        = ', rst%iseed_remd
      end if
    end select

    return

  end subroutine read_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    write_rst
  !> @brief        write data to restart file
  !! @authors      TM, CK
  !! @param[in]    file : unit number of restart file
  !! @param[in]    rst  : structure of restart information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_rst(file, rst)

    ! formal arguments
    integer,                 intent(in) :: file
    type(s_rst),             intent(in) :: rst

    ! local variables
    integer                  :: i
    character(HeaderLength)  :: head1, head2
    character(24)            :: name, date


    ! check memory allocation
    !
    if (.not. allocated(rst%coord))    &
      call error_msg('Write_Rst> not allocated: rst%coord')
    if (.not. allocated(rst%velocity)) &
      call error_msg('Write_Rst> not allocated: rst%velocity')

    ! make header
    !
    call fdate(date)
    call getlog(name)

    head1 = 'REMARKS CREATED BY GENESIS                                                      '
    head2 = 'REMARKS DATE: ' // date // ' CREATED BY USER: ' // name

    write(file) head1
    write(file) head2
    write(file) rst%num_atoms

    ! select restart file type
    !
    write(file) rst%rstfile_type

    select case(rst%rstfile_type)

    case(RstfileTypeMin)

      write(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      write(file) rst%energy, rst%delta_r
      write(file) rst%coord(1,:)
      write(file) rst%coord(2,:)
      write(file) rst%coord(3,:)

    case(RstfileTypeMd)

      write(file) rst%iseed, rst%num_deg_freedom
      write(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      write(file) rst%thermostat_momentum
      write(file) rst%barostat_momentum(1:3)
      write(file) rst%coord(1,:)
      write(file) rst%coord(2,:)
      write(file) rst%coord(3,:)
      write(file) rst%velocity(1,:)
      write(file) rst%velocity(2,:)
      write(file) rst%velocity(3,:)

    case(RstfileTypeRemd)

      write(file) rst%iseed, rst%num_deg_freedom
      write(file) rst%box_size_x, rst%box_size_y, rst%box_size_z
      write(file) rst%thermostat_momentum
      write(file) rst%barostat_momentum(1:3)
      write(file) rst%coord(1,:)
      write(file) rst%coord(2,:)
      write(file) rst%coord(3,:)
      write(file) rst%velocity(1,:)
      write(file) rst%velocity(2,:)
      write(file) rst%velocity(3,:)
      write(file) rst%iseed_remd
      write(file) rst%nreplicas
      write(file) rst%dimension
      write(file) rst%repid2parmsetid(:)
      do i = 1, rst%nreplicas
        write(file) rst%num_criteria (i,:,1)
        write(file) rst%num_criteria (i,:,2)
        write(file) rst%num_exchanges(i,:,1)
        write(file) rst%num_exchanges(i,:,2)
      end do

    end select

    return

  end subroutine write_rst

end module fileio_rst_mod
