!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   molecules_mod
!> @brief   define molecules information
!! @authors Yuji Sugita (YS), Takaharu Mori (TM), Chigusa Kobayashi (CK), 
!!          Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module molecules_mod

  use fileio_top_mod
  use fileio_par_mod
  use fileio_gpr_mod
  use fileio_pdb_mod
  use fileio_psf_mod
  use fileio_crd_mod
  use select_atoms_str_mod
  use molecules_str_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: define_molecules
  public  :: update_num_deg_freedom
  public  :: export_molecules
  private :: setup_molecule_charmm
  private :: setup_molecule_crd
  private :: setup_molecule_ref
  private :: setup_molecule_other
  private :: initialize_velocity
  private :: export_molecule_sel
  private :: export_molecule_all

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_molecules
  !> @brief        a driver subroutine for defining moleucles
  !! @authors      YS, CK, TM, NT
  !! @param[inout] molecule : structure of molecule
  !! @param[in]    pdb      : PDB information (optional)
  !! @param[in]    crd      : CRD information (optional)
  !! @param[in]    top      : TOP information (optional)
  !! @param[in]    par      : PAR information (optional)
  !! @param[in]    psf      : PSF information (optional)
  !! @param[in]    ref      : REF information (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_molecules(molecule, pdb, crd, top, par, gpr, psf, ref)

    ! formal arguments
    type(s_molecule),         intent(inout) :: molecule
    type(s_pdb),    optional, intent(in)    :: pdb
    type(s_crd),    optional, intent(in)    :: crd
    type(s_top),    optional, intent(in)    :: top
    type(s_par),    optional, intent(in)    :: par
    type(s_gpr),    optional, intent(in)    :: gpr
    type(s_psf),    optional, intent(in)    :: psf
    type(s_pdb),    optional, intent(in)    :: ref

    ! local variables
    integer        :: num_atoms_pdb, num_atoms_psf
    integer        :: num_atoms_crd, num_atoms_ref
    logical        :: present_pdb, present_crd, present_ref
    logical        :: present_top, present_par, present_psf, present_gpr


    present_pdb    = .false.
    present_crd    = .false.
    present_psf    = .false.
    present_ref    = .false.
    present_top    = .false.
    present_par    = .false.
    present_gpr    = .false.

    if (present(pdb)) then
      if (allocated(pdb%atom_no)) present_pdb = .true.
    end if
    if (present(crd)) then
      if (allocated(crd%atom_no)) present_crd = .true.
    end if
    if (present(psf)) then
      if (allocated(psf%atom_no)) present_psf = .true.
    end if
    if (present(ref)) then
      if (allocated(ref%atom_no)) present_ref = .true.
    end if
    if (present(top)) then
      if (allocated(top%atom_cls_no))   present_top = .true.
    end if
    if (present(par)) then
      if (allocated(par%bond_atom_cls)) present_par = .true.
    end if
    if (present(gpr)) then
      if (allocated(gpr%bond_list))     present_gpr = .true.
    end if


    ! Check about the number of atoms in PSF, PDB, CRD, and REF
    !
    if (present_pdb) num_atoms_pdb = size(pdb%atom_no)
    if (present_crd) num_atoms_crd = size(crd%atom_no)
    if (present_psf) num_atoms_psf = size(psf%atom_no)
    if (present_ref) num_atoms_ref = size(ref%atom_no)

    if (present_psf .and. present_pdb) then
      if (num_atoms_psf /= num_atoms_pdb) then
        write(MsgOut,*) 'Define_Molecules> ATOM in PSF = ', num_atoms_psf
        write(MsgOut,*) 'Define_Molecules> ATOM in PDB = ', num_atoms_pdb
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if

    if (present_psf .and. present_crd) then
      if (num_atoms_psf /= num_atoms_crd) then
        write(MsgOut,*) 'Define_Molecules> ATOM in PSF = ', num_atoms_psf
        write(MsgOut,*) 'Define_Molecules> ATOM in CRD = ', num_atoms_crd
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if

    if (present_pdb .and. present_crd) then
      if (num_atoms_pdb /= num_atoms_crd) then
        write(MsgOut,*) 'Define_Molecules> ATOM in PDB = ', num_atoms_pdb
        write(MsgOut,*) 'Define_Molecules> ATOM in CRD = ', num_atoms_crd
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if

    if (present_ref .and. present_pdb) then
      if (num_atoms_ref /= num_atoms_pdb) then
        write(MsgOut,*) 'Define_Molecules> ATOM in REF = ', num_atoms_ref
        write(MsgOut,*) 'Define_Molecules> ATOM in PDB = ', num_atoms_pdb
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if

    if (present_ref .and. present_crd) then
      if (num_atoms_ref /= num_atoms_crd) then
        write(MsgOut,*) 'Define_Molecules> ATOM in REF = ', num_atoms_ref
        write(MsgOut,*) 'Define_Molecules> ATOM in CRD = ', num_atoms_crd
        call error_msg('Define_Molecules> Number of atoms differs.')
      end if
    end if


    ! Setup molecule
    !
    if (present_top .and. present_par .and. present_psf) then  !< CHARMM

      if (.not. present_pdb .and. .not. present_crd) &
        call error_msg('Define_Molecules> coordinates are not present.')

      call setup_molecule_charmm(molecule, psf, pdb, crd, top, par)
      call setup_molecule_other (molecule)

    else if (present_top .and. present_gpr .and. present_psf) then  !< Go model

      if (.not. present_pdb .and. .not. present_crd) &
        call error_msg('Define_Molecules> coordinates are not present.')

      call setup_molecule_charmm(molecule, psf, pdb, crd)
      call setup_molecule_other (molecule)

    else if (present_psf) then

      if (.not. present_pdb .and. .not. present_crd) then
        call error_msg('Define_Molecules> coordinates are not present.')
      end if

      call setup_molecule_charmm(molecule, psf, pdb, crd)
      call setup_molecule_other (molecule)

    else

      if (.not. present_pdb .and. .not. present_crd) then
        call error_msg('Define_Molecules> coordinates are not present.')
      end if

      call setup_molecule_crd(molecule, pdb, crd)

    end if


    ! Assing velocities
    !
    call initialize_velocity(molecule)


    ! Setup reference coordinates in molecule
    !
    if (present_ref) then

      call setup_molecule_ref(molecule, ref)

    end if


    ! write summary of molecule information
    !
    if (main_rank) then

      write(MsgOut,'(A)') 'Define_Molecule> Summary of molecules'

      write(MsgOut,'(A20,I10,A20,I10)')                    &
           '  num_atoms       = ', molecule%num_atoms,     &
           '  num_bonds       = ', molecule%num_bonds
      write(MsgOut,'(A20,I10,A20,I10)')                    &
           '  num_angles      = ', molecule%num_angles,    &
           '  num_dihedrals   = ', molecule%num_dihedrals
      write(MsgOut,'(A20,I10,A20,I10)')                    &
           '  num_impropers   = ', molecule%num_impropers, &
           '  num_cmap_terms  = ', molecule%num_cmaps
      write(MsgOut,'(A20,I10,A20,I10)')                    &
           '  num_residues    = ', molecule%num_residues,  &
           '  num_molecules   = ', molecule%num_molecules
      write(MsgOut,'(A20,I10,A20,I10)')                    &
           '  num_segments    = ', molecule%num_segments,  &
           '  num_deg_freedom = ', molecule%num_deg_freedom
      write(MsgOut,'(A20,F10.3)')                          &
           '  total_charge    = ', molecule%total_charge
      write(MsgOut,'(A)') ' '
    end if


    return

  end subroutine define_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_num_deg_freedom
  !> @brief        update number of degree of freedom
  !! @authors      TM
  !! @param[in]    message         : update message
  !! @param[in]    add_ndegf       : added number of degree of freedom
  !! @param[inout] num_deg_freedom : number of degree of freedom
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_num_deg_freedom(message, add_ndegf, num_deg_freedom)

    ! formal arguments
    character(*),            intent(in)    :: message
    integer,                 intent(in)    :: add_ndegf
    integer,                 intent(inout) :: num_deg_freedom


    num_deg_freedom = num_deg_freedom + add_ndegf

    if (main_rank) then
      write(MsgOut,'(A)') &
           'Update_Num_Deg_Freedom> Number of degrees of freedom was updated'
      write(MsgOut,'(A20,I10,3A)') &
           '  num_deg_freedom = ', num_deg_freedom, &
           ' (',trim(message), ')'
      write(MsgOut,'(A)') &
           ' '
    end if

    return

  end subroutine update_num_deg_freedom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    export_molecules
  !> @brief        a driver subroutine for export moleucles
  !! @authors      NT
  !! @param[in]    molecule : molecule information
  !! @param[in]    selatoms : selection information  (optional)
  !! @param[out]   pdb      : PDB information        (optional)
  !! @param[out]   crd      : CRD information        (optional)
  !! @param[out]   top      : TOP information        (optional)
  !! @param[out]   par      : par information        (optional)
  !! @param[out]   psf      : PSF information        (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_molecules(molecule, selatoms, pdb, crd, top, par, gpr, psf)

    ! formal arguments
    type(s_molecule),           intent(in)    :: molecule
    type(s_selatoms), optional, intent(in)    :: selatoms
    type(s_pdb),      optional, intent(inout) :: pdb
    type(s_crd),      optional, intent(inout) :: crd
    type(s_top),      optional, intent(inout) :: top
    type(s_par),      optional, intent(inout) :: par
    type(s_gpr),      optional, intent(inout) :: gpr
    type(s_psf),      optional, intent(inout) :: psf


    if (present(pdb)) then

      if (present(selatoms)) then
        call export_molecule_sel(molecule, selatoms, pdb)
      else
        call export_molecule_all(molecule, pdb)
      end if

    end if

    if (present(crd)) then
      ! Not supported yet.
    end if

    if (present(top)) then
      ! Not supported yet.
    end if

    if (present(par)) then
      ! Not supported yet.
    end if

    if (present(gpr)) then
      ! Not supported yet.
    end if

    if (present(psf)) then
      ! Not supported yet.
    end if

    return

  end subroutine export_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_charmm
  !> @brief        setup molecules by CHARMM input informations
  !! @authors      YS, NT
  !! @param[inout] molecule : molecule information
  !! @param[in]    psf      : PSF information
  !! @param[in]    pdb      : PDB information (optional)
  !! @param[in]    crd      : CRD information (optional)
  !! @param[in]    top      : TOP information (optional)
  !! @param[in]    par      : PAR information (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_charmm(molecule, psf, pdb, crd, top, par)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule
    type(s_psf),             intent(in)    :: psf
    type(s_pdb),   optional, intent(in)    :: pdb
    type(s_crd),   optional, intent(in)    :: crd
    type(s_top),   optional, intent(in)    :: top
    type(s_par),   optional, intent(in)    :: par

    ! local variables
    integer                  :: natom
    integer                  :: nbond
    integer                  :: nangl
    integer                  :: ndihe
    integer                  :: nimpr
    integer                  :: ncmap
    integer                  :: i, j, k, icls
    logical                  :: present_pdb
    character(6)             :: acls
    character(4)             :: cres


    ! initialize
    !
    call init_molecules(molecule)

    ! allocation
    !
    natom = psf%num_atoms
    nbond = psf%num_bonds
    nangl = psf%num_angles
    ndihe = psf%num_dihedrals
    nimpr = psf%num_impropers
    ncmap = psf%num_cross_terms

    call alloc_molecules(molecule, MoleculeAtom, natom)
    call alloc_molecules(molecule, MoleculeBond, nbond)
    call alloc_molecules(molecule, MoleculeAngl, nangl)
    call alloc_molecules(molecule, MoleculeDihe, ndihe)
    call alloc_molecules(molecule, MoleculeImpr, nimpr)
    call alloc_molecules(molecule, MoleculeCmap, ncmap)

    ! copy information from PSF
    !
    molecule%atom_no(1:natom)          = psf%atom_no(1:natom)
    molecule%segment_name(1:natom)     = psf%segment_name(1:natom)
    molecule%residue_no(1:natom)       = psf%residue_no(1:natom)
    molecule%residue_name(1:natom)     = psf%residue_name(1:natom)
    molecule%molecule_no(1:natom)      = psf%molecule_no(1:natom)
    molecule%atom_name(1:natom)        = psf%atom_name(1:natom)
    molecule%atom_cls_name(1:natom)    = psf%atom_cls_name(1:natom)
    molecule%charge(1:natom)           = psf%charge(1:natom)
    molecule%mass(1:natom)             = psf%mass(1:natom)
    molecule%inv_mass(1:natom)         = 1.0_wp/psf%mass(1:natom)
    molecule%imove(1:natom)            = psf%imove(1:natom)

    molecule%bond_list(1:2,1:nbond)    = psf%bond_list(1:2,1:nbond)
    molecule%angl_list(1:3,1:nangl)    = psf%angl_list(1:3,1:nangl)
    molecule%dihe_list(1:4,1:ndihe)    = psf%dihe_list(1:4,1:ndihe)
    molecule%impr_list(1:4,1:nimpr)    = psf%impr_list(1:4,1:nimpr)
    molecule%cmap_list(1:8,1:ncmap)    = psf%cmap_list(1:8,1:ncmap)

    present_pdb = .false.
    if (present(pdb)) then
      if (allocated(pdb%atom_no)) present_pdb = .true.
    end if

    if (present_pdb) then

      ! copy information from PDB
      !
      molecule%chain_id(1:natom)         = pdb%chain_id(1:natom)
      molecule%atom_coord(1:3,1:natom)   = pdb%atom_coord(1:3,1:natom)
      molecule%atom_occupancy(1:natom)   = pdb%atom_occupancy(1:natom)
      molecule%atom_temp_factor(1:natom) = pdb%atom_temp_factor(1:natom)
       
    else

      ! copy information from CRD
      !
      molecule%chain_id(1:natom)         = 'A'
      molecule%atom_coord(1:3,1:natom)   = crd%atom_coord(1:3,1:natom)
      molecule%atom_occupancy(1:natom)   = 0.0_wp
      molecule%atom_temp_factor(1:natom) = 0.0_wp
       
    end if
 
    ! number of variables
    !
    molecule%num_atoms       = natom
    molecule%num_bonds       = nbond
    molecule%num_angles      = nangl
    molecule%num_dihedrals   = ndihe
    molecule%num_impropers   = nimpr
    molecule%num_cmaps       = ncmap
    molecule%num_deg_freedom = 3*natom

    ! redefine atom class name
    !
    if (present(top) .and. psf%type == PsfTypeCHARMM) then

      do i = 1, molecule%num_atoms

        cres = molecule%residue_name(i)
        icls = psf%atom_cls_no(i)

        molecule%atom_cls_name(i) = ''

        do j = 1, top%num_res_type
          if (cres == top%res_name(j)) then
            do k = top%res_atom_cls_bgn(j), top%res_atom_cls_end(j)
              if (icls == top%atom_cls_no(k)) then
                molecule%atom_cls_name(i) = top%atom_cls_name(k)
                goto 10
              end if
            end do
          end if
        end do

10      if (molecule%atom_cls_name(i) == '') &
          call error_msg('Setup_Molecule_Charmm> Unknown atom class no')

      end do
    end if

    ! define atom class number
    !
    if (present(par)) then

      do i = 1, molecule%num_atoms
        acls = molecule%atom_cls_name(i)
        do j = 1, par%num_atom_cls
          if (acls == par%nonb_atom_cls(j)) then
            molecule%atom_cls_no(i) = j
          end if
        end do
      end do

    end if

    return

  end subroutine setup_molecule_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_crd
  !> @brief        setup molecules by PDB/CRD information
  !! @authors      YS, NT
  !! @param[inout] molecule : molecule information
  !! @param[in]    pdb      : PDB information (optional)
  !! @param[in]    crd      : CRD information (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_crd(molecule, pdb, crd)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule
    type(s_pdb),   optional, intent(in)    :: pdb
    type(s_crd),   optional, intent(in)    :: crd

    ! local variables
    integer                  :: natom
    logical                  :: present_pdb


    ! initialize
    !
    call init_molecules(molecule)

    present_pdb = .false.
    if (present(pdb)) then
      if (allocated(pdb%atom_no)) present_pdb = .true.
    end if

    if (present_pdb) then

      ! copy information from PDB
      !

      natom = pdb%num_atoms
      call alloc_molecules(molecule, MoleculeAtom, natom)

      molecule%atom_no(1:natom)          = pdb%atom_no(1:natom)
      molecule%segment_name(1:natom)     = pdb%segment_name(1:natom)
      molecule%segment_no(1:natom)       = 0
      molecule%residue_no(1:natom)       = pdb%residue_no(1:natom)
      molecule%residue_name(1:natom)     = pdb%residue_name(1:natom)
      molecule%atom_name(1:natom)        = pdb%atom_name(1:natom)
      molecule%atom_cls_name(1:natom)    = ''
      molecule%atom_cls_no(1:natom)      = 0
      molecule%charge(1:natom)           = 0.0_wp
      molecule%mass(1:natom)             = 0.0_wp
      molecule%inv_mass(1:natom)         = 0.0_wp
      molecule%imove(1:natom)            = 0
      molecule%chain_id(1:natom)         = pdb%chain_id(1:natom)
      molecule%atom_coord(1:3,1:natom)   = pdb%atom_coord(1:3,1:natom)
      molecule%atom_occupancy(1:natom)   = pdb%atom_occupancy(1:natom)
      molecule%atom_temp_factor(1:natom) = pdb%atom_temp_factor(1:natom)

    else

      ! copy information from CRD
      !

      natom = crd%num_atoms
      call alloc_molecules(molecule, MoleculeAtom, natom)

      molecule%atom_no(1:natom)          = crd%atom_no(1:natom)
      molecule%segment_name(1:natom)     = crd%segment_name(1:natom)
      molecule%segment_no(1:natom)       = 0
      molecule%residue_no(1:natom)       = crd%residue_no(1:natom)
      molecule%residue_name(1:natom)     = crd%residue_name(1:natom)
      molecule%atom_name(1:natom)        = crd%atom_name(1:natom)
      molecule%atom_cls_name(1:natom)    = ''
      molecule%atom_cls_no(1:natom)      = 0
      molecule%charge(1:natom)           = 0.0_wp
      molecule%mass(1:natom)             = 0.0_wp
      molecule%inv_mass(1:natom)         = 0.0_wp
      molecule%imove(1:natom)            = 0
      molecule%chain_id(1:natom)         = 'A'
      molecule%atom_coord(1:3,1:natom)   = crd%atom_coord(1:3,1:natom)
      molecule%atom_occupancy(1:natom)   = 0.0_wp
      molecule%atom_temp_factor(1:natom) = 0.0_wp

    end if


    ! number of variables
    !
    molecule%num_atoms       = natom
    molecule%num_deg_freedom = 3*natom


    return

  end subroutine setup_molecule_crd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_ref
  !> @brief        setup molecules by Ref information
  !! @authors      CK
  !! @param[inout] molecule : molecule information
  !! @param[in]    ref      : ref information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_ref(molecule, ref)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule
    type(s_pdb),   optional, intent(in)    :: ref

    ! local variables
    integer                  :: natom


    if (molecule%num_atoms /= ref%num_atoms) then
      write(MsgOut,*) &
           'Setup_molecule_ref> ATOM in Molecule = ', molecule%num_atoms
      write(MsgOut,*) &
           'Setup_molecule_ref> ATOM in REF      = ', ref%num_atoms
      call error_msg('Setup_molecule_ref> Number of atoms differs.')
    end if

    natom = molecule%num_atoms

    call alloc_molecules(molecule, MoleculeRefc, natom)

    molecule%atom_refcoord(1:3,1:natom) = ref%atom_coord(1:3,1:natom)

    return

  end subroutine setup_molecule_ref

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_other
  !> @brief        setup other molecule informations
  !! @authors      NT
  !! @param[inout] molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_other(molecule)

    ! formal arguments
    !
    type(s_molecule),        intent(inout) :: molecule

    ! local variables
    !
    real(wp)                 :: mass, charge
    integer                  :: i, j
    integer                  :: nmol, imole, imole_pre, initial, final
    integer                  :: col, pcol
    integer                  :: src_col, dst_col, cur_col
    integer                  :: i1, i2
    integer                  :: resno_pre, resno_next
    character(4)             :: segment_pre, segment_next
    character(4)             :: residue_pre, residue_next

    integer, allocatable     :: color(:), color_range(:,:)


    ! total charge
    !
    charge = 0.0_wp
    do i = 1, molecule%num_atoms
      charge = charge + molecule%charge(i)
    end do
    molecule%total_charge = charge


    ! count the number of segments
    !
    do i = 1, molecule%num_atoms
      if (i == 1) then
        molecule%segment_no(1) = 1
        segment_pre = molecule%segment_name(1)
      else
        if (molecule%segment_name(i) /= segment_pre) then
          molecule%segment_no(i) = molecule%segment_no(i-1) + 1
          segment_pre            = molecule%segment_name(i)
        else
          molecule%segment_no(i) = molecule%segment_no(i-1)
        end if
      end if
    end do
    molecule%num_segments = molecule%segment_no(molecule%num_atoms)


    ! count the number of residues
    !
    do i = 1, molecule%num_atoms
      if (i == 1) then
        segment_pre = molecule%segment_name(i)
        residue_pre = molecule%residue_name(i)
        resno_pre   = molecule%residue_no(i)
        molecule%residue_c_no(i) = 1
      else
        segment_pre  = molecule%segment_name(i-1)
        residue_pre  = molecule%residue_name(i-1)
        resno_pre    = molecule%residue_no(i-1)
        segment_next = molecule%segment_name(i)
        residue_next = molecule%residue_name(i)
        resno_next   = molecule%residue_no(i)
        
        if ((segment_pre /= segment_next) .or. &
            (residue_pre /= residue_next) .or. &
            (resno_pre   /= resno_next)) then
          molecule%residue_c_no(i) = molecule%residue_c_no(i-1) + 1
        else
          molecule%residue_c_no(i) = molecule%residue_c_no(i-1)
        end if
      end if
    end do
    molecule%num_residues = molecule%residue_c_no(molecule%num_atoms)


    ! count the number of molecule
    !

    allocate(color(molecule%num_atoms), color_range(2,molecule%num_atoms/2))
    color(:)         = 0
    color_range(:,:) = 0
    
    ! coloring 
    cur_col = 0
    do i = 1, molecule%num_bonds

      i1 = molecule%bond_list(1,i)
      i2 = molecule%bond_list(2,i)

      if (color(i1) == 0 .and. color(i2) == 0) then

        cur_col = cur_col + 1
        color(i1) = cur_col
        color(i2) = cur_col

        color_range(1,cur_col) = min(i1, i2)
        color_range(2,cur_col) = max(i1, i2)

      else if (color(i1) == 0 .or. color(i2) == 0) then

        if (color(i1) /= 0) then
          
          col = color(i1)
          color(i2) = col
          color_range(1,col) = min(color_range(1,col),i2)
          color_range(2,col) = max(color_range(2,col),i2)

        else ! color(i2) /= 0

          col = color(i2)
          color(i1) = col
          color_range(1,col) = min(color_range(1,col),i1)
          color_range(2,col) = max(color_range(2,col),i1)

        end if

      else if (color(i1) /= color(i2)) then

        if (color(i1) < color(i2)) then
          src_col = color(i1)
          dst_col = color(i2)
        else
          src_col = color(i2)
          dst_col = color(i1)
        end if

        color_range(1,src_col) = min(color_range(1,src_col), &
                                     color_range(1,dst_col))
        color_range(2,src_col) = max(color_range(2,src_col), &
                                     color_range(2,dst_col))

        do j = color_range(1,dst_col), color_range(2,dst_col)
          if (color(j) == dst_col) &
            color(j) = src_col
        end do

        color_range(1,dst_col) = 0
        color_range(2,dst_col) = 0

      end if

    end do

    ! define molecule number
    nmol = 0
    pcol = -1

    do i = 1, molecule%num_atoms

      col = color(i)

      if (pcol /= col .or. col == 0) &
           nmol = nmol + 1

      molecule%molecule_no(i) = nmol

      pcol = col
    end do

    deallocate(color, color_range)

    molecule%num_molecules = nmol


    ! allocate molecule information
    !

    call alloc_molecules(molecule, MoleculeMole, nmol)


    ! define information per molecule
    !

    do i = 1, molecule%num_atoms
      if (i == 1) then
        imole = molecule%molecule_no(i)
        molecule%molecule_atom_no(imole) = i
      else
        imole_pre = molecule%molecule_no(i-1)  
        imole     = molecule%molecule_no(i)
        if (imole /= imole_pre) then
          molecule%molecule_atom_no(imole) = i
        end if
      end if
      molecule%molecule_name(imole) = molecule%segment_name(i)
    end do

    do i = 1, molecule%num_molecules
      initial  = molecule%molecule_atom_no(i)
      if (i /= molecule%num_molecules) then
        final = molecule%molecule_atom_no(i+1) - 1
      else
        final = molecule%num_atoms
      end if

      mass = 0.0_wp
      do j = initial, final
        mass = mass + molecule%mass(j)
      end do
      molecule%molecule_mass(i) = mass

    end do

    return

  end subroutine setup_molecule_other

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    initialize_velocity
  !> @brief        initialize velocities
  !! @authors      TM
  !! @param[inout] molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine initialize_velocity(molecule)

    ! formal arguments
    type(s_molecule),        intent(inout) :: molecule

    ! local variables
    integer                  :: natom

    natom = molecule%num_atoms
    molecule%atom_velocity(1:3,1:natom) = 0.0_wp

    return

  end subroutine initialize_velocity 

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    export_molecule_sel
  !> @brief        export molecule to pdb
  !! @authors      NT
  !! @param[in]    molecule : structure of molecule
  !! @param[in]    selatoms : information of atom selection
  !! @param[out]   pdb      : PDB information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_molecule_sel(molecule, selatoms, pdb)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_selatoms),        intent(in)    :: selatoms
    type(s_pdb),             intent(inout) :: pdb

    ! local variables
    integer                  :: i, idx, natom, nresi, nter, mole1, mole2
    integer,     allocatable :: ter_no(:)


    natom = size(selatoms%idx)
    nresi = 0
    nter  = 0

    allocate(ter_no(natom + 1))

    ! init pdb
    call init_pdb(pdb)

    pdb%hetatm_rec = .false.
    pdb%cryst_rec  = .false.
    pdb%model_rec  = .true.
    pdb%ter_rec    = .true.
    pdb%end_rec    = .true.
    pdb%segment    = .true.

    ! setup atom data
    call alloc_pdb(pdb, PdbAtom, natom)

    do i = 1, natom

      idx = selatoms%idx(i)

      pdb%hetatm(i)           = .false.
      pdb%atom_name(i)        = molecule%atom_name(idx)
      pdb%residue_name(i)     = molecule%residue_name(idx)
      pdb%segment_name(i)     = molecule%segment_name(idx)
      pdb%chain_id(i)         = molecule%chain_id(idx)
      pdb%atom_no(i)          = i
      pdb%residue_no(i)       = molecule%residue_no(idx)
      pdb%atom_coord(:,i)     = molecule%atom_coord(:,idx)
      pdb%atom_occupancy(i)   = 1.0_wp
      pdb%atom_temp_factor(i) = 0.0_wp

      nresi = max(nresi, pdb%residue_no(i))

      if (i /= natom) then

        mole1 = molecule%molecule_no(idx)
        mole2 = molecule%molecule_no(selatoms%idx(i + 1))

        if (mole1 /= mole2) then
          nter = nter + 1
          ter_no(nter) = i + 1
        end if

      end if

    end do

    nter = nter + 1
    ter_no(nter) = natom + 1

    ! setup ter data
    call alloc_pdb(pdb, PdbTerLine, nter)

    pdb%ter_line_no(1:nter) = ter_no(1:nter)


    if (natom > 999999) &
      pdb%atom_col7 = .true.

    if (nresi > 9999) then
      pdb%res_col5 = .true.
      if (nresi > 99999) &
        pdb%res_col6 = .true.
    end if

    deallocate(ter_no)

    return

  end subroutine export_molecule_sel

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    export_molecule_all
  !> @brief        export molecule to pdb
  !! @authors      NT
  !! @param[in]    molecule : molecule information
  !! @param[out]   pdb      : PDB information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine export_molecule_all(molecule, pdb)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_pdb),             intent(inout) :: pdb

    ! local variables
    integer                  :: i, natom, nresi, nter, mole1, mole2
    integer,     allocatable :: ter_no(:)


    natom = size(molecule%atom_name)
    nresi = 0
    nter  = 0

    allocate(ter_no(natom + 1))

    ! init pdb
    call init_pdb(pdb)

    pdb%hetatm_rec = .false.
    pdb%cryst_rec  = .false.
    pdb%model_rec  = .true.
    pdb%ter_rec    = .true.
    pdb%end_rec    = .true.
    pdb%segment    = .true.

    ! setup atom data
    call alloc_pdb(pdb, PdbAtom, natom)

    do i = 1, natom

      pdb%hetatm(i)           = .false.
      pdb%atom_name(i)        = molecule%atom_name(i)
      pdb%residue_name(i)     = molecule%residue_name(i)
      pdb%segment_name(i)     = molecule%segment_name(i)
      pdb%chain_id(i)         = molecule%chain_id(i)
      pdb%atom_no(i)          = i
      pdb%residue_no(i)       = molecule%residue_no(i)
      pdb%atom_coord(:,i)     = molecule%atom_coord(:,i)
      pdb%atom_occupancy(i)   = 1.0_wp
      pdb%atom_temp_factor(i) = 0.0_wp

      nresi = max(nresi, pdb%residue_no(i))

      if (i /= natom) then

        mole1 = molecule%molecule_no(i)
        mole2 = molecule%molecule_no(i + 1)

        if (mole1 /= mole2) then
          nter = nter + 1
          ter_no(nter) = i + 1
        end if

      end if

    end do

    nter = nter + 1
    ter_no(nter) = natom + 1

    ! setup ter data
    call alloc_pdb(pdb, PdbTerLine, nter)

    pdb%ter_line_no(1:nter) = ter_no(1:nter)


    if (natom > 999999) &
      pdb%atom_col7 = .true.

    if (nresi > 9999) then
      pdb%res_col5 = .true.
      if (nresi > 99999) &
        pdb%res_col6 = .true.
    end if

    deallocate(ter_no)

    return

  end subroutine export_molecule_all

end module molecules_mod
