!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   pr_run_prst_setup_mod
!> @brief   run output parallel I/O restart files
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module pr_run_prst_setup_mod

  use pr_setup_spdyn_peer_mod
  use pr_control_mod
  use pr_domain_index_file_mod
  use pr_cached_molecule_mod
  use sp_domain_mod
  use sp_parallel_io_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use fileio_str_mod
  use fileio_top_mod
  use fileio_par_mod
  use fileio_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use omp_lib
 
  implicit none
  private

  ! subroutines
  public  :: run_prst_setup
  private :: setup_mpi_dummy
  private :: setup_molecule_cache_file
  private :: setup_molecule_variables
  private :: setup_domain_variables
  private :: setup_domain_rst_files
  private :: setup_domain_index_list
  private :: output_domain_restart_files
  private :: output_pdb_file
  private :: setup_the

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_prst_setup
  !> @brief        run output parallel I/O rstart files
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_prst_setup(ctrl_data)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data

    ! local variables
    type(s_boundary)         :: boundary
    type(s_boundary)         :: boundary_rst
    type(s_constraints)      :: constraints
    type(s_enefunc)          :: enefunc
    type(s_dynvars)          :: dynvars
    type(s_dynamics)         :: dynamics
    type(s_top)              :: top
    type(s_par)              :: par


    ! setup domain information
    !

    write(MsgOut,'(A)') '[STEP2-1] Setup MPI dummy'
    write(MsgOut,'(A)') ' '

    call setup_mpi_dummy(ctrl_data)


    ! open par/top file
    !
    
    write(MsgOut,'(A)') '[STEP2-2] Open Par/Top file'
    write(MsgOut,'(A)') ' '

    call input_top(ctrl_data%inp_info%topfile, top)
    call input_par(ctrl_data%inp_info%parfile, par)
    call input_str(ctrl_data%inp_info%strfile, top, par)


    ! setup molecule cache file
    !

    write(MsgOut,'(A)') '[STEP2-3] Setup molecule cache file'
    write(MsgOut,'(A)') ' '

    call setup_molecule_cache_file(ctrl_data, top, par)


    ! setup molecule variables
    !

    write(MsgOut,'(A)') '[STEP2-4] Setup molecule variables'
    write(MsgOut,'(A)') ' '

    call setup_molecule_variables(ctrl_data)


    ! setup domain restart files
    !

    write(MsgOut,'(A)') '[STEP2-5] Setup domain restart files'
    write(MsgOut,'(A)') ' '

    call setup_domain_rst_files(ctrl_data, boundary_rst)


    ! setup domain variables (once)
    !

    write(MsgOut,'(A)') '[STEP2-6] Setup domain variables'
    write(MsgOut,'(A)') ' '

    call setup_domain_variables(ctrl_data, par, boundary_rst, boundary,&
                                enefunc, dynvars, dynamics, constraints)


    ! output PDB file
    !
    call output_pdb_file(ctrl_data)


    ! setup domain index list
    !

    write(MsgOut,'(A)') '[STEP2-7] Setup domain index list'
    write(MsgOut,'(A)') ' '

    call setup_domain_index_list(ctrl_data, boundary)


    ! output domain restart files
    !

    write(MsgOut,'(A)') '[STEP2-8] Output domain restart files'
    write(MsgOut,'(A)') ' '

    call output_domain_restart_files(ctrl_data, par,              &
                                     boundary, dynvars, dynamics, &
                                     enefunc, constraints)


    ! deallocate memory
    !

    write(MsgOut,'(A)') '[STEP2-9] Deallocate memory'
    write(MsgOut,'(A)') ' '

    call di_delete

    call cm_delete_selatoms
    call cm_delete_H_index
    call cm_delete_duplicate
    call cm_delete_solute_list_inv
    call cm_delete_solute_list
    call cm_delete_solute
    call cm_delete_water_list
    call cm_delete_molecules
    call cm_finalize(.not. ctrl_data%opt_info%keep_cache)

    call dealloc_constraints(constraints, ConstraintsBondGroup)
    call dealloc_enefunc(enefunc, EneFuncReff)
    call dealloc_par_all(par)
    call dealloc_top_all(top)

    return

  end subroutine run_prst_setup

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mpi_dummy
  !> @brief        setup MPI in dummy
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !! @param[inout] boundary  : boundary condition information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mpi_dummy(ctrl_data)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data


    my_world_rank = 0
    my_city_rank  = 0

    if (ctrl_data%opt_info%domain_xyz /= 0) then
      nproc_country = ctrl_data%opt_info%domain_xyz 
      nproc_city    = nproc_country
      nproc_world   = nproc_city
    else
      nproc_country = ctrl_data%bou_info%domain_x * &
                      ctrl_data%bou_info%domain_y * &
                      ctrl_data%bou_info%domain_z
      nproc_city    = nproc_country
      nproc_world   = nproc_city
    end if

    return

  end subroutine setup_mpi_dummy

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_cache_file
  !> @brief        setup molecule cache file
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !! @param[in]    top       : CHARMM top information
  !! @param[in]    par       : CHARMM par information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_cache_file(ctrl_data, top, par)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_top),             intent(in)    :: top
    type(s_par),             intent(in)    :: par

    ! local variables
    integer                  :: ic1, ic2
    character(200)           :: psffile, pdbfile, crdfile, reffile, prefix, file


    psffile = ctrl_data%inp_info%psffile
    pdbfile = ctrl_data%inp_info%pdbfile
    crdfile = ctrl_data%inp_info%crdfile
    reffile = ctrl_data%inp_info%reffile


    ! setup prefix
    !

    if (pdbfile /= '') then
      file = pdbfile
    else if (crdfile /= '') then
      file = crdfile
    else
      call error_msg('Setup_Molecule_Cache_File> coordinate is empty.')
    end if

    ic1 = scan(psffile, '/', .true.)
    if (ic1 == 0) &
      call error_msg('Setup_Molecule_Cache_File> PSF filename is empty. '// &
                     trim(psffile))
    ic2 = scan(file, '/', .true.)
    if (ic2 == 0) &
      call error_msg('Setup_Molecule_Cache_File> PDB/CRD filename is empty. '//&
                     trim(file))
    prefix = trim(psffile(ic1+1:)) // '_' // trim(file(ic2+1:))


    ! create s_molecule cache files
    !

    call cm_initialize(ctrl_data%opt_info%cachepath, prefix)

    call cm_create_molecules(top, par, psffile, pdbfile, crdfile, reffile)

    return

  end subroutine setup_molecule_cache_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_molecule_variables
  !> @brief        setup molecule variables
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_molecule_variables(ctrl_data)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data

    ! local variables
    integer                  :: i, i1, i2, j, col, pcol, nmol
    integer                  :: src_col, dst_col, cur_col
    integer, allocatable     :: color(:), color_range(:,:)


    ! count the number of molecules
    !

    allocate(color(cm_num_atoms), color_range(2,cm_num_atoms/2))
    color(:)         = 0
    color_range(:,:) = 0
    
    ! coloring 
    cur_col = 0
    do i = 1, cm_num_bonds

      i1 = cm_bond_list(1,i)
      i2 = cm_bond_list(2,i)

      if (color(i1) == 0 .and. color(i2) == 0) then

        cur_col = cur_col + 1
        color(i1) = cur_col
        color(i2) = cur_col

        color_range(1,cur_col) = min(i1, i2)
        color_range(2,cur_col) = max(i1, i2)

      else if (color(i1) == 0 .or. color(i2) == 0) then

        if (color(i1) /= 0) then
          
          col = color(i1)
          color(i2) = col
          color_range(1,col) = min(color_range(1,col),i2)
          color_range(2,col) = max(color_range(2,col),i2)

        else ! color(i2) /= 0

          col = color(i2)
          color(i1) = col
          color_range(1,col) = min(color_range(1,col),i1)
          color_range(2,col) = max(color_range(2,col),i1)

        end if

      else if (color(i1) /= color(i2)) then

        if (color(i1) < color(i2)) then
          src_col = color(i1)
          dst_col = color(i2)
        else
          src_col = color(i2)
          dst_col = color(i1)
        end if

        color_range(1,src_col) = min(color_range(1,src_col), &
                                     color_range(1,dst_col))
        color_range(2,src_col) = max(color_range(2,src_col), &
                                     color_range(2,dst_col))

        do j = color_range(1,dst_col), color_range(2,dst_col)
          if (color(j) == dst_col) &
            color(j) = src_col
        end do

        color_range(1,dst_col) = 0
        color_range(2,dst_col) = 0

      end if

    end do

    ! define molecule number
    nmol = 0
    pcol = -1

    do i = 1, cm_num_atoms

      col = color(i)

      if (pcol /= col .or. col == 0) &
           nmol = nmol + 1

      call cm_set_molecule_no(i, nmol)

      pcol = col
    end do

    deallocate(color, color_range)

    write(MsgOut,'(a,i9)') &
         'Setup_Molecule_Variables> # of molecules : ', nmol
    write(MsgOut,'(a)') ''

    return

  end subroutine setup_molecule_variables

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_domain_rst_files
  !> @brief        setup domain restart files
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !! @param[inout] boundary  : boundary information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_domain_rst_files(ctrl_data, boundary)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_boundary),        intent(inout) :: boundary

    ! local variables
    type(s_domain)           :: domain
    integer                  :: idom, ndom
    integer                  :: i, ix, ig
    integer                  :: nplace
    character(200)           :: filename
    logical                  :: bex


    call init_boundary(boundary)

    ! check restart files are input
    !
    if (ctrl_data%inp_info%rstfile == '') then

      call cm_recall_atom_coord
      call cm_recall_atom_velocity

      write(MsgOut,'(a)') '   ... skip'
      write(MsgOut,'(a)') ' '
      pio_restart = .false.
      return

    end if

    ! compute # of domains 
    !
    my_world_rank = 0
    my_city_rank  = 0

    do nplace = 1, 7
      filename = pio_get_ranked_filename(ctrl_data%inp_info%rstfile, nplace)
      inquire(file=filename, exist=bex)
      if (bex) &
        exit
    end do

    write(MsgOut,'(a,i10)')  'Restart file : # of places : ', nplace+1
    write(MsgOut,'(a)')      ' '

    call pio_read_domain_str(filename, boundary, domain)

    if (ctrl_data%opt_info%set_restart) pio_restart = .true.
    write(MsgOut,'(a)')      'Input domain information : '
    write(MsgOut,'(a,3i10)') '  domains (x,y,z) = ', boundary%num_domain(1), &
                                                     boundary%num_domain(2), &
                                                     boundary%num_domain(3)

    ndom = boundary%num_domain(1) * &
           boundary%num_domain(2) * &
           boundary%num_domain(3)


    ! backup original coordinates and velocities
    !

    call cm_backup_atom_coord
    call cm_backup_atom_velocity


    ! update coordinates and velocities
    !

    do idom = 1, ndom

      write(MsgOut,'(a,i6,a,i6)') '  read ', idom, '/', ndom

      my_world_rank = idom - 1
      my_city_rank  = idom - 1
      filename      = pio_get_ranked_filename(ctrl_data%inp_info%rstfile,nplace)

      call pio_read_domain_str(filename, boundary, domain)

!      do i = 1, size(domain%num_atom)
      do i = 1, domain%num_cell_local
        do ix = 1, domain%num_atom(i)
          ig = domain%id_l2g(ix,i)
          call cm_set_atom_coord(1, ig, domain%coord(1,ix, i))
          call cm_set_atom_coord(2, ig, domain%coord(2,ix, i))
          call cm_set_atom_coord(3, ig, domain%coord(3,ix, i))
          call cm_set_atom_velocity(1, ig, domain%velocity(1,ix, i))
          call cm_set_atom_velocity(2, ig, domain%velocity(2,ix, i))
          call cm_set_atom_velocity(3, ig, domain%velocity(3,ix, i))
        end do
      end do

    end do


    ! deallocate
    !

    call dealloc_domain_all(domain)

    my_world_rank = 0
    my_city_rank  = 0

    write(MsgOut,'(a)') ' '

    return

  end subroutine setup_domain_rst_files

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_domain_variables
  !> @brief        setup domain variables
  !! @authors      NT
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    boundary_rst: boundary condition information on restart
  !! @param[inout] boundary    : boundary condition information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamic variable information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] constraints : domain constraints information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_domain_variables(ctrl_data, par, &
                                    boundary_rst, boundary, enefunc, dynvars, &
                                    dynamics, constraints)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_par),             intent(in)    :: par
    type(s_boundary),        intent(in)    :: boundary_rst
    type(s_boundary),        intent(inout) :: boundary
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_constraints),     intent(inout) :: constraints

    ! local variables
    type(s_restraints)       :: restraints


    ! setups only to be related to the whole system
    !
    
    call setup_md_peer_0(ctrl_data,    &
                         par,          &
                         boundary_rst, &
                         boundary,     &
                         enefunc,      &
                         dynvars,      &
                         dynamics,     &
                         constraints,  &
                         restraints)

    write(MsgOut,'(a)') ' '

    call dealloc_restraints(restraints, RestraintsGroup)
    call dealloc_restraints(restraints, RestraintsFunc)

    return

  end subroutine setup_domain_variables

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_domain_index_list
  !> @brief        setup domain index list
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !! @param[in]    boundary  : boundary condition information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_domain_index_list(ctrl_data, boundary)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_boundary),        intent(in)    :: boundary
    

    write(MsgOut,'(a)') 'Setup_Domain_Index_List> '

    di_file_dir = ctrl_data%opt_info%cachepath


    call di_create(boundary)

    write(MsgOut,'(a)') ''

    return

  end subroutine setup_domain_index_list

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_domain_restart_files
  !> @brief        output domain restart files
  !! @authors      NT
  !! @param[in]    ctrl_data    : information of control parameters
  !! @param[in]    par          : CHARMM PAR information
  !! @param[in]    bundary      : boundary condition information
  !! @param[in]    dynvars      : dynamic variables information
  !! @param[in]    dynamics     : dynamics information
  !! @param[inout] enefuncO     : potential energy functions information
  !! @param[inout] constraintsO : constraints information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_domain_restart_files(ctrl_data, par,              &
                                         boundary, dynvars, dynamics, &
                                         enefuncO, constraintsO)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_par),             intent(in)    :: par
    type(s_boundary),        intent(in)    :: boundary
    type(s_dynvars),         intent(in)    :: dynvars
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_enefunc),         intent(inout) :: enefuncO
    type(s_constraints),     intent(inout) :: constraintsO
    
    ! local variables
    integer                  :: local_my_rank
    integer                  :: idom, ndom, tid, maxt
    character(200)           :: filename

    type(s_domain_index), allocatable :: domain_index(:)
    type(s_domain),       allocatable :: domain(:)
    type(s_enefunc),      allocatable :: enefunc(:)
    type(s_constraints),  allocatable :: constraints(:)
    type(s_pio_t0_info),  allocatable :: t0_info(:)


    ! allocate working variables per threads
    !
#ifdef OMP
    maxt = omp_get_max_threads()
#else
    maxt = 1
#endif
    allocate(domain_index(0:maxt-1), &
             domain      (0:maxt-1), &
             enefunc     (0:maxt-1), &
             constraints (0:maxt-1), &
             t0_info     (0:maxt-1))


    ! disable varbose standard output
    !
    main_rank = .false.


    ! update cache files with current memory states
    !     for multi-thread cache access

    call cm_sync_atom_coord
    call cm_sync_atom_velocity
    call cm_sync_molecule_no
    call cm_sync_water_list
    call cm_sync_solute
    call cm_sync_solute_list
    call cm_sync_solute_list_inv
    call cm_sync_duplicate
    call cm_sync_H_index


    ! start domain output
    !

    ndom = boundary%num_domain(1) * &
           boundary%num_domain(2) * &
           boundary%num_domain(3)


    !$omp parallel do private (local_my_rank, tid, filename)
    !

    do idom = 1, ndom

#ifdef OMP
      tid = omp_get_thread_num()
#else
      tid = 0
#endif
      write(MsgOut,*) 'tid:', tid, '  idom:', idom


      ! copy setup-once constraints, enefunc
      !

      enefunc(tid)     = enefuncO
      constraints(tid) = constraintsO


      ! get domain index information
      !

      call di_get_index(idom, domain_index(tid))


      ! setup domain
      !

      local_my_rank  = idom - 1

      call setup_md_peer(local_my_rank,     &
                         ctrl_data,         &
                         par,               &
                         domain_index(tid), &
                         boundary,          &
                         enefunc(tid),      &
                         constraints(tid),  &
                         domain(tid))


      ! create ranked filename
      !

      !$omp critical
      my_world_rank = local_my_rank
      my_city_rank  = local_my_rank
      filename      = pio_get_ranked_filename(ctrl_data%out_info%rstfile)
      !$omp end critical


      ! output domain restart file
      !

      call setup_the(t0_info(tid), ctrl_data, domain(tid))

      call pio_write_domain_rst(filename,         &
                                boundary,         &
                                domain(tid),      &
                                enefunc(tid),     &
                                constraints(tid), &
                                dynvars,          &
                                dynamics,         &
                                t0_info(tid))
    end do

    !
    !$omp end parallel do
    !$omp barrier

    write(MsgOut,'(a)') '  done.'
    write(MsgOut,'(a)') ''

    main_rank     = .true.
    nproc_world   = 1
    nproc_city    = 1
    nproc_country = 1

    deallocate(constraints, enefunc, domain, domain_index, t0_info)

    return

  end subroutine output_domain_restart_files

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output pdb file
  !> @brief        output cached molecule coordinates as PDB file
  !! @authors      NT
  !! @param[in]    ctrl_data : information of control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_pdb_file(ctrl_data)
    
    ! formal arguments
    type(s_ctrl_data),       intent(in) :: ctrl_data

    integer                  :: old_mno
    integer                  :: i, pdb


    if (ctrl_data%out_info%pdbfile == '') &
         return

    old_mno = cm_molecule_no(1)

    call open_file(pdb, ctrl_data%out_info%pdbfile, IOFileOutputReplace)

    do i = 1, cm_num_atoms

      if (old_mno /= cm_molecule_no(i)) then
        write(pdb, '("TER ",i7,6x,a4,i6,48x,a1)') &
                   mod(cm_atom_no(i)-1,1000000),  &
                   cm_residue_name(i), &
                   cm_residue_no(i),   &
                   ' '
        old_mno = cm_molecule_no(i)
      end if

      write(pdb, '("ATOM",i7,1x,a4,1x,a4,i6,3x,3f8.3,f6.2,f6.2,6x,a4)') &
                 mod(cm_atom_no(i)-1,1000000), &
                 cm_atom_name(i),    &
                 cm_residue_name(i), &
                 cm_residue_no(i),   &
                 cm_atom_coord(1,i), &
                 cm_atom_coord(2,i), &
                 cm_atom_coord(3,i), &
                 1.0_wp,             &
                 1.0_wp,             &
                 cm_segment_name(i)
    end do

    call close_file(pdb)

    return

  end subroutine output_pdb_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_the
  !> @brief        setup the parallel I/O local information
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_the(t0_info, ctrl_data, domain)

    ! formal arguments
    type(s_pio_t0_info),     intent(inout) :: t0_info
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_domain),  target, intent(in)    :: domain

    ! local variables
    integer                  :: nvar


    t0_info%input_topfile         = ctrl_data%inp_info%topfile
    t0_info%input_parfile         = ctrl_data%inp_info%parfile
    t0_info%energy_pairlistdist   = ctrl_data%ene_info%pairlistdist
    t0_info%energy_table          = ctrl_data%ene_info%table
    t0_info%energy_watermodel     = ctrl_data%ene_info%water_model
    t0_info%constraint_rigidbond  = ctrl_data%cons_info%rigid_bond
    t0_info%constraint_fastwater  = ctrl_data%cons_info%fast_water
    t0_info%constraint_watermodel = ctrl_data%cons_info%water_model
    t0_info%ensemble_type         = ctrl_data%ens_info%ensemble
    t0_info%boundary_boxsizex     = ctrl_data%bou_info%pbc_info%box_size_x
    t0_info%boundary_boxsizey     = ctrl_data%bou_info%pbc_info%box_size_y
    t0_info%boundary_boxsizez     = ctrl_data%bou_info%pbc_info%box_size_z
    t0_info%boundary_originx      = ctrl_data%bou_info%origin_x
    t0_info%boundary_originy      = ctrl_data%bou_info%origin_y
    t0_info%boundary_originz      = ctrl_data%bou_info%origin_z

    if (ctrl_data%opt_info%domain_xyz /= 0) then
      t0_info%boundary_domain_xyz = ctrl_data%opt_info%domain_xyz
    else
      t0_info%boundary_domain_xyz = &
                      ctrl_data%bou_info%domain_x * &
                      ctrl_data%bou_info%domain_y * &
                      ctrl_data%bou_info%domain_z
    end if

    if (allocated(ctrl_data%sel_info%groups)) then
      nvar = size(ctrl_data%sel_info%groups)
    else
      nvar = 0
    end if
    if (allocated(t0_info%selection_group)) then
      if (size(t0_info%selection_group) /= nvar) then
        deallocate(t0_info%selection_group)
        allocate  (t0_info%selection_group(nvar))
      end if
    else
      allocate(t0_info%selection_group(nvar))
    end if
    t0_info%selection_group(1:nvar) = ctrl_data%sel_info%groups(1:nvar)

    if (allocated(ctrl_data%res_info%function)) then
      nvar = size(ctrl_data%res_info%function)
    else
      nvar = 0
    end if
    if (allocated(t0_info%restraint_func)) then
      if (size(t0_info%restraint_func) /= nvar) then
        deallocate(t0_info%restraint_func,  &
                   t0_info%restraint_const, &
                   t0_info%restraint_index)
        allocate  (t0_info%restraint_func(nvar),  &
                   t0_info%restraint_const(nvar), &
                   t0_info%restraint_index(nvar))
      end if
    else
      allocate(t0_info%restraint_func(nvar),  &
               t0_info%restraint_const(nvar), &
               t0_info%restraint_index(nvar))
    end if
    t0_info%restraint_func(1:nvar)    = ctrl_data%res_info%function(1:nvar)
    t0_info%restraint_const(1:nvar)   = ctrl_data%res_info%constant(1:nvar)
    t0_info%restraint_index(1:nvar)   = ctrl_data%res_info%select_index(1:nvar)

    return

  end subroutine setup_the

end module pr_run_prst_setup_mod

