!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_assign_velocity_mod
!> @brief   generate initial velocities and remove trans- and rotational motions
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_assign_velocity_mod

  use random_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: initial_velocity
  public  :: stop_trans_rotation

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    initial_velocity
  !> @brief        generate initial velocities
  !! @authors      JJ
  !! @param[in]    firstt   : initial temperature
  !! @param[in]    natom    : number of atom
  !! @param[in]    ncell    : total number of cells in each domain
  !! @param[in]    id_g2l   : map of atom index (global -> local)
  !! @param[in]    mass     : mass in each domain
  !! @param[inout] iseed    : random number seed
  !! @param[inout] velocity : velocities generated
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine initial_velocity(firstt, natom, ncell, id_g2l, mass, &
                              iseed, velocity)

    ! formal arguments
    real(wp),                intent(in)    :: firstt
    integer,                 intent(in)    :: natom
    integer,                 intent(in)    :: ncell
    integer,                 intent(in)    :: id_g2l(:,:)
    real(wp),                intent(in)    :: mass(:,:)
    integer,                 intent(inout) :: iseed
    real(wp),                intent(inout) :: velocity(:,:,:)

    ! local variables
    real(wp)                 :: alpha1, beta1, alpha2, beta2, alpha3, beta3
    real(wp)                 :: sigma, myu, kBT
    integer                  :: i, ix, icel


    if (main_rank) then
      write(MsgOut,'(A)') 'Initial_Velocity> Generate initial velocities'
      write(MsgOut,'(A20,I10)')   '  iseed           = ', iseed
      write(MsgOut,'(A20,F10.3)') '  temperature     = ', firstt
      write(MsgOut,'(A)') ' '
    end if

    ! generate initial velocities 
    !
    kBT = KBOLTZ * firstt
    myu = 0.0_wp

    do i = 1, natom

       alpha1 = sqrt( -2.0_wp * log(random(iseed)) )
       beta1  = cos ( 2.0_wp * PI * random(iseed) )

       alpha2 = sqrt( -2.0_wp * log(random(iseed)) )
       beta2  = cos ( 2.0_wp * PI * random(iseed) )

       alpha3 = sqrt( -2.0_wp * log(random(iseed)) )
       beta3  = cos ( 2.0_wp * PI * random(iseed) )

       if (id_g2l(1,i) <= ncell .and. id_g2l(2,i) /=0) then
          icel  = id_g2l(1,i)
          ix    = id_g2l(2,i)
          sigma = sqrt( kBT / mass(ix,icel) )
          velocity(1,ix,icel) = sigma * alpha1 * beta1 + myu
          velocity(2,ix,icel) = sigma * alpha2 * beta2 + myu
          velocity(3,ix,icel) = sigma * alpha3 * beta3 + myu
       end if

    end do
   
    return 

  end subroutine initial_velocity

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    stop_trans_rotation
  !> @brief        remove trans and rotational motion about the center of mass
  !! @authors      JJ
  !! @param[in]    ncell      : number of cell
  !! @param[in]    natom      : number of atom in each cell
  !! @param[in]    stop_trans : flag for stop translational motion
  !! @param[in]    stop_rot   : flag for stop rotational motion
  !! @param[in]    mass       : mass in each domain
  !! @param[in]    coord      : coordinates in each domain
  !! @param[inout] velocity   : velocities in each domain
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine stop_trans_rotation(ncell, natom, stop_trans, stop_rot, &
                                 mass, coord, velocity)

    ! formal arguments
    integer,                 intent(in)    :: ncell
    integer,                 intent(in)    :: natom(:)
    logical,                 intent(in)    :: stop_trans
    logical,                 intent(in)    :: stop_rot
    real(wp),                intent(in)    :: mass(:,:)
    real(wp),                intent(in)    :: coord(:,:,:)
    real(wp),                intent(inout) :: velocity(:,:,:)

    ! local variables
    real(wp)                 :: massi, sum_m
    real(wp)                 :: sum_xm, sum_ym, sum_zm
    real(wp)                 :: x_cm, y_cm, z_cm
    real(wp)                 :: sum_vxm, sum_vym, sum_vzm
    real(wp)                 :: vx_cm, vy_cm, vz_cm
    real(wp)                 :: lx_cm, ly_cm, lz_cm
    real(wp)                 :: c1, c2, c3, v1, v2, v3
    real(wp)                 :: xx, xy, xz, yy, yz, zz, x, y, z
    real(wp)                 :: it(3,3), iit(3,3)
    real(wp)                 :: a(3,6), a_max, a_temp, coef, pivot
    real(wp)                 :: omg_x, omg_y, omg_z
    real(wp)                 :: ke_cm
    integer                  :: i, j, k, ix
    integer                  :: maxi


    ! calculate the coordinates of the center of mass
    !
    sum_m  = 0.0_wp
    sum_xm = 0.0_wp
    sum_ym = 0.0_wp
    sum_zm = 0.0_wp

    do i = 1, ncell
      do ix = 1, natom(i)
        massi  = mass(ix,i)
        sum_m  = sum_m  + massi
        sum_xm = sum_xm + coord(1,ix,i) * massi
        sum_ym = sum_ym + coord(2,ix,i) * massi
        sum_zm = sum_zm + coord(3,ix,i) * massi
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, sum_m, 1, mpi_wp_real,  &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, sum_xm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, sum_ym, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, sum_zm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
#endif

    x_cm = sum_xm / sum_m
    y_cm = sum_ym / sum_m
    z_cm = sum_zm / sum_m


    ! calculate the velocity about the center of mass
    ! 
    sum_vxm = 0.0_wp
    sum_vym = 0.0_wp
    sum_vzm = 0.0_wp

    do i = 1, ncell
      do ix = 1, natom(i)
        massi   = mass(ix,i)
        sum_vxm = sum_vxm + velocity(1,ix,i) * massi
        sum_vym = sum_vym + velocity(2,ix,i) * massi
        sum_vzm = sum_vzm + velocity(3,ix,i) * massi
      end do
    end do
   
#ifdef MPI
    call mpi_allreduce(mpi_in_place, sum_vxm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, sum_vym, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, sum_vzm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
#endif

    vx_cm = sum_vxm / sum_m
    vy_cm = sum_vym / sum_m
    vz_cm = sum_vzm / sum_m
    ke_cm = 0.5_wp * sum_m * (vx_cm*vx_cm + vy_cm*vy_cm + vz_cm*vz_cm)


    ! calculate the angular momentum (L) about the center of mass
    !    
    lx_cm = 0.0_wp
    ly_cm = 0.0_wp
    lz_cm = 0.0_wp

    do i = 1, ncell
      do ix = 1, natom(i)
        massi = mass(ix,i)
        c1 = coord(1,ix,i) - x_cm
        c2 = coord(2,ix,i) - y_cm
        c3 = coord(3,ix,i) - z_cm
        v1 = velocity(1,ix,i) - vx_cm
        v2 = velocity(2,ix,i) - vy_cm
        v3 = velocity(3,ix,i) - vz_cm
        lx_cm = lx_cm + (c2*v3 - c3*v2) * massi
        ly_cm = ly_cm + (c3*v1 - c1*v3) * massi
        lz_cm = lz_cm + (c1*v2 - c2*v1) * massi
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, lx_cm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, ly_cm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, lz_cm, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
#endif

    ! calculate the inertia tensor (I)
    !
    xx = 0.0_wp
    xy = 0.0_wp
    xz = 0.0_wp
    yy = 0.0_wp
    yz = 0.0_wp
    zz = 0.0_wp

    do i = 1, ncell
      do ix = 1, natom(i)
        massi = mass(ix,i)
        x  = coord(1,ix,i) - x_cm
        y  = coord(2,ix,i) - y_cm
        z  = coord(3,ix,i) - z_cm
        xx = xx + x * x * massi
        xy = xy + x * y * massi
        xz = xz + x * z * massi
        yy = yy + y * y * massi
        yz = yz + y * z * massi
        zz = zz + z * z * massi
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, xx, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, xy, 1, mpi_wp_real, & 
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, xz, 1, mpi_wp_real, & 
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, yy, 1, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, yz, 1, mpi_wp_real, & 
                       mpi_sum, mpi_comm_country, ierror)
    call mpi_allreduce(mpi_in_place, zz, 1, mpi_wp_real, & 
                       mpi_sum, mpi_comm_country, ierror)
#endif

    it(1,1) = yy + zz
    it(1,2) = -xy
    it(1,3) = -xz
    it(2,1) = -xy
    it(2,2) = xx + zz
    it(2,3) = -yz
    it(3,1) = -xz
    it(3,2) = -yz
    it(3,3) = xx + yy


    ! calculate the inverse matrix of the inertia tensor
    ! using Gauss-Jordan elimination method (with partial pivoting)
    ! 
    do i = 1, 3
      do j = 1, 3
        a(i,j) = it(i,j)
        if (i == j) a(i,j + 3) = 1.0_wp
        if (i /= j) a(i,j + 3) = 0.0_wp
      end do
    end do

    do k = 1, 3
      a_max = a(k,k)
      maxi = k
      do i = 1, 3
        if (abs( a(i,k) ) > a_max) then
          a_max = a(i,k)
          maxi = i
        end if
      end do

      do j = k, 6
        a_temp = a(k,j)
        a(k,j) = a(maxi,j)
        a(maxi,j) = a_temp
      end do

      pivot = 1.0_wp / a(k,k)
      do j = k, 6
        a(k,j) = a(k,j) * pivot
      end do
      do i = 1, 3
        if (i /= k) then
          coef = a(i,k)
          do j = k, 6
            a(i,j) = a(i,j) - a(k,j) * coef
          end do
        end if
      end do
    end do

    do i = 1, 3
      do j = 1, 3
        iit(i,j) = a(i,j + 3)
      end do
    end do

    ! calculate the angular velocity (OMG)
    !
    omg_x = iit(1,1)*lx_cm + iit(1,2)*ly_cm + iit(1,3)*lz_cm
    omg_y = iit(2,1)*lx_cm + iit(2,2)*ly_cm + iit(2,3)*lz_cm
    omg_z = iit(3,1)*lx_cm + iit(3,2)*ly_cm + iit(3,3)*lz_cm


    ! remove translational motion
    !
    if (stop_trans) then
      do i = 1, ncell
        do ix = 1, natom(i)
          velocity(1,ix,i) = velocity(1,ix,i) - vx_cm
          velocity(2,ix,i) = velocity(2,ix,i) - vy_cm
          velocity(3,ix,i) = velocity(3,ix,i) - vz_cm
        end do
      end do
    end if

    ! remove rotational motion
    !
    if (stop_rot) then
      do i = 1, ncell
        do ix = 1, natom(i)
          c1 = coord(1,ix,i) - x_cm
          c2 = coord(2,ix,i) - y_cm
          c3 = coord(3,ix,i) - z_cm
          velocity(1,ix,i) = velocity(1,ix,i) - (omg_y * c3 - omg_z * c2)
          velocity(2,ix,i) = velocity(2,ix,i) - (omg_z * c1 - omg_x * c3)
          velocity(3,ix,i) = velocity(3,ix,i) - (omg_x * c2 - omg_y * c1)
        end do
      end do
    end if

    return

  end subroutine stop_trans_rotation

end module sp_assign_velocity_mod
