!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_energy_pme
!> @brief   Smooth particle mesh ewald method
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_energy_pme_mod

  use sp_energy_table_linear_mod
  use sp_energy_table_point_mod
  use sp_energy_table_cubic_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use fft3d_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  real(wp),         save :: el_fact      ! e^2/(4*pi*eps) A*kcal/mol
  real(wp),         save :: alpha        ! alpha
  real(wp),         save :: alpha2m      ! -alpha^2
  real(wp),         save :: alpha2sp     ! 2*alpha/sqrt(pi)
  real(wp),         save :: vol_fact2    ! (2*pi/V)*el_fact
  real(wp),         save :: vol_fact4    ! (4*pi/V)*el_fact
  real(wp),         save :: u_self       ! Ewald self energy
  real(wp),         save :: box(3)       ! box size
  real(wp),         save :: box_inv(3)   ! Inverse of box size
  real(wp),         save :: bs_fact      ! B-spline factor (1/(n-1)...2)
  real(wp),         save :: bs_fact3     ! bs_fact^3
  real(wp),         save :: bs_fact3d    ! bs_fact^3*(n-1)
  real(wp),         save :: r_scale(3)   ! coordinate-scaling factor (I/L)
  integer,          save :: n_bs         ! Order of B-spline
  integer,          save :: ngrid(3)     ! Number of grid
  integer,          save :: nx           ! process number in x dimension
  integer,          save :: ny           ! process number in y dimension
  integer,          save :: nz           ! process number in z dimension
  integer,          save :: x_start
  integer,          save :: x_end
  integer,          save :: x_local  
  integer,          save :: x_start1
  integer,          save :: x_end1
  integer,          save :: x_local1 
  integer,          save :: y_start
  integer,          save :: y_end
  integer,          save :: y_local  
  integer,          save :: z_start
  integer,          save :: z_end
  integer,          save :: z_local  
  integer,          save :: nlocalx
  integer,          save :: nlocalx1
  integer,          save :: nlocaly
  integer,          save :: nlocalz
  integer,          save :: maxproc
  integer,          save :: ngridmax

  real(wp),         save, allocatable :: b2(:,:)        ! b^2(hx,hy,hz)
  real(wp),         save, allocatable :: gx(:)
  real(wp),         save, allocatable :: gy(:)
  real(wp),         save, allocatable :: gz(:,:)
  real(wp),         save, allocatable :: vir_fact(:,:,:)! -2*(1+G^2/4a)/G^2
  real(wp),         save, allocatable :: theta(:,:,:,:) ! F^-1[Theta](hz,hy,hx,procz)
  real(wp),         save, allocatable :: f(:,:,:)
  real(wp),         save, allocatable :: v(:,:,:)
  real(wp),         save, allocatable :: bsc(:,:,:,:)
  real(wp),         save, allocatable :: bscd(:,:,:,:)
  real(wp),         save, allocatable :: qdf(:,:,:,:)
  real(wp),         save, allocatable :: qdf_real(:)
  real(wp),         save, allocatable :: qdf_work(:,:)
  complex(wp),      save, allocatable :: ftqdf(:)
  complex(wp),      save, allocatable :: ftqdf_work(:,:)

  !$omp threadprivate(f)
  !$omp threadprivate(bsc)
  !$omp threadprivate(bscd)

#ifdef FFT3D
  complex(wp),      save, allocatable :: ftqdf2(:)
  complex(wp),      save, allocatable :: ftqdf3(:)
  complex(wp),      save, allocatable :: ftqdf4(:)
  complex(wp),      save, allocatable :: ftqdf5(:)
  complex(wp),      save, allocatable :: ftqdf_work2(:,:,:)
  complex(wp),      save, allocatable :: ftqdf_work3(:,:,:)
#endif

  ! parameters
  integer, parameter      :: NumIndex        = 45
  integer, parameter      :: Index(NumIndex) = (/ &
                              1,   2,   3,   4,   5,   6,   8,   9,  10,  12, &
                             15,  16,  18,  20,  24,  25,  27,  30,  32,  36, &
                             40,  45,  48,  50,  54,  60,  64,  72,  75,  80, &
                             81,  90,  96, 100, 120, 125, 128, 135, 144, 150, &
                            160, 162, 180, 192, 200/)

  ! subroutines
  public  :: setup_pme
  public  :: dealloc_pme
  public  :: compute_energy_nonbond_pme
  public  :: compute_energy_nonbond_pme_short
  public  :: compute_energy_nonbond_pme_long
  public  :: pme_pre
  private :: pme_nb14
  private :: pme_bond_corr_table
  private :: pme_bond_corr_linear
  private :: pme_bond_corr_point
  private :: pme_bond_corr
  private :: pme_real
  private :: pme_recip
  private :: b_spline_coef
  private :: b_spline_dev_coef

  public  :: erf_n
  public  :: erfc04

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_pme
  !> @brief        Setup for PME with domain decomposition
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    boundary : boundary information
  !! @param[inout] enefunc  : energy function
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_pme(domain, boundary, enefunc)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_boundary),        intent(in)    :: boundary
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: bcos, bsin, fact
    integer                  :: i, j, k, js
    integer                  :: localmax, ncell
    integer                  :: iproc(3), index_x, index_y, index_z
    integer                  :: remainder, quotient, expo
    integer                  :: alloc_stat

    real(wp),    allocatable :: bs(:)


    if (boundary%type /= BoundaryTypePBC) &
      call error_msg('Setup_PME> Error! BoundaryType is not PBC.')

    ncell    =  domain%num_cell_local + domain%num_cell_boundary

    ! Setting the number of grids (even number)
    !
    ngrid(1) = enefunc%pme_ngrid_x
    ngrid(2) = enefunc%pme_ngrid_y
    ngrid(3) = enefunc%pme_ngrid_z
    ngridmax = max(ngrid(1),ngrid(2),ngrid(3))

    ! grid partition
    !
    if (reciprocal_calc) then

      nx = boundary%num_domain(1)
      ny = boundary%num_domain(2)
      nz = boundary%num_domain(3)

      ! Check grid point in x direction
      !
      remainder = mod(ngrid(1),2*nx)

      if (remainder /= 0) ngrid(1) = ngrid(1) + 2*nx - remainder

      quotient = ngrid(1)/(2*nx)

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(1) = (2*nx) * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(1) = (2*nx) * quotient

      end if

#ifdef FFT3D

      ! Check grid point in y direction
      !
      remainder = mod(ngrid(2),ny*nz)

      if (remainder /= 0) ngrid(2) = ngrid(2) + nz*ny - remainder

      quotient = ngrid(2)/(nz*ny)

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(2) = (nz*ny) * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(2) = (nz*ny) * quotient

      end if

      ! Check grid point in z direction
      !
      remainder = mod(ngrid(3),nz*nx)

      if (remainder /= 0) ngrid(3) = ngrid(3) + nz*nx - remainder

      quotient = ngrid(3)/(nz*nx)

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(3) = (nz*nx) * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(3) = (nz*nx) * quotient

      end if

      remainder = mod(ngrid(3),nz*ny)

      if (remainder /= 0) ngrid(3) = ngrid(3) + nz*ny - remainder

      quotient = ngrid(3)/(nz*ny)

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(3) = (nz*ny) * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(3) = (nz*ny) * quotient

      end if

#else
      ! Check grid point in y direction
      !
      remainder = mod(ngrid(2),2*ny)

      if (remainder /= 0) ngrid(2) = ngrid(2) + 2*ny - remainder

      quotient = ngrid(2)/(2*ny)

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(2) = (2*ny) * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(2) = (2*ny) * quotient

      end if

      ! Check grid point in z direction
      !
      remainder = mod(ngrid(3),nz)

      if (remainder /= 0) ngrid(3) = ngrid(3) + nz - remainder

      quotient = ngrid(3)/nz

      if (quotient <= Index(NumIndex)) then

        do i = 1, NumIndex
          if (quotient <= Index(i)) exit
        end do

        quotient = Index(i)
        ngrid(3) = nz * quotient

      else

        expo = int(log(real(quotient,wp))/log(real(2,wp)))

        if (2**expo >= quotient) then
          quotient = 2**expo
        else
          quotient = 2**(expo+1)
        end if

        ngrid(3) = nz * quotient

      end if
#endif

      ! Show warning messages
      !
      if ((enefunc%pme_ngrid_x /= ngrid(1)) .or. &
          (enefunc%pme_ngrid_y /= ngrid(2)) .or. &
          (enefunc%pme_ngrid_z /= ngrid(3))) then

        if (main_rank) then
          write(MsgOut,'(A)') &
            ''
          write(MsgOut,'(A)') &
            'Setup_PME> PME grid number was adjusted automatically'
          write(MsgOut,'(A20,3I10)') &
            '  pme_ngrid(x,y,z)= ', ngrid(1), ngrid(2), ngrid(3)
          write(MsgOut,'(A)') &
            ''
        end if
      end if


      ngridmax = max(ngrid(1),ngrid(2),ngrid(3))

      ! index for each processor
      !

      iproc(1) = mod(my_city_rank, boundary%num_domain(1))
      iproc(2) = mod(my_city_rank/boundary%num_domain(1),boundary%num_domain(2))
      iproc(3) = my_city_rank/(boundary%num_domain(1)*boundary%num_domain(2))

      nlocalx  = int((ngrid(1)+nx-1)/nx)
      nlocaly  = int((ngrid(2)+ny-1)/ny)
      nlocalz  = int((ngrid(3)+nz-1)/nz)

      x_start  = (iproc(1))*nlocalx + 1
      y_start  = (iproc(2))*nlocaly + 1
      z_start  = (iproc(3))*nlocalz + 1

      x_end    = (iproc(1)+1)*nlocalx 
      y_end    = (iproc(2)+1)*nlocaly 
      z_end    = (iproc(3)+1)*nlocalz 

      x_local  = nlocalx
      y_local  = nlocaly
      z_local  = nlocalz

      localmax = max(nlocalx*nlocaly,nlocalx*nlocalz,nlocaly*nlocalz)
      index_x  = iproc(2)*nlocaly*nlocalz + iproc(3)
      index_y  = iproc(3)*nlocalx*nlocalz + iproc(1)
      index_z  = iproc(1)*nlocalx*nlocaly + iproc(2)

      ! new communicator according to the grid index
      !
#ifdef MPI
      call mpi_comm_split(mpi_comm_city, index_x,my_city_rank,grid_commx,ierror)
      call mpi_comm_size (grid_commx, nprocx, ierror)
      call mpi_comm_rank (grid_commx, my_x_rank, ierror)
      call mpi_comm_split(mpi_comm_city, index_y,my_city_rank,grid_commy,ierror)
      call mpi_comm_size (grid_commy, nprocy, ierror)
      call mpi_comm_rank (grid_commy, my_y_rank, ierror)
      call mpi_comm_split(mpi_comm_city, index_z,my_city_rank,grid_commz,ierror)
      call mpi_comm_size (grid_commz, nprocz, ierror)
      call mpi_comm_rank (grid_commz, my_z_rank, ierror)
#endif
      maxproc = max(nprocx,nprocy,nprocz,1)

      nlocalx1 = nlocalx/2 + 1
      if (iproc(1) == 0) then
        x_local1 = nlocalx1
        x_start1 = 1
        x_end1   = x_local1
      else
        x_local1 = nlocalx1 - 1
        x_start1 = iproc(1)*nlocalx/2 + 2
        x_end1   = x_start1 + x_local1 - 1
      end if
    end if

    ! Setting common parameters
    !
    el_fact  = ELECOEF/enefunc%dielec_const
    n_bs     = enefunc%pme_nspline
    alpha    = enefunc%pme_alpha
    alpha2m  = -alpha**2
    alpha2sp = 2.0_wp*alpha/sqrt(PI)

    if (reciprocal_calc) then

      j = 1
      do i = 1, n_bs - 2
        j = j*(n_bs - i)
      end do

      bs_fact   = 1.0_wp/real(j,wp)
      bs_fact3  = bs_fact**3
      bs_fact3d = bs_fact3 * real(n_bs - 1,wp)

      ! Preparing b2=b(h)^2, h shifted
      !
      allocate(bs(n_bs), b2(ngridmax,3))

      call b_spline_coef(n_bs, 1.0_wp, bs)

      do i = 1, n_bs - 1
        bs(i) = bs(i)*bs_fact
      end do

      do k = 1, 3

        fact = 2.0_wp * PI/real(ngrid(k),wp)

        do j = 1, ngrid(k)/2 + 1

          js = j - 1
          bcos = 0.0_wp
          bsin = 0.0_wp

          do i = 0, n_bs - 2
            bcos = bcos + bs(i+1) * cos(real(js*i,wp)*fact)
            bsin = bsin + bs(i+1) * sin(real(js*i,wp)*fact)
          end do

          b2(j, k) = 1.0_wp/(bcos**2 + bsin**2)

        end do

        do j = ngrid(k)/2 + 2, ngrid(k)

          js = j - 1 - ngrid(k)
          bcos = 0.0_wp
          bsin = 0.0_wp

          do i = 0, n_bs - 2
            bcos = bcos + bs(i+1) * cos(real(js*i,wp)*fact)
            bsin = bsin + bs(i+1) * sin(real(js*i,wp)*fact)
          end do

          b2(j, k) = 1.0_wp/(bcos**2 + bsin**2)

        end do

      end do

      if (n_bs/2 == 1) then
        do k = 1, 3
          b2(ngrid(k)/2 + 1,k) = 0.0_wp
        end do
      end if

      ! Prepareing theta = F^-1[theta](h), h shifted
      !
      allocate(gx(x_local1), gy(nlocaly), gz(nlocalz,nprocz), &
               vir_fact(nlocalz, nlocaly, x_local1),          &
               theta(nlocalz,nlocaly,x_local1,nprocz),        &
               v(3,MaxAtom,ncell),                            &
               qdf(nlocalx,nlocaly,nlocalz,nthread),          &
               qdf_real(nlocalx*nlocaly*nlocalz),             &
               qdf_work(nlocalx*nlocaly*nlocalz,maxproc),     &
               ftqdf(nlocalx1*nlocaly*nlocalz),               &
               ftqdf_work(nlocalx1*nlocaly*nlocalz,maxproc),  &
               stat = alloc_stat)
      if (alloc_stat /= 0) call error_msg_alloc

#ifdef FFT3D
      allocate(ftqdf2(nlocalx1*nlocaly*nlocalz),              &
               ftqdf3(nlocalx1*nlocaly*nlocalz),              &
               ftqdf4(nlocalx1*nlocaly*nlocalz),              &
               ftqdf5(nlocalx1*nlocaly*nlocalz),              &
               ftqdf_work2(ngrid(3),nlocalx1,nlocaly),        &
               ftqdf_work3(ngrid(3),nlocalx1,nlocaly),        &
               stat = alloc_stat)
      if (alloc_stat /= 0) call error_msg_alloc
#endif

      !$omp parallel private(alloc_stat)

      allocate(f(3,MaxAtom,ncell),                                   &
               bsc(n_bs,3,MaxAtom,ncell),bscd(n_bs,3,MaxAtom,ncell), &
               stat = alloc_stat)
      if (alloc_stat /= 0)   call error_msg_alloc

      !$omp end parallel

      deallocate(bs)

    end if

    return

  end subroutine setup_pme

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_pme
  !> @brief        deallocate all arrays used in PME
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_pme

    ! local variables
    integer :: dealloc_stat


    if (.not. allocated(f)) &
      return 

    !$omp parallel private(dealloc_stat)

    deallocate(f, bsc, bscd, stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc

    !$omp end parallel

#ifdef FFT3D
    deallocate(ftqdf2, ftqdf3, ftqdf4, ftqdf5, ftqdf_work2, ftqdf_work3, &
               stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc
#endif

    deallocate(b2, gx, gy, gz, vir_fact, theta, v, qdf, qdf_real, qdf_work, &
               ftqdf, ftqdf_work, stat = dealloc_stat)
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_pme

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pme
  !> @brief        Calculate nonbond energy by PME
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    pairlist : pair-list information
  !! @param[in]    boundary : boundary information
  !! @param[in]    npt      : flag for NPT or not
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pme(domain, enefunc, pairlist, boundary, &
                                      npt, nonb_ene, force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: npt
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)


    call timer(TimerNonBond, TimerOn)
 
    if (real_calc) then

      call timer(TimerPmeReal, TimerOn)

      if (enefunc%table%table) then
 
        if (enefunc%table%table_order == 3) then 

          call compute_energy_nonbond_nb14_table(                  &
                                        domain, enefunc,           &
                                        force, virial, eelec, evdw)

        else if (enefunc%table%table_order == 1) then

          call compute_energy_nonbond_nb14_table_linear(           &
                                        domain, enefunc,           &
                                        force, virial, eelec, evdw)

        else if (enefunc%table%table_order == 0) then

          call compute_energy_nonbond_nb14_table_linear(           &
                                        domain, enefunc,           &
                                        force, virial, eelec, evdw)

        end if

        if (nonb_ene) then

          if (enefunc%table%table_order == 3) then 

            if (enefunc%table%water_table) then

              call compute_energy_nonbond_table_solute(            &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

              call compute_energy_nonbond_table_water(             &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            else

              call compute_energy_nonbond_table(                   &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            end if

          else if (enefunc%table%table_order == 1) then

            if (enefunc%table%water_table) then

              call compute_energy_nonbond_table_solute_linear(     &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

              call compute_energy_nonbond_table_water_linear(      &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            else

              call compute_energy_nonbond_table_linear(            &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            end if

          else if (enefunc%table%table_order == 0) then

            if (enefunc%table%water_table) then

              call compute_energy_nonbond_table_solute_point(      &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

              call compute_energy_nonbond_table_water_point(       &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            else

              call compute_energy_nonbond_table_point(             &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            end if

          end if

        else

          if (enefunc%table%table_order == 3) then 

            if (enefunc%table%water_table) then

              call compute_force_nonbond_table_solute(             &
                                        domain, enefunc, pairlist, &
                                        force, virial)

              call compute_force_nonbond_table_water(              &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            else

              call compute_force_nonbond_table(domain, enefunc, pairlist, &
                                        force, virial)

            end if 

          else if (enefunc%table%table_order == 1) then

            if (enefunc%table%water_table) then

              call compute_force_nonbond_table_solute_linear(     &
                                        domain, enefunc,pairlist, &
                                        force, virial)

              call compute_force_nonbond_table_water_linear(       &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            else

              call compute_force_nonbond_table_linear(             &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            end if

          else if (enefunc%table%table_order == 0) then

            if (enefunc%table%water_table) then

              call compute_force_nonbond_table_solute_point(       &
                                        domain, enefunc, pairlist, &
                                        force, virial)

              call compute_force_nonbond_table_water_point(        &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            else

              call compute_force_nonbond_table_point(              &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            end if

          end if

        end if

        if (enefunc%table%table_order == 3) then 

          call pme_bond_corr_table(domain, enefunc, &
                                    force, virial, eelec)

        else if (enefunc%table%table_order == 1) then

          call pme_bond_corr_linear(domain, enefunc, &
                                        force, virial, eelec)

        else if (enefunc%table%table_order == 0) then

          call pme_bond_corr_point(domain, enefunc, &
                                        force, virial, eelec)

        end if

      else

        call pme_nb14(domain, enefunc, force, virial, eelec, evdw)

        call pme_real(domain, enefunc, pairlist, force, virial, eelec, evdw)

        call pme_bond_corr(domain, enefunc, force, virial, eelec)

      end if

      call timer(TimerPmeReal, TimerOff)

    end if

    if (reciprocal_calc) then

      call timer(TimerPmeRecip, TimerOn)

      if (npt) call pme_pre(domain, boundary)
      call pme_recip(domain, force, virial, eelec)

      call timer(TimerPmeRecip, TimerOff)

    end if

    ! Add self energy
    !
    eelec(1) = eelec(1) + u_self

    call timer(TimerNonBond, TimerOff)
 
    return

  end subroutine compute_energy_nonbond_pme

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pme_short
  !> @brief        Calculate nonbond energy (real part) by PME
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    pairlist : pair-list information
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pme_short(domain, enefunc, pairlist, &
                                           nonb_ene, force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)


    call timer(TimerNonBond, TimerOn)

    ! real part
    if (real_calc) then

      call timer(TimerPmeReal, TimerOn)

      if (enefunc%table%table) then

        if (enefunc%table%table_order == 3) then

          call compute_energy_nonbond_nb14_table(domain, enefunc, &
                                        force, virial, eelec, evdw)

        else if (enefunc%table%table_order == 1) then

          call compute_energy_nonbond_nb14_table_linear(domain, enefunc, &
                                        force, virial, eelec, evdw)
        end if

        if (nonb_ene) then

          if (enefunc%table%table_order == 3) then

            if (enefunc%table%water_table) then

              call compute_energy_nonbond_table_solute(            &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

              call compute_energy_nonbond_table_water(             &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            else

              call compute_energy_nonbond_table(                   &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            end if

          else if (enefunc%table%table_order == 1) then

            if (enefunc%table%water_table) then

              call compute_energy_nonbond_table_solute_linear(     &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

              call compute_energy_nonbond_table_water_linear(      &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            else

              call compute_energy_nonbond_table_linear(            &
                                        domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

            end if

          end if

        else

          if (enefunc%table%table_order == 3) then

            if (enefunc%table%water_table) then

              call compute_force_nonbond_table_solute(             &
                                        domain, enefunc, pairlist, &
                                        force, virial)

              call compute_force_nonbond_table_water(              &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            else

              call compute_force_nonbond_table(domain, enefunc, pairlist, &
                                        force, virial)

            end if

          else if (enefunc%table%table_order == 1) then

            if (enefunc%table%water_table) then

              call compute_force_nonbond_table_solute_linear(     &
                                        domain, enefunc,pairlist, &
                                        force, virial)

              call compute_force_nonbond_table_water_linear(       &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            else

              call compute_force_nonbond_table_linear(             &
                                        domain, enefunc, pairlist, &
                                        force, virial)

            end if

          end if

        end if

        if (enefunc%table%table_order == 3) then

          call pme_bond_corr_table(domain, enefunc, &
                                        force, virial, eelec)

          call pme_bond_corr_linear(domain, enefunc, &
                                        force, virial, eelec)

        end if

      else

        call pme_nb14(domain, enefunc, force, virial, eelec, evdw)

        call pme_real(domain, enefunc, pairlist, force, virial, eelec, evdw)

        call pme_bond_corr(domain, enefunc, force, virial, eelec)

      end if

      call timer(TimerPmeReal, TimerOff)

    end if

    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond_pme_short

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pme_long
  !> @brief        Calculate nonbond energy (reciprocal part) by PME
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    boundary : boundary information
  !! @param[in]    npt      : flag for NPT or not
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pme_long(domain, enefunc, boundary, &
                                             npt, force, virial, eelec)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: npt
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)


    call timer(TimerNonBond, TimerOn)

    ! reciprocal part
    if (reciprocal_calc) then

      call timer(TimerPmeRecip, TimerOn)

      if (npt) call pme_pre(domain, boundary)
      call pme_recip(domain, force, virial, eelec)

      call timer(TimerPmeRecip, TimerOff)

    end if

    ! Add self energy
    !
    eelec(1) = eelec(1) + u_self

    call timer(TimerNonBond, TimerOff)

    return

  end subroutine compute_energy_nonbond_pme_long

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_pre
  !> @brief        Prepare functions for PME calculation
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    boundary : boundary information
  !! @note         Extracted from setup_pme for NPT calculation
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_pre(domain, boundary)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_boundary),        intent(in)    :: boundary

    ! local variables
    integer                  :: i, j, k, is, js, ks, ix, iy, iz, iproc
    real(wp)                 :: fact, gfact(3), g2


    box(1) = boundary%box_size_x
    box(2) = boundary%box_size_y
    box(3) = boundary%box_size_z

    do k = 1, 3
      box_inv(k) = 1.0_wp / box(k)
      r_scale(k) = real(ngrid(k),wp) * box_inv(k)
    end do

    vol_fact2 = 2.0_wp * PI * el_fact * box_inv(1)*box_inv(2)*box_inv(3)
    vol_fact4 = 2.0_wp * vol_fact2


    ! Prepareing theta = F^-1[theta](h), h shifted
    ! Gx, Gy, Gz, and vir_fact=-2(1+G^2/4a)/G^2 are also prepared 
    ! for virial calculation
    !
    do k = 1, 3
      gfact(k) = 2.0_wp * PI * box_inv(k)
    end do

    fact = 0.25_wp / alpha2m

    do i = x_start1, x_end1

      ix = i - x_start1 + 1
      is = i - 1
      gx(ix) = gfact(1) * real(is,wp)

      do j = y_start, y_end

        iy = j - y_start + 1

        if (j <= ngrid(2)/2+1) then
          js = j - 1
        else
          js = j - 1 - ngrid(2)
        end if

        gy(iy) = gfact(2) * real(js,wp)

        do iproc = 1, nprocz
          do iz = 1, nlocalz 
            k = (iproc-1)*nlocalz + iz
            if (k <= ngrid(3)/2+1) then
              ks = k - 1
            else 
              ks = k - 1 - ngrid(3)
            end if
            gz(iz,iproc) = gfact(3) * real(ks,wp)
            g2 = gx(ix)**2 + gy(iy)**2 + gz(iz,iproc)**2
            if (iproc == (my_z_rank+1)) &
              vir_fact(iz,iy,ix) = -2.0_wp * (1.0_wp - g2 * fact)/g2
            theta(iz,iy,ix,iproc) = &
                 b2(i,1) * b2(j,2) * b2(k,3) * exp(g2 * fact)/g2
          end do

        end do
      end do
    end do

    if (my_city_rank == 0) &
      vir_fact(1,1,1) = 0.0_wp

    if (my_x_rank == 0 .and. my_y_rank == 0) &
      theta(1,1,1,1) = 0.0_wp

    ! Calculating self energy
    !
    u_self = 0.0_wp

    do i = 1, domain%num_cell_local
      do ix = 1, domain%num_atom(i)
        u_self = u_self + domain%charge(ix,i)**2
      end do 
    end do

    u_self = - u_self * el_fact * alpha/sqrt(PI)

    return

  end subroutine pme_pre

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_nb14
  !> @brief        calculate nonbonded14 energy 
  !  @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  !! @param[inout] evdw    : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_nb14(domain, enefunc, force, virial, eelec, evdw)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij, rij2, inv_r2, inv_rij
    real(wp)                 :: inv_r3, inv_r6, inv_r12
    real(wp)                 :: lj6, lj12
    real(wp)                 :: c1_switchup, c2_switchup
    real(wp)                 :: c1_switchdown, c2_switchdown, c6, c12
    real(wp)                 :: term_elec, term_elec1, switch, dswitch
    real(wp)                 :: term_lj12, term_lj6, term_lj
    real(wp)                 :: cutoff2, switchdist2
    real(wp)                 :: c1_switch, c2_switch, c4_switch, grad_coef
    real(wp)                 :: work(1:3)
    real(wp)                 :: viri_local(1:3,1:3)
    integer                  :: i, ix, iy, j, k, ij
    integer                  :: num_nb14, ini_nb14, fin_nb14
    integer                  :: ncell_local, id, omp_get_thread_num

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    real(wp),        pointer :: nb14_lj12(:,:), nb14_lj6(:,:)
    real(wp),        pointer :: cutoff, switchdist
    integer,         pointer :: cell_pair(:,:), atmcls(:,:), natom(:)
    integer,         pointer :: num_nb14_calc1(:,:), nb14_calc_list1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), nb14_calc_list(:,:)


    coord           => domain%coord
    charge          => domain%charge
    cell_pair       => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom

    cutoff          => enefunc%cutoffdist
    switchdist      => enefunc%switchdist
    nb14_lj12       => enefunc%nb14_lj12
    nb14_lj6        => enefunc%nb14_lj6
    num_nb14_calc1  => enefunc%num_nb14_calc1
    num_nb14_calc   => enefunc%num_nb14_calc
    nb14_calc_list1 => enefunc%nb14_calc_list1
    nb14_calc_list  => enefunc%nb14_calc_list

    ncell_local = domain%num_cell_local

    cutoff2     = cutoff * cutoff
    switchdist2 = switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup = cutoff2*cutoff2*cutoff2
      c2_switchup = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if


    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, ini_nb14, fin_nb14, num_nb14,       &
    !$omp         dij, rij2, rij, c2_switch, c4_switch, switch, dswitch,       &
    !$omp         inv_r6, inv_r12, term_elec1, term_lj12, term_lj6, term_lj,   &
    !$omp         grad_coef, work, term_elec, viri_local, lj12, lj6, inv_rij,  &
    !$omp         inv_r2, inv_r3)          
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_nb14 = 0

      do ix = 1, natom(i) - 1
        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc1(ix,i)
        num_nb14 = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = nb14_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij      = sqrt(rij2)

          ! cutoff
          !
          lj6      = nb14_lj6(atmcls(ix,i),atmcls(iy,i))
          lj12     = nb14_lj12(atmcls(ix,i),atmcls(iy,i))
          inv_r2   = 1.0_wp / rij2
          inv_rij  = 1.0_wp / rij
          inv_r3   = inv_r2 * inv_rij
          inv_r6   = inv_r3 * inv_r3

          if (enefunc%force_switch) then
            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6 - c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              grad_coef  = -12.0_wp * lj12* c12 *   &
                           (inv_r6 - c1_switchup)*inv_r6*inv_r2
              grad_coef  = grad_coef + 6.0_wp * lj6 * c6 *    &
                           (inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              grad_coef  = -12.0_wp * lj12 * inv_r12 * inv_r2
              grad_coef  = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if
          else
            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch * (cutoff2+2.0_wp*rij2-3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if
            inv_r12      = inv_r6 * inv_r6
            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            grad_coef    = ( dswitch*term_lj*rij   &
                           - switch * (12.0_wp*term_lj12-6.0_wp*term_lj6) ) * inv_r2
          end if

          ! electrostatic
          !
          term_elec1     = charge(ix,i)*charge(iy,i)*el_fact
          term_elec      = term_elec1*erfc04(alpha*rij)*inv_rij
          eelec(id+1)    = eelec(id+1) + term_elec 
          grad_coef      = grad_coef - &
                           (term_elec+alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2
          work(1:3) = grad_coef * dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
          
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do
      end do
    end do


    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_nb14 = 0

      do ix = 1, natom(i)
        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc(ix,ij)
        num_nb14 = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = nb14_calc_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij      = sqrt(rij2)

          ! cutoff
          !
          lj6   = nb14_lj6(atmcls(ix,i),atmcls(iy,j))
          lj12  = nb14_lj12(atmcls(ix,i),atmcls(iy,j))
          inv_r2   = 1.0_wp / rij2
          inv_rij  = 1.0_wp / rij
          inv_r3   = inv_r2 * inv_rij
          inv_r6   = inv_r3 * inv_r3

          if (enefunc%force_switch) then
            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6 - c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              grad_coef  = -12.0_wp * lj12* c12 *   &
                           (inv_r6 - c1_switchup)*inv_r6*inv_r2
              grad_coef  = grad_coef + 6.0_wp * lj6 * c6 *    &
                           (inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              grad_coef  = -12.0_wp * lj12 * inv_r12 * inv_r2
              grad_coef  = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if
          else
            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch*(cutoff2 + 2.0_wp*rij2-3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if
            inv_r12      = inv_r6 * inv_r6
            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            grad_coef    = ( dswitch*term_lj*rij   &
                           - switch*(12.0_wp*term_lj12-6.0_wp*term_lj6) )*inv_r2
          end if

          ! electrostatic
          !
          term_elec1     = charge(ix,i)*charge(iy,j)*el_fact
          term_elec      = term_elec1*erfc04(alpha*rij)*inv_rij
          eelec(id+1)    = eelec(id+1) + term_elec
          grad_coef      = grad_coef - &
                           (term_elec+alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2
          work(1:3)      = grad_coef * dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

      end do
    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine pme_nb14

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_table
  !> @brief        calculate bond correction term in PME
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_table(domain, enefunc, force, virial, eelec)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij2, rij
    real(wp)                 :: R, h00, h10, h01, h11
    real(wp)                 :: term_elec
    real(wp)                 :: grad_coef
    real(wp)                 :: work(1:3), viri_local(1:3,1:3)
    integer                  :: i, ix, iy, j, k, ij, L, iwater
    integer                  :: num_excl, ini_excl, fin_excl
    integer                  :: id, omp_get_thread_num
    integer                  :: ncell_local

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    real(wp),        pointer :: table_elec(:), table_delec(:)
    real(wp),        pointer :: dtable_elec(:), dtable_delec(:)
    real(wp),        pointer :: density
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: natom(:), nwater(:), water_list(:,:,:)
    integer,         pointer :: num_nonb_excl1(:,:), nonb_excl_list1(:,:)
    integer,         pointer :: num_nonb_excl(:,:), nonb_excl_list(:,:)


    cell_pair       => domain%cell_pairlist1
    natom           => domain%num_atom
    nwater          => domain%num_water
    water_list      => domain%water_list
    coord           => domain%coord
    charge          => domain%charge

    table_elec      => enefunc%table%table_ecor
    table_delec     => enefunc%table%table_decor
    dtable_elec     => enefunc%table%dtable_ecor
    dtable_delec    => enefunc%table%dtable_decor
    density         => enefunc%table%density
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nonb_excl   => enefunc%num_nonb_excl
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nonb_excl_list  => enefunc%nonb_excl_list

    ncell_local = domain%num_cell_local


    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, ini_excl, fin_excl, num_excl,       &
    !$omp         dij, rij2, L, R, h00, h10, h01, h11, grad_coef, work,        &
    !$omp         term_elec, viri_local)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_excl = 0

      do ix = 1, natom(i) - 1
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = rij2 * density
          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          term_elec = table_elec(L)*h00 + table_elec(L+1)*h01              &
                    + dtable_elec(L)*h10 + dtable_elec(L+1)*h11
          term_elec  = charge(ix,i) * charge(iy,i) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L)*h00 + table_delec(L+1)*h01            &
                    + dtable_delec(L)*h10 + dtable_delec(L+1)*h11
          grad_coef = charge(ix,i) * charge(iy,i) * term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
          
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do
      end do
 
      if (enefunc%table%water_table) then
        do iwater = 1, nwater(i)
          do j = 1, 2
            do k = j+1, 3
              ix   = water_list(j,iwater,i)
              iy   = water_list(k,iwater,i)
              dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
              rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2  = rij2 * density
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
              h10  = R*(1.0_wp-R)*(1.0_wp-R)
              h01  = R*R*(3.0_wp-2.0_wp*R)
              h11  = R*R*(R-1.0_wp)
        
              term_elec = table_elec(L)*h00 + table_elec(L+1)*h01              &
                        + dtable_elec(L)*h10 + dtable_elec(L+1)*h11
              term_elec  = charge(ix,i) * charge(iy,i) * term_elec
              eelec(id+1) = eelec(id+1) + term_elec
        
              term_elec = table_delec(L)*h00 + table_delec(L+1)*h01            &
                        + dtable_delec(L)*h10 + dtable_delec(L+1)*h11
              grad_coef = charge(ix,i) * charge(iy,i) * term_elec
        
              work(1:3) = grad_coef*dij(1:3)
        
              force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
        
              ! virial
              !
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do
        end do
      end if

    end do

    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_excl = 0

      do ix = 1, natom(i)
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij = sqrt(rij2)

          ! energy and gradient
          !
          rij2  = rij2 * density
          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          term_elec = table_elec(L)*h00 + table_elec(L+1)*h01              &
                    + dtable_elec(L)*h10 + dtable_elec(L+1)*h11
          term_elec  = charge(ix,i) * charge(iy,j) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L)*h00 + table_delec(L+1)*h01            &
                    + dtable_delec(L)*h10 + dtable_delec(L+1)*h11
          grad_coef = charge(ix,i) * charge(iy,j) * term_elec

          work(1:3)= grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

      end do
    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel
  
    return

  end subroutine pme_bond_corr_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_linear
  !> @brief        calculate bond correction term in PME
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_linear(domain, enefunc, force, virial, eelec)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij2
    real(wp)                 :: R
    real(wp)                 :: term_elec, grad_coef
    real(wp)                 :: work(1:3)
    real(wp)                 :: viri_local(1:3,1:3), cutoff2
    integer                  :: i, ix, iy, j, k, ij, L, iwater
    integer                  :: num_excl, ini_excl, fin_excl
    integer                  :: id, omp_get_thread_num
    integer                  :: ncell_local

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    real(wp),        pointer :: table_elec(:), table_delec(:)
    real(wp),        pointer :: density, cutoff
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: natom(:), nwater(:), water_list(:,:,:)
    integer,         pointer :: num_nonb_excl1(:,:), nonb_excl_list1(:,:)
    integer,         pointer :: num_nonb_excl(:,:), nonb_excl_list(:,:)


    cell_pair       => domain%cell_pairlist1
    natom           => domain%num_atom
    nwater          => domain%num_water
    water_list      => domain%water_list
    coord           => domain%coord
    charge          => domain%charge

    table_elec      => enefunc%table%table_ecor
    table_delec     => enefunc%table%table_decor
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nonb_excl   => enefunc%num_nonb_excl
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nonb_excl_list  => enefunc%nonb_excl_list

    ncell_local = domain%num_cell_local
    cutoff2     = cutoff*cutoff

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, ini_excl, fin_excl, num_excl,       &
    !$omp         dij, rij2, L, R, grad_coef, work, term_elec, viri_local)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_excl = 0

      do ix = 1, natom(i) - 1
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = cutoff2*density/rij2
          L    = int(rij2)
          R    = rij2 - L

          term_elec = table_elec(L) + R*(table_elec(L+1)-table_elec(L))
          term_elec  = charge(ix,i) * charge(iy,i) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L) + R*(table_delec(L+1)-table_delec(L))
          grad_coef = charge(ix,i) * charge(iy,i) * term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
          
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do
      end do
 
      if (enefunc%table%water_table) then
        do iwater = 1, nwater(i)
          do j = 1, 2
            do k = j+1, 3
              ix   = water_list(j,iwater,i)
              iy   = water_list(k,iwater,i)
              dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
              rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2  = cutoff2*density/rij2 
              L    = int(rij2)
              R    = rij2 - L
       
              term_elec = table_elec(L) + R*(table_elec(L+1)-table_elec(L))
              term_elec  = charge(ix,i) * charge(iy,i) * term_elec
              eelec(id+1) = eelec(id+1) + term_elec
       
              term_elec = table_delec(L) + R*(table_delec(L+1)-table_delec(L))
              grad_coef = charge(ix,i) * charge(iy,i) * term_elec
       
              work(1:3) = grad_coef*dij(1:3)
       
              force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
       
              ! virial
              !
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do
        end do
      end if

    end do

    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_excl = 0

      do ix = 1, natom(i)
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = cutoff2*density/rij2
          L    = int(rij2)
          R    = rij2 - L

          term_elec = table_elec(L) + R*(table_elec(L+1)-table_elec(L))

          term_elec  = charge(ix,i) * charge(iy,j) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L) + R*(table_delec(L+1)-table_delec(L))
          grad_coef = charge(ix,i) * charge(iy,j) * term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

      end do
    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine pme_bond_corr_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_point 
  !> @brief        calculate bond correction term in PME
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_point(domain, enefunc, force, virial, eelec)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij2
    real(wp)                 :: R
    real(wp)                 :: term_elec, grad_coef
    real(wp)                 :: work(1:3)
    real(wp)                 :: viri_local(1:3,1:3), cutoff2
    integer                  :: i, ix, iy, j, k, ij, L, iwater
    integer                  :: num_excl, ini_excl, fin_excl
    integer                  :: id, omp_get_thread_num
    integer                  :: ncell_local

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    real(wp),        pointer :: table_elec(:), table_delec(:)
    real(wp),        pointer :: density, cutoff
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: natom(:), nwater(:), water_list(:,:,:)
    integer,         pointer :: num_nonb_excl1(:,:), nonb_excl_list1(:,:)
    integer,         pointer :: num_nonb_excl(:,:), nonb_excl_list(:,:)


    cell_pair       => domain%cell_pairlist1
    natom           => domain%num_atom
    nwater          => domain%num_water
    water_list      => domain%water_list
    coord           => domain%coord
    charge          => domain%charge

    table_elec      => enefunc%table%table_ecor
    table_delec     => enefunc%table%table_decor
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nonb_excl   => enefunc%num_nonb_excl
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nonb_excl_list  => enefunc%nonb_excl_list

    ncell_local = domain%num_cell_local
    cutoff2     = cutoff*cutoff

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, ini_excl, fin_excl, num_excl,       &
    !$omp         dij, rij2, L, R, grad_coef, work, term_elec, viri_local)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_excl = 0

      do ix = 1, natom(i) - 1
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = cutoff2*density/rij2
          L    = int(rij2)
          L    = min(L,table_maxpoint)

          term_elec = table_elec(L) 
          term_elec  = charge(ix,i) * charge(iy,i) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L) 
          grad_coef = charge(ix,i) * charge(iy,i) * term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
          
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do
      end do
 
      do iwater = 1, nwater(i)
        do j = 1, 2
          do k = j+1, 3
            ix   = water_list(j,iwater,i)
            iy   = water_list(k,iwater,i)
            dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
            rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
            rij2  = cutoff2*density/rij2 
            L    = int(rij2)
            L    = min(L,table_maxpoint)
  
            term_elec = table_elec(L) 
            term_elec  = charge(ix,i) * charge(iy,i) * term_elec
            eelec(id+1) = eelec(id+1) + term_elec

            term_elec = table_delec(L) 
            grad_coef = charge(ix,i) * charge(iy,i) * term_elec
  
            work(1:3) = grad_coef*dij(1:3)
  
            force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
            force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
  
            ! virial
            !
            viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
            viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
            viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
            viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
            viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
            viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

          end do
        end do
      end do

    end do

    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_excl = 0

      do ix = 1, natom(i)
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = cutoff2*density/rij2
          L    = int(rij2)
          L    = min(L,table_maxpoint)

          term_elec = table_elec(L) 

          term_elec  = charge(ix,i) * charge(iy,j) * term_elec
          eelec(id+1) = eelec(id+1) + term_elec

          term_elec = table_delec(L) 
          grad_coef = charge(ix,i) * charge(iy,j) * term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

      end do
    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine pme_bond_corr_point

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr
  !> @brief        calculate bond correction term in PME
  !  @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr(domain, enefunc, force, virial, eelec)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij2, rij
    real(wp)                 :: qij, uij, fij(1:3)
    real(wp)                 :: grad_coef
    real(wp)                 :: work(1:3)
    real(wp)                 :: viri_local(1:3,1:3)
    integer                  :: i, ix, iy, j, k, ij, L, iwater
    integer                  :: num_excl, ini_excl, fin_excl
    integer                  :: ncell_local, id, omp_get_thread_num

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: natom(:)
    integer,         pointer :: num_nonb_excl1(:,:), nonb_excl_list1(:,:)
    integer,         pointer :: num_nonb_excl(:,:), nonb_excl_list(:,:)


    cell_pair       => domain%cell_pairlist1
    natom           => domain%num_atom
    coord           => domain%coord
    charge          => domain%charge

    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nonb_excl   => enefunc%num_nonb_excl
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nonb_excl_list  => enefunc%nonb_excl_list

    ncell_local     =  domain%num_cell_local


    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, ini_excl, fin_excl, num_excl,       &
    !$omp         dij, rij2, rij, grad_coef, qij, uij, fij, work, viri_local)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_excl = 0

      do ix = 1, natom(i) - 1
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list1(k,i)

          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij = sqrt(rij2)

          qij = charge(ix,i)*charge(iy,i)
          uij = qij * erf_n(alpha*rij)/rij
          eelec(id+1) = eelec(id+1) - uij * el_fact

          fij(1:3) = (uij-alpha2sp*qij*exp(alpha2m*rij2))*dij(1:3)/rij2
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - fij(1:3)*el_fact
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + fij(1:3)*el_fact
          
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*fij(1)*el_fact
          viri_local(2,1) = viri_local(2,1) + dij(2)*fij(1)*el_fact
          viri_local(3,1) = viri_local(3,1) + dij(3)*fij(1)*el_fact
          viri_local(2,2) = viri_local(2,2) + dij(2)*fij(2)*el_fact
          viri_local(3,2) = viri_local(3,2) + dij(3)*fij(2)*el_fact
          viri_local(3,3) = viri_local(3,3) + dij(3)*fij(3)*el_fact
        end do
      end do

    end do

    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_excl = 0

      do ix = 1, natom(i)
        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        do k = ini_excl, fin_excl

          iy = nonb_excl_list(k,ij)

          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij = sqrt(rij2)

          qij = charge(ix,i)*charge(iy,j)
          uij = qij * erf_n(alpha*rij)/rij
          eelec(id+1) = eelec(id+1) - uij * el_fact

          fij(1:3) = (uij-alpha2sp*qij*exp(alpha2m*rij2))*dij(1:3)/rij2
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - fij(1:3)*el_fact
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + fij(1:3)*el_fact

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*fij(1)*el_fact
          viri_local(2,1) = viri_local(2,1) + dij(2)*fij(1)*el_fact
          viri_local(3,1) = viri_local(3,1) + dij(3)*fij(1)*el_fact
          viri_local(2,2) = viri_local(2,2) + dij(2)*fij(2)*el_fact
          viri_local(3,2) = viri_local(3,2) + dij(3)*fij(2)*el_fact
          viri_local(3,3) = viri_local(3,3) + dij(3)*fij(3)*el_fact

        end do

      end do
    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel
  
    return

  end subroutine pme_bond_corr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_real
  !> @brief        calculate PME real part + Lennard-Jones
  !  @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    pairlist : pairlist information
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_real(domain, enefunc, pairlist, &
                      force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    type(s_pairlist),target, intent(in)    :: pairlist
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2, rij
    real(wp)                  :: inv_rij, inv_r2, inv_r6, inv_r3, inv_r12
    real(wp)                  :: lj6, lj12, term_lj12, term_lj6, term_lj
    real(wp)                  :: term_elec1, term_elec, switch, dswitch
    real(wp)                  :: cutoff2, switchdist2
    real(wp)                  :: c1_switch, c2_switch, c4_switch, grad_coef
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown, c6, c12
    real(wp)                  :: el_fact, alpha, alpha2m, alpha2sp
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: elec_temp, evdw_temp
    real(wp)                  :: dij_list(3,100)
    real(wp)                  :: force_local(3), rij2_list(100), table(12)
    real(wp)                  :: force_localj(3,100)
    integer                   :: i, ix, iy, j, k, ij, iix, iwater, list
    integer                   :: num_nb15, ini_nb15, fin_nb15, num_count
    integer                   :: id, omp_get_thread_num, j_list(100)
    integer                   :: ncell, ncell_local

    real(wp),         pointer :: coord(:,:,:), charge(:,:), cell_move(:,:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: nonb_lj12(:,:), nonb_lj6(:,:)
    real(wp),         pointer :: cutoff, switchdist
    integer,          pointer :: cell_pairlist(:,:)
    integer,          pointer :: natom(:), atmcls(:,:)
    integer,          pointer :: nb15_list(:,:), nb15_cell(:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)


    cell_pairlist   => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom
    coord           => domain%coord
    charge          => domain%charge
    trans1          => domain%trans_vec
    trans2          => domain%translated
    cell_move       => domain%cell_move

    cutoff          => enefunc%cutoffdist
    switchdist      => enefunc%switchdist
    nonb_lj12       => enefunc%nonb_lj12
    nonb_lj6        => enefunc%nonb_lj6

    nb15_list       => pairlist%nb15_list
    nb15_cell       => pairlist%nb15_cell
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list

    ncell_local = domain%num_cell_local
    ncell       = domain%num_cell_local + domain%num_cell_boundary

    cutoff2     = cutoff * cutoff
    switchdist2 = switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup = cutoff2*cutoff2*cutoff2
      c2_switchup = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    el_fact  = ELECOEF/enefunc%dielec_const
    alpha    = enefunc%pme_alpha
    alpha2m  = -alpha**2
    alpha2sp = 2.0_wp*alpha/sqrt(PI)

    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, num_nb15, rtmp, qtmp, ini_nb15, fin_nb15, k, iy,  &
    !$omp         dij, rij2, rij, c2_switch, c4_switch, switch, dswitch,       &
    !$omp         inv_rij, inv_r2, inv_r6, inv_r12, term_elec1, term_elec,     &
    !$omp         term_lj12, term_lj6, term_lj, inv_r3, grad_coef, work,       &
    !$omp         viri_local, iwater, ij, j, trans_x, trans_y, trans_z, iix,   &
    !$omp         list, num_count, j_list, dij_list, rij2_list, force_local,   &
    !$omp         force_localj, lj6, lj12, elec_temp, evdw_temp)

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        trans2(1,ix,i) = coord(1,ix,i) + trans1(1,ix,i)
        trans2(2,ix,i) = coord(2,ix,i) + trans1(2,ix,i)
        trans2(3,ix,i) = coord(3,ix,i) + trans1(3,ix,i)
      end do
    end do

    !$omp barrier

    viri_local(1:3,1:3) = 0.0_wp
    elec_temp = 0.0_wp
    evdw_temp = 0.0_wp

    ! energy within a cell(solute-solute)
    !
    do i = id+1, ncell_local, nthread

      num_nb15 = 0
      do ix = 1, natom(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)

        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc1(ix,i)
        num_nb15  = fin_nb15
        num_count = 0

        do k = ini_nb15, fin_nb15

          iy = nb15_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count) = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count) = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count

          dij(1:3) = dij_list(1:3,k)
          iy   = j_list(k)
          rij2  = rij2_list(k)
          rij = sqrt(rij2)

          ! van der Waals
          !
          lj6   = nonb_lj6(atmcls(ix,i),atmcls(iy,i))
          lj12  = nonb_lj12(atmcls(ix,i),atmcls(iy,i))
          rij      = sqrt(rij2)
          inv_r2   = 1.0_wp / rij2
          inv_rij  = 1.0_wp / rij
          inv_r3   = inv_r2 * inv_rij
          inv_r6   = inv_r3 * inv_r3
          inv_r12  = inv_r6 * inv_r6

          if (enefunc%force_switch) then
            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6 - c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              grad_coef  = -12.0_wp * lj12* c12 *   &
                           (inv_r6 - c1_switchup)*inv_r6*inv_r2
              grad_coef  = grad_coef + 6.0_wp * lj6 * c6 *    &
                           (inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              grad_coef  = -12.0_wp * lj12 * inv_r12*inv_r2
              grad_coef  = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if
          else
            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch*(cutoff2 + 2.0_wp*rij2-3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if
            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            grad_coef    = (dswitch * term_lj * rij   &
                         - switch * (12.0_wp*term_lj12-6.0_wp*term_lj6) )*inv_r2
          end if

          ! electrostatic
          !
          term_elec1     = charge(ix,i)*charge(iy,i)*el_fact
          term_elec      = term_elec1*erfc04(alpha*rij)*inv_rij
          eelec(id+1)    = eelec(id+1) + term_elec
          grad_coef      = grad_coef - &
                           (term_elec+alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2
          work(1:3)      = grad_coef * dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)
#ifdef KCOMP
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
#else
          force_localj(1:3,k)  = work(1:3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

#ifdef KCOMP
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif
      end do

    end do


    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cell(ij)

        ix   = nb15_list(iix,ij)
        rtmp(1:3) = trans2(1:3,ix,i) + cell_move(1:3,j,i)
        qtmp = charge(ix,i)

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calc(ix,ij)
        num_nb15 = fin_nb15

        num_count = 0
        do k = ini_nb15, fin_nb15
          iy = nb15_calc_list(k,ij)
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j) 
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count) = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count) = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
!DIR$IVDEP
        do k = 1, num_count
          dij(1:3) = dij_list(1:3,k)
          iy   = j_list(k)
          rij2  = rij2_list(k)
          rij = sqrt(rij2)

          ! van der Waals
          !
          lj6   = nonb_lj6(atmcls(ix,i),atmcls(iy,j))
          lj12  = nonb_lj12(atmcls(ix,i),atmcls(iy,j))
          rij      = sqrt(rij2)
          inv_r2   = 1.0_wp / rij2
          inv_rij  = 1.0_wp / rij
          inv_r3   = inv_r2 * inv_rij
          inv_r6   = inv_r3 * inv_r3
          inv_r12  = inv_r6 * inv_r6

          if (enefunc%force_switch) then
            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6 - c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              grad_coef  = -12.0_wp * lj12* c12 *   &
                           (inv_r6 - c1_switchup)*inv_r6*inv_r2
              grad_coef  = grad_coef + 6.0_wp * lj6 * c6 *    &
                           (inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              grad_coef  = -12.0_wp * lj12 * inv_r12 * inv_r2
              grad_coef  = grad_coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if
          else
            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch*(cutoff2 + 2.0_wp*rij2-3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if
            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            grad_coef    = (dswitch * term_lj * rij   &
                         - switch * (12.0_wp*term_lj12-6.0_wp*term_lj6) )*inv_r2
          end if

          ! electrostatic
          !
          term_elec1     = charge(ix,i)*charge(iy,j)*el_fact
          term_elec      = term_elec1*erfc04(alpha*rij)*inv_rij
          eelec(id+1)    = eelec(id+1) + term_elec
          grad_coef      = grad_coef - &
                           (term_elec+alpha2sp*term_elec1*exp(alpha2m*rij2))*inv_r2
          work(1:3)      = grad_coef * dij(1:3)

          work(1:3) = grad_coef*dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)

          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)

      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    eelec(id+1) = eelec(id+1) + elec_temp
    evdw(id+1)  = evdw(id+1)  + evdw_temp

    !$omp end parallel

    return

  end subroutine pme_real

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_recip
  !> @brief        Calculate PME reciprocal part with domain decomposition
  !! @authors      JJ
  !! @param[in]    domain : domain information
  !! @param[inout] force  : forces of target systems
  !! @param[inout] virial : virial term of target systems
  !! @param[inout] eelec  : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_recip(domain, force, virial, eelec)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)

    ! local variables
    real(wp)                 :: elec_temp
    real(wp)                 :: dv(3), bsc_tmp(3,n_bs), bscd_tmp(3,n_bs)
    real(wp)                 :: viri_local(1:3,1:3), temp
    real(wp)                 :: tqq
    real(wp)                 :: f_1, f_2, f_3
    integer                  :: i, j, k, k1, icel, iproc, id
    integer                  :: ix, iv, iy, iz, ixs, iys, izs
    integer                  :: ixyz, nix, niy, niz
    integer                  :: ixx, iyy, izz, ii(3)
    integer                  :: vx_tmp(n_bs), vy_tmp(n_bs), vz_tmp(n_bs)
    integer                  :: ncell, omp_get_thread_num

    complex(wp), allocatable :: work1(:), work2(:)
    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    integer,         pointer :: natom(:)


    coord  => domain%coord
    charge => domain%charge
    natom  => domain%num_atom

    ncell  = domain%num_cell_local + domain%num_cell_boundary

    ! Initializing the energy and force
    !
    qdf_real(1:nlocalx*nlocaly*nlocalz) = 0.0_wp

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, iv, ix, iy, iz, icel, k, ii, dv, izz, izs, iyy,    &
    !$omp         iys, ixx, ixs, viri_local, iproc, tqq, k1, temp, work1,      &
    !$omp         work2, elec_temp, f_1, f_2, f_3, ixyz, vx_tmp, vy_tmp,       &
    !$omp         vz_tmp, nix, niy, niz, bsc_tmp, bscd_tmp)
    !

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! initialization
    !
    elec_temp = 0.0_wp

    allocate(work1(ngridmax), work2(2*ngridmax))

    ! Translation to scaled fractional coordinates (v(x,y,z))
    !
    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        v(1:3,ix,i) = coord(1:3,ix,i) * r_scale(1:3)
        v(1:3,ix,i) = v(1:3,ix,i) + real(ngrid(1:3)/2,wp) &
                    - real(ngrid(1:3),wp)*anint(v(1:3,ix,i)/real(ngrid(1:3),wp))
      end do
    end do

    ! Initialization of Q-fanction
    !
    do iz = 1, nlocalz
      do iy = 1, nlocaly
        do ix = 1, nlocalx
          qdf(ix,iy,iz,id+1) = 0.0_wp
        end do
      end do
    end do

    ! Calculateing Q-fanction
    !
    do icel = id+1, ncell, nthread
      do i = 1, natom(icel)

        do k = 1, 3
          ii(k) = int(v(k,i,icel))
        end do

        do k = 1,3
          dv(k) = v(k,i,icel) - real(ii(k),wp)
          call b_spline_dev_coef(n_bs, dv(k), bsc(1,k,i,icel), bscd(1,k,i,icel)) !TODO bsc(1:n_bs,k,i,icel) ?
        end do

        nix = 0 
        niy = 0 
        niz = 0 

        do ixyz = 1, n_bs
          ixx = ii(1) - ixyz + 1
          iyy = ii(2) - ixyz + 1
          izz = ii(3) - ixyz + 1
          if (ixx < 0) then
            ixs = ixx + ngrid(1) + 1
          else if (ixx == ngrid(1)) then
            ixs = ixx
          else
            ixs = ixx + 1
          end if

          if (iyy < 0) then
            iys = iyy + ngrid(2) + 1
          else if (iyy == ngrid(2)) then
            iys = iyy
          else
            iys = iyy + 1
          end if

          if (izz < 0) then
            izs = izz + ngrid(3) + 1
          else if (izz == ngrid(3)) then
            izs = izz
          else
            izs = izz + 1
          end if

          if (ixs >= x_start .and. ixs <= x_end) then
            nix = nix + 1
            vx_tmp(nix) = ixs - x_start + 1
            bsc_tmp(1,nix) = bsc(ixyz,1,i,icel)
          end if

          if (iys >= y_start .and. iys <= y_end) then
            niy = niy + 1
            vy_tmp(niy) = iys - y_start + 1
            bsc_tmp(2,niy) = bsc(ixyz,2,i,icel)
          end if

          if (izs >= z_start .and. izs <= z_end) then
            niz = niz + 1
            vz_tmp(niz) = izs - z_start + 1
            bsc_tmp(3,niz) = bsc(ixyz,3,i,icel)
          end if
        end do

        do iz = 1, niz
          izs = vz_tmp(iz)
          do iy = 1, niy
            iys = vy_tmp(iy)
            do ix = 1, nix
              ixs = vx_tmp(ix)
              qdf(ixs,iys,izs,id+1) = qdf(ixs,iys,izs,id+1)           &
                                    + bsc_tmp(1,ix)*bsc_tmp(2,iy)     &
                                     * bsc_tmp(3,iz)*charge(i,icel)   &
                                     * bs_fact3
            end do
          end do
        end do
      end do
    end do

    !$omp barrier
    do iproc = 1, nthread
      !$omp do
      do iz = 1, nlocalz
        do iy = 1, nlocaly
          k = (iy-1)*nlocalx + (iz-1)*nlocalx*nlocaly
          do iv = 1, nlocalx, 16
            do ix = iv, min0(iv+15,nlocalx)
              qdf_real(k+ix) = qdf_real(k+ix) + qdf(ix,iy,iz,iproc)
            end do
          end do
        end do
      end do
    end do

#ifdef FFT3D
    ! FFT (forward)
    !
    !$omp master
#ifdef MPI
    call mpi_alltoall(qdf_real, nlocalx*nlocaly*nlocalz/nprocx, mpi_wp_real,   &
                      qdf_work, nlocalx*nlocaly*nlocalz/nprocx, mpi_wp_real,   &
                      grid_commx, ierror)
#else
    qdf_work(1:nlocalx*nlocaly*nlocalz,1) = qdf_real(1:nlocalx*nlocaly*nlocalz)
#endif
    !$omp end master
    !$omp barrier
    call fft3d_tuned(qdf_work, ftqdf, ftqdf2, ftqdf3, ftqdf4, ftqdf5,          &
                     ftqdf_work, ftqdf_work, ftqdf_work, ftqdf_work,           &
                     ftqdf_work2, ftqdf_work3, work1, work2, nlocalx, nlocaly, &
                     nlocalz, nlocalx1, x_local1, ngrid(1), ngrid(2), ngrid(3),&
                     nprocx, nprocy, nprocz, id, nthread, my_x_rank, my_y_rank,&
                     my_z_rank, x_start1, x_end1, y_start, y_end, z_start,     &
                     z_end, grid_commx, grid_commy, grid_commz)
#else
    ! FFT (forward)
    !
    !$omp master
#ifdef MPI
    call mpi_allgather(qdf_real, nlocalx*nlocaly*nlocalz, mpi_wp_real,         &
                       qdf_work, nlocalx*nlocaly*nlocalz, mpi_wp_real,         &
                       grid_commx, ierror) 
#else
    qdf_work(1:nlocalx*nlocaly*nlocalz,1) = qdf_real(1:nlocalx*nlocaly*nlocalz)
#endif
    !$omp end master
    !$omp barrier

    call fft3d(qdf_work, ftqdf, ftqdf, ftqdf, ftqdf_work, ftqdf_work,          &
               ftqdf_work, work1, work2, nlocalx, nlocaly,                     &
               nlocalz, nlocalx1, x_local1, ngrid(1), ngrid(2), ngrid(3),      &
               nprocx, nprocy, nprocz, id, nthread, my_x_rank, x_start1,       &
               x_end1, y_start, y_end, z_start, z_end, grid_commx, grid_commy, &
               grid_commz)
#endif

    ! Energy calculation
    !
    viri_local(1:3,1:3) = 0.0_wp

    iproc = my_z_rank + 1
#ifdef FFT3D
    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          izz = iz + (iproc-1)*nlocalz
          tqq = real(ftqdf_work3(izz,ix,iy),wp)**2 + &
                imag(ftqdf_work3(izz,ix,iy))**2
          tqq = tqq * theta(iz,iy,ix,iproc) * vol_fact4
          elec_temp = elec_temp + tqq

          ! virial
          ! 
          viri_local(1,1) = viri_local(1,1) +  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gx(ix)*gx(ix))
          viri_local(2,1) = viri_local(2,1) +  &
                            tqq*vir_fact(iz,iy,ix)*gy(iy)*gx(ix)
          viri_local(3,1) = viri_local(3,1) +  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gx(ix)
          viri_local(2,2) = viri_local(2,2) +  &
                            tqq*(1.0_wp + vir_fact(iz,iy,ix)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) +  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) +  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gz(iz,iproc)*gz(iz,iproc))

        end do
      end do
    end do

    if (my_x_rank == 0) then
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          izz = iz + (iproc-1)*nlocalz
          tqq = real(ftqdf_work3(izz,1,iy),wp)**2 + &
                imag(ftqdf_work3(izz,1,iy))**2
          tqq = tqq * theta(iz,iy,1,iproc) * vol_fact2
          elec_temp = elec_temp - tqq
          viri_local(1,1) = viri_local(1,1) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,1)*gx(1)*gx(1))
          viri_local(2,1) = viri_local(2,1) -  &
                            tqq*vir_fact(iz,iy,1)*gy(iy)*gx(1)
          viri_local(3,1) = viri_local(3,1) -  &
                            tqq*vir_fact(iz,iy,1)*gz(iz,iproc)*gx(1)
          viri_local(2,2) = viri_local(2,2) -  &
                            tqq*(1.0_wp+ vir_fact(iz,iy,1)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) -  &
                            tqq*vir_fact(iz,iy,1)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,1)*gz(iz,iproc)*gz(iz,iproc))
        end do
      end do
    end if

    if (my_x_rank == (nprocx-1)) then
      ix = x_local1
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          izz = iz + (iproc-1)*nlocalz
          tqq = real(ftqdf_work3(izz,ix,iy),wp)**2 + &
                imag(ftqdf_work3(izz,ix,iy))**2
          tqq = tqq * theta(iz,iy,ix,iproc) * vol_fact2
          elec_temp = elec_temp - tqq
          viri_local(1,1) = viri_local(1,1) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gx(ix)*gx(ix))
          viri_local(2,1) = viri_local(2,1) -  &
                            tqq*vir_fact(iz,iy,ix)*gy(iy)*gx(ix)
          viri_local(3,1) = viri_local(3,1) -  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gx(ix)
          viri_local(2,2) = viri_local(2,2) -  &
                            tqq*(1.0_wp + vir_fact(iz,iy,ix)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) -  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gz(iz,iproc)*gz(iz,iproc))
        end do
      end do
    end if

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) + viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) + viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) + viri_local(i,i)
    end do

    ! F^-1[Th]*F^-1[Q] (=X)
    !
    !$omp barrier
    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread
        k = (iy-1)*nlocalz + (ix-1)*nlocalz*nlocaly
        do iproc = 1, nprocz
          do iz = 1, nlocalz
            izz = iz + (iproc-1)*nlocalz
            ftqdf_work(k+iz,iproc) = ftqdf_work3(izz,ix,iy)    &
                                   * cmplx(theta(iz,iy,ix,iproc))
          end do
        end do
      end do
    end do

   !$omp barrier

    call bfft3d_tuned(qdf_work, qdf_real, ftqdf, ftqdf2, ftqdf3, ftqdf_work3,  &
                      ftqdf_work2, ftqdf_work, work1, work2, nlocalx, nlocaly, &
                      nlocalz, nlocalx1, x_local1, ngrid(1), ngrid(2),         &
                      ngrid(3), nprocx, nprocy, nprocz, id, nthread, my_x_rank,&
                      my_y_rank, my_z_rank, x_start1, x_end1, y_start, y_end,  &
                      z_start, z_end, grid_commx, grid_commy, grid_commz)

#else

    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          k = iz + (iy-1)*nlocalz + (ix-1)*nlocalz*nlocaly
          tqq = real(ftqdf_work(k,iproc),wp)**2 + &
                imag(ftqdf_work(k,iproc))**2
          tqq = tqq * theta(iz,iy,ix,iproc) * vol_fact4
          elec_temp = elec_temp + tqq

          ! virial
          ! 
          viri_local(1,1) = viri_local(1,1) +  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gx(ix)*gx(ix))
          viri_local(2,1) = viri_local(2,1) +  &
                            tqq*vir_fact(iz,iy,ix)*gy(iy)*gx(ix)
          viri_local(3,1) = viri_local(3,1) +  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gx(ix)
          viri_local(2,2) = viri_local(2,2) +  &
                            tqq*(1.0_wp + vir_fact(iz,iy,ix)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) +  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) +  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gz(iz,iproc)*gz(iz,iproc))
        end do
      end do
    end do

    if (my_x_rank == 0) then
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          k = iz + (iy-1)*nlocalz
          tqq = real(ftqdf_work(k,iproc),wp)**2 + imag(ftqdf_work(k,iproc))**2
          tqq = tqq * theta(iz,iy,1,iproc) * vol_fact2
          elec_temp = elec_temp - tqq
          viri_local(1,1) = viri_local(1,1) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,1)*gx(1)*gx(1))
          viri_local(2,1) = viri_local(2,1) -  &
                            tqq*vir_fact(iz,iy,1)*gy(iy)*gx(1)
          viri_local(3,1) = viri_local(3,1) -  &
                            tqq*vir_fact(iz,iy,1)*gz(iz,iproc)*gx(1)
          viri_local(2,2) = viri_local(2,2) -  &
                            tqq*(1.0_wp+ vir_fact(iz,iy,1)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) -  &
                            tqq*vir_fact(iz,iy,1)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,1)*gz(iz,iproc)*gz(iz,iproc))
        end do
      end do
    end if

    if (my_x_rank == (nprocx-1)) then
      ix = x_local1
      do iy = id+1, nlocaly, nthread
        do iz = 1, nlocalz
          k = iz + (iy-1)*nlocalz + (x_local1-1)*nlocalz*nlocaly
          tqq = real(ftqdf_work(k,iproc),wp)**2 + &
                imag(ftqdf_work(k,iproc))**2
          tqq = tqq * theta(iz,iy,ix,iproc) * vol_fact2
          elec_temp = elec_temp - tqq
          viri_local(1,1) = viri_local(1,1) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gx(ix)*gx(ix))
          viri_local(2,1) = viri_local(2,1) -  &
                            tqq*vir_fact(iz,iy,ix)*gy(iy)*gx(ix)
          viri_local(3,1) = viri_local(3,1) -  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gx(ix)
          viri_local(2,2) = viri_local(2,2) -  &
                            tqq*(1.0_wp + vir_fact(iz,iy,ix)*gy(iy)*gy(iy))
          viri_local(3,2) = viri_local(3,2) -  &
                            tqq*vir_fact(iz,iy,ix)*gz(iz,iproc)*gy(iy)
          viri_local(3,3) = viri_local(3,3) -  &
                            tqq*(1.0_wp+vir_fact(iz,iy,ix)*gz(iz,iproc)*gz(iz,iproc))
        end do
      end do
    end if

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) + viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) + viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) + viri_local(i,i)
    end do

    ! F^-1[Th]*F^-1[Q] (=X)
    !
    !$omp barrier
    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread
        k = (iy-1)*nlocalz + (ix-1)*nlocalz*nlocaly
        do iproc = 1, nprocz
          ftqdf_work(k+1:k+nlocalz,iproc) = ftqdf_work(k+1:k+nlocalz,iproc)    &
                                          * cmplx(theta(1:nlocalz,iy,ix,iproc))
        end do
      end do
    end do
    !$omp barrier

    call bfft3d(qdf_work, qdf_real, ftqdf, ftqdf, ftqdf, ftqdf_work, ftqdf_work,          &
               ftqdf_work, work1, work2, nlocalx, nlocaly,     &
               nlocalz, nlocalx1, x_local1, ngrid(1), ngrid(2), ngrid(3),      &
               nprocx, nprocy, nprocz, id, nthread, my_x_rank, x_start1,       &
               x_end1, y_start, y_end, z_start, z_end, grid_commx, grid_commy, &
               grid_commz)

#endif

    ! X is saved on qdf
    !
    do iz = id+1, nlocalz, nthread
      do iy = 1, nlocaly
        do ix = 1, nlocalx
          k = ix + (iy-1)*nlocalx + (iz-1)*nlocalx*nlocaly
          qdf(ix,iy,iz,1) = qdf_real(k)
        end do
      end do
    end do

    !$omp barrier
    ! Gradient
    !
    do icel = id+1, ncell, nthread
      do i = 1, natom(icel)

        f_1 = 0.0_wp
        f_2 = 0.0_wp
        f_3 = 0.0_wp

        do k = 1, 3
          ii(k) = int(v(k,i,icel))
        end do

        nix=0
        niy=0
        niz=0

        do ixyz = 1, n_bs

          ixx = ii(1) - ixyz + 1
          iyy = ii(2) - ixyz + 1
          izz = ii(3) - ixyz + 1

          if (ixx < 0) then
            ixs = ixx + ngrid(1) + 1
          else if (ixx == ngrid(1)) then
            ixs = ixx
          else
            ixs = ixx + 1
          end if

          if (iyy < 0) then
            iys = iyy + ngrid(2) + 1
          else if (iyy == ngrid(2)) then
            iys = iyy
          else
            iys = iyy + 1
          end if

          if (izz < 0) then
            izs = izz + ngrid(3) + 1
          else if (izz == ngrid(3)) then
            izs = izz
          else
            izs = izz + 1
          end if

          if (ixs >= x_start .and. ixs <= x_end) then
            nix = nix + 1
            vx_tmp(nix) = ixs - x_start + 1
            bsc_tmp(1,nix)  = bsc(ixyz,1,i,icel)
            bscd_tmp(1,nix) = bscd(ixyz,1,i,icel)
          end if

          if (iys >= y_start .and. iys <= y_end) then
            niy = niy + 1
            vy_tmp(niy) = iys - y_start + 1
            bsc_tmp(2,niy)  = bsc(ixyz,2,i,icel)
            bscd_tmp(2,niy) = bscd(ixyz,2,i,icel)
          end if

          if (izs >= z_start .and. izs <= z_end) then
            niz = niz + 1
            vz_tmp(niz) = izs - z_start + 1
            bsc_tmp(3,niz)  = bsc(ixyz,3,i,icel)
            bscd_tmp(3,niz) = bscd(ixyz,3,i,icel)
          end if
        end do

        do iz = 1, niz
          izs = vz_tmp(iz)
          do iy = 1, niy
            iys = vy_tmp(iy)
            do ix = 1, nix
              ixs = vx_tmp(ix)

              f_1 = f_1                                  &
                  + bscd_tmp(1,ix)*bsc_tmp(2,iy)         &
                  * bsc_tmp(3,iz)*qdf(ixs,iys,izs,1)
              f_2 = f_2                                  &
                  + bsc_tmp(1,ix)*bscd_tmp(2,iy)         &
                  * bsc_tmp(3,iz)*qdf(ixs,iys,izs,1)
              f_3 = f_3                                  &
                  + bsc_tmp(1,ix)*bsc_tmp(2,iy)          &
                  * bscd_tmp(3,iz)*qdf(ixs,iys,izs,1)
            end do
          end do
        end do

        force(1,i,icel,id+1) = force(1,i,icel,id+1)                          &
             -f_1 * charge(i,icel) * r_scale(1) * vol_fact4 * bs_fact3d
        force(2,i,icel,id+1) = force(2,i,icel,id+1)                          &
             -f_2 * charge(i,icel) * r_scale(2) * vol_fact4 * bs_fact3d
        force(3,i,icel,id+1) = force(3,i,icel,id+1)                          &
             -f_3 * charge(i,icel) * r_scale(3) * vol_fact4 * bs_fact3d
  
      end do
    end do

    !$omp barrier

    deallocate(work1, work2)

    eelec(id+1) = eelec(id+1) + elec_temp

    !$omp end parallel
 
    return

  end subroutine pme_recip

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    b_spline_coef
  !> @brief        Calculate B-spline coefficients
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine b_spline_coef(n, u, M)

    ! formal arguments
    integer,                 intent(in)    :: n
    real(wp),                intent(in)    :: u
    real(wp),                intent(out)   :: M(n)

    ! local variables
    integer   :: i,j
    real(wp)  :: u_3, u_2

    if (n .eq. 4) then

      u_3 = u*u*u
      u_2 = u*u
      M(1) = u_3
      M(2) = -3.0_wp*u_3 + 3.0_wp*u_2 + 3.0_wp*u + 1.0_wp 
      M(3) = 3.0_wp*u_3 - 6.0_wp*u_2 + 4.0_wp
      M(4) = (1.0_wp-u)*(1.0_wp-u)*(1.0_wp-u)

    else

      M(1) = u
      M(2) = 1.0_wp - u

      do j = 3, n
        M(j) = M(j-1) * (1.0_wp-u)
        do i = 1, j - 2
          M(j - i) = (real(j-i-1,wp) + u) * M(j-i) &
                   + (real(i+1,wp) - u) * M(j-i-1)
        end do
        M(1) = M(1) * u
      end do
 
    end if

    return

  end subroutine b_spline_coef

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    b_spline_dev_coef
  !> @brief        Calculate B-spline coefficients and its derivatives
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine b_spline_dev_coef(n, u, M, dM)

    ! formal arguments
    integer,                 intent(in)    :: n
    real(wp),                intent(in)    :: u
    real(wp),                intent(out)   :: M(n)
    real(wp),                intent(out)   :: dM(n)

    ! local variables
    integer                  :: i,j
    real(wp)                 :: u_3, u_2


    if (n .eq. 4) then

      u_3 = u*u*u
      u_2 = u*u
      M(1) = u_3
      M(2) = -3.0_wp*u_3 + 3.0_wp*u_2 + 3.0_wp*u + 1.0_wp 
      M(3) = 3.0_wp*u_3 - 6.0_wp*u_2 + 4.0_wp
      M(4) = (1.0_wp-u)*(1.0_wp-u)*(1.0_wp-u)
      dM(1) = u_2
      dM(2) = -3.0_wp*u_2 + 2.0_wp*u + 1.0_wp
      dM(3) = 3.0_wp*u_2 - 4.0_wp*u
      dM(4) = -(1.0_wp-u)*(1.0_wp-u)

    else

      M(1) = u
      M(2) = 1.0_wp - u

      do j = 3, n-1
        M(j) = M(j-1) * (1.0_wp-u)
        do i = 1, j-2
          M(j-i) = (real(j-i-1,wp)+u) * M(j-i) &
                 + (real(i+1,wp)-u) * M(j-i-1)
        end do
        M(1) = M(1) * u
      end do

      dM(1) = M(1)
      do i = 2, n - 1
        dM(i) = M(i) - M(i-1)
      end do
      dM(n) = - M(n-1)

      M(n) = M(n-1) * (1.0_wp-u)
      do i = 1, n - 2
        M(n-i) = (real(n-i-1,wp)+u) * M(n-i) &
               + (real(i+1,wp)-u) * M(n-i-1)
      end do
      M(1) = M(1) * u

    end if

    return

  end subroutine b_spline_dev_coef

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      erf_n
  !> @brief        error function
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function erf_n (x) result(erf)

    ! formal arguments
    real(wp),                intent(in)    :: x

    ! return value
    real(wp)                 :: erf


    erf = 1.0_wp - erfc04(x)

    return

  end function erf_n

!-------------------------------------------------------------------------------
! Complementary error function
! Taken from SUN's FDLIBM version 5.2 and translated from c to fortran.
!-------------------------------------------------------------------------------
! Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
!
! Developed at SunSoft, a Sun Microsystems, Inc. business.
! Permission to use, copy, modify, and distribute this
! software is freely granted, provided that this notice 
! is preserved.
!-------------------------------------------------------------------------------
! Definition:
!------------
!                       x
!                2      |\
! erf(x)  =  ---------  | exp(-t*t)dt
!             sqrt(pi) \| 
!                       0
!
! erfc(x) =  1 - erf(x)
!
! Note that  erf(-x) = -erf(x)
!            erfc(-x) = 2 - erfc(x)
!
!
! Method:
!--------
!
! 1. For |x| in [0, 0.84375]
!    erf(x)  = x + x*R(x^2)
!    erfc(x) = 1 - erf(x)           if x in [-.84375,0.25]
!            = 0.5 + ((0.5-x)-x*R)  if x in [0.25,0.84375]
!    where R = P/Q where P is an odd poly of degree 8 and
!                        Q is an odd poly of degree 10.
!                                                 -57.90
!                        | R - (erf(x)-x)/x | <= 2
!
!
!    Remark. The formula is derived by noting
!            erf(x) = (2/sqrt(pi))*(x - x^3/3 + x^5/10 - x^7/42 + ....)
!    and that
!            2/sqrt(pi) = 1.128379167095512573896158903121545171688
!    is close to one. The interval is chosen because the fix
!    point of erf(x) is near 0.6174 (i.e., erf(x)=x when x is
!    near 0.6174), and by some experiment, 0.84375 is chosen to
!    guarantee the error is less than one ulp for erf.
!
! 2. For |x| in [0.84375,1.25], let s = |x| - 1, and c = 0.84506291151
!    rounded to single (24 bits)
!    erf(x)  = sign(x) * (c  + P1(s)/Q1(s))
!    erfc(x) = (1-c)  - P1(s)/Q1(s) if x > 0
!              1+(c+P1(s)/Q1(s))    if x < 0
!              |P1/Q1 - (erf(|x|)-c)| <= 2**-59.06
!
!    Remark: here we use the taylor series expansion at x=1.
!            erf(1+s) = erf(1) + s*Poly(s)
!                     = 0.845.. + P1(s)/Q1(s)
!            That is, we use rational approximation to approximate
!            erf(1+s) - (c = (single)0.84506291151)
!    Note that |P1/Q1|< 0.078 for x in [0.84375,1.25] where 
!    P1(s) = degree 6 poly in s
!    Q1(s) = degree 6 poly in s
!
! 3. For x in [1.25,1/0.35(~2.857143)], 
!    erfc(x) = (1/x)*exp(-x*x-0.5625+R1/S1)
!    erf(x)  = 1 - erfc(x)
!    where 
!    R1(z) = degree 7 poly in z, (z=1/x^2)
!    S1(z) = degree 8 poly in z
!
! 4. For x in [1/0.35,28]
!    erfc(x) = (1/x)*exp(-x*x-0.5625+R2/S2) if x > 0
!            = 2.0 - (1/x)*exp(-x*x-0.5625+R2/S2) if -6<x<0
!            = 2.0 - tiny(if x <= -6)
!    erf(x)  = sign(x)*(1.0 - erfc(x)) if x < 6, else
!    erf(x)  = sign(x)*(1.0 - tiny)
!    where
!    R2(z) = degree 6 poly in z, (z=1/x^2)
!    S2(z) = degree 7 poly in z
!
!    Note1: To compute exp(-x*x-0.5625+R/S), let s be a single
!           precision number and s := x; then
!           -x*x = -s*s + (s-x)*(s+x)
!            exp(-x*x-0.5626+R/S) = exp(-s*s-0.5625)*exp((s-x)*(s+x)+R/S)
!    Note2: Here 4 and 5 make use of the asymptotic series
!                      exp(-x*x)
!           erfc(x) ~ ---------- * ( 1 + Poly(1/x^2) )
!                     x*sqrt(pi)
!           We use rational approximation to approximate
!           g(s) = f(1/x^2) = log(erfc(x)*x) - x*x + 0.5625
!           Here is the error bound for R1/S1 and R2/S2
!           |R1/S1 - f(x)|  < 2**(-62.57)
!           |R2/S2 - f(x)|  < 2**(-61.52)
!
! 5. For inf > x >= 28
!    erf(x)  = sign(x) *(1 - tiny)  (raise inexact)
!    erfc(x) = tiny*tiny (raise underflow) if x > 0
!            = 2 - tiny if x<0
!
! 7. Special case:
!    erf(0)  = 0, erf(inf)  = 1, erf(-inf) = -1,
!    erfc(0) = 1, erfc(inf) = 0, erfc(-inf) = 2, 
!    erfc/erf(NaN) is NaN
!-------------------------------------------------------------------------------

  function erfc04(x) result(erfc)

    real(wp),         intent(in)    :: x
    real(wp)                        :: erfc
    real(wp)                        :: ax,p,q,r,s,y,z

    real(wp),         parameter     :: zero1=  0.0_wp
    real(wp),         parameter     :: half1=  0.5_wp
    real(wp),         parameter     :: one1 =  1.0_wp
    real(wp),         parameter     :: two1 =  2.0_wp
    real(wp),         parameter     :: erx  =  8.45062911510467529297e-01_wp
    ! Coefficients for approximation to  erf on [0,0.84375]
    real(wp),         parameter     :: efx  =  1.28379167095512586316e-01_wp
    real(wp),         parameter     :: efx8 =  1.02703333676410069053e+00_wp
    real(wp),         parameter     :: pp0  =  1.28379167095512558561e-01_wp
    real(wp),         parameter     :: pp1  = -3.25042107247001499370e-01_wp
    real(wp),         parameter     :: pp2  = -2.84817495755985104766e-02_wp
    real(wp),         parameter     :: pp3  = -5.77027029648944159157e-03_wp
    real(wp),         parameter     :: pp4  = -2.37630166566501626084e-05_wp
    real(wp),         parameter     :: qq1  =  3.97917223959155352819e-01_wp
    real(wp),         parameter     :: qq2  =  6.50222499887672944485e-02_wp
    real(wp),         parameter     :: qq3  =  5.08130628187576562776e-03_wp
    real(wp),         parameter     :: qq4  =  1.32494738004321644526e-04_wp
    real(wp),         parameter     :: qq5  = -3.96022827877536812320e-06_wp
    ! Coefficients for approximation to  erf  in [0.84375,1.25] 
    real(wp),         parameter     :: pa0  = -2.36211856075265944077e-03_wp
    real(wp),         parameter     :: pa1  =  4.14856118683748331666e-01_wp
    real(wp),         parameter     :: pa2  = -3.72207876035701323847e-01_wp
    real(wp),         parameter     :: pa3  =  3.18346619901161753674e-01_wp
    real(wp),         parameter     :: pa4  = -1.10894694282396677476e-01_wp
    real(wp),         parameter     :: pa5  =  3.54783043256182359371e-02_wp
    real(wp),         parameter     :: pa6  = -2.16637559486879084300e-03_wp
    real(wp),         parameter     :: qa1  =  1.06420880400844228286e-01_wp
    real(wp),         parameter     :: qa2  =  5.40397917702171048937e-01_wp
    real(wp),         parameter     :: qa3  =  7.18286544141962662868e-02_wp
    real(wp),         parameter     :: qa4  =  1.26171219808761642112e-01_wp
    real(wp),         parameter     :: qa5  =  1.36370839120290507362e-02_wp
    real(wp),         parameter     :: qa6  =  1.19844998467991074170e-02_wp
    ! Coefficients for approximation to  erfc in [1.25,1/0.35]
    real(wp),         parameter     :: ra0  = -9.86494403484714822705e-03_wp
    real(wp),         parameter     :: ra1  = -6.93858572707181764372e-01_wp
    real(wp),         parameter     :: ra2  = -1.05586262253232909814e+01_wp
    real(wp),         parameter     :: ra3  = -6.23753324503260060396e+01_wp
    real(wp),         parameter     :: ra4  = -1.62396669462573470355e+02_wp
    real(wp),         parameter     :: ra5  = -1.84605092906711035994e+02_wp
    real(wp),         parameter     :: ra6  = -8.12874355063065934246e+01_wp
    real(wp),         parameter     :: ra7  = -9.81432934416914548592e+00_wp
    real(wp),         parameter     :: sa1  =  1.96512716674392571292e+01_wp
    real(wp),         parameter     :: sa2  =  1.37657754143519042600e+02_wp
    real(wp),         parameter     :: sa3  =  4.34565877475229228821e+02_wp
    real(wp),         parameter     :: sa4  =  6.45387271733267880336e+02_wp
    real(wp),         parameter     :: sa5  =  4.29008140027567833386e+02_wp
    real(wp),         parameter     :: sa6  =  1.08635005541779435134e+02_wp
    real(wp),         parameter     :: sa7  =  6.57024977031928170135e+00_wp
    real(wp),         parameter     :: sa8  = -6.04244152148580987438e-02_wp
    ! Coefficients for approximation to  erfc in [1/.35,28]
    real(wp),         parameter     :: rb0  = -9.86494292470009928597e-03_wp
    real(wp),         parameter     :: rb1  = -7.99283237680523006574e-01_wp
    real(wp),         parameter     :: rb2  = -1.77579549177547519889e+01_wp
    real(wp),         parameter     :: rb3  = -1.60636384855821916062e+02_wp
    real(wp),         parameter     :: rb4  = -6.37566443368389627722e+02_wp
    real(wp),         parameter     :: rb5  = -1.02509513161107724954e+03_wp
    real(wp),         parameter     :: rb6  = -4.83519191608651397019e+02_wp
    real(wp),         parameter     :: sb1  =  3.03380607434824582924e+01_wp
    real(wp),         parameter     :: sb2  =  3.25792512996573918826e+02_wp
    real(wp),         parameter     :: sb3  =  1.53672958608443695994e+03_wp
    real(wp),         parameter     :: sb4  =  3.19985821950859553908e+03_wp
    real(wp),         parameter     :: sb5  =  2.55305040643316442583e+03_wp
    real(wp),         parameter     :: sb6  =  4.74528541206955367215e+02_wp
    real(wp),         parameter     :: sb7  = -2.24409524465858183362e+01_wp

    ax = abs(x)

    if (ax < 0.84375_wp) then

      if (ax < epsilon(x)) then
        erfc = one1 - x
      else

        z = x**2
        r = pp0 + z*(pp1 + z*(pp2 + z*(pp3 + z*pp4)))
        s = one1 + z*(qq1 + z*(qq2 + z*(qq3 + z*(qq4 + z*qq5))))
        y = r/s

        if (x < 0.25_wp) then
          erfc =  one1 - (x + x*y)
        else
          r = x*y
          r = r + (x - half1)
          erfc = half1 - r 
        end if

      end if

    else if (ax < 1.25_wp) then

      s = ax - one1
      p = pa0 + s*(pa1 + s*(pa2 + s*(pa3 + s*(pa4 + s*(pa5 + s*pa6)))))
      q = one1 + s*(qa1 + s*(qa2 + s*(qa3 + s*(qa4 + s*(qa5 + s*qa6)))))

      if (x > zero1) then
        z  = one1 - erx
        erfc = z - p/q
      else
        z = erx + p/q
        erfc = one1 + z
      end if

    else if (ax < 28.0_wp) then

      s = one1/(ax**2)

      if (ax < 2.857143_wp) then
        p = ra0 + s*(ra1 + s*(ra2 + s*(ra3 + s*(ra4 +   &
             s*(ra5 + s*(ra6 + s*ra7))))))
        q = one1 + s*(sa1 + s*(sa2 + s*(sa3 + s*(sa4 +   &
             s*(sa5 + s*(sa6 + s*(sa7 + s*sa8)))))))
      else

        if (x < -6.0_wp) then

          erfc = two1
          return

        end if

        p = rb0 + s*(rb1 + s*(rb2 + s*(rb3 + s*(rb4 + s*(rb5 + s*rb6)))))
        q = one1 + s*(sb1 + s*(sb2 + s*(sb3 + s*(sb4 +   &
             s*(sb5 + s*(sb6 + s*sb7))))))

      end if

      z = ax
      r = exp(-z**2-0.5625_wp+(z-ax)*(z+ax)+p/q)

      if (x > zero1) then
        erfc = r/x
      else
        erfc = two1 + r/x
      end if

    else

      if (x > zero1) then
        erfc = zero1
      else
        erfc = two1
      end if

    end if

    return

  end function erfc04

end module sp_energy_pme_mod
