!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_md_respa_mod
!> @brief   perform molecular dynamics simulation with velocity verlet
!!          and respa algorithm
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_md_respa_mod

  use sp_output_mod
  use sp_update_domain_mod
  use sp_assign_velocity_mod
  use sp_dynvars_mod
  use sp_constraints_mod
  use sp_communicate_mod
  use sp_energy_mod
  use sp_output_str_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_ensemble_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use random_mod
  use messages_mod
  use timers_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: vverlet_respa_dynamics
  private :: initial_vverlet
  private :: control_temp_pres_vver1
  private :: control_temp_pres_vver2
  private :: nosehoover_thermostat_vverlet

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    vverlet_respa_dynamics
  !> @brief        velocity verlet integrator using respa
  !! @authors      JJ
  !! @param[inout] output      : output information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamic variables information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : non-bond pair list information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] constraints : bond constraint information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] comm        : information of communication
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine vverlet_respa_dynamics(output, domain, enefunc, dynvars, &
                                    dynamics, pairlist, boundary,     &
                                    constraints, ensemble, comm)

    ! formal arguments
    type(s_output),          intent(inout) :: output
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynvars), target, intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_boundary),        intent(inout) :: boundary
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_comm),            intent(inout) :: comm

    ! local variables
    real(wp)                 :: simtim, temperature, factor
    real(wp)                 :: dt_short, dt_long
    integer                  :: i, j, k, jx, nsteps
    integer                  :: istep, multistep
    integer                  :: iseed
    logical                  :: npt

    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: coord_old(:,:,:), vel(:,:,:), mass(:,:)
    real(wp),        pointer :: force(:,:,:), force_omp(:,:,:,:)
    real(wp),        pointer :: force_short(:,:,:), force_long(:,:,:)
    integer,         pointer :: ncell, natom(:)


    ncell       => domain%num_cell_local
    natom       => domain%num_atom
    mass        => domain%mass
    coord       => domain%coord
    coord_ref   => domain%coord_ref
    coord_old   => domain%coord_old
    force       => domain%force
    force_short => domain%force_short
    force_long  => domain%force_long
    vel         => domain%velocity

    temperature =  ensemble%temperature
    nsteps      =  dynamics%nsteps
    multistep   =  dynamics%elec_long_period
    dt_short    =  dynamics%timestep/AKMA_PS
    dt_long     =  real(multistep,wp)*dt_short
    simtim      =  dynamics%initial_time
    iseed       =  dynamics%iseed


    npt = (ensemble%ensemble == EnsembleNPT)

    ! Check restart
    !
    if (dynamics%restart) then

      call communicate_coor(domain, comm)

      call compute_energy_short(domain, enefunc, pairlist, boundary, coord, &
                                mod(0,dynamics%eneout_period)== 0,       &
                                dynvars%energy, &
                                force_short,    &
                                force_omp,      &
                                dynvars%virial, &
                                dynvars%virial_extern)

      call compute_energy_long (domain, enefunc, pairlist, boundary, coord, &
                                npt, dynvars%energy, force_long, force_omp, &
                                dynvars%virial)

      call communicate_force(domain, comm, force_short)

      call communicate_force(domain, comm, force_long)

    else

      call initial_velocity(temperature,           &
                            domain%num_atom_all,   &
                            domain%num_cell_local, &
                            domain%id_g2l,         &
                            domain%mass,           &
                            iseed,                 &
                            domain%velocity)

      call stop_trans_rotation(domain%num_cell_local,         &
                               domain%num_atom,               &
                               dynamics%stop_com_translation, &
                               dynamics%stop_com_rotation,    &
                               domain%mass,                   &
                               domain%coord,                  &
                               domain%velocity)

      call initial_vverlet(npt, output, enefunc, dynamics,            &
                           pairlist, boundary, ensemble, constraints, &
                           domain, dynvars, comm)

    end if

    do i = 1, nsteps / multistep

      simtim = simtim + dt_long * AKMA_PS
      dynvars%time = simtim
      dynvars%step = i * multistep

      call timer(TimerIntegrator, TimerOn)

      ! half step temperature correction
      !
      do k = 1, ncell
        do jx = 1, natom(k)
          coord_old(1:3,jx,k) = coord(1:3,jx,k)
        end do
      end do

      if (ensemble%ensemble == EnsembleNVT) then
    
        call control_temp_pres_vver1(dt_long, dynamics, ensemble, domain, &
                                     constraints, boundary, dynvars)

      end if 

      ! half step velocities with force_long
      !
      !$omp parallel do private(k, jx, factor)
      do k = 1, ncell
        do jx = 1, natom(k)
          factor = 0.5_wp * dt_long / mass(jx,k)
          vel(1:3,jx,k) = vel(1:3,jx,k) + factor * force_long(1:3,jx,k)
        end do
      end do
      !$omp end parallel do

      call timer(TimerIntegrator, TimerOff)

      ! Calculation of short range forces
      !
      do j = 1, multistep

        call timer(TimerIntegrator, TimerOn)

        ! decide the current steps
        !
        istep = (i-1) * multistep + j

        ! half steps velocities and full step coordinates with force_short
        !
        !$omp parallel do private(k, jx, factor)
        do k = 1, ncell
          do jx = 1, natom(k)
            coord_ref(1:3,jx,k) = coord(1:3,jx,k)
            factor = 0.5_wp * dt_short / mass(jx,k)
            vel(1:3,jx,k) = vel(1:3,jx,k) + factor*force_short(1:3,jx,k)
            coord(1:3,jx,k) = coord(1:3,jx,k) + dt_short*vel(1:3,jx,k)
          end do
        end do
        !$omp end parallel do

        call timer(TimerIntegrator, TimerOff)

        ! Constraint (RATTLE VV1 for short)
        ! 
        if (constraints%rigid_bond) &
          call compute_constraints(ConstraintModeLEAP, .true., dt_short, &
                                   coord_ref, domain, constraints,         &
                                   coord, vel, dynvars%virial_const)

        ! cell migration and update cell pairlist
        !
        call timer(TimerComm1, TimerOn)

        if (dynamics%nbupdate_period > 0 .and. &
            j == multistep .and. i > 1) then

          call domain_interaction_update_md(istep, dynamics, domain, enefunc, &
                                          pairlist, boundary, constraints, comm)

        end if

        call timer(TimerComm1, TimerOff)

        ! short range forces
        !
        call communicate_coor(domain, comm)

        call compute_energy_short(domain, enefunc, pairlist, boundary, coord, &
                                  mod(istep,dynamics%eneout_period) == 0,  &
                                  dynvars%energy, &
                                  force_short,    &
                                  force_omp,      &
                                  dynvars%virial, &
                                  dynvars%virial_extern)

        call communicate_force(domain, comm, force_short)
        
        ! full step velocities with foce_short
        !
        call timer(TimerIntegrator, TimerOn)

        !$omp parallel do private(k, jx, factor)
        do k = 1, ncell
          do jx = 1, natom(k)
            factor = 0.5_wp * dt_short / mass(jx,k)
            vel(1:3,jx,k) = vel(1:3,jx,k) + factor * force_short(1:3,jx,k)
          end do
        end do
        !$omp end parallel do

        call timer(TimerIntegrator, TimerOff)

        ! Constraint (RATTLE VV2 for short)
        !
        if (constraints%rigid_bond) then
          call compute_constraints(ConstraintModeVVER2, .false., dt_short, &
                                   coord_ref, domain, constraints,         &
                                   coord, vel, dynvars%virial_const)

          dynvars%virial(1:3,1:3) = dynvars%virial(1:3,1:3) &
                                  + dynvars%virial_const(1:3,1:3)
        end if

      end do

      ! long range forces
      !
      call communicate_coor(domain, comm)

      call compute_energy_long(domain, enefunc, pairlist, boundary, coord, &
                               npt, dynvars%energy, force_long, force_omp, &
                               dynvars%virial)

      call communicate_force(domain, comm, force_long)


      ! full step velocities with force_long
      !
      call timer(TimerIntegrator, TimerOn)

      !$omp parallel do private(k, jx, factor)
      do k = 1, ncell
        do jx = 1, natom(k)
          factor = 0.5_wp * dt_long / mass(jx,k)
          vel(1:3,jx,k) = vel(1:3,jx,k) + factor * force_long(1:3,jx,k)
        end do
      end do
      !$omp end parallel do

      call timer(TimerIntegrator, TimerOff)

      ! constraint-corrected full step velocities
      !
      if (constraints%rigid_bond) then

        call compute_constraints(ConstraintModeVVER2, .false., dt_long, &
                                 coord_old, domain, constraints,        &
                                 coord, vel, dynvars%virial_const)

        dynvars%virial(1:3,1:3) = dynvars%virial(1:3,1:3) &
                                + dynvars%virial_const(1:3,1:3)
      end if

      ! temperature-corrected full step velocities
      !
      if (ensemble%ensemble == EnsembleNVT) then
  
        call control_temp_pres_vver2(dt_long, dynamics, ensemble, domain, &
                                     constraints, boundary, dynvars)

      end if

      ! energy output
      !
      if (dynamics%eneout_period > 0) then
        if (mod(istep,dynamics%eneout_period) == 0 ) then
          do k = 1, ncell
            do jx = 1, natom(k)
              force(1:3,jx,k) = force_short(1:3,jx,k) + force_long(1:3,jx,k)
            end do
          end do
        end if
      end if

      call output_md(output, enefunc, dynamics, boundary, ensemble, iseed, &
                     dynvars, domain)

    end do

    return

  end subroutine vverlet_respa_dynamics

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    initial_vverlet
  !> @brief        compute the first step (0+dt)
  !! @authors      JJ
  !! @param[in]    npt         : flag for NPT or not
  !! @param[in]    output      : output information
  !! @param[in]    enefunc     : potential energy functions information
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    pairlist    : pairlist information
  !! @param[in]    boundary    : boundary information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] dynvars     : dynamic variables information
  !! @param[inout] comm        : communication information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

   subroutine initial_vverlet(npt, output, enefunc, dynamics, pairlist, &
                              boundary, ensemble, constraints, domain,  &
                              dynvars, comm)

    ! formal arguments
    logical,                 intent(in)    :: npt
    type(s_output),          intent(in)    :: output
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_constraints),     intent(inout) :: constraints
    type(s_domain),  target, intent(inout) :: domain
    type(s_dynvars), target, intent(inout) :: dynvars
    type(s_comm),            intent(inout) :: comm

    ! local variables
    real(wp)                 :: factor, temperature, energy(18), temp(18)
    real(wp)                 :: imass, simtim, dt, friction
    integer                  :: i, ix, j, jx, ncell, k

    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:), mass(:,:)
    real(wp),        pointer :: force(:,:,:), force_omp(:,:,:,:)
    real(wp),        pointer :: force_short(:,:,:), force_long(:,:,:)
    integer,         pointer :: natom(:)


    natom       => domain%num_atom
    coord       => domain%coord
    coord_ref   => domain%coord_ref
    force       => domain%force
    force_short => domain%force_short
    force_long  => domain%force_long 
    force_omp   => domain%force_omp
    vel         => domain%velocity
    vel_ref     => domain%velocity_ref
    mass        => domain%mass

    dt          =  dynamics%timestep/AKMA_PS
    simtim      =  dynamics%initial_time
    temperature =  ensemble%temperature
    friction    =  ensemble%gamma_t * AKMA_PS
    ncell       =  domain%num_cell_local

    dynvars%time = simtim
    dynvars%step = 0

    ! save coordinates(0) and velocities(0)
    ! if rigid-body on, update coordinates(0)
    !
    do i = 1, ncell
      do ix = 1, natom(i)
        coord_ref(1:3,ix,i) = coord(1:3,ix,i)
        vel_ref(1:3,ix,i)   = vel(1:3,ix,i)
      end do
    end do

    if (constraints%rigid_bond) then

      call compute_constraints(ConstraintModeLEAP, .false., dt, coord_ref, &
                               domain, constraints, coord, vel,            &
                               dynvars%virial_const)

      do i = 1, ncell
        do ix = 1, natom(i)
          coord_ref(1:3,ix,i) = coord(1:3,ix,i)
        end do
      end do
    end if

    ! calculate energy(0) and forces(0)
    !
    call communicate_coor(domain, comm)

    call compute_energy_short(domain, enefunc, pairlist, boundary, coord, &
                              .true.,         &
                              dynvars%energy, &
                              force_short,    &
                              force_omp,      &
                              dynvars%virial, &
                              dynvars%virial_extern)

    call compute_energy_long (domain, enefunc, pairlist, boundary, coord, &
                              npt, dynvars%energy, force_long, force_omp, &
                              dynvars%virial)

    call communicate_force(domain, comm, force_short)

    call communicate_force(domain, comm, force_long)

    do i = 1, ncell
      do ix = 1, natom(i)
        force(1:3,ix,i) = force_short(1:3,ix,i) + force_long(1:3,ix,i)
      end do
    end do

    ! if rigid-body on, update velocity(0 + dt/2 and 0 - dt/2)
    !
    if (constraints%rigid_bond) then

      ! calculate velocities(0 + dt/2) and coordinates(0 + dt)
      ! update coordinates(0 + dt) and velocities(0 + dt/2)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          imass = 1.0_wp/mass(ix,i)
          vel(1:3,ix,i) = vel_ref(1:3,ix,i) + 0.5_wp*dt*force(1:3,ix,i)*imass
          coord(1:3,ix,i) = coord_ref(1:3,ix,i) + dt*vel(1:3,ix,i)
        end do
      end do

      call compute_constraints(ConstraintModeLEAP, .true., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               dynvars%virial_const)

      ! calculate velocities(0 - dt/2) and coordinates(0 - dt)
      ! update coordinates(0 - dt) and velocities(0 - dt/2)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          imass = 1.0_wp/mass(ix,i)
          vel_ref(1:3,ix,i) = vel_ref(1:3,ix,i) - 0.5_wp*dt*force(1:3,ix,i)*imass
          coord(1:3,ix,i) = coord_ref(1:3,ix,i) - dt*vel_ref(1:3,ix,i)
        end do
      end do

      call compute_constraints(ConstraintModeLEAP, .true., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               dynvars%virial_const)

      ! calculate velocity(0)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel_ref(1:3,ix,i) = 0.5_wp*(vel(1:3,ix,i) + vel_ref(1:3,ix,i))
        end do
      end do

      ! vel <= updated velocities (0) and coord <= constrained
      ! coordinates(0)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel(1:3,ix,i) = vel_ref(1:3,ix,i)
          coord(1:3,ix,i) = coord_ref(1:3,ix,i)
        end do
      end do

    end if

    ! output dynvars(0)
    !
    dynvars%time = 0.0_wp
    dynvars%step = 0

    call compute_dynvars(dynamics, boundary, ensemble, domain, dynvars)

    call output_dynvars(output, enefunc, dynvars)

    ! at this point
    !   coord, velocity, and force are at 0

    return

  end subroutine initial_vverlet

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    control_temp_pres_vver1
  !> @brief        driver to control temperature and pressure
  !! @authors      JJ
  !! @param[in]    dt          : delta t
  !! @param[in]    dynamics    : dynamics information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control_temp_pres_vver1(dt, dynamics, ensemble, domain, &
                                     constraints, boundary, dynvars)

    ! formal arguments
    real(wp),                intent(in)    :: dt
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_domain),          intent(inout) :: domain
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_dynvars),         intent(inout) :: dynvars

    ! Nose-Hoover thermostating may have to be skipped at mdstep = 1
    ! and Langevin dynamics may have to be Newtonian dynamics at mdstep = 1
    ! to get the identical trajectory with LEAP

    select case (ensemble%tpcontrol)

    case (TpcontrolLangevin)

    case (TpcontrolNoseHoover)

      call nosehoover_thermostat_vverlet(dt, dynamics, ensemble, domain, &
                                         dynvars)

    end select

    return

  end subroutine control_temp_pres_vver1

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    control_temp_pres_vver2
  !> @brief        driver to control temperature and pressure
  !! @authors      JJ
  !! @param[in]    dt          : delta t
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control_temp_pres_vver2(dt, dynamics, ensemble, domain, &
                                     constraints, boundary, dynvars)

    ! formal arguments
    real(wp),                intent(in)    :: dt
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),          intent(inout) :: domain
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_dynvars),         intent(inout) :: dynvars


    ! This subroutine may have to be skipped at mdstep = 1
    ! to get the identical trajectory with LEAP

    select case(ensemble%tpcontrol)

    case (TpcontrolBerendsen)

      ! Berendsen thermostat
      !

    case (TpcontrolAndersen)

      ! Andersen thermostat
      !

    case (TpcontrolEvans)

      ! Evans thermostat
      !

    case (TpcontrolNoseHoover)

      ! Nose-Hoover thermostat
      !
      call nosehoover_thermostat_vverlet(dt, dynamics, ensemble, domain, &
                                         dynvars)

    end select

    return

  end subroutine control_temp_pres_vver2

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    nosehoover_thermostat_vverlet
  !> @brief        control temperature using Nose-Hoover thermostat
  !! @authors      JJ
  !! @param[in]    dt       : delta t
  !! @param[in]    dynamics : dynamics information
  !! @param[in]    ensemble : ensemble information
  !! @param[inout] domain   : domain information
  !! @param[inout] dynvars  : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine nosehoover_thermostat_vverlet(dt, dynamics, ensemble, domain, &
                                           dynvars)

    ! formal arguments
    real(wp),                intent(in)    :: dt
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),  target, intent(inout) :: domain
    type(s_dynvars), target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: temp0, tau_t
    real(wp)                 :: ekf, tempf, factor1
    integer                  :: i, ix
    integer                  :: num_degree, ncell

    real(wp),        pointer :: vel(:,:,:), mass(:,:)
    real(wp),        pointer :: moment
    integer,         pointer :: natom(:)


    natom      => domain%num_atom
    vel        => domain%velocity
    mass       => domain%mass
    moment     => dynvars%thermostat_momentum

    temp0      =  ensemble%temperature
    tau_t      =  ensemble%tau_t/AKMA_PS
    num_degree =  domain%num_deg_freedom
    ncell      =  domain%num_cell_local

    ! calculate temperature(t or t + dt)
    !
    ekf = 0.0_wp
    do i = 1, ncell
      do ix = 1, natom(i)
        ekf = ekf + mass(ix,i)*(vel(1,ix,i)**2+vel(2,ix,i)**2+vel(3,ix,i)**2)
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, ekf, 1, mpi_wp_real, mpi_sum, &
                       mpi_comm_country, ierror)
#endif

    tempf = ekf/(domain%num_deg_freedom*KBOLTZ)

    ! calculate modified velocities(t or t + dt)
    !
    factor1 = tempf/temp0  - 1
    factor1 = factor1/(tau_t**2)

    ! update the Nose-Hoover factor
    !
    moment = moment + factor1*dt/4.0_wp

    ! scale velocities according to the updated Nose-Hoover factor
    !
    do i = 1, ncell
      do ix = 1, natom(i)
        vel(1:3,ix,i) = exp(-moment*dt/2.0_wp)*vel(1:3,ix,i)
      end do
    end do

    ! calculate temperature with updated velocities
    !
    ekf = 0.0_wp
    do i = 1, ncell
      do ix = 1, natom(i)
        ekf = ekf + mass(ix,i)*(vel(1,ix,i)**2+vel(2,ix,i)**2+vel(3,ix,i)**2)
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, ekf, 1, mpi_wp_real, mpi_sum, &
                       mpi_comm_country, ierror)
#endif

    tempf = ekf/(domain%num_deg_freedom*KBOLTZ)

    ! calculate modified velocities(t or t + dt)
    !
    factor1 = tempf/temp0  - 1
    factor1 = factor1/(tau_t**2)

    ! update the Nose-Hoover factor
    !
    moment = moment + factor1*dt/4.0_wp

    return

  end subroutine nosehoover_thermostat_vverlet

end module sp_md_respa_mod
