!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_md_vverlet_mod
!> @brief   perform molecular dynamics simulation with velocity verlet algorithm
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_md_vverlet_mod

  use sp_output_mod
  use sp_update_domain_mod
  use sp_assign_velocity_mod
  use sp_dynvars_mod
  use sp_constraints_mod
  use sp_communicate_mod
  use sp_energy_mod
  use sp_output_str_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_ensemble_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use random_mod
  use messages_mod
  use timers_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: vverlet_dynamics
  private :: initial_vverlet
  private :: control_temp_pres_vver1
  private :: control_temp_pres_vver2
  private :: nosehoover_thermostat_vv1
  private :: nosehoover_thermostat_vverlet
  private :: langevin_thermostat_vv1
  private :: langevin_thermostat_vv2
  private :: langevin_barostat_vv1
  private :: langevin_barostat_vv2
  private :: update_barostat
  private :: reduce_pres
  private :: bcast_boxsize

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    vverlet_dynamics
  !> @brief        velocity verlet integrator
  !! @authors      JJ
  !! @param[inout] output      : output information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamic variables information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : non-bond pair list information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] constraints : bond constraint information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] comm        : information of communication
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine vverlet_dynamics(output, domain, enefunc, dynvars, dynamics, &
                              pairlist, boundary, constraints, ensemble, comm)

    ! formal arguments
    type(s_output),          intent(inout) :: output
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynvars), target, intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_boundary),        intent(inout) :: boundary
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_comm),            intent(inout) :: comm

    ! local variables
    real(wp)                 :: simtim, dt, temperature, factor
    real(wp)                 :: energy(18), temp(18), temp_prev(18)
    integer                  :: ncell, nb
    integer                  :: i, j, k, jx, nsteps
    integer                  :: iseed, num_degree
    integer                  :: istart, iend
    logical                  :: npt

    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:)
    real(wp),        pointer :: force(:,:,:), mass(:,:), force_omp(:,:,:,:)
    integer,         pointer :: natom(:)


    natom       => domain%num_atom
    mass        => domain%mass
    coord       => domain%coord
    coord_ref   => domain%coord_ref
    force       => domain%force
    force_omp   => domain%force_omp
    vel         => domain%velocity
    vel_ref     => domain%velocity_ref

    ncell       =  domain%num_cell_local
    nb          =  domain%num_cell_boundary
    nsteps      =  dynamics%nsteps
    istart      =  dynamics%istart_step
    iend        =  dynamics%iend_step
    dt          =  dynamics%timestep/AKMA_PS
    simtim      =  dynamics%initial_time
    iseed       =  dynamics%iseed
    temperature =  ensemble%temperature

    ! initialize
    !
    simtim = 0.0_wp

    if (ensemble%ensemble == EnsembleNPT  .or. &
        ensemble%ensemble == EnsembleNPAT .or. &
        ensemble%ensemble == EnsembleNPgT) then
      npt = .true.
    else
      npt = .false.
    end if


    ! Check restart
    !
    if (.not. dynamics%restart) then

      call initial_velocity(temperature,           &
                            domain%num_atom_all,   &
                            domain%num_cell_local, &
                            domain%id_g2l,         &
                            domain%mass,           &
                            iseed,                 &
                            domain%velocity)

      call stop_trans_rotation(domain%num_cell_local,         &
                               domain%num_atom,               &
                               dynamics%stop_com_translation, &
                               dynamics%stop_com_rotation,    &
                               domain%mass,                   &
                               domain%coord,                  &
                               domain%velocity)

      call initial_vverlet(npt, output, enefunc, dynamics,       &
                           pairlist, boundary, ensemble, constraints, &
                           domain, dynvars, comm)

    else

      ! After 2nd cycle of REMD simulation (istart /= 1), this is skipped
      !
      if (istart == 1) then
        call communicate_coor(domain, comm)

        call compute_energy(domain, enefunc, pairlist, boundary, coord, &
                            npt, .false., mod(i,dynamics%eneout_period)==0, &
                            dynvars%energy, &
                            force,          &
                            force_omp,      &
                            dynvars%virial, &
                            dynvars%virial_extern)

        call communicate_force(domain, comm, force)
      end if

    end if


    ! Main loop
    !
    do i = istart, iend

      simtim = simtim + dynamics%timestep
      dynvars%time = simtim
      dynvars%step = i

      call timer(TimerIntegrator, TimerOn)

      do k = 1, ncell
        do jx = 1, natom(k)
          coord_ref(1:3,jx,k) = coord(1:3,jx,k)
          vel_ref  (1:3,jx,k) = vel  (1:3,jx,k)
        end do
      end do

      ! Compute velocities(t + 1/2 dt) and coordinates(t + dt)
      !
      if (ensemble%tpcontrol /= TpcontrolNoseHoover .and. &
          ensemble%tpcontrol /= TpcontrolLangevin) then

        ! VV1
        !
        !$omp parallel do private(k, jx, factor)
        do k = 1, ncell
          do jx = 1, natom(k)
            factor = 0.5_wp*dt / mass(jx,k)
            vel  (1:3,jx,k) = vel  (1:3,jx,k) + factor * force(1:3,jx,k)
            coord(1:3,jx,k) = coord(1:3,jx,k) + dt*vel(1:3,jx,k)
          end do
        end do
        !$omp end parallel do

        ! RATTLE VV1
        !
        if (constraints%rigid_bond) &
          call compute_constraints(ConstraintModeLEAP, .true., dt, coord_ref,&
                                   domain, constraints, coord, vel, &
                                   dynvars%virial_const)
          dynvars%virial_const(1:3,1:3) = 2.0_wp * dynvars%virial_const(1:3,1:3)

      else

        call control_temp_pres_vver1(dynamics, i, ensemble, domain, constraints, &
                                     boundary, dynvars)

      end if

      call timer(TimerIntegrator, TimerOff)
      ! update cell and pairlist
      !
      if (dynamics%nbupdate_period > 0 .and. i > 1) &
        call domain_interaction_update_md(i-1, dynamics, domain, enefunc, &
                                          pairlist, boundary, constraints, comm)


      ! calculate potential energy(t + dt), force(t + dt), and virial(t + dt)
      !
      call communicate_coor(domain, comm)

      call compute_energy(domain, enefunc, pairlist, boundary, coord, &
                          npt, .false., mod(i,dynamics%eneout_period)==0, &
                          dynvars%energy, &
                          force,          &
                          force_omp,      &
                          dynvars%virial, &
                          dynvars%virial_extern)

      call communicate_force(domain, comm, force)

      call timer(TimerIntegrator, TimerOn)

      if (ensemble%tpcontrol == TpcontrolLangevin) then

        if (ensemble%ensemble == EnsembleNPT) then

          call langevin_barostat_vv2  (dynamics, ensemble, domain, &
                                       constraints, boundary, dynvars)

        else if (ensemble%ensemble == EnsembleNVT) then

          call langevin_thermostat_vv2(dynamics, ensemble, domain, &
                                       constraints, dynvars)

        end if

      else

        ! Compute velocities(t + dt)
        !   v(t+dt) = v(t+1/2dt) + 0.5dt*F(t+dt)/m
        !
        !$omp parallel do private(k, jx, factor)
        do k = 1, ncell
          do jx = 1, natom(k)
            factor = 0.5_wp*dt / mass(jx,k)
            vel(1:3,jx,k) = vel(1:3,jx,k) + factor*force(1:3,jx,k)
          end do
        end do
        !$omp end parallel do

        ! Velocty constraint (RATTLE VV2)
        !   coord     is at t + dt, vel (unconstrained) is at t + dt
        !   coord_ref is at t,      vel_ref             is at t
        !   compute constrained velocities(t+dt)
        !   add constraint virial(t+dt)
        !   Note that constraint virial(t) is added to virial(t+dt) in leapfrog
        !             which is correct?
        !
        if (constraints%rigid_bond) then

          call compute_constraints(ConstraintModeVVER2, .false., dt, coord_ref,&
                                   domain, constraints, coord, vel, &
                                   dynvars%virial_const)

        end if

        ! control temperature VV2
        !
        if (ensemble%ensemble /= EnsembleNVE) then
  
          call control_temp_pres_vver2(dynamics, ensemble, domain, &
                                       constraints, boundary, dynvars)
        end if
        
      end if

      ! Remove translational and rotational motion about COM(t + dt)
      !
      if (dynamics%stoptr_period > 0) then

        if (mod(i,dynamics%stoptr_period) == 0) then

            call stop_trans_rotation(domain%num_cell_local,         &
                                     domain%num_atom,               &
                                     dynamics%stop_com_translation, &
                                     dynamics%stop_com_rotation,    &
                                     domain%mass,                   &
                                     domain%coord,                  &
                                     domain%velocity)
        end if

      end if

      call timer(TimerIntegrator, TimerOff)

      ! OUTPUT energy(t + dt), trajectory(t + dt), and restart data
      !   coord     is at t + dt, coord_ref    is at t
      !   vel       is at t + dt, vel_ref      is at t
      !   box_size  is at ??????, box_size_ref is at ??????
      !

      call output_md(output, enefunc, dynamics, boundary, ensemble, iseed, &
                     dynvars, domain)

      ! output parallel I/O restart
      !
      call output_prst_md(output, enefunc, dynamics, boundary, &
                                  dynvars, domain, constraints)

    end do


    return

  end subroutine vverlet_dynamics

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    initial_vverlet
  !> @brief        compute the first step (0+dt)
  !! @authors      JJ
  !! @param[in]    npt         : flag for NPT or not
  !! @param[in]    output      : output information
  !! @param[in]    enefunc     : potential energy functions information
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    pairlist    : pairlist information
  !! @param[in]    boundary    : boundary information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] dynvars     : dynamic variables information
  !! @param[inout] comm        : communication information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

   subroutine initial_vverlet(npt, output, enefunc, dynamics, pairlist, &
                              boundary, ensemble, constraints, domain,  &
                              dynvars, comm)

    ! formal arguments
    logical,                 intent(in)    :: npt
    type(s_output),          intent(in)    :: output
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_constraints),     intent(inout) :: constraints
    type(s_domain),  target, intent(inout) :: domain
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_comm),            intent(inout) :: comm

    ! local variables
    real(wp)                 :: factor, temperature, energy(18), temp(18)
    real(wp)                 :: imass, simtim, dt, friction
    integer                  :: i, ix, j, jx, ncell, k

    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:)
    real(wp),        pointer :: force(:,:,:), mass(:,:), force_omp(:,:,:,:)
    integer,         pointer :: natom(:)


    natom       => domain%num_atom
    coord       => domain%coord
    coord_ref   => domain%coord_ref
    force       => domain%force
    force_omp   => domain%force_omp
    vel         => domain%velocity
    vel_ref     => domain%velocity_ref
    mass        => domain%mass

    ncell       =  domain%num_cell_local
    temperature =  ensemble%temperature
    friction    =  ensemble%gamma_t * AKMA_PS
    dt          =  dynamics%timestep/AKMA_PS
    simtim      =  dynamics%initial_time

    dynvars%time = simtim
    dynvars%step = 0


    ! save coordinates(0) and velocities(0)
    ! if rigid-body on, update coordinates(0)
    !
    do j = 1, ncell
      do jx = 1, natom(j)
        coord_ref(1:3,jx,j) = coord(1:3,jx,j)
        vel_ref  (1:3,jx,j) = vel  (1:3,jx,j)
      end do
    end do

    if (constraints%rigid_bond) then

      call compute_constraints(ConstraintModeLEAP, .false., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               dynvars%virial_const)

      do i = 1, ncell
        do ix = 1, natom(i)
          coord_ref(1:3,ix,i) = coord(1:3,ix,i)
        end do
      end do

    end if

    ! calculate energy(0) and forces(0)
    !
    call communicate_coor(domain, comm)

    call compute_energy(domain, enefunc, pairlist, boundary, coord,  &
                        npt, .false., .true., &
                        dynvars%energy, &
                        force,          &
                        force_omp,      &
                        dynvars%virial, &
                        dynvars%virial_extern)

    call communicate_force(domain, comm, force)

    ! if rigid-body on, update velocity(0 + dt/2 and 0 - dt/2)
    !
    if (constraints%rigid_bond) then

      ! calculate velocities(0 + dt/2) and coordinates(0 + dt)
      ! update coordinates(0 + dt) and velocities(0 + dt/2)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          imass = 1.0_wp/mass(ix,i)
          vel  (1:3,ix,i) = vel_ref  (1:3,ix,i) + 0.5_wp*dt*force(1:3,ix,i)*imass
          coord(1:3,ix,i) = coord_ref(1:3,ix,i) + dt*vel(1:3,ix,i)
        end do
      end do

      call compute_constraints(ConstraintModeLEAP, .false., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               dynvars%virial_const)
      dynvars%virial_const(1:3,1:3) = 2.0_wp * dynvars%virial_const(1:3,1:3)
      !
      ! please check
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel(1:3,ix,i) = (coord(1:3,ix,i) - coord_ref(1:3,ix,i))/dt
        end do
      end do

      ! calculate velocities(0 - dt/2) and coordinates(0 - dt)
      ! update coordinates(0 - dt) and velocities(0 - dt/2)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          imass = 1.0_wp/mass(ix,i)
          vel_ref(1:3,ix,i) = vel_ref(1:3,ix,i) - 0.5_wp*dt*force(1:3,ix,i)*imass
          coord  (1:3,ix,i) = coord_ref(1:3,ix,i) - dt*vel_ref(1:3,ix,i)
        end do
      end do

      call compute_constraints(ConstraintModeLEAP, .false., -dt, coord_ref, &
                               domain, constraints, coord, vel_ref, &
                               dynvars%virial_const)
      dynvars%virial_const(1:3,1:3) = 2.0_wp * dynvars%virial_const(1:3,1:3)

      !
      ! please check
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel_ref(1:3,ix,i) = (coord_ref(1:3,ix,i) - coord(1:3,ix,i))/dt
        end do
      end do

      ! calculate velocity(0)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel_ref(1:3,ix,i) = 0.5_wp*(vel(1:3,ix,i) + vel_ref(1:3,ix,i))
        end do
      end do

      ! vel <= updated velocities (0) and coord <= constrained coordinates(0)
      !
      do i = 1, ncell
        do ix = 1, natom(i)
          vel(1:3,ix,i) = vel_ref(1:3,ix,i)
          coord(1:3,ix,i) = coord_ref(1:3,ix,i)
        end do
      end do

    end if

    ! output dynvars(0)
    !
    call compute_dynvars(dynamics, boundary, ensemble, domain, dynvars)

    call output_dynvars(output, enefunc, dynvars)

    dynamics%restart = .true.

    return

  end subroutine initial_vverlet

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    control_temp_pres_vver1
  !> @brief        driver to control temperature and pressure
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control_temp_pres_vver1(dynamics, istep, ensemble, domain, &
                                     constraints, boundary, dynvars)

    ! formal arguments
    type(s_dynamics),        intent(in)    :: dynamics
    integer,                 intent(in)    :: istep
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_domain),          intent(inout) :: domain
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_dynvars),         intent(inout) :: dynvars

    ! local variable
    integer :: alloc_stat, ncell
 
    ! Nose-Hoover thermostating may have to be skipped at mdstep = 1
    ! and Langevin dynamics may have to be Newtonian dynamics at mdstep = 1
    ! to get the identical trajectory with LEAP

    if (ensemble%tpcontrol == TpcontrolLangevin) then

      if (ensemble%ensemble == EnsembleNVT) then

        if (istep == 1) then
          ncell = domain%num_cell_local
          allocate(ensemble%random_force(3,MaxAtom,ncell), stat=alloc_stat) 
          if (alloc_stat /= 0) call error_msg_alloc
        end if
        call langevin_thermostat_vv1(dynamics, istep, ensemble, domain,  &
                                     constraints, dynvars)

      else if (ensemble%ensemble == EnsembleNPT) then

        if (istep == 1) then
          ncell = domain%num_cell_local
          allocate(ensemble%random_force(3,MaxAtom,ncell), stat=alloc_stat) 
          if (alloc_stat /= 0) call error_msg_alloc
        end if
        call langevin_barostat_vv1  (dynamics, istep, ensemble, domain,  &
                                     constraints, boundary, dynvars)

      end if

    else if (ensemble%tpcontrol == TpcontrolNoseHoover) then

      if (ensemble%ensemble == EnsembleNVT) then

        call nosehoover_thermostat_vv1(dynamics, ensemble, domain, constraints,&
                                       dynvars)
      end if

    end if

    return

  end subroutine control_temp_pres_vver1

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    control_temp_pres_vver2
  !> @brief        driver to control temperature and pressure
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine control_temp_pres_vver2(dynamics, ensemble, domain, constraints, &
                                     boundary, dynvars)

    ! formal arguments
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),          intent(inout) :: domain
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_dynvars),         intent(inout) :: dynvars


    ! This subroutine may have to be skipped at mdstep = 1
    ! to get the identical trajectory with LEAP

    select case (ensemble%tpcontrol)

    case (TpcontrolBerendsen)

      ! Berendsen thermostat
      !

    case (TpcontrolAndersen)

      ! Andersen thermostat
      !

    case (TpcontrolEvans)

      ! Evans thermostat
      !

    case (TpcontrolNoseHoover)

      ! Nose-Hoover thermostat
      ! 
      call nosehoover_thermostat_vverlet(dynamics, ensemble, domain, dynvars)

    end select

    return

  end subroutine control_temp_pres_vver2

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    nosehoover_thermostat_vv1
  !> @brief        control temperature using Nose-Hoover thermostat at VV1
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine nosehoover_thermostat_vv1(dynamics, ensemble, domain, &
                                       constraints, dynvars)

    ! formal arguments
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),  target, intent(inout) :: domain
    type(s_constraints),     intent(inout) :: constraints
    type(s_dynvars),         intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: dt
    real(wp)                 :: factor
    integer                  :: i, ix
    integer                  :: num_degree, ncell

    real(wp),        pointer :: vel(:,:,:), coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: force(:,:,:), mass(:,:)
    integer,         pointer :: natom(:)


    dt         =  dynamics%timestep/AKMA_PS
    ncell      =  domain%num_cell_local

    natom      => domain%num_atom
    vel        => domain%velocity
    coord      => domain%coord
    coord_ref  => domain%coord_ref
    force      => domain%force
    mass       => domain%mass


    ! thermostat
    !
    call nosehoover_thermostat_vverlet(dynamics, ensemble, domain, dynvars)

    ! VV1
    !
    !$omp parallel do private(i, ix, factor)
    do i = 1, ncell
      do ix = 1, natom(i)
        factor = 0.5_wp*dt / mass(ix,i)
        vel  (1:3,ix,i) = vel  (1:3,ix,i) + factor * force(1:3,ix,i)
        coord(1:3,ix,i) = coord(1:3,ix,i) + dt*vel(1:3,ix,i)
      end do
    end do
    !$omp end parallel do

    ! RATTLE VV1
    !
    if (constraints%rigid_bond) &
      call compute_constraints(ConstraintModeLEAP, .true., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               dynvars%virial_const)
      dynvars%virial_const(1:3,1:3) = 2.0_wp * dynvars%virial_const(1:3,1:3)

    return

  end subroutine nosehoover_thermostat_vv1

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    nosehoover_thermostat_vverlet
  !> @brief        control temperature using Nose-Hoover thermostat
  !! @authors      JJ
  !! @param[in]    dynamics : dynamics information
  !! @param[in]    ensemble : ensemble information
  !! @param[inout] domain   : domain information
  !! @param[inout] dynvars  : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine nosehoover_thermostat_vverlet(dynamics, ensemble, domain, dynvars)

    ! formal arguments
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),  target, intent(inout) :: domain
    type(s_dynvars), target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: temp0, tau_t, dt
    real(wp)                 :: ekf, tempf, factor1
    integer                  :: i, ix
    integer                  :: num_degree, ncell

    real(wp),        pointer :: vel(:,:,:), mass(:,:)
    real(wp),        pointer :: moment
    integer,         pointer :: natom(:)


    dt          =  dynamics%timestep/AKMA_PS
    temp0       =  ensemble%temperature
    tau_t       =  ensemble%tau_t/AKMA_PS
    num_degree  =  domain%num_deg_freedom
    ncell       =  domain%num_cell_local

    natom       => domain%num_atom
    vel         => domain%velocity
    mass        => domain%mass
    moment      => dynvars%thermostat_momentum

    ! calculate temperature(t or t + dt)
    !
    ekf = 0.0_wp
    do i = 1, ncell
      do ix = 1, natom(i)
        ekf = ekf + &
             mass(ix,i)*(vel(1,ix,i)*vel(1,ix,i) + vel(2,ix,i)*vel(2,ix,i) + &
                         vel(3,ix,i)*vel(3,ix,i))
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, ekf, 1, mpi_wp_real, mpi_sum, &
                       mpi_comm_country, ierror)
#endif

    tempf = ekf/(domain%num_deg_freedom*KBOLTZ)

    ! calculate modified velocities(t or t + dt)
    !
    factor1 = tempf/temp0  - 1
    factor1 = factor1/(tau_t**2)

    ! update the Nose-Hoover factor
    !
    moment = moment + factor1*dt/4.0_wp

    ! scale velocities according to the updated Nose-Hoover factor
    ! 
    do i = 1, ncell
      do ix = 1, natom(i)
        vel(1:3,ix,i) = exp(-moment*dt/2.0_wp)*vel(1:3,ix,i)
      end do
    end do

    ! calculate temperature with updated velocities
    !
    ekf = 0.0_wp
    do i = 1, ncell
      do ix = 1, natom(i)
        ekf = ekf + &
             mass(ix,i)*(vel(1,ix,i)**2 + vel(2,ix,i)**2 + vel(3,ix,i)**2)
      end do
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, ekf, 1, mpi_wp_real, mpi_sum, &
                       mpi_comm_country, ierror)
#endif

    tempf = ekf/(domain%num_deg_freedom*KBOLTZ)

    ! calculate modified velocities(t or t + dt)
    !
    factor1 = tempf/temp0  - 1
    factor1 = factor1/(tau_t**2)

    ! update the Nose-Hoover factor
    !
    moment = moment + factor1*dt/4.0_wp

    return

  end subroutine nosehoover_thermostat_vverlet

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    langevin_thermostat_vv1
  !> @brief        control temperature using Langevin thermostat
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine langevin_thermostat_vv1(dynamics, istep, ensemble, domain,  &
                                     constraints, dynvars)

    ! formal arguments
    type(s_dynamics),         intent(in)    :: dynamics
    integer,                  intent(in)    :: istep
    type(s_ensemble), target, intent(in)    :: ensemble
    type(s_domain),   target, intent(inout) :: domain
    type(s_constraints),      intent(inout) :: constraints
    type(s_dynvars),  target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: imass, dt, half_dt, inv_dt
    real(wp)                 :: temp0, gamma_t, scale_v
    real(wp)                 :: factor, sigma
    real(wp)                 :: rsq, v1, v2, grandom(1:3)
    real(wp)                 :: kBT, vel_tmp(1:3)
    integer                  :: j, jx, k, ncell

    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:)
    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: force(:,:,:), mass(:,:)
    real(wp),        pointer :: random_f(:,:,:)
    real(wp),        pointer :: temporary(:,:,:,:)
    real(wp),        pointer :: virial(:,:), viri_const(:,:)
    integer,         pointer :: natom(:)


    dt            =  dynamics%timestep/AKMA_PS
    half_dt       =  dt/2.0_wp
    inv_dt        =  1.0_wp/dt
    temp0         =  ensemble%temperature
    gamma_t       =  ensemble%gamma_t *AKMA_PS
    random_f      => ensemble%random_force
    ncell         =  domain%num_cell_local

    natom         => domain%num_atom
    mass          => domain%mass
    coord         => domain%coord
    coord_ref     => domain%coord_ref
    vel           => domain%velocity
    vel_ref       => domain%velocity_ref
    force         => domain%force
    temporary     => domain%force_omp
    virial        => dynvars%virial
    viri_const    => dynvars%virial_const

    ! setup variables
    !
    kBT      = KBOLTZ * temp0

    ! scale factor for velocities
    !
    half_dt = dt / 2.0_wp
    scale_v = exp(- gamma_t*half_dt)

    ! random force
    !
    if (istep == 1) then

      factor   = 2.0_wp*gamma_t*KBOLTZ*temp0/half_dt

      do j = 1, ncell
        do jx = 1, natom(j)

          sigma = sqrt(mass(jx,j) * factor)
          rsq = 2.0_wp

          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(1) = rsq * v1
          rsq = 2.0_wp

          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(2) = rsq * v1
          rsq = 2.0_wp

          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(3) = rsq * v1
          random_f(1:3,jx,j) = sigma*grandom(1:3)

        end do
      end do

    end if

    ! thermostat1 (scale velocities, cooling)
    !
    do j = 1, ncell
      do jx = 1, natom(j)
        vel(1:3,jx,j) = vel_ref(1:3,jx,j)*scale_v
      end do
    end do

    ! VV1
    !
    do j = 1, ncell
      do jx = 1, natom(j)
        factor = half_dt / mass(jx,j)
        vel  (1:3,jx,j) = vel(1:3,jx,j) + factor*force(1:3,jx,j)
        vel  (1:3,jx,j) = vel(1:3,jx,j) + factor*random_f(1:3,jx,j)
        coord(1:3,jx,j) = coord_ref(1:3,jx,j)+vel(1:3,jx,j)*dt
      end do
    end do

    ! Coordinate constraint (RATTLE VV1)
    !
    if (constraints%rigid_bond) then
      do j = 1, ncell
        do jx = 1, natom(j)
          temporary(1,jx,j,1) = coord(1,jx,j)
          temporary(2,jx,j,1) = coord(2,jx,j)
          temporary(3,jx,j,1) = coord(3,jx,j)
        end do
      end do

      call compute_constraints(ConstraintModeLEAP, .false., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               viri_const)
      dynvars%virial_const(1:3,1:3) = 2.0_wp * dynvars%virial_const(1:3,1:3)

      do j = 1, ncell
        do jx = 1, natom(j)
          vel_tmp(1:3) = (coord(1:3,jx,j)-temporary(1:3,jx,j,1)) * inv_dt
          vel(1:3,jx,j) = vel(1:3,jx,j) + vel_tmp(1:3)
        end do
      end do
    end if

    return

  end subroutine langevin_thermostat_vv1

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    langevin_thermostat_vv2
  !> @brief        Langevin thermostat and barostat
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine langevin_thermostat_vv2(dynamics, ensemble, domain, constraints, &
                                     dynvars)

    ! formal arguments
    type(s_dynamics),         intent(in)    :: dynamics
    type(s_ensemble), target, intent(in)    :: ensemble
    type(s_domain),   target, intent(inout) :: domain
    type(s_constraints),      intent(inout) :: constraints
    type(s_dynvars),  target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: dt, inv_dt, temp0
    real(wp)                 :: scale_v, factor
    real(wp)                 :: gamma_t
    real(wp)                 :: sigma
    real(wp)                 :: v1, v2, rsq, grandom(1:3)
    real(wp)                 :: half_dt, quart_dt
    integer                  :: i, j, jx, ncell

    real(wp),        pointer :: mass(:,:), viri_const(:,:)
    real(wp),        pointer :: random_f(:,:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:), force(:,:,:)
    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    integer,         pointer :: natom(:)


    dt         =  dynamics%timestep/AKMA_PS
    inv_dt     =  1.0_wp/dt
    temp0      =  ensemble%temperature
    gamma_t    =  ensemble%gamma_t * AKMA_PS
    random_f   => ensemble%random_force
    ncell      =  domain%num_cell_local

    mass       => domain%mass
    natom      => domain%num_atom
    coord      => domain%coord
    coord_ref  => domain%coord_ref
    vel        => domain%velocity
    vel_ref    => domain%velocity_ref
    force      => domain%force
    viri_const => dynvars%virial_const

    ! time step
    !
    half_dt = dt / 2.0_wp

    ! random force
    !
    factor   = 2.0_wp*gamma_t*KBOLTZ*temp0/dt
    do j = 1, ncell
      do jx = 1, natom(j)

        sigma = sqrt(mass(jx,j) * factor)
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(1) = rsq * v1
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(2) = rsq * v1
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(3) = rsq * v1
        random_f(1:3,jx,j) = sigma*grandom(1:3)

      end do
    end do

    ! VV2
    !
    do j = 1, ncell
      do jx = 1, natom(j)
        factor = half_dt/mass(jx,j)
        vel(1:3,jx,j) = vel(1:3,jx,j) + factor*random_f(1:3,jx,j)
        vel(1:3,jx,j) = vel(1:3,jx,j) + factor*force(1:3,jx,j)
      end do
    end do

    ! RATTLE VV2
    !
    if (constraints%rigid_bond) &
      call compute_constraints(ConstraintModeVVER2, .false., dt, coord_ref, &
                               domain, constraints, coord, vel, &
                               viri_const)

    ! thermostat
    !
    scale_v = exp(-gamma_t*half_dt)
    do j = 1, ncell
      do jx = 1, natom(j)
        vel(1:3,jx,j) = vel(1:3,jx,j)*scale_v
      end do
    end do

    return

  end subroutine langevin_thermostat_vv2

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    langevin_barostat_vv1
  !> @brief        Langevin thermostat and barostat
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine langevin_barostat_vv1(dynamics, istep, ensemble, domain,   &
                                   constraints, boundary, dynvars)

    ! formal arguments
    type(s_dynamics),         intent(in)    :: dynamics
    integer,                  intent(in)    :: istep
    type(s_ensemble), target, intent(inout) :: ensemble
    type(s_domain),   target, intent(inout) :: domain
    type(s_constraints),      intent(inout) :: constraints
    type(s_boundary),         intent(inout) :: boundary
    type(s_dynvars),  target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: dt, inv_dt, temp0, press0, d_ndegf
    real(wp)                 :: kin(1:3), ekin
    real(wp)                 :: vel_tmp(1:3)
    real(wp)                 :: volume, press(1:3)
    real(wp)                 :: pressxy, pressxyz
    real(wp)                 :: crdx, crdy, crdz, factor
    real(wp)                 :: bmoment_ref(3), scale_b(1:3)
    real(wp)                 :: gamma_t, gamma_p
    real(wp)                 :: sigma
    real(wp)                 :: v1, v2, rsq, grandom(1:3)
    real(wp)                 :: half_dt, quart_dt, size_scale(1:3), vel_scale
    real(wp)                 :: virial_constraint(3,3), virial_sum(3)
    integer                  :: i, j, ij, jx, i_ndegf, maxiter
    integer                  :: ncell, nboundary

    real(wp),        pointer :: pmass, pforce(:), random_f(:,:,:)
    real(wp),        pointer :: mass(:,:)
    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: temporary(:,:,:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:), force(:,:,:)
    real(wp),        pointer :: virial(:,:), viri_const(:,:)
    real(wp),        pointer :: bmoment(:)
    integer,         pointer :: natom(:)


    dt         =  dynamics%timestep/AKMA_PS
    inv_dt     =  1.0_wp/dt
    temp0      =  ensemble%temperature
    press0     =  ensemble%pressure * ATMOS_P
    gamma_t    =  ensemble%gamma_t * AKMA_PS
    gamma_p    =  ensemble%gamma_p * AKMA_PS
    pmass      => ensemble%pmass
    pforce     => ensemble%pforce
    random_f   => ensemble%random_force
    i_ndegf    =  domain%num_deg_freedom
    d_ndegf    =  real(i_ndegf,wp)
    ncell      =  domain%num_cell_local
    nboundary  =  domain%num_cell_boundary

    natom      => domain%num_atom
    mass       => domain%mass
    coord      => domain%coord
    coord_ref  => domain%coord_ref
    vel        => domain%velocity
    vel_ref    => domain%velocity_ref
    force      => domain%force
    temporary  => domain%force_omp
    virial     => dynvars%virial
    viri_const => dynvars%virial_const
    bmoment    => dynvars%barostat_momentum

    ! save box size(t+dt) and compute volume(t+dt)
    !
    boundary%box_size_x_ref = boundary%box_size_x
    boundary%box_size_y_ref = boundary%box_size_y
    boundary%box_size_z_ref = boundary%box_size_z
    volume = boundary%box_size_x * boundary%box_size_y * boundary%box_size_z

    ! time step
    !
    half_dt  = dt / 2.0_wp
    quart_dt = dt / 4.0_wp

    ! maximum iteration
    !
    if (constraints%rigid_bond) then
      maxiter = 4
    else
      maxiter = 1
    end if

    ! constraint virial
    !
    virial_constraint(1:3,1:3) = 0.0_wp

    ! barostate coefficient
    !
    bmoment_ref(1:3) = bmoment(1:3)

    ! pmass and stochastic force (pforce)
    !
    if (istep == 1) then

      if (ensemble%isotropy == IsotropyISO) then

        pmass = real(i_ndegf+3,wp)*KBOLTZ*temp0 / (2.0_wp*PI*gamma_p)**2
        sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
        pforce(1) = sigma * gauss()
        pforce(2) = pforce(1)
        pforce(3) = pforce(1)

      else if (ensemble%isotropy == IsotropySEMI_ISO) then

        pmass = real(i_ndegf+3,wp)*KBOLTZ*temp0 / (3.0_wp*(2.0_wp*PI*gamma_p)**2)
        sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
        pforce(1) = sigma * gauss()
        pforce(2) = pforce(1)
        pforce(3) = sigma * gauss()

      else if (ensemble%isotropy == IsotropyANISO) then

        pmass = real(i_ndegf+3,wp)*KBOLTZ*temp0 / (3.0_wp*(2.0_wp*PI*gamma_p)**2)
        sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
        pforce(1) = sigma * gauss()
        pforce(2) = sigma * gauss()
        pforce(3) = sigma * gauss()

      else if (ensemble%isotropy == IsotropyXY_Fixed) then

        pmass = real(i_ndegf+3,wp)*KBOLTZ*temp0 / (3.0_wp*(2.0_wp*PI*gamma_p)**2)
        sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
        pforce(1) = 0.0_wp
        pforce(2) = 0.0_wp
        pforce(3) = sigma * gauss()

      end if

#ifdef MPI
      call mpi_bcast(pforce, 3, mpi_wp_real, 0, mpi_comm_country, ierror)
#endif

      ! random force
      !
      factor   = 2.0_wp*gamma_t*KBOLTZ*temp0/half_dt

      do j = 1, ncell
        do jx = 1, natom(j)

          sigma = sqrt(mass(jx,j) * factor)
          rsq = 2.0_wp

          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(1) = rsq * v1
          rsq = 2.0_wp
  
          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(2) = rsq * v1
          rsq = 2.0_wp

          do while (rsq >= 1.0_wp)
            v1  = 2.0_wp*grnd() - 1.0_wp
            v2  = 2.0_wp*grnd() - 1.0_wp
            rsq = v1*v1 + v2*v2
          end do

          rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
          grandom(3) = rsq * v1
          random_f(1:3,jx,j) = sigma*grandom(1:3)

        end do
      end do

    end if

    do i = 1, maxiter

      ! thermostat1 (scale velocities)
      !
      vel_scale = exp(-gamma_t*quart_dt)
      do j = 1, ncell
        do jx = 1, natom(j)
          vel(1:3,jx,j) = vel_ref(1:3,jx,j)*vel_scale
        end do
      end do

      ! compute kinetic energy(t+dt)
      !   v(t+dt) = 0.5*(v(t+3/2dt) + v(t+1/2dt))
      !
      kin(1:3) = 0.0_wp

      do j = 1, ncell
        do jx = 1, natom(j)
          kin(1:3)  = kin(1:3) + mass(jx,j) * vel(1:3,jx,j)*vel(1:3,jx,j)
        end do
      end do
      ekin   = 0.5_wp * (kin(1)+kin(2)+kin(3))

      ! virial + virial_constraint
      !
      virial_sum(1) = virial(1,1) + virial_constraint(1,1)
      virial_sum(2) = virial(2,2) + virial_constraint(2,2)
      virial_sum(3) = virial(3,3) + virial_constraint(3,3)

      call reduce_pres(kin, ekin, virial_sum)

      ! compute pressure(t+dt) in the unit of kcal/mol*A3
      !
      press(1:3) = (kin(1:3) + virial_sum(1:3))/volume
      pressxyz = (press(1)+press(2)+press(3))/3.0_wp
      pressxy  = (press(1)+press(2))/2.0_wp

      ! update barostat
      !
      call update_barostat(ensemble, boundary, bmoment_ref, pforce, ensemble%isotropy, press(1), &
                           press(2), press(3), pressxyz, pressxy, press0,    &
                           volume, d_ndegf, pmass, gamma_p, ekin, half_dt,   &
                           quart_dt, natom, ncell, vel, bmoment)

      ! thermostat2 (scale velocities)
      !
      vel_scale = exp(-gamma_t*quart_dt)
      do j = 1, ncell
        do jx = 1, natom(j)
          vel(1:3,jx,j) = vel(1:3,jx,j)*vel_scale
        end do
      end do

      ! VV1
      !
      size_scale(1:3) = exp(bmoment(1:3)*dt)
      do j = 1, ncell
        do jx = 1, natom(j)
          factor = half_dt / mass(jx,j)
          vel  (1:3,jx,j) = vel(1:3,jx,j) + factor*force(1:3,jx,j)
          vel  (1:3,jx,j) = vel(1:3,jx,j) + factor*random_f(1:3,jx,j)
          coord(1:3,jx,j) = size_scale(1:3)*coord_ref(1:3,jx,j)+vel(1:3,jx,j)*dt
        end do
      end do

      ! RATTLE VV1
      !
      if (constraints%rigid_bond) then
        do j = 1, ncell
          do jx = 1, natom(j)
            temporary(1:3,jx,j,1) = coord(1:3,jx,j)
          end do
        end do
        call compute_constraints(ConstraintModeLEAP, .false., dt, coord_ref, &
                                 domain, constraints, coord, vel, &
                                 viri_const)

        virial_constraint(1:3,1:3) = virial_constraint(1:3,1:3) &
                                   + 2.0_wp * viri_const(1:3,1:3)

        do j = 1, ncell
          do jx = 1, natom(j)
            vel_tmp(1:3) = (coord(1:3,jx,j)-temporary(1:3,jx,j,1)) * inv_dt
            vel(1:3,jx,j) = vel(1:3,jx,j) + vel_tmp(1:3)
            force(1:3,jx,j) = force(1:3,jx,j) + mass(jx,j)*vel_tmp(1:3)/half_dt
          end do
        end do
      end if

    end do

    viri_const(1:3,1:3) = virial_constraint(1:3,1:3)

    ! compute box size(t+2dt)
    !   size(t+2dt) = exp[eta(t+3/2dt)*dt] * size(t+dt)
    !
    scale_b(1:3) = exp(bmoment(1:3)*dt)
    boundary%box_size_x = scale_b(1) * boundary%box_size_x_ref
    boundary%box_size_y = scale_b(2) * boundary%box_size_y_ref
    boundary%box_size_z = scale_b(3) * boundary%box_size_z_ref

    call bcast_boxsize(boundary%box_size_x, boundary%box_size_y, &
                       boundary%box_size_z)

    boundary%cell_size_x = boundary%box_size_x / real(boundary%num_cells_x,wp)
    boundary%cell_size_y = boundary%box_size_y / real(boundary%num_cells_y,wp)
    boundary%cell_size_z = boundary%box_size_z / real(boundary%num_cells_z,wp)

    ! update boudary conditions
    !
    dynvars%barostat_momentum(1:3) = bmoment(1:3)
    do j = 1, ncell+nboundary
      do jx = 1, natom(j)
        domain%trans_vec(1:3,jx,j) = domain%trans_vec(1:3,jx,j) * scale_b(1:3)
      end do
    end do

    do ij = 1, maxcell

      i = domain%cell_pairlist1(1,ij)
      j = domain%cell_pairlist1(2,ij)
      if (domain%cell_move(1,j,i) > 0.0_wp) &
        domain%cell_move(1,j,i) = boundary%box_size_x
      if (domain%cell_move(2,j,i) > 0.0_wp) &
        domain%cell_move(2,j,i) = boundary%box_size_y
      if (domain%cell_move(3,j,i) > 0.0_wp) &
        domain%cell_move(3,j,i) = boundary%box_size_z
      if (domain%cell_move(1,j,i) < 0.0_wp) &
        domain%cell_move(1,j,i) = -boundary%box_size_x
      if (domain%cell_move(2,j,i) < 0.0_wp) &
        domain%cell_move(2,j,i) = -boundary%box_size_y
      if (domain%cell_move(3,j,i) < 0.0_wp) &
        domain%cell_move(3,j,i) = -boundary%box_size_z
    end do

    return

  end subroutine langevin_barostat_vv1

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    langevin_barostat_vv2
  !> @brief        Langevin thermostat and barostat
  !! @authors      JJ
  !! @param[in]    dynamics    : dynamics information
  !! @param[in]    ensemble    : ensemble information
  !! @param[inout] domain      : domain information
  !! @param[inout] constraints : constraints information
  !! @param[inout] boundary    : boundary information
  !! @param[inout] dynvars     : dynamic variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine langevin_barostat_vv2(dynamics, ensemble, domain, constraints, &
                                   boundary, dynvars)

    ! formal arguments
    type(s_dynamics),         intent(in)    :: dynamics
    type(s_ensemble), target, intent(inout) :: ensemble
    type(s_domain),   target, intent(inout) :: domain
    type(s_constraints),      intent(inout) :: constraints
    type(s_boundary),         intent(inout) :: boundary
    type(s_dynvars),  target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: dt, inv_dt, temp0, press0, d_ndegf
    real(wp)                 :: kin(1:3), ekin
    real(wp)                 :: vel_tmp(1:3)
    real(wp)                 :: volume, press(1:3)
    real(wp)                 :: pressxy, pressxyz
    real(wp)                 :: factor
    real(wp)                 :: bmoment_ref(3)
    real(wp)                 :: gamma_t, gamma_p
    real(wp)                 :: sigma, virial_constraint(1:3,1:3)
    real(wp)                 :: virial_sum(1:3)
    real(wp)                 :: v1, v2, rsq, grandom(1:3)
    real(wp)                 :: half_dt, quart_dt, size_scale(1:3), vel_scale
    real(wp)                 :: vel_change(1:3)
    integer                  :: i, j, jx, i_ndegf, ncell, maxiter

    real(wp),        pointer :: pmass, pforce(:), random_f(:,:,:)
    real(wp),        pointer :: mass(:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:)
    real(wp),        pointer :: force(:,:,:), force_add(:,:,:,:)
    real(wp),        pointer :: coord(:,:,:), coord_ref(:,:,:)
    real(wp),        pointer :: virial(:,:), viri_const(:,:)
    real(wp),        pointer :: bmoment(:)
    integer,         pointer :: natom(:)


    dt         =  dynamics%timestep/AKMA_PS
    inv_dt     =  1.0_wp/dt
    temp0      =  ensemble%temperature
    press0     =  ensemble%pressure * ATMOS_P
    gamma_t    =  ensemble%gamma_t * AKMA_PS
    gamma_p    =  ensemble%gamma_p * AKMA_PS
    pmass      => ensemble%pmass
    pforce     => ensemble%pforce
    random_f   => ensemble%random_force
    i_ndegf    =  domain%num_deg_freedom
    d_ndegf    =  real(i_ndegf,wp)
    ncell      =  domain%num_cell_local

    mass       => domain%mass
    natom      => domain%num_atom
    coord      => domain%coord
    coord_ref  => domain%coord_ref
    vel        => domain%velocity
    vel_ref    => domain%velocity_ref
    force      => domain%force
    force_add  => domain%force_omp
    virial     => dynvars%virial
    viri_const => dynvars%virial_const
    bmoment    => dynvars%barostat_momentum

    ! save box size(t+dt) and compute volume(t+dt)
    !
    boundary%box_size_x_ref = boundary%box_size_x
    boundary%box_size_y_ref = boundary%box_size_y
    boundary%box_size_z_ref = boundary%box_size_z
    volume = boundary%box_size_x * boundary%box_size_y * boundary%box_size_z

    ! time step
    !
    half_dt = dt / 2.0_wp
    quart_dt = dt / 4.0_wp

    ! initial constraint virial
    !
    virial_constraint(1:3,1:3) = 0.0_wp

    ! initial constraint force
    !
    do j = 1, ncell
      do jx = 1, natom(j)
        force_add(1:3,jx,j,1) = 0.0_wp
      end do
    end do
 
    ! maximum iteration
    !
    if (constraints%rigid_bond) then
      maxiter = 4
    else 
      maxiter = 1
    end if

    ! barostate coefficient
    !
    bmoment_ref(1:3) = bmoment(1:3)

    ! pmass and stochastic force (pforce)
    !
    if (ensemble%isotropy == IsotropyISO) then

      sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
      pforce(1) = sigma * gauss()
      pforce(2) = pforce(1)
      pforce(3) = pforce(1)

    else if (ensemble%isotropy == IsotropySEMI_ISO) then

      sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
      pforce(1) = sigma * gauss()
      pforce(2) = pforce(1)
      pforce(3) = sigma * gauss()

    else if (ensemble%isotropy == IsotropyANISO) then

      sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
      pforce(1) = sigma * gauss()
      pforce(2) = sigma * gauss()
      pforce(3) = sigma * gauss()

    else if (ensemble%isotropy == IsotropyXY_Fixed) then

      sigma = sqrt(2.0_wp*gamma_p*pmass*KBOLTZ*temp0/dt)
      pforce(1) = 0.0_wp
      pforce(2) = 0.0_wp
      pforce(3) = sigma * gauss()

    end if

#ifdef MPI
    call mpi_bcast(pforce, 3, mpi_wp_real, 0, mpi_comm_country, ierror)
#endif

    ! random force
    !
    factor   = 2.0_wp*gamma_t*KBOLTZ*temp0/dt
    do j = 1, ncell
      do jx = 1, natom(j)

        sigma = sqrt(mass(jx,j) * factor)
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(1) = rsq * v1
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(2) = rsq * v1
        rsq = 2.0_wp

        do while (rsq >= 1.0_wp)
          v1  = 2.0_wp*grnd() - 1.0_wp
          v2  = 2.0_wp*grnd() - 1.0_wp
          rsq = v1*v1 + v2*v2
        end do

        rsq   = sqrt(-2.0_wp * log(rsq) / rsq)
        grandom(3) = rsq * v1
        random_f(1:3,jx,j) = sigma*grandom(1:3)

      end do
    end do

    do j = 1, ncell
      do jx = 1, natom(j)
        vel_ref(1:3,jx,j) = vel(1:3,jx,j)
      end do
    end do

    ! iteration of barostats
    !
    do i = 1, maxiter

      do j = 1, ncell
        do jx = 1, natom(j)
          vel(1:3,jx,j) = vel_ref(1:3,jx,j)
        end do
      end do

      ! VV2
      !
      do j = 1, ncell
        do jx = 1, natom(j)
          factor = half_dt / mass(jx,j)
          vel(1:3,jx,j) = vel(1:3,jx,j) + factor*force(1:3,jx,j)
          vel(1:3,jx,j) = vel(1:3,jx,j) + factor*force_add(1:3,jx,j,1)
          vel(1:3,jx,j) = vel(1:3,jx,j) + factor*random_f(1:3,jx,j)
        end do
      end do

      ! RATTLE VV2
      !
      if (constraints%rigid_bond) then
        do j = 1, ncell
          do jx = 1, natom(j)
            coord_ref(1:3,jx,j) = vel(1:3,jx,j)
          end do
        end do
        call compute_constraints(ConstraintModeVVER2, .false., dt, coord_ref, &
                                 domain, constraints, coord, vel, &
                                 viri_const)
        do j = 1, ncell
          do jx = 1, natom(j)
            vel_change(1:3) = vel(1:3,jx,j) - coord_ref(1:3,jx,j)
            force_add(1:3,jx,j,1) = force_add(1:3,jx,j,1)                  &
                                  + mass(jx,j)*vel_change(1:3)/half_dt
          end do
        end do
       
        ! constraint virial
        ! 
        if (i == maxiter) then
          do j = 1, ncell
            do jx = 1, natom(j)
              virial_constraint(1,1) = virial_constraint(1,1)             + &
                                       force_add(1,jx,j,1)*coord(1,jx,j)
              virial_constraint(2,2) = virial_constraint(2,2)             + &
                                       force_add(2,jx,j,1)*coord(2,jx,j)
              virial_constraint(3,3) = virial_constraint(3,3)             + &
                                       force_add(3,jx,j,1)*coord(3,jx,j)
            end do
          end do
        end if

      end if

      ! termostat
      !
      vel_scale = exp(-gamma_t*quart_dt)
      do j = 1, ncell
        do jx = 1, natom(j)
          vel(1:3,jx,j) = vel(1:3,jx,j)*vel_scale
        end do
      end do

      ! compute kinetic energy(t+dt)
      !   v(t+dt) = 0.5*(v(t+3/2dt) + v(t+1/2dt))
      !
      kin(1:3) = 0.0_wp

      do j = 1, ncell
        do jx = 1, natom(j)
          kin(1:3)  = kin(1:3) + mass(jx,j) * vel(1:3,jx,j)*vel(1:3,jx,j)
        end do
      end do

      ekin   = 0.5_wp * (kin(1)+kin(2)+kin(3))
      virial_sum(1) = virial(1,1) + virial_constraint(1,1)
      virial_sum(2) = virial(2,2) + virial_constraint(2,2)
      virial_sum(3) = virial(3,3) + virial_constraint(3,3)

      call reduce_pres(kin, ekin, virial_sum)

      ! compute pressure(t+dt) in the unit of kcal/mol*A3
      !
      press(1:3) = (kin(1:3) + virial_sum(1:3))/volume
      pressxyz = (press(1)+press(2)+press(3))/3.0_wp
      pressxy  = (press(1)+press(2))/2.0_wp

      ! update barostat
      !
      call update_barostat(ensemble, boundary, bmoment_ref, pforce, ensemble%isotropy, press(1),  &
                           press(2), press(3), pressxyz, pressxy, press0,     &
                           volume, d_ndegf, pmass, gamma_p, ekin, half_dt,    &
                           quart_dt, natom, ncell, vel, bmoment)

      ! thermostat
      !
      do j = 1, ncell
        do jx = 1, natom(j)
          vel(1:3,jx,j) = vel(1:3,jx,j)*vel_scale
        end do
      end do

    end do

    viri_const(1:3,1:3) = virial_constraint(1:3,1:3)

    return

  end subroutine langevin_barostat_vv2

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_barostat
  !> @brief        update barostat parameter bmoment (eta)
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_barostat(ensemble, boundary, bmoment_ref, pforce, isotropy, &
                             pressx, pressy, pressz,      &
                             pressxyz, pressxy, press0,   &
                             volume, d_ndegf, pmass,      &
                             gamma_p, ekin, half_dt,      &
                             quart_dt, natom, ncell, vel, bmoment)

    ! formal arguments
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: bmoment_ref(:)
    real(wp),                intent(in)    :: pforce(:)
    integer,                 intent(in)    :: isotropy
    real(wp),                intent(in)    :: ekin
    real(wp),                intent(in)    :: pressx
    real(wp),                intent(in)    :: pressy
    real(wp),                intent(in)    :: pressz
    real(wp),                intent(in)    :: pressxyz
    real(wp),                intent(in)    :: pressxy
    real(wp),                intent(in)    :: press0
    real(wp),                intent(in)    :: volume
    real(wp),                intent(in)    :: d_ndegf
    real(wp),                intent(in)    :: pmass
    real(wp),                intent(in)    :: gamma_p
    real(wp),                intent(in)    :: quart_dt
    real(wp),                intent(in)    :: half_dt
    integer,                 intent(in)    :: natom(:)
    integer,                 intent(in)    :: ncell
    real(wp),                intent(inout) :: vel(:,:,:)
    real(wp),                intent(inout) :: bmoment(:)

    ! local variable
    real(wp)         :: gamma0, pressxy0
    real(wp)         :: vel_scale(1:3)
    integer          :: i, ix

    gamma0 = ensemble%gamma*ATMOS_P*100.0_wp/1.01325_wp

    ! eta(t+1/4dt)
    !
    bmoment(1:3) = exp(-gamma_p*quart_dt/2.0_wp)*bmoment_ref(1:3)

    ! eta(t+1/4dt) is scaled according to pressure
    !
    if (isotropy == IsotropyISO) then

      bmoment(1) = bmoment(1) + quart_dt*(3.0_wp*volume*(pressxyz - press0) &
                              + 6.0_wp*ekin/d_ndegf + pforce(1))/pmass
      bmoment(2) = bmoment(1)
      bmoment(3) = bmoment(1)
      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropySEMI_ISO) then

      if (ensemble%ensemble == EnsembleNPT) then
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressxy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressxy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz  - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      else if (ensemble%ensemble == EnsembleNPgT) then
        pressxy0 = press0 - gamma0 / boundary%box_size_z
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressxy - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressxy - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz  - press0)   &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      end if

      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropyANISO) then

      if (ensemble%ensemble == EnsembleNPT) then
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressx - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      else if (ensemble%ensemble == EnsembleNPgT) then
        pressxy0 = press0 - gamma0 / boundary%box_size_z
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressx - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressy - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0)   &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      end if

      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropyXY_Fixed) then

      bmoment(1) = 0.0_wp
      bmoment(2) = 0.0_wp
      bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0) &
                              + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    end if

    ! velocities are scaled according to scaled eta(t+1/4dt)
    !
    vel_scale(1:3) = exp(-half_dt*bmoment(1:3)*(1.0_wp+3.0_wp/d_ndegf))
    do i = 1, ncell
      do ix = 1, natom(i)
        vel(1:3,ix,i) = vel(1:3,ix,i) * vel_scale(1:3)
      end do
    end do

    ! eta(t+1/4dt) is scaled
    !
    bmoment(1:3) = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    ! eta(t+1/2dt)
    !
    if (isotropy == IsotropyISO) then

      bmoment(1) = bmoment(1) + quart_dt*(3.0_wp*volume*(pressxyz - press0) &
                              + 6.0_wp*ekin/d_ndegf + pforce(1))/pmass
      bmoment(2) = bmoment(1)
      bmoment(3) = bmoment(1)
      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropySEMI_ISO) then

      if (ensemble%ensemble == EnsembleNPT) then
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressxy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressxy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz  - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      else if (ensemble%ensemble == EnsembleNPgT) then
        pressxy0 = press0 - gamma0 / boundary%box_size_z
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressxy - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressxy - pressxy0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz  - press0)   &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      end if

      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropyANISO) then

      if (ensemble%ensemble == EnsembleNPT) then
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressx - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressy - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0) &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      else if (ensemble%ensemble == EnsembleNPgT) then
        pressxy0 = press0 - gamma0 / boundary%box_size_z
        bmoment(1) = bmoment(1) + quart_dt*(volume*(pressx - pressxy0)  &
                                + 2.0_wp*ekin/d_ndegf + pforce(1))/pmass
        bmoment(2) = bmoment(2) + quart_dt*(volume*(pressy - pressxy0)  &
                                + 2.0_wp*ekin/d_ndegf + pforce(2))/pmass
        bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0)    &
                                + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      end if

      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    else if (isotropy == IsotropyXY_Fixed) then

      bmoment(1) = 0.0_wp
      bmoment(2) = 0.0_wp
      bmoment(3) = bmoment(3) + quart_dt*(volume*(pressz - press0) &
                              + 2.0_wp*ekin/d_ndegf + pforce(3))/pmass
      bmoment(1:3)  = exp(-gamma_p*quart_dt/2.0_wp)*bmoment(1:3)

    end if

    return

  end subroutine update_barostat

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reduce_pres
  !> @brief
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reduce_pres(val1, val2, val3)

    ! formal arguments
    real(wp),                intent(inout) :: val1(:), val2, val3(:)

#ifdef MPI

    ! local variables
    real(wp)                      :: before_reduce(7), after_reduce(7)


    before_reduce(1:3) = val1(1:3)
    before_reduce(4)   = val2
    before_reduce(5:7) = val3(1:3)

    call mpi_allreduce(before_reduce, after_reduce, 7, mpi_wp_real, &
                       mpi_sum, mpi_comm_country, ierror)

    val1(1:3)    = after_reduce(1:3)
    val2         = after_reduce(4)
    val3(1:3)    = after_reduce(5:7)

#endif

    return

  end subroutine reduce_pres

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    bcast_boxsize
  !> @brief
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine bcast_boxsize(val1, val2, val3)

    ! formal arguments
    real(wp),                intent(inout) :: val1, val2, val3

#ifdef MPI

    ! local variables
    real(wp)                      :: list(3)

    list(1) = val1
    list(2) = val2
    list(3) = val3

    call mpi_bcast(list, 3, mpi_wp_real, 0, mpi_comm_country, ierror)

    val1 = list(1)
    val2 = list(2)
    val3 = list(3)

#endif

    return

  end subroutine bcast_boxsize

end module sp_md_vverlet_mod
