!--------1---------2---------3---------4---------5---------6---------7---------8
! 
!  Module   sp_migration_mod
!> @brief   migration of atoms in each domain                 
!! @authors Jaewoon Jung (JJ) 
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_migration_mod

  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_domain_str_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public :: update_outgoing_solute
  public :: update_outgoing_atom
  public :: update_outgoing_water
  public :: update_outgoing_HGr
  public :: update_incoming_solute
  public :: update_incoming_atom
  public :: update_incoming_water
  public :: update_incoming_HGr

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_solute
  !> @brief        check solute particles going other cells
  !! @authors      JJ
  !! @param[in]    boundary : boundary condition information
  !! @param[inout] domain   : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_solute(boundary, domain)

    ! formal arguments
    type(s_boundary), target, intent(in)    :: boundary
    type(s_domain),   target, intent(inout) :: domain

    ! local variable
    real(wp)                  :: x_shift, y_shift, z_shift
    real(wp)                  :: move(3)
    integer                   :: i, k, ix, icx, icy, icz, icel, ncel
    integer                   :: icel_local, icel_bd

    real(wp),         pointer :: bsize_x, bsize_y, bsize_z
    real(wp),         pointer :: csize_x, csize_y, csize_z
    real(wp),         pointer :: coord(:,:,:), velocity(:,:,:)
    real(wp),         pointer :: charge(:,:), mass(:,:)
    real(wp),         pointer :: buf_real(:,:,:)
    integer,          pointer :: ncel_x, ncel_y, ncel_z, ncel_local, ncel_bd
    integer,          pointer :: nsolute(:), cell_g2l(:), cell_g2b(:)
    integer,          pointer :: atmcls(:,:), id_l2g(:,:)
    integer,          pointer :: ptl_add(:), ptl_exit(:), ptlindex(:,:)
    integer,          pointer :: buf_int(:,:,:)


    bsize_x    => boundary%box_size_x
    bsize_y    => boundary%box_size_y
    bsize_z    => boundary%box_size_z
    ncel_x     => boundary%num_cells_x
    ncel_y     => boundary%num_cells_y
    ncel_z     => boundary%num_cells_z
    csize_x    => boundary%cell_size_x
    csize_y    => boundary%cell_size_y
    csize_z    => boundary%cell_size_z

    ncel_local => domain%num_cell_local
    ncel_bd    => domain%num_cell_boundary
    nsolute    => domain%num_solute
    cell_g2l   => domain%cell_g2l
    cell_g2b   => domain%cell_g2b
    coord      => domain%coord
    velocity   => domain%velocity
    charge     => domain%charge
    mass       => domain%mass
    atmcls     => domain%atom_cls_no
    id_l2g     => domain%id_l2g
    ptl_add    => domain%ptl_add
    ptl_exit   => domain%ptl_exit
    ptlindex   => domain%ptl_exit_index
    buf_int    => domain%buf_integer
    buf_real   => domain%buf_real


    ! initializaiton
    !
    ncel = ncel_local + ncel_bd
    ptl_exit(1:ncel_local) = 0
    ptl_add(1:ncel) = 0

    ! Check outgoing particles
    !
    do i = 1, ncel_local

      k = 0
      do ix = 1, nsolute(i)

        x_shift = coord(1,ix,i) - boundary%origin_x
        y_shift = coord(2,ix,i) - boundary%origin_y
        z_shift = coord(3,ix,i) - boundary%origin_z

        !coordinate shifted to the first quadrant and set into the boundary box
        move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
        move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
        move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)

        x_shift = x_shift + move(1)
        y_shift = y_shift + move(2)
        z_shift = z_shift + move(3)

        !assign which cell
        icx = int(x_shift/csize_x)
        icy = int(y_shift/csize_y)
        icz = int(z_shift/csize_z)
        if (icx == ncel_x) icx = icx - 1
        if (icy == ncel_y) icy = icy - 1
        if (icz == ncel_z) icz = icz - 1
        icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y
        icel_local = cell_g2l(icel)
        icel_bd    = cell_g2b(icel)

        if (icel_local /= i) then

          ptl_exit(i) = ptl_exit(i) + 1
          ptlindex(ptl_exit(i),i) = ix

          if (icel_local /= 0) then

            ptl_add (icel_local) = ptl_add(icel_local) + 1
            buf_real(1,ptl_add(icel_local),icel_local) = coord(1,ix,i)
            buf_real(2,ptl_add(icel_local),icel_local) = coord(2,ix,i)
            buf_real(3,ptl_add(icel_local),icel_local) = coord(3,ix,i)
            buf_real(4,ptl_add(icel_local),icel_local) = velocity(1,ix,i)
            buf_real(5,ptl_add(icel_local),icel_local) = velocity(2,ix,i)
            buf_real(6,ptl_add(icel_local),icel_local) = velocity(3,ix,i)
            buf_real(7,ptl_add(icel_local),icel_local) = charge(ix,i)
            buf_real(8,ptl_add(icel_local),icel_local) = mass(ix,i)
            buf_int (1,ptl_add(icel_local),icel_local) = atmcls(ix,i)
            buf_int (2,ptl_add(icel_local),icel_local) = id_l2g(ix,i)

          else if (icel_bd /= 0) then

            icel_bd = icel_bd + ncel_local
            ptl_add (icel_bd) = ptl_add(icel_bd) + 1
            buf_real(1,ptl_add(icel_bd),icel_bd) = coord(1,ix,i)
            buf_real(2,ptl_add(icel_bd),icel_bd) = coord(2,ix,i)
            buf_real(3,ptl_add(icel_bd),icel_bd) = coord(3,ix,i)
            buf_real(4,ptl_add(icel_bd),icel_bd) = velocity(1,ix,i)
            buf_real(5,ptl_add(icel_bd),icel_bd) = velocity(2,ix,i)
            buf_real(6,ptl_add(icel_bd),icel_bd) = velocity(3,ix,i)
            buf_real(7,ptl_add(icel_bd),icel_bd) = charge(ix,i)
            buf_real(8,ptl_add(icel_bd),icel_bd) = mass(ix,i)
            buf_int (1,ptl_add(icel_bd),icel_bd) = atmcls(ix,i)
            buf_int (2,ptl_add(icel_bd),icel_bd) = id_l2g(ix,i)

          end if

        end if

       end do
    end do

    return

  end subroutine update_outgoing_solute

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_atom
  !> @brief        check particles (not bonded to hydrogen) going other cells
  !! @authors      JJ
  !! @param[in]    boundary    : boundary condition information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_atom(boundary, constraints, domain)

    ! formal arguments
    type(s_boundary),    target, intent(in)    :: boundary
    type(s_constraints), target, intent(in)    :: constraints
    type(s_domain),      target, intent(inout) :: domain

    ! local variable
    real(wp)                     :: x_shift, y_shift, z_shift
    real(wp)                     :: move(3)
    integer                      :: i, k, ix, icx, icy, icz, icel, ncel
    integer                      :: icel_local, icel_bd

    real(wp),            pointer :: bsize_x, bsize_y, bsize_z
    real(wp),            pointer :: csize_x, csize_y, csize_z
    real(wp),            pointer :: coord(:,:,:), velocity(:,:,:)
    real(wp),            pointer :: charge(:,:), mass(:,:)
    real(wp),            pointer :: buf_real(:,:,:)
    integer,             pointer :: ncel_x, ncel_y, ncel_z, ncel_local, ncel_bd
    integer,             pointer :: nsolute(:), cell_g2l(:), cell_g2b(:)
    integer,             pointer :: atmcls(:,:), id_l2g(:,:)
    integer,             pointer :: ptl_add(:), ptl_exit(:), ptlindex(:,:)
    integer,             pointer :: buf_int(:,:,:)


    bsize_x    => boundary%box_size_x
    bsize_y    => boundary%box_size_y
    bsize_z    => boundary%box_size_z
    ncel_x     => boundary%num_cells_x
    ncel_y     => boundary%num_cells_y
    ncel_z     => boundary%num_cells_z
    csize_x    => boundary%cell_size_x
    csize_y    => boundary%cell_size_y
    csize_z    => boundary%cell_size_z

    nsolute    => constraints%No_HGr

    ncel_local => domain%num_cell_local
    ncel_bd    => domain%num_cell_boundary
    cell_g2l   => domain%cell_g2l
    cell_g2b   => domain%cell_g2b
    coord      => domain%coord
    velocity   => domain%velocity
    charge     => domain%charge
    mass       => domain%mass
    atmcls     => domain%atom_cls_no
    id_l2g     => domain%id_l2g
    ptl_add    => domain%ptl_add
    ptl_exit   => domain%ptl_exit
    ptlindex   => domain%ptl_exit_index
    buf_int    => domain%buf_integer
    buf_real   => domain%buf_real


    ! initializaiton
    !
    ncel = ncel_local + ncel_bd
    ptl_exit(1:ncel_local) = 0
    ptl_add(1:ncel) = 0

    ! Check outgoing particles
    !
    do i = 1, ncel_local

      k = 0
      do ix = 1, nsolute(i)

        x_shift = coord(1,ix,i) - boundary%origin_x
        y_shift = coord(2,ix,i) - boundary%origin_y
        z_shift = coord(3,ix,i) - boundary%origin_z

        !coordinate shifted to the first quadrant and set into the boundary box
        move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
        move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
        move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)

        x_shift = x_shift + move(1)
        y_shift = y_shift + move(2)
        z_shift = z_shift + move(3)

        !assign which cell
        icx = int(x_shift/csize_x)
        icy = int(y_shift/csize_y)
        icz = int(z_shift/csize_z)
        if (icx == ncel_x) icx = icx - 1
        if (icy == ncel_y) icy = icy - 1
        if (icz == ncel_z) icz = icz - 1
        icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y
        icel_local = cell_g2l(icel)
        icel_bd    = cell_g2b(icel)

        if (icel_local /= i) then

          ptl_exit(i) = ptl_exit(i) + 1
          ptlindex(ptl_exit(i),i) = ix

          if (icel_local /= 0) then

            ptl_add (icel_local) = ptl_add(icel_local) + 1
            buf_real(1,ptl_add(icel_local),icel_local) = coord(1,ix,i)
            buf_real(2,ptl_add(icel_local),icel_local) = coord(2,ix,i)
            buf_real(3,ptl_add(icel_local),icel_local) = coord(3,ix,i)
            buf_real(4,ptl_add(icel_local),icel_local) = velocity(1,ix,i)
            buf_real(5,ptl_add(icel_local),icel_local) = velocity(2,ix,i)
            buf_real(6,ptl_add(icel_local),icel_local) = velocity(3,ix,i)
            buf_real(7,ptl_add(icel_local),icel_local) = charge(ix,i)
            buf_real(8,ptl_add(icel_local),icel_local) = mass(ix,i)
            buf_int (1,ptl_add(icel_local),icel_local) = atmcls(ix,i)
            buf_int (2,ptl_add(icel_local),icel_local) = id_l2g(ix,i)

          else if (icel_bd /= 0) then

            icel_bd = icel_bd + ncel_local
            ptl_add (icel_bd) = ptl_add(icel_bd) + 1
            buf_real(1,ptl_add(icel_bd),icel_bd) = coord(1,ix,i)
            buf_real(2,ptl_add(icel_bd),icel_bd) = coord(2,ix,i)
            buf_real(3,ptl_add(icel_bd),icel_bd) = coord(3,ix,i)
            buf_real(4,ptl_add(icel_bd),icel_bd) = velocity(1,ix,i)
            buf_real(5,ptl_add(icel_bd),icel_bd) = velocity(2,ix,i)
            buf_real(6,ptl_add(icel_bd),icel_bd) = velocity(3,ix,i)
            buf_real(7,ptl_add(icel_bd),icel_bd) = charge(ix,i)
            buf_real(8,ptl_add(icel_bd),icel_bd) = mass(ix,i)
            buf_int (1,ptl_add(icel_bd),icel_bd) = atmcls(ix,i)
            buf_int (2,ptl_add(icel_bd),icel_bd) = id_l2g(ix,i)

          end if

        end if

       end do
    end do

    return

  end subroutine update_outgoing_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_water
  !> @brief        check water particles going/remaining cells
  !! @authors      JJ
  !! @param[in]    boundary : boundary condition information
  !! @param[inout] domain   : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_water(boundary, domain)

    ! formal arguments
    type(s_boundary), target, intent(in)    :: boundary
    type(s_domain),   target, intent(inout) :: domain

    ! local variable
    real(wp)                  :: x_shift, y_shift, z_shift
    real(wp)                  :: move(3)
    integer                   :: i, k, iwater, list, ix
    integer                   :: icx, icy, icz, icel, ncel
    integer                   :: icel_local, icel_bd

    real(wp),         pointer :: bsize_x, bsize_y, bsize_z
    real(wp),         pointer :: csize_x, csize_y, csize_z
    real(wp),         pointer :: coord(:,:,:), velocity(:,:,:)
    real(wp),         pointer :: water_move_real(:,:,:), water_stay_real(:,:,:)
    integer,          pointer :: ncel_x, ncel_y, ncel_z, ncel_local, ncel_bd
    integer,          pointer :: nwater(:), water_list(:,:,:)
    integer,          pointer :: cell_g2l(:), cell_g2b(:), id_l2g(:,:)
    integer,          pointer :: water_move(:), water_stay(:)
    integer,          pointer :: water_move_int(:,:,:), water_stay_int(:,:,:)


    bsize_x         => boundary%box_size_x
    bsize_y         => boundary%box_size_y
    bsize_z         => boundary%box_size_z
    ncel_x          => boundary%num_cells_x
    ncel_y          => boundary%num_cells_y
    ncel_z          => boundary%num_cells_z
    csize_x         => boundary%cell_size_x
    csize_y         => boundary%cell_size_y
    csize_z         => boundary%cell_size_z

    ncel_local      => domain%num_cell_local
    ncel_bd         => domain%num_cell_boundary
    nwater          => domain%num_water 
    water_list      => domain%water_list
    cell_g2l        => domain%cell_g2l
    cell_g2b        => domain%cell_g2b
    coord           => domain%coord
    velocity        => domain%velocity
    id_l2g          => domain%id_l2g
    water_move      => domain%water%move
    water_stay      => domain%water%stay
    water_move_real => domain%water%move_real
    water_stay_real => domain%water%stay_real
    water_move_int  => domain%water%move_integer
    water_stay_int  => domain%water%stay_integer


    ! initializaiton
    !
    ncel = ncel_local + ncel_bd
    water_stay(1:ncel_local) = 0
    water_move(1:ncel) = 0

    ! Check outgoing particles
    !
    do i = 1, ncel_local

      k = 0
      do iwater = 1, nwater(i)
        
        ix = water_list(1,iwater,i)
        x_shift = coord(1,ix,i) - boundary%origin_x
        y_shift = coord(2,ix,i) - boundary%origin_y
        z_shift = coord(3,ix,i) - boundary%origin_z

        !coordinate shifted to the first quadrant and set into the boundary box
        move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
        move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
        move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)

        x_shift = x_shift + move(1)
        y_shift = y_shift + move(2)
        z_shift = z_shift + move(3)

        !assign which cell
        icx = int(x_shift/csize_x)
        icy = int(y_shift/csize_y)
        icz = int(z_shift/csize_z)
        if (icx == ncel_x) icx = icx - 1
        if (icy == ncel_y) icy = icy - 1
        if (icz == ncel_z) icz = icz - 1
        icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y
        icel_local = cell_g2l(icel)
        icel_bd    = cell_g2b(icel)

        ! atom index
        if (icel_local /= i) then

          if (icel_local /= 0) then

            water_move(icel_local) = water_move(icel_local) + 1

            do list = 1, 3
              ix = water_list(list,iwater,i)
              water_move_real(6*(list-1)+1,water_move(icel_local),icel_local) &
                 = coord(1,ix,i)
              water_move_real(6*(list-1)+2,water_move(icel_local),icel_local) &
                 = coord(2,ix,i)
              water_move_real(6*(list-1)+3,water_move(icel_local),icel_local) &
                 = coord(3,ix,i)
              water_move_real(6*(list-1)+4,water_move(icel_local),icel_local) &
                 = velocity(1,ix,i)
              water_move_real(6*(list-1)+5,water_move(icel_local),icel_local) &
                 = velocity(2,ix,i)
              water_move_real(6*(list-1)+6,water_move(icel_local),icel_local) &
                 = velocity(3,ix,i)
              water_move_int(list,water_move(icel_local),icel_local)          &
                 = id_l2g(ix,i)
            end do

          else if (icel_bd /= 0) then

            icel_bd = icel_bd + ncel_local
            water_move(icel_bd) = water_move(icel_bd) + 1

            do list = 1, 3
              ix = water_list(list,iwater,i)
              water_move_real(6*(list-1)+1,water_move(icel_bd),icel_bd)       &
                 = coord(1,ix,i)
              water_move_real(6*(list-1)+2,water_move(icel_bd),icel_bd)       &
                 = coord(2,ix,i)
              water_move_real(6*(list-1)+3,water_move(icel_bd),icel_bd)       &
                 = coord(3,ix,i)
              water_move_real(6*(list-1)+4,water_move(icel_bd),icel_bd)       &
                 = velocity(1,ix,i)
              water_move_real(6*(list-1)+5,water_move(icel_bd),icel_bd)       &
                 = velocity(2,ix,i)
              water_move_real(6*(list-1)+6,water_move(icel_bd),icel_bd)       &
                 = velocity(3,ix,i)
              water_move_int(list,water_move(icel_bd),icel_bd) = id_l2g(ix,i)
            end do
          end if

        else 

          water_stay(i) = water_stay(i) + 1

          do list = 1, 3
            ix = water_list(list,iwater,i)
            water_stay_real(6*(list-1)+1,water_stay(i),i) = coord(1,ix,i)
            water_stay_real(6*(list-1)+2,water_stay(i),i) = coord(2,ix,i)
            water_stay_real(6*(list-1)+3,water_stay(i),i) = coord(3,ix,i)
            water_stay_real(6*(list-1)+4,water_stay(i),i) = velocity(1,ix,i)
            water_stay_real(6*(list-1)+5,water_stay(i),i) = velocity(2,ix,i)
            water_stay_real(6*(list-1)+6,water_stay(i),i) = velocity(3,ix,i)
            water_stay_int(list,water_stay(i),i) = id_l2g(ix,i)
          end do
        
        end if

      end do

    end do

    return

  end subroutine update_outgoing_water

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_HGr
  !> @brief        check hydrogen-bonded particles going/remaining cells
  !! @authors      JJ
  !! @param[in]    boundary    : boundary condition information
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_HGr(boundary, constraints, domain)

    ! formal arguments
    type(s_boundary),    target, intent(in)    :: boundary
    type(s_constraints), target, intent(inout) :: constraints
    type(s_domain),      target, intent(inout) :: domain

    ! local variable
    real(wp)                     :: x_shift, y_shift, z_shift
    real(wp)                     :: move(3)
    integer                      :: i, j, k, list, ix
    integer                      :: num_move, num_stay
    integer                      :: icx, icy, icz, icel, ncel
    integer                      :: icel_local, icel_bd

    real(wp),            pointer :: bsize_x, bsize_y, bsize_z
    real(wp),            pointer :: csize_x, csize_y, csize_z
    real(wp),            pointer :: coord(:,:,:), vel(:,:,:)
    real(wp),            pointer :: charge(:,:), mass(:,:)
    real(wp),            pointer :: HGr_bond_dist(:,:,:,:)
    real(wp),            pointer :: HGr_move_real(:,:,:,:)
    real(wp),            pointer :: HGr_stay_real(:,:,:,:)
    integer,             pointer :: ncel_x, ncel_y, ncel_z
    integer,             pointer :: ncel_local, ncel_bd
    integer,             pointer :: connect
    integer,             pointer :: atmcls(:,:)
    integer,             pointer :: HGr_local(:,:), HGr_bond_list(:,:,:,:)
    integer,             pointer :: cell_g2l(:), cell_g2b(:), id_l2g(:,:)
    integer,             pointer :: HGr_move(:,:), HGr_stay(:,:)
    integer,             pointer :: HGr_move_int(:,:,:,:), HGr_stay_int(:,:,:,:)


    bsize_x       => boundary%box_size_x
    bsize_y       => boundary%box_size_y
    bsize_z       => boundary%box_size_z
    ncel_x        => boundary%num_cells_x
    ncel_y        => boundary%num_cells_y
    ncel_z        => boundary%num_cells_z
    csize_x       => boundary%cell_size_x
    csize_y       => boundary%cell_size_y
    csize_z       => boundary%cell_size_z

    HGr_local     => constraints%HGr_local
    HGr_bond_list => constraints%HGr_bond_list
    HGr_move      => constraints%HGr_move
    HGr_stay      => constraints%HGr_stay
    HGr_bond_dist => constraints%HGr_bond_dist
    HGr_move_real => constraints%HGr_move_real
    HGr_stay_real => constraints%HGr_stay_real
    HGr_move_int  => constraints%HGr_move_int
    HGr_stay_int  => constraints%HGr_stay_int
    connect       => constraints%connect

    ncel_local    => domain%num_cell_local
    ncel_bd       => domain%num_cell_boundary
    cell_g2l      => domain%cell_g2l
    cell_g2b      => domain%cell_g2b
    coord         => domain%coord
    vel           => domain%velocity
    charge        => domain%charge
    mass          => domain%mass
    id_l2g        => domain%id_l2g
    atmcls        => domain%atom_cls_no


    ! initializaiton
    !
    ncel = ncel_local + ncel_bd
    HGr_stay(1:connect,1:ncel_local) = 0
    HGr_move(1:connect,1:ncel) = 0

    ! Check outgoing particles
    !
    do i = 1, ncel_local
      do j = 1, connect
        do k = 1, HGr_local(j,i)

          ix = HGr_bond_list(1,k,j,i)
          x_shift = coord(1,ix,i) - boundary%origin_x
          y_shift = coord(2,ix,i) - boundary%origin_y
          z_shift = coord(3,ix,i) - boundary%origin_z

          move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
          move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
          move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)

          x_shift = x_shift + move(1)
          y_shift = y_shift + move(2)
          z_shift = z_shift + move(3)

          !assign which cell
          icx = int(x_shift/csize_x)
          icy = int(y_shift/csize_y)
          icz = int(z_shift/csize_z)
          if (icx == ncel_x) icx = icx - 1
          if (icy == ncel_y) icy = icy - 1
          if (icz == ncel_z) icz = icz - 1
          icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y
          icel_local = cell_g2l(icel)
          icel_bd    = cell_g2b(icel)

          ! atom index
          if (icel_local /= i) then
            if (icel_local /= 0) then

              HGr_move(j,icel_local) = HGr_move(j,icel_local) + 1
              num_move = HGr_move(j,icel_local)

              do list = 1, j+1
                ix = HGr_bond_list(list,k,j,i)
                HGr_move_real(9*(list-1)+1,num_move,j,icel_local) = &
                     coord(1,ix,i)
                HGr_move_real(9*(list-1)+2,num_move,j,icel_local) = &
                     coord(2,ix,i)
                HGr_move_real(9*(list-1)+3,num_move,j,icel_local) = &
                     coord(3,ix,i)
                HGr_move_real(9*(list-1)+4,num_move,j,icel_local) = vel(1,ix,i)
                HGr_move_real(9*(list-1)+5,num_move,j,icel_local) = vel(2,ix,i)
                HGr_move_real(9*(list-1)+6,num_move,j,icel_local) = vel(3,ix,i)
                HGr_move_real(9*(list-1)+7,num_move,j,icel_local) = charge(ix,i)
                HGr_move_real(9*(list-1)+8,num_move,j,icel_local) = mass(ix,i)
                HGr_move_real(9*(list-1)+9,num_move,j,icel_local) = &
                     HGr_bond_dist(list,k,j,i)
                HGr_move_int(2*(list-1)+1,num_move,j,icel_local)  = atmcls(ix,i)
                HGr_move_int(2*(list-1)+2,num_move,j,icel_local)  = id_l2g(ix,i)
              end do

            else if (icel_bd /= 0) then

              icel_bd = icel_bd + ncel_local
              HGr_move(j,icel_bd) = HGr_move(j,icel_bd) + 1
              num_move = HGr_move(j,icel_bd)

              do list = 1, j+1
                ix = HGr_bond_list(list,k,j,i)
                hgr_move_real(9*(list-1)+1,num_move,j,icel_bd) = coord(1,ix,i)
                hgr_move_real(9*(list-1)+2,num_move,j,icel_bd) = coord(2,ix,i)
                hgr_move_real(9*(list-1)+3,num_move,j,icel_bd) = coord(3,ix,i)
                hgr_move_real(9*(list-1)+4,num_move,j,icel_bd) = vel(1,ix,i)
                hgr_move_real(9*(list-1)+5,num_move,j,icel_bd) = vel(2,ix,i)
                hgr_move_real(9*(list-1)+6,num_move,j,icel_bd) = vel(3,ix,i)
                hgr_move_real(9*(list-1)+7,num_move,j,icel_bd) = charge(ix,i)
                hgr_move_real(9*(list-1)+8,num_move,j,icel_bd) = mass(ix,i)
                HGr_move_real(9*(list-1)+9,num_move,j,icel_bd) = &
                     HGr_bond_dist(list,k,j,i)
                hgr_move_int(2*(list-1)+1,num_move,j,icel_bd)  = atmcls(ix,i)
                hgr_move_int(2*(list-1)+2,num_move,j,icel_bd)  = id_l2g(ix,i)
              end do
            end if

          else

            hgr_stay(j,i) = hgr_stay(j,i) + 1
            num_stay = hgr_stay(j,i)

            do list = 1, j+1
              ix = HGr_bond_list(list,k,j,i)
              hgr_stay_real(9*(list-1)+1,num_stay,j,i) = coord(1,ix,i)
              hgr_stay_real(9*(list-1)+2,num_stay,j,i) = coord(2,ix,i)
              hgr_stay_real(9*(list-1)+3,num_stay,j,i) = coord(3,ix,i)
              hgr_stay_real(9*(list-1)+4,num_stay,j,i) = vel(1,ix,i)
              hgr_stay_real(9*(list-1)+5,num_stay,j,i) = vel(2,ix,i)
              hgr_stay_real(9*(list-1)+6,num_stay,j,i) = vel(3,ix,i)
              hgr_stay_real(9*(list-1)+7,num_stay,j,i) = charge(ix,i)
              hgr_stay_real(9*(list-1)+8,num_stay,j,i) = mass(ix,i)
              hgr_stay_real(9*(list-1)+9,num_stay,j,i) = &
                   HGr_bond_dist(list,k,j,i)
              hgr_stay_int (2*(list-1)+1,num_stay,j,i) = atmcls(ix,i)
              hgr_stay_int (2*(list-1)+2,num_stay,j,i) = id_l2g(ix,i)
            end do
        
          end if

        end do
      end do
    end do

    return

  end subroutine update_outgoing_HGr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_solute
  !> @brief        check solute particles incoming to each cell
  !! @authors      JJ
  !! @param[inout] domain : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_solute(domain)

    ! formal arguments
    type(s_domain),  target, intent(inout) :: domain

    ! local variables
    integer                  :: i, j, k, ix, kx
    logical                  :: insert

    real(wp),        pointer :: coord(:,:,:), velocity(:,:,:)
    real(wp),        pointer :: charge(:,:), mass(:,:)
    real(wp),        pointer :: buf_real(:,:,:)
    integer,         pointer :: ncel_local, nsolute(:)
    integer,         pointer :: atmcls(:,:), id_l2g(:,:), id_g2l(:,:)
    integer,         pointer :: ptl_add(:), ptl_exit(:), ptlindex(:,:)
    integer,         pointer :: buf_int(:,:,:)


    ncel_local => domain%num_cell_local
    nsolute    => domain%num_solute
    coord      => domain%coord
    velocity   => domain%velocity
    charge     => domain%charge
    mass       => domain%mass
    atmcls     => domain%atom_cls_no
    id_l2g     => domain%id_l2g
    id_g2l     => domain%id_g2l
    ptl_add    => domain%ptl_add
    ptl_exit   => domain%ptl_exit
    ptlindex   => domain%ptl_exit_index
    buf_int    => domain%buf_integer
    buf_real   => domain%buf_real


    ! incoming particles
    !
    do i = 1, ncel_local

#ifdef DEBUG
      if (ptl_add(i)+nsolute(i)-ptl_exit(i) > MaxAtom) &
        call error_msg('Debug: Update_Incoming_Solute> atom is exceed MaxAtom')
#endif 

      ! when the number of coming particles is larger than that of outgoing ones
      !
      if (ptl_add(i) >= ptl_exit(i)) then

        do k = 1, ptl_exit(i)
          coord(1,ptlindex(k,i),i)    = buf_real(1,k,i)
          coord(2,ptlindex(k,i),i)    = buf_real(2,k,i)
          coord(3,ptlindex(k,i),i)    = buf_real(3,k,i)
          velocity(1,ptlindex(k,i),i) = buf_real(4,k,i)
          velocity(2,ptlindex(k,i),i) = buf_real(5,k,i)
          velocity(3,ptlindex(k,i),i) = buf_real(6,k,i)
          charge(ptlindex(k,i),i)     = buf_real(7,k,i)
          mass(ptlindex(k,i),i)       = buf_real(8,k,i)
          atmcls(ptlindex(k,i),i)     = buf_int (1,k,i)
          id_l2g(ptlindex(k,i),i)     = buf_int (2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ptlindex(k,i)

        end do

        do k = ptl_exit(i)+1, ptl_add(i)
          ix = k + nsolute(i) - ptl_exit(i)
          coord(1,ix,i)               = buf_real(1,k,i)
          coord(2,ix,i)               = buf_real(2,k,i)
          coord(3,ix,i)               = buf_real(3,k,i)
          velocity(1,ix,i)            = buf_real(4,k,i)
          velocity(2,ix,i)            = buf_real(5,k,i)
          velocity(3,ix,i)            = buf_real(6,k,i)
          charge(ix,i)                = buf_real(7,k,i)
          mass(ix,i)                  = buf_real(8,k,i)
          atmcls(ix,i)                = buf_int (1,k,i)
          id_l2g(ix,i)                = buf_int (2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ix
        end do

        nsolute(i) = nsolute(i) + ptl_add(i) - ptl_exit(i)

      ! when the number of coming particles is less than that of outgoing ones
      !
      else

        do k = 1, ptl_add(i)
          coord(1,ptlindex(k,i),i)    = buf_real(1,k,i)
          coord(2,ptlindex(k,i),i)    = buf_real(2,k,i)
          coord(3,ptlindex(k,i),i)    = buf_real(3,k,i)
          velocity(1,ptlindex(k,i),i) = buf_real(4,k,i)
          velocity(2,ptlindex(k,i),i) = buf_real(5,k,i)
          velocity(3,ptlindex(k,i),i) = buf_real(6,k,i)
          charge(ptlindex(k,i),i)     = buf_real(7,k,i)
          mass(ptlindex(k,i),i)       = buf_real(8,k,i)
          atmcls(ptlindex(k,i),i)     = buf_int (1,k,i)
          id_l2g(ptlindex(k,i),i)     = buf_int (2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ptlindex(k,i)
        end do

        j = 0
        ix = nsolute(i)
        k = ptl_add(i) + 1

        do while (j < (ptl_exit(i)-ptl_add(i)))

          insert = .true.
          do kx = k, ptl_exit(i)
            if (ix == ptlindex(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = ptlindex(k,i)
            coord(1,kx,i)          = coord(1,ix,i)
            coord(2,kx,i)          = coord(2,ix,i)
            coord(3,kx,i)          = coord(3,ix,i)
            velocity(1,kx,i)       = velocity(1,ix,i)
            velocity(2,kx,i)       = velocity(2,ix,i)
            velocity(3,kx,i)       = velocity(3,ix,i)
            charge(kx,i)           = charge(ix,i)
            mass(kx,i)             = mass(ix,i)
            atmcls(kx,i)           = atmcls(ix,i)
            id_l2g(kx,i)           = id_l2g(ix,i)
            id_g2l(1,id_l2g(kx,i)) = i
            id_g2l(2,id_l2g(kx,i)) = kx

            j = j + 1
            k = k + 1
          end if
          ix = ix - 1

        end do
        nsolute(i) = nsolute(i) + ptl_add(i) - ptl_exit(i)

      end if
    end do

    return

  end subroutine update_incoming_solute

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_atom
  !> @brief        check particles (not bonded to hydrogen) incoming to each
  !!               cell
  !! @authors      JJ
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_atom(constraints, domain)

    ! formal arguments
    type(s_constraints), target, intent(inout) :: constraints
    type(s_domain),      target, intent(inout) :: domain

    ! local variables
    integer                      :: i, j, k, ix, kx
    logical                      :: insert

    real(wp),         pointer    :: coord(:,:,:), velocity(:,:,:)
    real(wp),         pointer    :: charge(:,:), mass(:,:)
    real(wp),         pointer    :: buf_real(:,:,:)
    integer,          pointer    :: ncel_local, nsolute(:)
    integer,          pointer    :: atmcls(:,:), id_l2g(:,:), id_g2l(:,:)
    integer,          pointer    :: ptl_add(:), ptl_exit(:), ptlindex(:,:)
    integer,          pointer    :: buf_int(:,:,:)


    nsolute    => constraints%No_HGr

    ncel_local => domain%num_cell_local
    coord      => domain%coord
    velocity   => domain%velocity
    charge     => domain%charge
    mass       => domain%mass
    atmcls     => domain%atom_cls_no
    id_l2g     => domain%id_l2g
    id_g2l     => domain%id_g2l
    ptl_add    => domain%ptl_add
    ptl_exit   => domain%ptl_exit
    ptlindex   => domain%ptl_exit_index
    buf_int    => domain%buf_integer
    buf_real   => domain%buf_real


    ! incoming particles
    !
    do i = 1, ncel_local

#ifdef DEBUG
      if (ptl_add(i)+nsolute(i)-ptl_exit(i) > MaxAtom) &
        call error_msg('Debug: Update_Incoming_Atom> atom is exceed MaxAtom')
#endif 

      ! when the number of coming particles is larger than that of outgoing ones
      !
      if (ptl_add(i) >= ptl_exit(i)) then

        do k = 1, ptl_exit(i)
          coord(1,ptlindex(k,i),i)    = buf_real(1,k,i)
          coord(2,ptlindex(k,i),i)    = buf_real(2,k,i)
          coord(3,ptlindex(k,i),i)    = buf_real(3,k,i)
          velocity(1,ptlindex(k,i),i) = buf_real(4,k,i)
          velocity(2,ptlindex(k,i),i) = buf_real(5,k,i)
          velocity(3,ptlindex(k,i),i) = buf_real(6,k,i)
          charge(ptlindex(k,i),i)     = buf_real(7,k,i)
          mass(ptlindex(k,i),i)       = buf_real(8,k,i)
          atmcls(ptlindex(k,i),i)     = buf_int(1,k,i)
          id_l2g(ptlindex(k,i),i)     = buf_int(2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ptlindex(k,i)
        end do

        do k = ptl_exit(i)+1, ptl_add(i)
          ix = k + nsolute(i) - ptl_exit(i)
          coord(1,ix,i)               = buf_real(1,k,i)
          coord(2,ix,i)               = buf_real(2,k,i)
          coord(3,ix,i)               = buf_real(3,k,i)
          velocity(1,ix,i)            = buf_real(4,k,i)
          velocity(2,ix,i)            = buf_real(5,k,i)
          velocity(3,ix,i)            = buf_real(6,k,i)
          charge(ix,i)                = buf_real(7,k,i)
          mass(ix,i)                  = buf_real(8,k,i)
          atmcls(ix,i)                = buf_int (1,k,i)
          id_l2g(ix,i)                = buf_int (2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ix
        end do
        nsolute(i) = nsolute(i) + ptl_add(i) - ptl_exit(i)

      ! when the number of coming particles is less than that of outgoing ones
      !
      else

        do k = 1, ptl_add(i)
          coord(1,ptlindex(k,i),i)    = buf_real(1,k,i)
          coord(2,ptlindex(k,i),i)    = buf_real(2,k,i)
          coord(3,ptlindex(k,i),i)    = buf_real(3,k,i)
          velocity(1,ptlindex(k,i),i) = buf_real(4,k,i)
          velocity(2,ptlindex(k,i),i) = buf_real(5,k,i)
          velocity(3,ptlindex(k,i),i) = buf_real(6,k,i)
          charge(ptlindex(k,i),i)     = buf_real(7,k,i)
          mass(ptlindex(k,i),i)       = buf_real(8,k,i)
          atmcls(ptlindex(k,i),i)     = buf_int (1,k,i)
          id_l2g(ptlindex(k,i),i)     = buf_int (2,k,i)
          id_g2l(1,buf_int(2,k,i))    = i
          id_g2l(2,buf_int(2,k,i))    = ptlindex(k,i)
        end do

        j  = 0
        ix = nsolute(i)
        k  = ptl_add(i) + 1

        do while (j < (ptl_exit(i)-ptl_add(i)))

          insert = .true.

          do kx = k, ptl_exit(i)
            if (ix == ptlindex(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = ptlindex(k,i)
            coord(1,kx,i)          = coord(1,ix,i)
            coord(2,kx,i)          = coord(2,ix,i)
            coord(3,kx,i)          = coord(3,ix,i)
            velocity(1,kx,i)       = velocity(1,ix,i)
            velocity(2,kx,i)       = velocity(2,ix,i)
            velocity(3,kx,i)       = velocity(3,ix,i)
            charge(kx,i)           = charge(ix,i)
            mass(kx,i)             = mass(ix,i)
            atmcls(kx,i)           = atmcls(ix,i)
            id_l2g(kx,i)           = id_l2g(ix,i)
            id_g2l(1,id_l2g(kx,i)) = i
            id_g2l(2,id_l2g(kx,i)) = kx

            j = j + 1
            k = k + 1

          end if
          ix = ix - 1

        end do
        nsolute(i) = nsolute(i) + ptl_add(i) - ptl_exit(i)

      end if
    end do

    return

  end subroutine update_incoming_atom

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_water
  !> @brief        check water particles incoming cells
  !! @authors      JJ
  !! @param[inout] domain   : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_water(domain)

    ! formal arguments
    type(s_domain),  target, intent(inout) :: domain

    ! local variable
    integer                  :: i, k, iwater, jwater, ix

    real(wp),        pointer :: coord(:,:,:), velocity(:,:,:)
    real(wp),        pointer :: water_move_real(:,:,:), water_stay_real(:,:,:)
    real(wp),        pointer :: water_charge(:), water_mass(:)
    real(wp),        pointer :: charge(:,:), mass(:,:)
    integer,         pointer :: ncel_local
    integer,         pointer :: nsolute(:), natom(:), nwater(:)
    integer,         pointer :: water_list(:,:,:)
    integer,         pointer :: id_l2g(:,:), id_g2l(:,:)
    integer,         pointer :: water_move(:), water_stay(:)
    integer,         pointer :: water_move_int(:,:,:), water_stay_int(:,:,:)
    integer,         pointer :: water_atmcls(:), atmcls(:,:)


    ncel_local      => domain%num_cell_local
    nsolute         => domain%num_solute 
    nwater          => domain%num_water
    natom           => domain%num_atom
    water_list      => domain%water_list
    coord           => domain%coord
    velocity        => domain%velocity
    charge          => domain%charge
    mass            => domain%mass
    atmcls          => domain%atom_cls_no
    id_l2g          => domain%id_l2g
    id_g2l          => domain%id_g2l
    water_atmcls    => domain%water%atom_cls_no
    water_charge    => domain%water%charge
    water_mass      => domain%water%mass  
    water_move      => domain%water%move
    water_stay      => domain%water%stay
    water_move_real => domain%water%move_real
    water_stay_real => domain%water%stay_real
    water_move_int  => domain%water%move_integer
    water_stay_int  => domain%water%stay_integer

 
    do i = 1, ncel_local

#ifdef DEBUG
      if (water_stay(i)+water_move(i) > MaxWater) then
        call error_msg('Debug: Update_Incoming_Water> water is exceed MaxWater')
      end if
#endif 

      do iwater = 1, water_stay(i)
        do k = 1, 3
          water_list(k,iwater,i) = nsolute(i) + 3*(iwater-1) + k
          ix = water_list(k,iwater,i)

          coord(1,ix,i)          = water_stay_real(6*(k-1)+1,iwater,i)
          coord(2,ix,i)          = water_stay_real(6*(k-1)+2,iwater,i)
          coord(3,ix,i)          = water_stay_real(6*(k-1)+3,iwater,i)
          velocity(1,ix,i)       = water_stay_real(6*(k-1)+4,iwater,i)
          velocity(2,ix,i)       = water_stay_real(6*(k-1)+5,iwater,i)
          velocity(3,ix,i)       = water_stay_real(6*(k-1)+6,iwater,i)
          id_l2g(ix,i)           = water_stay_int (k,iwater,i)  
          id_g2l(1,id_l2g(ix,i)) = i
          id_g2l(2,id_l2g(ix,i)) = ix
          charge(ix,i)           = water_charge(k)
          mass(ix,i)             = water_mass(k)
          atmcls(ix,i)           = water_atmcls(k)
        end do
      end do

      do iwater = 1, water_move(i)
        jwater = iwater + water_stay(i)
        do k = 1, 3
          water_list(k,jwater,i) = nsolute(i) + 3*(jwater-1) + k
          ix = water_list(k,jwater,i)
          coord(1,ix,i)          = water_move_real(6*(k-1)+1,iwater,i)
          coord(2,ix,i)          = water_move_real(6*(k-1)+2,iwater,i)
          coord(3,ix,i)          = water_move_real(6*(k-1)+3,iwater,i)
          velocity(1,ix,i)       = water_move_real(6*(k-1)+4,iwater,i)
          velocity(2,ix,i)       = water_move_real(6*(k-1)+5,iwater,i)
          velocity(3,ix,i)       = water_move_real(6*(k-1)+6,iwater,i)
          id_l2g(ix,i)           = water_move_int (k,iwater,i)
          id_g2l(1,id_l2g(ix,i)) = i
          id_g2l(2,id_l2g(ix,i)) = ix
          charge(ix,i)           = water_charge(k)
          mass(ix,i)             = water_mass(k)
          atmcls(ix,i)           = water_atmcls(k)
        end do
      end do

      nwater(i) = water_stay(i) + water_move(i)
      natom(i)  = nsolute(i) + 3*nwater(i)
    end do
  
    return

  end subroutine update_incoming_water

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_HGr
  !> @brief        check hydrogen-bonded particles incoming cells
  !! @authors      JJ
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_HGr(constraints, domain)

    ! formal variable
    type(s_constraints), target, intent(inout)  :: constraints
    type(s_domain),      target, intent(inout)  :: domain

    ! local variable
    integer                      :: i, j, k, k1, list, ix, connect
    integer                      :: num_atom

    real(wp),         pointer    :: coord(:,:,:), vel(:,:,:)
    real(wp),         pointer    :: hgr_move_real(:,:,:,:)
    real(wp),         pointer    :: hgr_stay_real(:,:,:,:)
    real(wp),         pointer    :: charge(:,:), mass(:,:)
    real(wp),         pointer    :: HGr_bond_dist(:,:,:,:)
    integer,          pointer    :: ncel_local
    integer,          pointer    :: nsolute(:)
    integer,          pointer    :: id_l2g(:,:), id_g2l(:,:), atmcls(:,:)
    integer,          pointer    :: No_HGr(:), HGr_local(:,:)
    integer,          pointer    :: HGr_bond_list(:,:,:,:)
    integer,          pointer    :: HGr_move(:,:), HGr_stay(:,:)
    integer,          pointer    :: HGr_move_int(:,:,:,:), HGr_stay_int(:,:,:,:)


    No_HGr        => constraints%no_hgr
    HGr_local     => constraints%HGr_local
    HGr_bond_list => constraints%HGr_bond_list 
    HGr_move      => constraints%HGr_move
    HGr_stay      => constraints%HGr_stay
    HGr_bond_dist => constraints%HGr_bond_dist
    HGr_move_real => constraints%HGr_move_real
    HGr_stay_real => constraints%HGr_stay_real
    HGr_move_int  => constraints%HGr_move_int
    HGr_stay_int  => constraints%HGr_stay_int
    connect       =  constraints%connect
 
    ncel_local    => domain%num_cell_local
    nsolute       => domain%num_solute 
    coord         => domain%coord
    vel           => domain%velocity
    charge        => domain%charge
    mass          => domain%mass
    atmcls        => domain%atom_cls_no
    id_l2g        => domain%id_l2g
    id_g2l        => domain%id_g2l


    do i = 1, ncel_local

      num_atom = No_HGr(i)

      do j = 1, connect

#ifdef DEBUG
        if (HGr_stay(j,i)+Hgr_move(j,i) > HGroupMax) then
          call error_msg('Debug: Update_Incoming_HGr> HGr is exceed HGroupMax')
        end if
#endif 

        do k = 1, HGr_stay(j,i)
          do list = 1, j+1
            num_atom = num_atom + 1
            HGr_bond_list(list,k,j,i) = num_atom
            ix = num_atom
            coord(1,ix,i)             = HGr_stay_real(9*(list-1)+1,k,j,i) 
            coord(2,ix,i)             = HGr_stay_real(9*(list-1)+2,k,j,i) 
            coord(3,ix,i)             = HGr_stay_real(9*(list-1)+3,k,j,i) 
            vel(1,ix,i)               = HGr_stay_real(9*(list-1)+4,k,j,i) 
            vel(2,ix,i)               = HGr_stay_real(9*(list-1)+5,k,j,i) 
            vel(3,ix,i)               = HGr_stay_real(9*(list-1)+6,k,j,i) 
            charge(ix,i)              = HGr_stay_real(9*(list-1)+7,k,j,i)
            mass(ix,i)                = HGr_stay_real(9*(list-1)+8,k,j,i)
            HGr_bond_dist(list,k,j,i) = HGr_stay_real(9*(list-1)+9,k,j,i)
            atmcls(ix,i)              = HGr_stay_int (2*(list-1)+1,k,j,i) 
            id_l2g(ix,i)              = HGr_stay_int (2*(list-1)+2,k,j,i) 
            id_g2l(1,id_l2g(ix,i))    = i
            id_g2l(2,id_l2g(ix,i))    = ix
          end do
        end do

        do k = 1, HGr_move(j,i)
          k1 = k + HGr_stay(j,i)
          do list = 1, j+1
            num_atom = num_atom + 1
            HGr_bond_list(list,k1,j,i) = num_atom
            ix = num_atom
            coord(1,ix,i)             = HGr_move_real(9*(list-1)+1,k,j,i)
            coord(2,ix,i)             = HGr_move_real(9*(list-1)+2,k,j,i)
            coord(3,ix,i)             = HGr_move_real(9*(list-1)+3,k,j,i)
            vel(1,ix,i)               = HGr_move_real(9*(list-1)+4,k,j,i)
            vel(2,ix,i)               = HGr_move_real(9*(list-1)+5,k,j,i)
            vel(3,ix,i)               = HGr_move_real(9*(list-1)+6,k,j,i)
            charge(ix,i)              = HGr_move_real(9*(list-1)+7,k,j,i)
            mass(ix,i)                = HGr_move_real(9*(list-1)+8,k,j,i)
            HGr_bond_dist(list,k1,j,i)= HGr_move_real(9*(list-1)+9,k,j,i)
            atmcls(ix,i)              = HGr_move_int (2*(list-1)+1,k,j,i)
            id_l2g(ix,i)              = HGr_move_int (2*(list-1)+2,k,j,i)
            id_g2l(1,id_l2g(ix,i))    = i
            id_g2l(2,id_l2g(ix,i))    = ix
          end do
        end do

        HGr_local(j,i) = HGr_stay(j,i) + HGr_move(j,i)

      end do

      nsolute(i) = num_atom

    end do
  
    return

  end subroutine update_incoming_HGr

end module sp_migration_mod
