!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_pairlist_mod
!> @brief   set pairlist for nonbonded interactions
!! @authors Jaewoon Jung (JJ)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_pairlist_mod

  use sp_pairlist_str_mod
  use sp_boundary_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use molecules_str_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: setup_pairlist
  public  :: update_pairlist_table
  public  :: update_pairlist_pbc

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_pairlist
  !> @brief        initialize/allocate/setup pairlist for nonbonded interactions
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    domain   : domain information
  !! @param[inout] pairlist : pair-list information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_pairlist(enefunc, domain, pairlist)

    ! formal arguments
    type(s_enefunc),         intent(in)    :: enefunc 
    type(s_domain),          intent(in)    :: domain 
    type(s_pairlist),        intent(inout) :: pairlist
    

    call init_pairlist(pairlist)

    pairlist%pairlistdist = enefunc%pairlistdist
   
    if (enefunc%table%water_table) then 

      call alloc_pairlist(pairlist, PairListTable, domain%num_cell_local)

      call timer(TimerPairList, TimerOn)
      call update_pairlist_table(enefunc, domain, pairlist)
      call timer(TimerPairList, TimerOff)

    else

      call alloc_pairlist(pairlist, PairListNoTable, domain%num_cell_local)

      call timer(TimerPairList, TimerOn)
      call update_pairlist_pbc(enefunc, domain, pairlist)
      call timer(TimerPairList, TimerOff)

    end if

    return

  end subroutine setup_pairlist

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_pairlist_table
  !> @brief        update pairlist among solute atoms in each domain  
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    domain   : domain information
  !! @param[inout] pairlist : pair-list information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_pairlist_table(enefunc, domain, pairlist)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_domain),   target, intent(in)    :: domain
    type(s_pairlist), target, intent(inout) :: pairlist

    ! local variables
    real(wp)                  :: pairdist2, waterdist2, dij(1:3), rij2
    real(wp)                  :: rtmp(1:3)
    logical                   :: nb15_calc
    integer                   :: i, j, ij, k, ix, iy, iwater, jwater, list
    integer                   :: id, omp_get_thread_num, num_nb15_cell
    integer                   :: num_excl, ini_excl, fin_excl
    integer                   :: num_nb14, ini_nb14, fin_nb14
    integer                   :: num_nb15_pre, num_nb15

    real(wp),         pointer :: coord(:,:,:), trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: pairdist, cell_move(:,:,:)
    integer,          pointer :: num_nonb_excl1(:,:), num_nb14_calc1(:,:)
    integer,          pointer :: num_nonb_excl(:,:), num_nb14_calc(:,:)
    integer,          pointer :: nonb_excl_list1(:,:), nb14_calc_list1(:,:)
    integer,          pointer :: nonb_excl_list(:,:), nb14_calc_list(:,:)
    integer,          pointer :: ncell, nboundary, natom(:)
    integer,          pointer :: nsolute(:), nwater(:), water_list(:,:,:)
    integer,          pointer :: cell_pair(:,:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: num_nb15_calcw(:,:), num_nb15_calcww(:,:)
    integer,          pointer :: nb15_list(:,:), nb15_listw(:,:)
    integer,          pointer :: nb15_listww(:,:)
    integer,          pointer :: nb15_cell(:), nb15_cellw(:), nb15_cellww(:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)
    integer,          pointer :: nb15_calc_water(:,:)


    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nb14_calc_list1 => enefunc%nb14_calc_list1
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nb14_calc   => enefunc%num_nb14_calc
    nonb_excl_list  => enefunc%nonb_excl_list
    nb14_calc_list  => enefunc%nb14_calc_list

    ncell           => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    water_list      => domain%water_list
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    cell_pair       => domain%cell_pairlist1
    cell_move       => domain%cell_move

    pairdist        => pairlist%pairlistdist
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc
    num_nb15_calcw  => pairlist%num_nb15_calcw
    num_nb15_calcww => pairlist%num_nb15_calcww
    nb15_list       => pairlist%nb15_list
    nb15_listw      => pairlist%nb15_listw
    nb15_listww     => pairlist%nb15_listww
    nb15_cell       => pairlist%nb15_cell
    nb15_cellw      => pairlist%nb15_cellw
    nb15_cellww     => pairlist%nb15_cellww
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list
    nb15_calc_water => pairlist%nb15_calc_water

    pairdist2       =  pairdist * pairdist
    waterdist2      = (pairdist+2.0_wp) * (pairdist+2.0_wp)

    do i = 1, ncell+nboundary
      do ix = 1, natom(i)
        trans2(1:3,ix,i) = coord(1:3,ix,i) + trans1(1:3,ix,i)
      end do
    end do

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, j, iy, k, ij, iwater, jwater, ini_excl, num_excl, &
    !$omp         ini_nb14, num_nb14, num_nb15, num_nb15_pre, fin_excl,        &
    !$omp         fin_nb14, nb15_calc, num_nb15_cell, rtmp, dij, rij2, list)
    !

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! make a pairlist among solutes in the same cell
    !
    do i = id+1, ncell, nthread

      ini_excl = 0
      num_excl = 0
      ini_nb14 = 0
      num_nb14 = 0
      num_nb15 = 0
      num_nb15_pre = 0

#ifdef DEBUG
      if (nsolute(i) > MaxAtom) &
        call error_msg( &
            'Debug: Update_Pairlist_Table> nsolute is exceed MaxAtom')
#endif

      do ix = 1, nsolute(i) - 1

        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc1(ix,i)
        num_nb14 = fin_nb14

        do iy = ix + 1, nsolute(i)

          nb15_calc = .true.

          do k = ini_excl, fin_excl
            if (iy == nonb_excl_list1(k,i)) then
              nb15_calc = .false.
              exit
            end if
          end do

          do k = ini_nb14, fin_nb14
            if (iy == nb14_calc_list1(k,i)) then
              nb15_calc = .false.
              exit
            end if
          end do

          if (nb15_calc) then
            num_nb15 = num_nb15 + 1
#ifdef DEBUG
            if (num_nb15 > MaxNb15) &
              call error_msg( &
                   'Debug: Update_Pairlist_Table> num_nb15 is exceed MaxNb15')
#endif
            nb15_calc_list1(num_nb15,i) = iy
          end if

        end do

        num_nb15_calc1(ix,i) = num_nb15 - num_nb15_pre
        num_nb15_pre = num_nb15

      end do
    end do

    ! Make a pairlist between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

#ifdef DEBUG
      if (nsolute(i) > MaxAtom .or. nsolute(j) > MaxAtom) &
        call error_msg( &
             'Debug: Update_Pairlist_Table> nsolute is exceed MaxAtom')

      if (nwater(i) > MaxWater .or. nwater(j) > MaxWater) &
        call error_msg( &
             'Debug: Update_Pairlist_Table> nwater is exceed MaxWater')
#endif

      ini_excl = 0
      num_excl = 0
      ini_nb14 = 0
      num_nb14 = 0
      num_nb15 = 0
      num_nb15_pre = 0
      num_nb15_cell = 0

      ! Pairlist between solutes
      !
      do ix = 1, nsolute(i)

        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc(ix,ij)
        num_nb14 = fin_nb14

        rtmp(1:3) = trans2(1:3,ix,i) + cell_move(1:3,j,i)

        do iy = 1, nsolute(j)

          nb15_calc = .true.

          do k = ini_excl, fin_excl
            if (iy == nonb_excl_list(k,ij)) then
              nb15_calc = .false.
              exit
            end if
          end do

          do k = ini_nb14, fin_nb14
            if (iy == nb14_calc_list(k,ij)) then
              nb15_calc = .false.
              exit
            end if
          end do

          if (nb15_calc) then

            dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j)
            rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

            ! store interaction table
            !
            if (rij2 < pairdist2) then
              num_nb15 = num_nb15 + 1

#ifdef DEBUG
              if (num_nb15 > MaxNb15) &
                call error_msg( &
                    'Debug: Update_Pairlist_Table> num_nb15 is exceed MaxNb15')
#endif
              nb15_calc_list(num_nb15,ij) = iy
            end if
          end if

        end do

        do iy = nsolute(j)+1, natom(j)

          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j)
          rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < pairdist2) then

            num_nb15 = num_nb15 + 1

#ifdef DEBUG
            if (num_nb15 > MaxNb15) &
              call error_msg( &
                'Debug: Update_Pairlist_Table> num_nb15 is exceed MaxNb15')
#endif
            nb15_calc_list(num_nb15,ij) = iy
          end if

        end do

        num_nb15_calc(ix,ij) = num_nb15 - num_nb15_pre

        if (num_nb15 /= num_nb15_pre) then
          num_nb15_cell = num_nb15_cell + 1
          nb15_list(num_nb15_cell,ij) = ix
        end if

        num_nb15_pre = num_nb15

      end do

      nb15_cell(ij) = num_nb15_cell

      ! Pairlist between solutes and solvent with different cells
      !
      num_nb15_cell = 0
      do iwater = 1, nwater(i)

        ix = nsolute(i) + 3*(iwater-1) + 1
        rtmp = trans2(1:3,ix,i) + cell_move(1:3,j,i)

        do iy = 1, nsolute(j)

          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j)
          rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < waterdist2) then
            num_nb15 = num_nb15 + 1

#ifdef DEBUG
            if (num_nb15 > MaxNb15) &
              call error_msg( &
                   'Debug: Update_Pairlist_Table> num_nb15 is exceed MaxNb15')
#endif
            nb15_calc_list(num_nb15,ij) = iy
          end if

        end do

        num_nb15_calcw(iwater,ij) = num_nb15 - num_nb15_pre

        if (num_nb15 /= num_nb15_pre) then
          num_nb15_cell = num_nb15_cell + 1
          nb15_listw(num_nb15_cell,ij) = iwater
        end if

        num_nb15_pre = num_nb15

      end do

      nb15_cellw(ij) = num_nb15_cell

    end do

    ! Make a pairlist between solvents of different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

#ifdef DEBUG
      if (nwater(i) > MaxWater .or. nwater(j) > MaxWater) &
        call error_msg( &
             'Debug: Update_Pairlist_Table> nwater is exceed MaxWater')
#endif

      num_nb15 = 0
      num_nb15_pre = 0
      num_nb15_cell = 0

      do iwater = 1, nwater(i)

        ix = nsolute(i) + 3*(iwater-1) + 1
        rtmp(1:3) = trans2(1:3,ix,i) + cell_move(1:3,j,i)

        do jwater = 1, nwater(j)

          iy = nsolute(j) + 3*(jwater-1) + 1
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j)
          rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! store interaction table
          !
          if (rij2 < waterdist2) then
            num_nb15 = num_nb15 + 1

#ifdef DEBUG
            if (num_nb15 > MaxNb15Water) &
              call error_msg( &
              'Debug: Update_Pairlist_Table> num_nb15 is exceed MaxNb15Water')
#endif
            nb15_calc_water(num_nb15,ij) = jwater

          end if

        end do

        num_nb15_calcww(iwater,ij) = num_nb15 - num_nb15_pre

        if (num_nb15 /= num_nb15_pre) then
          num_nb15_cell = num_nb15_cell + 1
          nb15_listww(num_nb15_cell,ij) = iwater
        end if

        num_nb15_pre = num_nb15

      end do

      nb15_cellww(ij) = num_nb15_cell

    end do

    !$omp end parallel

    return

  end subroutine update_pairlist_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_pairlist_pbc
  !> @brief        update pairlist in each domain with periodic boundary 
  !!               condition 
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    domain   : domain information
  !! @param[inout] pairlist : pair-list information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_pairlist_pbc(enefunc, domain, pairlist)
  
    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_domain),   target, intent(in)    :: domain  
    type(s_pairlist), target, intent(inout) :: pairlist

    ! local variables
    real(wp)                  :: pairdist2
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: rtmp(1:3)

    integer                   :: i, j, ij, k, ix, iy
    integer                   :: num_nb15_pre, num_nb15
    integer                   :: num_nb15_total, num_nb15_total1
    integer                   :: num_excl, ini_excl, fin_excl
    integer                   :: num_nb14, ini_nb14, fin_nb14
    logical                   :: nb15_calc
    integer                   :: id, omp_get_thread_num, num_nb15_cell

    real(wp),         pointer :: coord(:,:,:), cell_move(:,:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    integer,          pointer :: num_nonb_excl1(:,:), num_nb14_calc1(:,:)
    integer,          pointer :: num_nonb_excl(:,:), num_nb14_calc(:,:)
    integer,          pointer :: nonb_excl_list1(:,:), nb14_calc_list1(:,:)
    integer,          pointer :: nonb_excl_list(:,:), nb14_calc_list(:,:)
    integer,          pointer :: natom(:), cell_pairlist1(:,:)
    integer,          pointer :: ncell, nboundary
    integer,          pointer :: nb15_cell(:), nb15_list(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)

    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list1 => enefunc%nonb_excl_list1
    nb14_calc_list1 => enefunc%nb14_calc_list1
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nb14_calc   => enefunc%num_nb14_calc
    nonb_excl_list  => enefunc%nonb_excl_list
    nb14_calc_list  => enefunc%nb14_calc_list

    ncell           => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    natom           => domain%num_atom
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    cell_pairlist1  => domain%cell_pairlist1
    cell_move       => domain%cell_move

    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc
    nb15_list       => pairlist%nb15_list
    nb15_cell       => pairlist%nb15_cell
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list

    pairdist2       =  pairlist%pairlistdist * pairlist%pairlistdist
    num_nb15_total  =  0

    !$omp parallel                                                             &
    !$omp private(id, i, ix, ini_excl, num_excl, ini_nb14, num_nb14, num_nb15, &
    !$omp         num_nb15_pre, fin_excl, fin_nb14, iy, k, nb15_calc, ij, j,   &
    !$omp         rtmp, dij, rij2, num_nb15_cell)                              &
    !$omp reduction(+:num_nb15_total)

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    do i = id+1, ncell+nboundary, nthread
      do ix = 1, natom(i)
        trans2(1:3,ix,i) = coord(1:3,ix,i) + trans1(1:3,ix,i)
      end do
    end do

    !$omp barrier

    ! make a pairlist in the same cell
    !
    do i = id+1, ncell, nthread

      ini_excl = 0
      num_excl = 0
      ini_nb14 = 0
      num_nb14 = 0
      num_nb15 = 0
      num_nb15_pre = 0

#ifdef DEBUG
      if (natom(i) > MaxAtom) &
        call error_msg( &
             'Debug: Update_Pairlist_Pbc> natom(cell) is exceed MaxAtom')
#endif

      do ix = 1, natom(i) - 1

        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl1(ix,i)
        num_excl = fin_excl

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc1(ix,i)
        num_nb14 = fin_nb14

        do iy = ix + 1, natom(i)

          nb15_calc = .true.
          do k = ini_excl, fin_excl
            if (iy == nonb_excl_list1(k,i)) then
              nb15_calc = .false.
              exit
            end if
          end do

          do k = ini_nb14, fin_nb14
            if (iy == nb14_calc_list1(k,i)) then
              nb15_calc = .false.
              exit
            end if
          end do

          if (nb15_calc) then

            ! store interaction table
            !
            num_nb15 = num_nb15 + 1
#ifdef DEBUG
            if (num_nb15 > MaxNb15) &
              call error_msg( &
                   'Debug: Update_Pairlist_Pbc> num_nb15 is exceed MaxNb15')
#endif
            nb15_calc_list1(num_nb15,i) = iy
          end if
        end do

        num_nb15_calc1(ix,i) = num_nb15 - num_nb15_pre
        num_nb15_total = num_nb15_total + num_nb15 - num_nb15_pre
        num_nb15_pre = num_nb15

      end do
    end do 

    ! Make a pairlist between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist1(1,ij)
      j = cell_pairlist1(2,ij)

#ifdef DEBUG
      if (natom(i) > MaxAtom .or. natom(j) > MaxAtom) &
        call error_msg( &
             'Debug: Update_Pairlist_Pbc> natom(cell) is exceed MaxAtom')
#endif

      ini_excl = 0
      num_excl = 0
      ini_nb14 = 0
      num_nb14 = 0
      num_nb15 = 0
      num_nb15_pre = 0
      num_nb15_cell = 0

      do ix = 1, natom(i)

        ini_excl = num_excl + 1
        fin_excl = num_excl + num_nonb_excl(ix,ij)
        num_excl = fin_excl

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc(ix,ij)
        num_nb14 = fin_nb14

        rtmp(1:3) = trans2(1:3,ix,i) + cell_move(1:3,j,i)

        do iy = 1, natom(j)

          nb15_calc = .true.

          do k = ini_excl, fin_excl
            if (iy == nonb_excl_list(k,ij)) then
              nb15_calc = .false.
              exit
            end if
          end do

          do k = ini_nb14, fin_nb14
            if (iy == nb14_calc_list(k,ij)) then
              nb15_calc = .false.
              exit
            end if
          end do

          if (nb15_calc) then

            dij(1:3) = rtmp(1:3) - trans2(1:3,iy,j) 

            rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

            ! store interaction table
            !
            if (rij2 < pairdist2) then 

              num_nb15 = num_nb15 + 1

#ifdef DEBUG
              if (num_nb15 > MaxNb15) &
                call error_msg( &
                     'Debug: Update_Pairlist_Pbc> num_nb15 is exceed MaxNb15')
#endif
              nb15_calc_list(num_nb15,ij) = iy
            end if
          end if
        end do

        num_nb15_calc(ix,ij) = num_nb15 - num_nb15_pre

        if (num_nb15 /= num_nb15_pre) then
          num_nb15_cell = num_nb15_cell + 1
          nb15_list(num_nb15_cell,ij) = ix
        end if

        num_nb15_total = num_nb15_total + num_nb15 - num_nb15_pre
        num_nb15_pre = num_nb15

      end do

      nb15_cell(ij) = num_nb15_cell

    end do

    !$omp end parallel

#ifdef MPI
    call mpi_reduce(num_nb15_total, num_nb15_total1, 1, mpi_integer, mpi_sum, &
         0, mpi_comm_country, ierror)
#endif

    return

  end subroutine update_pairlist_pbc

end module sp_pairlist_mod
