!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_parallel_io_md
!> @brief   Parallel I/O module
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_parallel_io_mod

  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_ensemble_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use fileio_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! parameters
  integer,       parameter      :: PioFileHeaderMarker = 12345

  ! structures
  type, public :: s_pio_t0_info

    ! input files
    character(200)              :: input_topfile
    character(200)              :: input_parfile

    ! energy settings
    real(wp)                    :: energy_pairlistdist
    logical                     :: energy_table
    character(5)                :: energy_watermodel
    
    ! constraint settings
    logical                     :: constraint_rigidbond
    logical                     :: constraint_fastwater
    character(5)                :: constraint_watermodel

    ! ensemble settings
    integer                     :: ensemble_type

    ! boundary settings
    real(wp)                    :: boundary_boxsizex
    real(wp)                    :: boundary_boxsizey
    real(wp)                    :: boundary_boxsizez
    real(wp)                    :: boundary_originx
    real(wp)                    :: boundary_originy
    real(wp)                    :: boundary_originz
    integer                     :: boundary_domain_xyz

    ! selection settings    
    character(1000),allocatable :: selection_group(:)

    ! restraint settings
    integer,        allocatable :: restraint_func(:)
    character(256), allocatable :: restraint_const(:)
    character(256), allocatable :: restraint_index(:)

  end type s_pio_t0_info

  ! global variables
  logical,              public  :: pio_restart  = .false.

  type(s_constraints),  public, save, target :: pio_constraints
  type(s_dynvars),      public, save, target :: pio_dynvars
  type(s_dynamics),     public, save, target :: pio_dynamics
  type(s_pio_t0_info),  public, save, target :: pio_t0_info

  logical                       :: pio_crdtrj_hdr = .true.
  logical                       :: pio_veltrj_hdr = .true.

  integer,          allocatable :: tmp_atom_idx(:)
  real(wp),         allocatable :: tmp_atom_crd(:,:)


  ! subroutines
  public  :: pio_write_domain_trj
  public  :: pio_write_domain_rst
  public  :: pio_read_domain_rst
  public  :: pio_read_domain_str
  public  :: pio_check_compatible
  public  :: pio_check_ranked_file
  public  :: pio_get_ranked_filename
  public  :: pio_get_file_endian
  public  :: pio_open_file

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_write_domain_trj
  !> @brief        write trajectory for domain
  !! @authors      NT
  !! @param[in]    unit_no    : file unit number
  !! @param[in]    boundary   : boundary information
  !! @param[in]    domain     : domain information
  !! @param[in]    trajectory : trajectory information
  !! @param[in]    snapshot   : number of snapshot
  !! @param[in]    crd_trj    : flag for coordinate or trajectory
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_write_domain_trj(unit_no,    &
                                  boundary,   &
                                  domain,     &
                                  trajectory, &
                                  snapshot,   &
                                  crd_trj)

    ! formal arguments
    integer,                 intent(in)    :: unit_no
    type(s_boundary),        intent(in)    :: boundary
    type(s_domain),          intent(in)    :: domain
    real(wp),                intent(in)    :: trajectory(:,:,:)
    integer,                 intent(in)    :: snapshot
    logical,                 intent(in)    :: crd_trj

    ! local variables
    real(wp)                 :: box(6)
    integer                  :: i, ix, natom_domain


    ! write header
    !
    if (      crd_trj .and. pio_crdtrj_hdr .or. &
        .not. crd_trj .and. pio_veltrj_hdr) then

      ! my rank number
      write(unit_no) my_world_rank
      ! # of all atoms
      write(unit_no) domain%num_atom_all
      ! # of snapshot
      write(unit_no) snapshot

      if (crd_trj) then
        pio_crdtrj_hdr = .false.
      else
        pio_veltrj_hdr = .false.
      end if

    end if

    if (.not. allocated(tmp_atom_idx)) &
      allocate(tmp_atom_idx(domain%num_cell_local * MaxAtom))

    if (.not. allocated(tmp_atom_crd)) &
      allocate(tmp_atom_crd(3,domain%num_cell_local * MaxAtom))

    ! write body
    !
    natom_domain = 1
    do i = 1, domain%num_cell_local
      do ix = 1, domain%num_atom(i)
        tmp_atom_idx(    natom_domain) = domain%id_l2g(ix,i)
        tmp_atom_crd(1:3,natom_domain) = trajectory(1:3,ix,i)
        natom_domain = natom_domain + 1
      end do
    end do

    natom_domain = natom_domain - 1

    box = 0.0_wp
    box(1) = boundary%box_size_x_ref
    box(3) = boundary%box_size_y_ref
    box(6) = boundary%box_size_z_ref

    ! box information
    write(unit_no) box
    ! # of domain atoms
    write(unit_no) natom_domain
    ! global atom indices
    write(unit_no) tmp_atom_idx(    1:natom_domain)
    ! atom coordinates
    write(unit_no) tmp_atom_crd (1:3,1:natom_domain)

    return

  end subroutine pio_write_domain_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_write_domain_rst
  !> @brief        write restart information for domain
  !! @authors      NT
  !! @param[in]    filename    : restart filename 
  !! @param[in]    boundary    : boundary information
  !! @param[in]    domain      : domain information
  !! @param[in]    enefunc     : potential energy functions information
  !! @param[in]    constraints : constraints information (optional)
  !! @param[in]    dynvars     : dynvars information (optional)
  !! @param[in]    dynamics    : dynamics information (optional)
  !! @param[in]    t0_info     : t=0 information (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_write_domain_rst(filename,    &
                                  boundary,    &
                                  domain,      &
                                  enefunc,     &
                                  constraints, &
                                  dynvars,     &
                                  dynamics,    &
                                  t0_info)
                             
    ! formal arguments
    character(*),                   intent(in) :: filename
    type(s_boundary),               intent(in) :: boundary
    type(s_domain),                 intent(in) :: domain
    type(s_enefunc),                intent(in) :: enefunc
    type(s_constraints),  optional, intent(in) :: constraints
    type(s_dynvars),      optional, intent(in) :: dynvars
    type(s_dynamics),     optional, intent(in) :: dynamics
    type(s_pio_t0_info),  optional, intent(in) :: t0_info

    ! local variables
    integer               :: file
    integer               :: i, j, ncell, nvar, nvar1, nvar2, nvar3


    !$omp critical
    call pio_open_file(file, filename, IOFileOutputReplace)
    !$omp end critical


    ! write maximum value for allocatable variables
    !

    ! sp_domain_str
    write(file) MaxAtom
    write(file) MaxWater
    write(file) MaxMove
    write(file) MaxWaterMove

    ! sp_enefunc_str
    write(file) MaxBond
    write(file) MaxAngle
    write(file) MaxDihe
    write(file) MaxImpr
    write(file) MaxCmap
    write(file) BondMove
    write(file) AngleMove
    write(file) DiheMove
    write(file) ImprMove
    write(file) CmapMove

    ! sp_pairlist_str
    write(file) MaxNb15
    write(file) MaxNb15Water

    ! sp_constraints_str
    write(file) HGroupMax
    write(file) HGrpMaxMove


    ! write boundary information
    !

    write(file) boundary%box_size_x
    write(file) boundary%box_size_y
    write(file) boundary%box_size_z
    write(file) boundary%origin_x
    write(file) boundary%origin_y
    write(file) boundary%origin_z
    write(file) boundary%box_size_x_ref
    write(file) boundary%box_size_y_ref
    write(file) boundary%box_size_z_ref
    write(file) boundary%num_domain(1:3)
    write(file) boundary%num_cells_x
    write(file) boundary%num_cells_y
    write(file) boundary%num_cells_z


    ! write s_domain information
    !

    write(file) domain%num_deg_freedom
    write(file) domain%num_atom_all

    ! domain_str::Dynvar variables
    if (allocated(domain%num_atom)) then
      ncell = size(domain%num_atom(:))
    else
      ncell = 0
    end if
    write(file) ncell
    write(file) domain%num_atom   (1:ncell)
    write(file) domain%num_atom_t0(1:ncell)
    write(file) domain%num_solute (1:ncell)
    write(file) domain%num_water  (1:ncell)

    do i = 1, ncell

      ! 'MaxAtom' arrays
      nvar = domain%num_atom(i)
      write(file) domain%coord      (1:3, 1:nvar, i)
      write(file) domain%velocity   (1:3, 1:nvar, i)
      write(file) domain%trans_vec  (1:3, 1:nvar, i)
      write(file) domain%charge     (     1:nvar, i)
      write(file) domain%mass       (     1:nvar, i)
      write(file) domain%id_l2g     (     1:nvar, i)
      write(file) domain%atom_cls_no(     1:nvar, i)
      write(file) domain%solute_list(     1:nvar, i)

      ! 'MaxWater' arrays
      nvar = domain%num_water(i)
      write(file) domain%water_list (1:3, 1:nvar, i)

    end do

    ! domain_str::Global variables
    if (allocated(domain%id_g2l)) then
      nvar = size(domain%id_g2l(1,:))
    else
      nvar = 0
    end if
    write(file) nvar
    !! write size only
    

    ! write s_enefunc information
    !

    ! enefunc_str::Bond variables
    ! enefunc_str::Angl variables
    ! enefunc_str::Dihe variables
    ! enefunc_str::Impr variables

    ncell = domain%num_cell_local
    write(file) ncell
    write(file) enefunc%num_bond    (1:ncell)
    write(file) enefunc%num_angle   (1:ncell)
    write(file) enefunc%num_dihedral(1:ncell)
    write(file) enefunc%num_improper(1:ncell)

    do i = 1, ncell

      nvar = enefunc%num_bond(i)
      write(file) enefunc%bond_list       (1:2, 1:nvar, i)
      write(file) enefunc%bond_force_const(     1:nvar, i)
      write(file) enefunc%bond_dist_min   (     1:nvar, i)

      nvar = enefunc%num_angle(i)
      write(file) enefunc%angle_list       (1:3, 1:nvar, i)
      write(file) enefunc%angle_force_const(     1:nvar, i)
      write(file) enefunc%angle_theta_min  (     1:nvar, i)
      write(file) enefunc%urey_force_const (     1:nvar, i)
      write(file) enefunc%urey_rmin        (     1:nvar, i)

      nvar = enefunc%num_dihedral(i)
      write(file) enefunc%dihe_list       (1:4, 1:nvar, i)
      write(file) enefunc%dihe_force_const(     1:nvar, i)
      write(file) enefunc%dihe_periodicity(     1:nvar, i)
      write(file) enefunc%dihe_phase      (     1:nvar, i)

      nvar = enefunc%num_improper(i)
      write(file) enefunc%impr_list       (1:4, 1:nvar, i)
      write(file) enefunc%impr_force_const(     1:nvar, i)
      write(file) enefunc%impr_phase      (     1:nvar, i)

    end do

    ! enefunc_str::EneFuncRest variables

    write(file) enefunc%restraint

    if (allocated(enefunc%num_restraint)) then
      ncell = size(enefunc%num_restraint(:))
    else
      ncell = 0
    end if
    write(file) ncell
    write(file) enefunc%num_restraint(1:ncell)

    do i = 1, ncell

      nvar = enefunc%num_restraint(i)
      write(file) enefunc%restraint_atom (     1:nvar, i)
      write(file) enefunc%restraint_force(1:4, 1:nvar, i)
      write(file) enefunc%restraint_coord(1:3, 1:nvar, i)

    end do

    ! enefunc_str::Cmap variables

    if (allocated(enefunc%cmap_coef)) then
      nvar1 = size(enefunc%cmap_coef(1,1,:,1,1))
      nvar2 = size(enefunc%cmap_coef(1,1,1,1,:))
    else
      nvar1 = 0
      nvar2 = 0
    end if
    write(file) nvar1
    write(file) nvar2
    write(file) enefunc%num_cmap(1:ncell)

    do i = 1, ncell

      nvar = enefunc%num_cmap(i)
      write(file) enefunc%cmap_list(1:8, 1:nvar, i)
      write(file) enefunc%cmap_type(     1:nvar, i)

    end do

    write(file) enefunc%cmap_resolution(1:nvar2)
    write(file) enefunc%cmap_coef      (1:4, 1:4, 1:nvar1, 1:nvar1, 1:nvar2)

    ! enefunc_str::nonbond parameters
    
    nvar = enefunc%num_atom_cls

    write(file) nvar
    write(file) enefunc%nb14_lj12(1:nvar,1:nvar)
    write(file) enefunc%nb14_lj6 (1:nvar,1:nvar)
    write(file) enefunc%nonb_lj12(1:nvar,1:nvar)
    write(file) enefunc%nonb_lj6 (1:nvar,1:nvar)
    write(file) lj_coef          (1:2,   1:nvar)


    ! enefunc_str::other variables

    write(file) enefunc%table%atom_cls_no_O
    write(file) enefunc%table%atom_cls_no_H
    write(file) enefunc%table%charge_O
    write(file) enefunc%table%charge_H
    write(file) enefunc%table%mass_O
    write(file) enefunc%table%mass_H
    write(file) enefunc%table%water_model


    ! write s_constraints information
    !

    if (present(constraints)) then

      write(file) constraints%num_bonds
      write(file) constraints%connect
      write(file) constraints%water_rHH
      write(file) constraints%water_rOH
      write(file) constraints%water_massO
      write(file) constraints%water_massH

      ! domain_constraint_str::ConstraintsDomainBond
      if (allocated(constraints%No_HGr)) then
        nvar  = size(constraints%No_HGr(:))
        nvar1 = size(constraints%HGr_local(:,1))
        nvar2 = size(constraints%HGr_bond_list(:,1,1,1))
      else
        nvar  = 0
        nvar1 = 0
        nvar2 = 0
      end if

      write(file) nvar
      write(file) nvar1
      write(file) nvar2
      write(file) constraints%No_HGr   (         1:nvar)
      write(file) constraints%HGr_local(1:nvar1, 1:nvar)

      do i = 1, nvar
        do j = 1, nvar1
          nvar3 = constraints%HGr_local(j,i)
          write(file) nvar3
          write(file) constraints%HGr_bond_list(1:nvar2, 1:nvar3, j, i)
          write(file) constraints%HGr_bond_dist(1:nvar2, 1:nvar3, j, i)
        end do
      end do

    else

      write(file) pio_constraints%num_bonds
      write(file) pio_constraints%connect
      write(file) pio_constraints%water_rHH
      write(file) pio_constraints%water_rOH
      write(file) pio_constraints%water_massO
      write(file) pio_constraints%water_massH

      ! domain_constraint_str::ConstraintsDomainBond
      if (allocated(pio_constraints%No_HGr)) then
        nvar  = size(pio_constraints%No_HGr(:))
        nvar1 = size(pio_constraints%HGr_local(:,1))
        nvar2 = size(pio_constraints%HGr_bond_list(:,1,1,1))
      else
        nvar  = 0
        nvar1 = 0
        nvar2 = 0
      end if

      write(file) nvar
      write(file) nvar1
      write(file) nvar2
      write(file) pio_constraints%No_HGr   (         1:nvar)
      write(file) pio_constraints%HGr_local(1:nvar1, 1:nvar)
  
      do i = 1, nvar
        do j = 1, nvar1
          nvar3 = constraints%HGr_local(j,i)
          write(file) nvar3
          write(file) pio_constraints%HGr_bond_list(1:nvar2, 1:nvar3, j, i)
          write(file) pio_constraints%HGr_bond_dist(1:nvar2, 1:nvar3, j, i)
        end do
      end do

    end if


    ! dynvars information
    !

    if (present(dynvars)) then

      write(file) dynvars%thermostat_momentum
      write(file) dynvars%barostat_momentum(1:3)

    else

      write(file) pio_dynvars%thermostat_momentum
      write(file) pio_dynvars%barostat_momentum(1:3)

    end if


    ! dynamics information
    !

    if (present(dynamics)) then

      write(file) dynamics%iseed

    else

      write(file) pio_dynamics%iseed

    end if


    ! t0 information
    !
    if (present(t0_info)) then

      write(file) t0_info%input_topfile
      write(file) t0_info%input_parfile
      write(file) t0_info%energy_pairlistdist
      write(file) t0_info%energy_table
      write(file) t0_info%energy_watermodel
      write(file) t0_info%constraint_rigidbond
      write(file) t0_info%constraint_fastwater
      write(file) t0_info%constraint_watermodel
      write(file) t0_info%ensemble_type
      write(file) t0_info%boundary_boxsizex
      write(file) t0_info%boundary_boxsizey
      write(file) t0_info%boundary_boxsizez
      write(file) t0_info%boundary_originx
      write(file) t0_info%boundary_originy
      write(file) t0_info%boundary_originz
      write(file) t0_info%boundary_domain_xyz

      if (allocated(t0_info%selection_group)) then
        nvar = size(t0_info%selection_group)
      else
        nvar = 0
      end if
      write(file) nvar
      write(file) t0_info%selection_group(1:nvar)

      if (allocated(t0_info%restraint_func)) then
        nvar = size(t0_info%restraint_func)
      else
        nvar = 0
      end if
      write(file) nvar
      write(file) t0_info%restraint_func (1:nvar)
      write(file) t0_info%restraint_const(1:nvar)
      write(file) t0_info%restraint_index(1:nvar)

    else

      write(file) pio_t0_info%input_topfile
      write(file) pio_t0_info%input_parfile
      write(file) pio_t0_info%energy_pairlistdist
      write(file) pio_t0_info%energy_table
      write(file) pio_t0_info%energy_watermodel
      write(file) pio_t0_info%constraint_rigidbond
      write(file) pio_t0_info%constraint_fastwater
      write(file) pio_t0_info%constraint_watermodel
      write(file) pio_t0_info%ensemble_type
      write(file) pio_t0_info%boundary_boxsizex
      write(file) pio_t0_info%boundary_boxsizey
      write(file) pio_t0_info%boundary_boxsizez
      write(file) pio_t0_info%boundary_originx
      write(file) pio_t0_info%boundary_originy
      write(file) pio_t0_info%boundary_originz
      write(file) pio_t0_info%boundary_domain_xyz

      if (allocated(pio_t0_info%selection_group)) then
        nvar = size(pio_t0_info%selection_group)
      else
        nvar = 0
      end if
      write(file) nvar
      write(file) pio_t0_info%selection_group(1:nvar)

      if (allocated(pio_t0_info%restraint_func)) then
        nvar = size(pio_t0_info%restraint_func)
      else
        nvar = 0
      end if
      write(file) nvar
      write(file) pio_t0_info%restraint_func (1:nvar)
      write(file) pio_t0_info%restraint_const(1:nvar)
      write(file) pio_t0_info%restraint_index(1:nvar)

    end if


    ! write restart information
    !       (This domain restart file is RESTART or not.)
    write(file) pio_restart


    !$omp critical
    call close_file(file)
    !$omp end critical

    return

  end subroutine pio_write_domain_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_read_domain_rst
  !> @brief        read restart information for domain
  !! @authors      NT
  !! @param[in]    filename    : restart filename 
  !! @param[inout] boundary    : boundary information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] constraints : constraints information (optional)
  !! @param[inout] dynvars     : dynvars information (optional)
  !! @param[inout] dynamics    : dynamics information (optional)
  !! @param[inout] t0_info     : t=0 information (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_read_domain_rst(filename,    &
                                 boundary,    &
                                 domain,      &
                                 enefunc,     &
                                 constraints, &
                                 dynvars,     &
                                 dynamics,    &
                                 t0_info)
                             
    ! formal arguments
    character(*),                   intent(in)    :: filename
    type(s_boundary),               intent(inout) :: boundary
    type(s_domain),                 intent(inout) :: domain
    type(s_enefunc),                intent(inout) :: enefunc
    type(s_constraints),  optional, intent(inout) :: constraints
    type(s_dynvars),      optional, intent(inout) :: dynvars
    type(s_dynamics),     optional, intent(inout) :: dynamics
    type(s_pio_t0_info),  optional, intent(inout) :: t0_info

    ! local variables
    integer               :: file
    integer               :: i, j, ncell, nvar, nvar1, nvar2, nvar3
    integer               :: ix, icel, ig


    ! initialize
    !
    call init_boundary(boundary)
    call init_domain  (domain)
    call init_enefunc (enefunc)
    if (present(constraints)) call init_constraints(constraints)
    if (present(dynvars))     call init_dynvars    (dynvars)
    if (present(dynamics))    call init_dynamics   (dynamics)


    call pio_open_file(file, filename, IOFileInput)


    ! read maximum value for allocatable variables
    !

    ! sp_domain_str
    read(file) MaxAtom
    read(file) MaxWater
    read(file) MaxMove
    read(file) MaxWaterMove

    ! sp_enefunc_str
    read(file) MaxBond
    read(file) MaxAngle
    read(file) MaxDihe
    read(file) MaxImpr
    read(file) MaxCmap
    read(file) BondMove
    read(file) AngleMove
    read(file) DiheMove
    read(file) ImprMove
    read(file) CmapMove

    ! sp_pairlist_str
    read(file) MaxNb15
    read(file) MaxNb15Water

    ! sp_constraints_str
    read(file) HGroupMax
    read(file) HGrpMaxMove


    ! read boundary information
    !

    read(file) boundary%box_size_x
    read(file) boundary%box_size_y
    read(file) boundary%box_size_z
    read(file) boundary%origin_x
    read(file) boundary%origin_y
    read(file) boundary%origin_z
    read(file) boundary%box_size_x_ref
    read(file) boundary%box_size_y_ref
    read(file) boundary%box_size_z_ref
    read(file) boundary%num_domain(1:3)
    read(file) boundary%num_cells_x
    read(file) boundary%num_cells_y
    read(file) boundary%num_cells_z


    ! read s_domain information
    !

    read(file) domain%num_deg_freedom
    read(file) domain%num_atom_all

    ! domain_str::Dynvar variables
    read(file) ncell
    call alloc_domain(domain, DomainDynvar, ncell, 1, 1)
    read(file) domain%num_atom   (1:ncell)
    read(file) domain%num_atom_t0(1:ncell)
    read(file) domain%num_solute (1:ncell)
    read(file) domain%num_water  (1:ncell)

    do i = 1, ncell

      ! 'MaxAtom' arrays
      nvar = domain%num_atom(i)
      read(file) domain%coord      (1:3, 1:nvar, i)
      read(file) domain%velocity   (1:3, 1:nvar, i)
      read(file) domain%trans_vec  (1:3, 1:nvar, i)
      read(file) domain%charge     (     1:nvar, i)
      read(file) domain%mass       (     1:nvar, i)
      read(file) domain%id_l2g     (     1:nvar, i)
      read(file) domain%atom_cls_no(     1:nvar, i)
      read(file) domain%solute_list(     1:nvar, i)

      ! 'MaxWater' arrays
      nvar = domain%num_water(i)
      read(file) domain%water_list (1:3, 1:nvar, i)

    end do

    ! domain_str::Global variables
    read(file) nvar
    call alloc_domain(domain, DomainGlobal, nvar, 1, 1)
    !! Re-compute id_g2l by id_l2g
    do icel = 1, ncell
      do ix = 1, domain%num_atom(icel)
        ig = domain%id_l2g(ix, icel)
        domain%id_g2l(1, ig) = icel
        domain%id_g2l(2, ig) = ix
      end do
    end do
    

    ! read s_enefunc information
    !

    ! enefunc_str::Bond variables
    ! enefunc_str::Angl variables
    ! enefunc_str::Dihe variables
    ! enefunc_str::Impr variables

    read(file) ncell
    call alloc_enefunc(enefunc, EneFuncBase, ncell, ncell)
    call alloc_enefunc(enefunc, EneFuncBond, ncell, ncell)
    call alloc_enefunc(enefunc, EneFuncAngl, ncell, ncell)
    call alloc_enefunc(enefunc, EneFuncDihe, ncell, ncell)
    call alloc_enefunc(enefunc, EneFuncImpr, ncell, ncell)
    read(file) enefunc%num_bond    (1:ncell)
    read(file) enefunc%num_angle   (1:ncell)
    read(file) enefunc%num_dihedral(1:ncell)
    read(file) enefunc%num_improper(1:ncell)

    do i = 1, ncell

      nvar = enefunc%num_bond(i)
      read(file) enefunc%bond_list       (1:2, 1:nvar, i)
      read(file) enefunc%bond_force_const(     1:nvar, i)
      read(file) enefunc%bond_dist_min   (     1:nvar, i)

      nvar = enefunc%num_angle(i)
      read(file) enefunc%angle_list       (1:3, 1:nvar, i)
      read(file) enefunc%angle_force_const(     1:nvar, i)
      read(file) enefunc%angle_theta_min  (     1:nvar, i)
      read(file) enefunc%urey_force_const (     1:nvar, i)
      read(file) enefunc%urey_rmin        (     1:nvar, i)

      nvar = enefunc%num_dihedral(i)
      read(file) enefunc%dihe_list       (1:4, 1:nvar, i)
      read(file) enefunc%dihe_force_const(     1:nvar, i)
      read(file) enefunc%dihe_periodicity(     1:nvar, i)
      read(file) enefunc%dihe_phase      (     1:nvar, i)

      nvar = enefunc%num_improper(i)
      read(file) enefunc%impr_list       (1:4, 1:nvar, i)
      read(file) enefunc%impr_force_const(     1:nvar, i)
      read(file) enefunc%impr_phase      (     1:nvar, i)

    end do

    ! enefunc_str::EneFuncRest variables

    read(file) enefunc%restraint

    read(file) ncell
    call alloc_enefunc(enefunc, EneFuncRest, ncell, ncell)
    read(file) enefunc%num_restraint(1:ncell)

    do i = 1, ncell

      nvar = enefunc%num_restraint(i)
      read(file) enefunc%restraint_atom (     1:nvar, i)
      read(file) enefunc%restraint_force(1:4, 1:nvar, i)
      read(file) enefunc%restraint_coord(1:3, 1:nvar, i)

    end do

    ! enefunc_str::Cmap variables

    read(file) nvar1
    read(file) nvar2
    call alloc_enefunc(enefunc, EneFuncCmap, ncell, nvar1, nvar2)
    read(file) enefunc%num_cmap(1:ncell)
    
    do i = 1, ncell

      nvar = enefunc%num_cmap(i)
      read(file) enefunc%cmap_list  (1:8, 1:nvar, i)
      read(file) enefunc%cmap_type  (     1:nvar, i)

    end do

    read(file) enefunc%cmap_resolution(1:nvar2)
    read(file) enefunc%cmap_coef      (1:4, 1:4, 1:nvar1, 1:nvar1, 1:nvar2)


    ! enefunc_str::nonbond parameters
    
    read(file) nvar
    call alloc_enefunc(enefunc, EneFuncNbon, nvar)

    read(file) enefunc%nb14_lj12(1:nvar,1:nvar)
    read(file) enefunc%nb14_lj6 (1:nvar,1:nvar)
    read(file) enefunc%nonb_lj12(1:nvar,1:nvar)
    read(file) enefunc%nonb_lj6 (1:nvar,1:nvar)
    read(file) lj_coef          (1:2,   1:nvar)

    enefunc%num_atom_cls = nvar


    ! enefunc_str::other variables

    read(file) enefunc%table%atom_cls_no_O
    read(file) enefunc%table%atom_cls_no_H
    read(file) enefunc%table%charge_O
    read(file) enefunc%table%charge_H
    read(file) enefunc%table%mass_O
    read(file) enefunc%table%mass_H
    read(file) enefunc%table%water_model


    ! read s_constraints information
    !
    if (present(constraints)) then

      read(file) constraints%num_bonds
      read(file) constraints%connect
      read(file) constraints%water_rHH
      read(file) constraints%water_rOH
      read(file) constraints%water_massO
      read(file) constraints%water_massH

      ! domain_constraint_str::ConstraintsDomainBond
      read(file) nvar
      read(file) nvar1
      read(file) nvar2  ! == nvar1 + 1

      call alloc_constraints(constraints, ConstraintsDomainBond, nvar, nvar1)
      read(file) constraints%No_HGr   (         1:nvar)
      read(file) constraints%HGr_local(1:nvar1, 1:nvar)
  
      do i = 1, nvar
        do j = 1, nvar1
          read(file) nvar3
          read(file) constraints%HGr_bond_list(1:nvar2, 1:nvar3, j, i)
          read(file) constraints%HGr_bond_dist(1:nvar2, 1:nvar3, j, i)
        end do
      end do

    else

      read(file) pio_constraints%num_bonds
      read(file) pio_constraints%connect
      read(file) pio_constraints%water_rHH
      read(file) pio_constraints%water_rOH
      read(file) pio_constraints%water_massO
      read(file) pio_constraints%water_massH

      ! domain_constraint_str::ConstraintsDomainBond
      read(file) nvar
      read(file) nvar1
      read(file) nvar2  ! == nvar1 + 1

      call alloc_constraints(pio_constraints,ConstraintsDomainBond,nvar,nvar1)
      read(file) pio_constraints%No_HGr   (         1:nvar)
      read(file) pio_constraints%HGr_local(1:nvar1, 1:nvar)
  
      do i = 1, nvar
        do j = 1, nvar1
          read(file) nvar3
          read(file) pio_constraints%HGr_bond_list(1:nvar2, 1:nvar3, j, i)
          read(file) pio_constraints%HGr_bond_dist(1:nvar2, 1:nvar3, j, i)
        end do
      end do

    end if


    ! dynvars
    !

    if (present(dynvars)) then

      read(file) dynvars%thermostat_momentum
      read(file) dynvars%barostat_momentum(1:3)

    else

      read(file) pio_dynvars%thermostat_momentum
      read(file) pio_dynvars%barostat_momentum(1:3)

    end if


    ! dynamics
    !

    if (present(dynamics)) then

      read(file) dynamics%iseed

    else

      read(file) pio_dynamics%iseed

    end if


    ! t0 information
    !

    if (present(t0_info)) then

      read(file) t0_info%input_topfile
      read(file) t0_info%input_parfile
      read(file) t0_info%energy_pairlistdist
      read(file) t0_info%energy_table
      read(file) t0_info%energy_watermodel
      read(file) t0_info%constraint_rigidbond
      read(file) t0_info%constraint_fastwater
      read(file) t0_info%constraint_watermodel
      read(file) t0_info%ensemble_type
      read(file) t0_info%boundary_boxsizex
      read(file) t0_info%boundary_boxsizey
      read(file) t0_info%boundary_boxsizez
      read(file) t0_info%boundary_originx
      read(file) t0_info%boundary_originy
      read(file) t0_info%boundary_originz
      read(file) t0_info%boundary_domain_xyz

      read(file) nvar
      if (allocated(t0_info%selection_group)) &
        deallocate(t0_info%selection_group)
      allocate(t0_info%selection_group(nvar))
      read(file) t0_info%selection_group(1:nvar)

      read(file) nvar
      if (allocated(t0_info%restraint_func))  &
        deallocate(t0_info%restraint_func,    &
                   t0_info%restraint_const,   &
                   t0_info%restraint_index)
      allocate(t0_info%restraint_func(nvar),  &
               t0_info%restraint_const(nvar), &
               t0_info%restraint_index(nvar))
      read(file) t0_info%restraint_func(1:nvar)
      read(file) t0_info%restraint_const(1:nvar)
      read(file) t0_info%restraint_index(1:nvar)

    else

      read(file) pio_t0_info%input_topfile
      read(file) pio_t0_info%input_parfile
      read(file) pio_t0_info%energy_pairlistdist
      read(file) pio_t0_info%energy_table
      read(file) pio_t0_info%energy_watermodel
      read(file) pio_t0_info%constraint_rigidbond
      read(file) pio_t0_info%constraint_fastwater
      read(file) pio_t0_info%constraint_watermodel
      read(file) pio_t0_info%ensemble_type
      read(file) pio_t0_info%boundary_boxsizex
      read(file) pio_t0_info%boundary_boxsizey
      read(file) pio_t0_info%boundary_boxsizez
      read(file) pio_t0_info%boundary_originx
      read(file) pio_t0_info%boundary_originy
      read(file) pio_t0_info%boundary_originz
      read(file) pio_t0_info%boundary_domain_xyz

      read(file) nvar
      if (allocated(pio_t0_info%selection_group)) &
        deallocate(pio_t0_info%selection_group)
      allocate(pio_t0_info%selection_group(nvar))
      read(file) pio_t0_info%selection_group(1:nvar)

      read(file) nvar
      if (allocated(pio_t0_info%restraint_func))  &
        deallocate(pio_t0_info%restraint_func,    &
                   pio_t0_info%restraint_const,   &
                   pio_t0_info%restraint_index)
      allocate(pio_t0_info%restraint_func(nvar),  &
               pio_t0_info%restraint_const(nvar), &
               pio_t0_info%restraint_index(nvar))
      read(file) pio_t0_info%restraint_func(1:nvar)
      read(file) pio_t0_info%restraint_const(1:nvar)
      read(file) pio_t0_info%restraint_index(1:nvar)

    end if


    ! read restart information
    !       (Domain restart file is START or RESTART)
    read(file) pio_restart


    call close_file(file)

    return

  end subroutine pio_read_domain_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_read_domain_str
  !> @brief        read domain information
  !! @authors      NT
  !! @param[in]    filename : restart filename 
  !! @param[inout] boundary : boundary information
  !! @param[inout] domain   : domain information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_read_domain_str(filename, boundary, domain)
                             
    ! formal arguments
    character(*),            intent(in)    :: filename
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain

    ! local variables
    integer                  :: file
    integer                  :: i, ncell, nvar
    integer                  :: ix, icel, ig


    call pio_open_file(file, filename, IOFileInput)


    ! read maximum value for allocatable variables
    !

    ! sp_domain_str
    read(file) MaxAtom
    read(file) MaxWater
    read(file) MaxMove
    read(file) MaxWaterMove

    ! sp_enefunc_str
    read(file) MaxBond
    read(file) MaxAngle
    read(file) MaxDihe
    read(file) MaxImpr
    read(file) MaxCmap
    read(file) BondMove
    read(file) AngleMove
    read(file) DiheMove
    read(file) ImprMove
    read(file) CmapMove

    ! sp_pairlist_str
    read(file) MaxNb15
    read(file) MaxNb15Water

    ! sp_constraints_str
    read(file) HGroupMax
    read(file) HGrpMaxMove


    ! read boundary information
    !

    read(file) boundary%box_size_x
    read(file) boundary%box_size_y
    read(file) boundary%box_size_z
    read(file) boundary%origin_x
    read(file) boundary%origin_y
    read(file) boundary%origin_z
    read(file) boundary%box_size_x_ref
    read(file) boundary%box_size_y_ref
    read(file) boundary%box_size_z_ref
    read(file) boundary%num_domain(1:3)
    read(file) boundary%num_cells_x
    read(file) boundary%num_cells_y
    read(file) boundary%num_cells_z


    ! read s_domain information
    !

    read(file) domain%num_deg_freedom
    read(file) domain%num_atom_all

    ! domain_str::Dynvar variables
    read(file) ncell
    call alloc_domain(domain, DomainDynvar, ncell, 1, 1)
    read(file) domain%num_atom   (1:ncell)
    read(file) domain%num_atom_t0(1:ncell)
    read(file) domain%num_solute (1:ncell)
    read(file) domain%num_water  (1:ncell)

    do i = 1, ncell

      ! 'MaxAtom' arrays
      nvar = domain%num_atom(i)
      read(file) domain%coord      (1:3, 1:nvar, i)
      read(file) domain%velocity   (1:3, 1:nvar, i)
      read(file) domain%trans_vec  (1:3, 1:nvar, i)
      read(file) domain%charge     (     1:nvar, i)
      read(file) domain%mass       (     1:nvar, i)
      read(file) domain%id_l2g     (     1:nvar, i)
      read(file) domain%atom_cls_no(     1:nvar, i)
      read(file) domain%solute_list(     1:nvar, i)

      ! 'MaxWater' arrays
      nvar = domain%num_water(i)
      read(file) domain%water_list (1:3, 1:nvar, i)

    end do

    ! domain_str::Global variables
    read(file) nvar
    call alloc_domain(domain, DomainGlobal, nvar, 1, 1)
    !! Re-compute id_g2l by id_l2g
    do icel = 1, ncell
      do ix = 1, domain%num_atom(icel)
        ig = domain%id_l2g(ix, icel)
        domain%id_g2l(1, ig) = icel
        domain%id_g2l(2, ig) = ix
      end do
    end do

    ! read ncell_local
    read(file) domain%num_cell_local

    !$omp critical
    call close_file(file)
    !$omp end critical

    return

  end subroutine pio_read_domain_str

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_check_compatible
  !> @brief        check md input parameters are compatible between t=0 and t=n
  !! @authors      NT
  !! @param[in]    energy_pairlistdist   :      [ENERGY] pairlistdist
  !! @param[in]    energy_table          :      [ENERGY] table
  !! @param[in]    energy_watermodel     :      [ENERGY] watermodel
  !! @param[in]    constraint_rigidbond  : [CONSTRAINTS] rigidbond
  !! @param[in]    constraint_fastwater  : [CONSTRAINTS] fastwater
  !! @param[in]    constraint_watermodel : [CONSTRAINTS] watermodel
  !! @param[in]    ensemble_type         :    [ENSEMBLE] ensemble
  !! @param[in]    boundary_boxsizex     :    [BOUNDARY] boxsizex
  !! @param[in]    boundary_boxsizey     :    [BOUNDARY] boxsizey
  !! @param[in]    boundary_boxsizez     :    [BOUNDARY] boxsizez
  !! @param[in]    boundary_originx      :    [BOUNDARY] originx
  !! @param[in]    boundary_originy      :    [BOUNDARY] originy
  !! @param[in]    boundary_originz      :    [BOUNDARY] originz
  !! @param[in]    boundary_domain_x     :    [BOUNDARY] domain_x
  !! @param[in]    boundary_domain_y     :    [BOUNDARY] domain_y
  !! @param[in]    boundary_domain_z     :    [BOUNDARY] domain_z
  !! @param[in]    selection_group       :   [SELECTION] group000
  !! @param[in]    restraint_func        :  [RESTRAINTS] func000
  !! @param[in]    restraint_const       :  [RESTRAINTS] const000
  !! @param[in]    restraint_index       :  [RESTRAINTS] index000
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_check_compatible(energy_pairlistdist,   &
                                  energy_table,          &
                                  energy_watermodel,     &
                                  constraint_rigidbond,  &
                                  constraint_fastwater,  &
                                  constraint_watermodel, &
                                  ensemble_type,         &
                                  boundary_boxsizex,     &
                                  boundary_boxsizey,     &
                                  boundary_boxsizez,     &
                                  boundary_originx,      & 
                                  boundary_originy,      & 
                                  boundary_originz,      & 
                                  boundary_domain_x,     &
                                  boundary_domain_y,     &
                                  boundary_domain_z,     &
                                  selection_group,       &
                                  restraint_func,        &
                                  restraint_const,       &
                                  restraint_index)

    ! formal arguments
    real(wp),                     intent(in)   :: energy_pairlistdist
    logical,                      intent(in)   :: energy_table
    character(*),                 intent(in)   :: energy_watermodel
    logical,                      intent(in)   :: constraint_rigidbond
    logical,                      intent(in)   :: constraint_fastwater
    character(*),                 intent(in)   :: constraint_watermodel
    integer,                      intent(in)   :: ensemble_type
    real(wp),                     intent(in)   :: boundary_boxsizex
    real(wp),                     intent(in)   :: boundary_boxsizey
    real(wp),                     intent(in)   :: boundary_boxsizez
    real(wp),                     intent(in)   :: boundary_originx
    real(wp),                     intent(in)   :: boundary_originy
    real(wp),                     intent(in)   :: boundary_originz
    integer,                      intent(in)   :: boundary_domain_x
    integer,                      intent(in)   :: boundary_domain_y
    integer,                      intent(in)   :: boundary_domain_z
    character(1000), allocatable, intent(in)   :: selection_group(:)
    integer,         allocatable, intent(in)   :: restraint_func(:)
    character(256),  allocatable, intent(in)   :: restraint_const(:)
    character(256),  allocatable, intent(in)   :: restraint_index(:)
    
    ! local variables
    integer                :: ip, i, s1, s2
    character(200)         :: file1, file2
    character(50)          :: valstr
    logical                :: lerr, lwrn


    if (.not. main_rank) &
      return

    lerr = .false.
    lwrn = .false.

    if (pio_t0_info%energy_pairlistdist /= energy_pairlistdist) then

      write(MsgOut,'(a,f8.2)') &
           'Pio_Check_Compatible> ERROR : [Energy] pairlist dist: must be ', &
           pio_t0_info%energy_pairlistdist
      lerr = .true.

    end if

    if (pio_t0_info%energy_table .and. .not. energy_table .or. &
        .not. pio_t0_info%energy_table .and. energy_table) then

      if (pio_t0_info%energy_table) then
        valstr = 'YES'
      else
        valstr = 'NO'
      end if

      write(MsgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Energy] table:         must be ', &
           trim(valstr)
      lerr = .true.

    end if

    if (pio_t0_info%energy_watermodel /= energy_watermodel) then

      write(MsgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Energy] watermodel:    must be ', &
           trim(pio_t0_info%energy_watermodel)
      lerr = .true.

    end if

    if (pio_t0_info%constraint_rigidbond .and. .not. constraint_rigidbond .or. &
        .not. pio_t0_info%constraint_rigidbond .and. constraint_rigidbond) then

      if (pio_t0_info%constraint_rigidbond) then
        valstr = 'YES'
      else
        valstr = 'NO'
      end if
      write(MsgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Constraint]rigid-bond: must be ', &
           trim(valstr)
      lerr = .true.

    end if

    if (pio_t0_info%constraint_fastwater .and. .not. constraint_fastwater .or. &
        .not. pio_t0_info%constraint_fastwater .and. constraint_fastwater) then

      if (pio_t0_info%constraint_fastwater) then
        valstr = 'YES'
      else
        valstr = 'NO'
      end if
      write(msgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Constraint]fast water: must be ', &
           trim(valstr)
      lerr = .true.

    end if

    if (pio_t0_info%constraint_watermodel /= constraint_watermodel) then

      write(MsgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Constraint]watermodel: must be ', &
           trim(pio_t0_info%constraint_watermodel)
      lerr = .true.

    end if

    if (pio_t0_info%ensemble_type /= ensemble_type) then

      write(MsgOut,'(a,a)') &
           'Pio_Check_Compatible> ERROR : [Ensemble]ensemble: must be ', &
           EnsembleTypes(pio_t0_info%ensemble_type)
      lerr = .true.

    end if

    if (pio_t0_info%boundary_boxsizex /= boundary_boxsizex .or. &
        pio_t0_info%boundary_boxsizey /= boundary_boxsizey .or. &
        pio_t0_info%boundary_boxsizez /= boundary_boxsizez) then

      write(MsgOut,'(a,3f8.2)') &
           'Pio_Check_Compatible> WARNING : [Boundary] box size:    t=0 : ', &
           pio_t0_info%boundary_boxsizex, &
           pio_t0_info%boundary_boxsizey, &
           pio_t0_info%boundary_boxsizez
      lwrn = .true.

    end if

    if (pio_t0_info%boundary_originx /= boundary_originx .or. &
        pio_t0_info%boundary_originy /= boundary_originy .or. &
        pio_t0_info%boundary_originz /= boundary_originz) then

      write(MsgOut,'(a,3f8.2)') &
           'Pio_Check_Compatible> WARNING : [Boundary] origin:      t=0 : ', &
           pio_t0_info%boundary_originx, &
           pio_t0_info%boundary_originy, &
           pio_t0_info%boundary_originz
      lwrn = .true.

    end if

    if (boundary_domain_x == 0 .and. &
        boundary_domain_y == 0 .and. &
        boundary_domain_z == 0) then

      if (pio_t0_info%boundary_domain_xyz /= nproc_world) then
        write(MsgOut,'(a,i8)') &
           'Pio_Check_Compatible> ERROR : # of process :          must be ', &
           pio_t0_info%boundary_domain_xyz
        lerr = .true.
      end if

    else

      if (pio_t0_info%boundary_domain_xyz /= boundary_domain_x * &
                                             boundary_domain_y * &
                                             boundary_domain_z) then
        write(MsgOut,'(a,i8)') &
           'Pio_Check_Compatible> ERROR : [Constraint]domain xyz : must be ',&
           pio_t0_info%boundary_domain_xyz
        lerr = .true.
      end if

    end if

    s1 = 0
    s2 = 0
    if (allocated(pio_t0_info%selection_group)) &
      s1 = size(pio_t0_info%selection_group)
    if (allocated(selection_group)) &
      s2 = size(selection_group)

    if (s1 /= s2) then
      
      write(MsgOut,'(a,i8)') &
           'Pio_Check_Compatible> ERROR : [Selection] # of groups : must be ',&
           s1
      lerr = .true.

    else

      do i = 1, s1
        if (pio_t0_info%selection_group(i) /= selection_group(i)) then
          write(MsgOut,'(a,i8,a,a)') &
               'Pio_Check_Compatible> ERROR : [Selection] group ', i, &
               ' : must be ', trim(pio_t0_info%selection_group(i))
          lerr = .true.

        end if
      end do

    end if

    s1 = 0
    s2 = 0
    if (allocated(pio_t0_info%restraint_func)) &
      s1 = size(pio_t0_info%restraint_func)
    if (allocated(restraint_func)) &
      s2 = size(restraint_func)

    if (s1 /= s2) then
      
      write(MsgOut,'(a,i8)') &
           'Pio_Check_Compatible> ERROR : [Restraint] # of funcs : must be ',&
           s1
      lerr = .true.

    else

      do i = 1, s1
        if (pio_t0_info%restraint_func(i) /= restraint_func(i)) then
          write(MsgOut,'(a,i8,a,i8)') &
               'Pio_Check_Compatible> ERROR : [Restraint] func ', i, &
               ' : must be ', pio_t0_info%restraint_func(i)
          lerr = .true.

        end if
      end do

      do i = 1, s1
        if (pio_t0_info%restraint_const(i) /= restraint_const(i)) then
          write(MsgOut,'(a,i8,a,a)') &
               'Pio_Check_Compatible> ERROR : [Restraint] const ', i, &
               ' : must be ', trim(pio_t0_info%restraint_const(i))
          lerr = .true.

        end if
      end do

      do i = 1, s1
        if (pio_t0_info%restraint_index(i) /= restraint_index(i)) then
          write(MsgOut,'(a,i8,a,a)') &
               'Pio_Check_Compatible> ERROR : [Restraint] index ', i, &
               ' : must be ', trim(pio_t0_info%restraint_index(i))
          lerr = .true.

        end if
      end do

    end if


    if (lerr) &
      call error_msg('Pio_Check_Compatible> ERROR')

    if (lwrn) &
      write(MsgOut,*) ' '

    return

  end subroutine pio_check_compatible

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      pio_check_ranked_file
  !> @brief        check ranked file name or not
  !! @authors      NT
  !! @param[in]    filename : file name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function pio_check_ranked_file(filename)

    ! function
    logical                  :: pio_check_ranked_file

    ! formal arguments
    character(100),          intent(in) :: filename

    ! local variables
    integer                  :: ci1, ci2


    ! check filename
    !
    ci1 = scan(filename, '(', .true.)
    ci2 = scan(filename, ')', .true.)

    pio_check_ranked_file = (ci1 /= 0 .and. ci2 /= 0 .and. ci1 < ci2)

    return

  end function pio_check_ranked_file

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      pio_get_ranked_filename
  !> @brief        get ranked file name
  !! @authors      NT
  !! @param[in]    filename : file name
  !! @param[out]   nplace   : number of places
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function pio_get_ranked_filename(filename, nplace)

    ! function
    character(100)           :: pio_get_ranked_filename

    ! formal arguments
    character(*),            intent(in) :: filename
    integer,       optional, intent(in) :: nplace

    ! local variables
    integer                  :: ci1, ci2, ip
    character(50)            :: fmt_str

    ! constants
    integer,       parameter :: Places (7) = &
         (/10, 100, 1000, 10000, 100000, 1000000, 10000000/)


    if (present(nplace)) then
      ip = nplace
    else
      do ip = 1, size(Places)
        if (nproc_world < Places(ip)) &
          exit
      end do
    end if

    ! check filename
    !
    pio_get_ranked_filename = filename
    ci1 = scan(filename, '(', .true.)
    ci2 = scan(filename, ')', .true.)

    if (ci1 == 0 .or. ci2 ==0 .or. ci1 > ci2) then
      call error_msg( &
      'Pio_Get_Ranked_Filename> Filename is not correctly ranked in [OUTPUT]')
    end if

    write(fmt_str,'(A,I0,A,I0,A)') '(A,I',ip,'.',ip,',A)'

    write(pio_get_ranked_filename,fmt=fmt_str)   &
         trim(filename(1:ci1-1)), my_world_rank, &
         trim(filename(ci2+1:))

    return

  end function pio_get_ranked_filename

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      pio_get_file_endian
  !> @brief      
  !! @authors      NT
  !! @param[in]  
  !! @param[out] 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function pio_get_file_endian(filename)

    ! function
    integer          :: pio_get_file_endian

    ! formal arguments
    character(*),    intent(in) :: filename

    ! local variables
    integer          :: file, ival


    file = get_unit_no()

    open(file, &
         file    = filename,        &
         status  = 'old',           &
         form    = 'unformatted',   &
         convert = 'little_endian', &
         access  = 'direct',        &
         recl    = 4)
    read(file,rec=1) ival

    if (ival /= 4) then

      close(file)

      open(file, &
           file    = filename,      &
           status  = 'old',         &
           form    = 'unformatted', &
           convert = 'big_endian',  &
           access  = 'direct',      &
           recl    = 4)
      read(file,rec=1) ival

      if (ival /= 4) &
        call error_msg('Pio_Get_File_Endian> unknown file format.')

      pio_get_file_endian = IOFileBigEndian

    else

      pio_get_file_endian = IOFileLittleEndian

    end if

    close(file)
    call free_unit_no(file)

    return

  end function pio_get_file_endian

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pio_open_file
  !> @brief        open domain decomposition restart file
  !! @authors      NT
  !! @param[inout] file     : file unit number
  !! @param[in]    filename : file name
  !! @param[in]    in_out   : input or output
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pio_open_file(file, filename, in_out)

    ! formal arguments
    integer,                 intent(inout) :: file
    character(*),            intent(in)    :: filename
    integer,                 intent(in)    :: in_out

    ! local variables
    integer                  :: imark, ist


    select case(in_out)

    case (IOFileInput)

      call open_binary_file(file, filename, in_out, &
                            pio_get_file_endian(filename))
      read(file, iostat=ist) imark

      if (ist /= 0 .or. imark /= PioFileHeaderMarker) &
        call error_msg('Pio_Open_File> Unknown file format :'//&
        trim(filename))

    case (IOFileOutputNew, IOFileOutputReplace)

      call open_binary_file(file, filename, in_out)
      write(file) PioFileHeaderMarker

    case default

      call error_msg('Pio_Open_File> Unknown I/O mode')

    end select

    return

  end subroutine pio_open_file

end module sp_parallel_io_mod
