!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_setup_spdyn
!> @brief   setup variables and structures in MD (DD) simulaton
!! @authors Jaewoon Jung (JJ), Takaharu Mori (TM), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_setup_spdyn_mod

  use sp_control_mod
  use sp_restart_mod
  use sp_output_mod
  use sp_input_mod
  use sp_minimize_mod
  use sp_dynamics_mod
  use sp_dynvars_mod
  use sp_domain_mod
  use sp_ensemble_mod
  use sp_enefunc_table_mod
  use sp_restraints_mod
  use sp_constraints_mod
  use sp_boundary_mod
  use sp_pairlist_mod
  use sp_enefunc_mod
  use sp_energy_mod
  use sp_energy_pme_mod
  use sp_parallel_io_mod
  use sp_communicate_mod
  use sp_output_str_mod
  use sp_minimize_str_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_ensemble_str_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_energy_str_mod
  use sp_domain_str_mod
  use sp_remd_str_mod
  use sp_remd_mod
  use molecules_mod
  use molecules_str_mod
  use fileio_localres_mod
  use fileio_top_mod
  use fileio_par_mod
  use fileio_gpr_mod
  use fileio_psf_mod
  use fileio_pdb_mod
  use fileio_crd_mod
  use fileio_rst_mod
  use messages_mod
  use timers_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: setup_spdyn_md
  public  :: setup_spdyn_min
  public  :: setup_spdyn_remd
  public  :: setup_spdyn_md_pio
  public  :: setup_spdyn_min_pio
  private :: read_parallel_io_rst
  private :: save_parallel_io_t0

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_spdyn_md
  !> @brief        setup variables and structures in MD simulation
  !! @authors      JJ
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   output      : information of output
  !! @param[out]   molecule    : information of molecules
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   pairlist    : information of nonbonded pairlist
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   dynamics    : information of molecular dynamics
  !! @param[out]   constraints : information of constraints
  !! @param[out]   ensemble    : information of ensemble
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   domain      : information of each domain
  !! @param[out]   comm        : communicator for domain
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_spdyn_md(ctrl_data, output, molecule, enefunc, pairlist,    &
                           dynvars, dynamics, constraints, ensemble, boundary, &
                           domain, comm)

    ! formal arguments
    type(s_ctrl_data),       intent(inout) :: ctrl_data
    type(s_output),          intent(inout) :: output
    type(s_molecule),        intent(inout) :: molecule
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain
    type(s_comm),            intent(inout) :: comm

    ! local variables
    type(s_restraints)       :: restraints
    type(s_top)              :: top
    type(s_par)              :: par
    type(s_gpr)              :: gpr
    type(s_psf)              :: psf
    type(s_pdb)              :: pdb
    type(s_crd)              :: crd
    type(s_rst)              :: rst
    type(s_pdb)              :: ref
    type(s_localres)         :: localres


    ! setup parallel I/O
    !

    if (pio_check_ranked_file(ctrl_data%inp_info%rstfile)) then

      call setup_spdyn_md_pio(ctrl_data, output, enefunc, pairlist,     &
                              dynvars, dynamics, constraints, ensemble, &
                              boundary, domain, comm)
      return

    end if


    ! read input files
    !
    call input_md(ctrl_data%inp_info, top, par, psf, pdb, crd, rst, ref, &
                  localres)


    ! define molecules
    !
    call define_molecules(molecule, pdb, crd, top, par, gpr, psf, ref)

    call dealloc_pdb_all(pdb)
    call dealloc_crd_all(crd)
    call dealloc_top_all(top)
    call dealloc_gpr_all(gpr)
    call dealloc_psf_all(psf)
    call dealloc_pdb_all(ref)


    ! restart coordinates, velocity and boundary
    !
    if (rst%rstfile_type /= RstfileTypeUndef) then
      call setup_restart_pre(rst, molecule)
    end if


    ! set parameters for boundary condition
    !
    call setup_boundary(ctrl_data%bound_info,            &
                        ctrl_data%ene_info%table,        &
                        ctrl_data%ene_info%pairlistdist, &
                        ctrl_data%ene_info%water_model,  &
                        ctrl_data%ens_info%ensemble,     &
                        ctrl_data%cons_info%rigid_bond,  &
                        rst, boundary)


    ! set parameters for domain 
    !
    call setup_domain(ctrl_data%ene_info,  &
                      ctrl_data%cons_info, &
                      par, boundary, molecule, enefunc, constraints, domain)


    ! set parameters for communication
    !
    call setup_communicate(boundary, domain, comm)
    call setup_communicate_size(domain, comm)


    ! set parameters for restraints
    !
    call setup_restraints(ctrl_data%res_info, &
                          ctrl_data%sel_info, &
                          molecule, restraints)


    ! setup enefunc in each domain
    !
    call define_enefunc(ctrl_data%ene_info, &
                        par, localres, molecule, constraints, restraints, &
                        domain, enefunc)

    call dealloc_localres(localres, LocalRestraint)


    ! set parameters for pairlist
    !
    call setup_pairlist(enefunc, domain, pairlist)


    ! set parameters for PME
    !
    if (ctrl_data%ene_info%electrostatic == ElectrostaticPME) then
      call setup_pme(domain, boundary, enefunc)
      call pme_pre  (domain, boundary)
    end if


    ! set parameters for dynamic variables
    !
    call setup_dynvars(dynvars)


    ! set parameters for dynamics
    !
    call setup_dynamics(ctrl_data%dyn_info,   &
                        ctrl_data%bound_info, &
                        ctrl_data%res_info, molecule, dynamics)


    ! set parameters for ensemble
    !
    call setup_ensemble(ctrl_data%ens_info, dynamics, ensemble)



    ! set parameters for constraints
    !
    call setup_constraints(ctrl_data%cons_info, &
                           par, molecule, enefunc, constraints)

    call dealloc_par_all(par)
    call dealloc_restraints_all(restraints)
    call dealloc_molecules_all(molecule)


    ! set output
    !
    call setup_output_md(ctrl_data%out_info, dynamics, output)


    ! restart other variables
    !
    if (rst%rstfile_type /= RstfileTypeUndef) then
      call setup_restart_post(rst, dynamics, dynvars)
      call dealloc_rst_all(rst)
    end if


    ! save t0 information ( case of normal start and parallel restart out )
    !
    if (pio_check_ranked_file(output%rstfile)) then
      call save_parallel_io_t0(ctrl_data, boundary, .true.)
    end if

    domain%num_deg_freedom = molecule%num_deg_freedom

    ! change step number
    !
    if (dynamics%restart .and. dynamics%integrator == IntegratorLEAP) then
     dynvars%step = 1
    else
     dynvars%step = 0
    endif

    return

  end subroutine setup_spdyn_md

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_spdyn_min
  !> @brief        setup variables and structures in minimization
  !! @authors      TM, JJ
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   output      : information of output
  !! @param[out]   molecule    : information of molecules
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   pairlist    : information of nonbonded pairlist
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   minimize    : information of minimize
  !! @param[out]   constraints : information of constraints
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   domain      : information of each domain
  !! @param[out]   comm        : communicator for domain
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_spdyn_min(ctrl_data, output, molecule, enefunc, pairlist, &
                             dynvars, minimize, constraints, boundary,       &
                             domain, comm)

    ! formal arguments
    type(s_ctrl_data),       intent(inout) :: ctrl_data
    type(s_output),          intent(inout) :: output
    type(s_molecule),        intent(inout) :: molecule
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_minimize),        intent(inout) :: minimize
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain
    type(s_comm),            intent(inout) :: comm

    ! local variables
    type(s_restraints)       :: restraints
    type(s_top)              :: top
    type(s_par)              :: par
    type(s_gpr)              :: gpr
    type(s_crd)              :: crd
    type(s_psf)              :: psf
    type(s_pdb)              :: pdb
    type(s_rst)              :: rst
    type(s_pdb)              :: ref
    type(s_localres)         :: localres
 
 
    ! setup parallel I/O
    !

    if (pio_check_ranked_file(ctrl_data%inp_info%rstfile)) then

      call setup_spdyn_min_pio(ctrl_data, output, enefunc, pairlist,     &
                               dynvars, minimize, constraints, boundary, &
                               domain, comm)
      return

    end if


    ! read input files
    !
    call input_min(ctrl_data%inp_info, top, par, psf, pdb, crd, rst, ref, &
                   localres)


    ! define molecules
    !
    call define_molecules(molecule, pdb, crd, top, par, gpr, psf, ref)

    call dealloc_pdb_all(pdb)
    call dealloc_crd_all(crd)
    call dealloc_top_all(top)
    call dealloc_psf_all(psf)
    call dealloc_pdb_all(ref)


    ! set parameters for boundary condition
    !
    call setup_boundary(ctrl_data%bound_info,            &
                        ctrl_data%ene_info%table,        &
                        ctrl_data%ene_info%pairlistdist, &
                        ctrl_data%ene_info%water_model,  &
                        ctrl_data%ens_info%ensemble,     &
                        ctrl_data%cons_info%rigid_bond,  &
                        rst, boundary)


    ! set parameters for domain
    !
    call setup_domain(ctrl_data%ene_info,  &
                      ctrl_data%cons_info, &
                      par, boundary, molecule, enefunc, constraints, domain)


    ! set parameters for communication
    !
    call setup_communicate(boundary, domain, comm)
    call setup_communicate_size(domain, comm)


    ! set parameters for restraints
    !
    call setup_restraints(ctrl_data%res_info, &
                          ctrl_data%sel_info, &
                          molecule, restraints)


    ! setup enefunc in each domain
    !
    call define_enefunc(ctrl_data%ene_info, &
                        par, localres, molecule, constraints, restraints, &
                        domain, enefunc)

    call dealloc_restraints_all(restraints)
    call dealloc_molecules_all(molecule)
    call dealloc_localres(localres, LocalRestraint)
    call dealloc_par_all(par)


    ! set parameters for pairlist
    !
    call setup_pairlist(enefunc, domain, pairlist)


    ! set parameters for PME
    !
    if (ctrl_data%ene_info%electrostatic == ElectrostaticPME) then
      call setup_pme(domain, boundary, enefunc)
      call pme_pre  (domain, boundary)
    end if


    ! set parameters for dynamic variables
    !
    call setup_dynvars(dynvars)


    ! set parameters for minimize
    !
    call setup_minimize(ctrl_data%min_info, minimize)


    ! set output
    !
    call setup_output_min(ctrl_data%out_info, minimize, output)


    ! save t0 information ( case of normal start and parallel restart out )
    !
    if (pio_check_ranked_file(output%rstfile)) then
      call save_parallel_io_t0(ctrl_data, boundary, .false.)
    end if

    domain%num_deg_freedom = molecule%num_deg_freedom

    ! change step number
    !
    dynvars%step = 0

    return

  end subroutine setup_spdyn_min

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_spdyn_remd
  !> @brief        setup variables and structures in MD simulation
  !! @authors      TM
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   output      : information of output
  !! @param[out]   molecule    : information of molecules
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   pairlist    : information of nonbonded pairlist
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   dynamics    : information of molecular dynamics
  !! @param[out]   constraints : information of constraints
  !! @param[out]   ensemble    : information of ensemble
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   domain      : information of each domain
  !! @param[out]   comm        : communicator for domain
  !! @param[out]   remd        : information of remd
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_spdyn_remd(ctrl_data, output, molecule, enefunc, pairlist,    &
                              dynvars, dynamics, constraints, ensemble, boundary,&
                              domain, comm, remd)

    ! formal arguments
    type(s_ctrl_data),       intent(inout) :: ctrl_data
    type(s_output),          intent(inout) :: output
    type(s_molecule),        intent(inout) :: molecule
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain
    type(s_comm),            intent(inout) :: comm
    type(s_remd),            intent(inout) :: remd

    ! local variables
    type(s_restraints)       :: restraints
    type(s_top)              :: top
    type(s_par)              :: par
    type(s_gpr)              :: gpr
    type(s_psf)              :: psf
    type(s_pdb)              :: pdb
    type(s_crd)              :: crd
    type(s_rst)              :: rst
    type(s_pdb)              :: ref
    type(s_localres)         :: localres


    ! setup parallel I/O
    !

    if (pio_check_ranked_file(ctrl_data%inp_info%rstfile)) then

      call setup_spdyn_md_pio(ctrl_data, output, enefunc, pairlist,     &
                              dynvars, dynamics, constraints, ensemble, &
                              boundary, domain, comm)
      return

    end if


    ! read input files
    !
    call input_remd(ctrl_data%inp_info, top, par, psf, pdb, crd, rst, ref, &
                    localres)

    ! define molecules
    !
    call define_molecules(molecule, pdb, crd, top, par, gpr, psf, ref)

    call dealloc_pdb_all(pdb)
    call dealloc_crd_all(crd)
    call dealloc_top_all(top)
    call dealloc_gpr_all(gpr)
    call dealloc_psf_all(psf)
    call dealloc_pdb_all(ref)


    ! restart coordinates, velocity and boundary
    !
    if (rst%rstfile_type /= RstfileTypeUndef) then
      call setup_restart_pre(rst, molecule)
    end if


    ! set parameters for boundary condition
    !
    call setup_boundary(ctrl_data%bound_info,            &
                        ctrl_data%ene_info%table,        &
                        ctrl_data%ene_info%pairlistdist, &
                        ctrl_data%ene_info%water_model,  &
                        ctrl_data%ens_info%ensemble,     &
                        ctrl_data%cons_info%rigid_bond,  &
                        rst, boundary)


    ! set parameters for domain 
    !
    call setup_domain(ctrl_data%ene_info,  &
                      ctrl_data%cons_info, &
                      par, boundary, molecule, enefunc, constraints, domain)


    ! set parameters for communication
    !
    call setup_communicate(boundary, domain, comm)
    call setup_communicate_size(domain, comm)


    ! set parameters for restraints
    !
    call setup_restraints(ctrl_data%res_info, &
                          ctrl_data%sel_info, &
                          molecule, restraints)


    ! setup enefunc in each domain
    !
    call define_enefunc(ctrl_data%ene_info, &
                        par, localres, molecule, constraints, restraints, &
                        domain, enefunc)

    call dealloc_localres(localres, LocalRestraint)


    ! set parameters for pairlist
    !
    call setup_pairlist(enefunc, domain, pairlist)


    ! set parameters for PME
    !
    if (ctrl_data%ene_info%electrostatic == ElectrostaticPME) then
      call setup_pme(domain, boundary, enefunc)
      call pme_pre  (domain, boundary)
    end if


    ! set parameters for dynamic variables
    !
    call setup_dynvars(dynvars)


    ! set parameters for dynamics
    !
    call setup_dynamics(ctrl_data%dyn_info,   &
                        ctrl_data%bound_info, &
                        ctrl_data%res_info, molecule, dynamics)


    ! set parameters for ensemble
    !
    call setup_ensemble(ctrl_data%ens_info, dynamics, ensemble)



    ! set parameters for constraints
    !
    call setup_constraints(ctrl_data%cons_info, &
                           par, molecule, enefunc, constraints)

    call dealloc_par_all(par)


    ! setup remd
    !
    call setup_remd(ctrl_data%rep_info, rst, boundary, dynamics, molecule, &
                    domain, restraints, ensemble, enefunc, remd)
    call dealloc_restraints_all(restraints)
    call dealloc_molecules_all(molecule)


    ! set output
    !
    call setup_output_remd(ctrl_data%out_info, dynamics, output)


    ! restart other variables
    !
    if (rst%rstfile_type /= RstfileTypeUndef) then
      call setup_restart_post(rst, dynamics, dynvars)
      call dealloc_rst_all(rst)
    end if


    ! save t0 information ( case of normal start and parallel restart out )
    !
    if (pio_check_ranked_file(output%rstfile)) then
      call save_parallel_io_t0(ctrl_data, boundary, .true.)
    end if

    domain%num_deg_freedom = molecule%num_deg_freedom

    ! change step number
    !
    if (dynamics%restart .and. dynamics%integrator == IntegratorLEAP) then
     dynvars%step = 1
    else
     dynvars%step = 0
    endif

    return

  end subroutine setup_spdyn_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_spdyn_md_pio
  !> @brief        setup variables and structures in MD simulation
  !! @authors      JJ
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   output      : information of output
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   pairlist    : information of nonbonded pairlist
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   dynamics    : information of molecular dynamics
  !! @param[out]   constraints : information of constraints
  !! @param[out]   ensemble    : information of ensemble
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   domain      : information of each domain
  !! @param[out]   comm        : communicator for domain
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_spdyn_md_pio(ctrl_data, output, enefunc, pairlist,     &
                                dynvars, dynamics, constraints, ensemble, &
                                boundary, domain, comm)

    ! formal arguments
    type(s_ctrl_data),       intent(inout) :: ctrl_data
    type(s_output),          intent(inout) :: output
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain
    type(s_comm),            intent(inout) :: comm

    ! local variables
    type(s_localres)         :: localres


    ! read domain restart file (Parallel I/O)
    !
    call read_parallel_io_rst(ctrl_data,   &
                              domain,      &
                              enefunc,     &
                              boundary,    &
                              constraints, &
                              dynvars,     &
                              dynamics)

    !
    ! determine restart or not

    dynamics%restart=pio_restart

    ! read input files
    !
    if (ctrl_data%inp_info%local_resfile /= '') &
    call input_localres(ctrl_data%inp_info%local_resfile, localres)


    ! set parameters for boundary condition
    !
    call setup_boundary_pio(ctrl_data%bound_info,            &
                            ctrl_data%ene_info%table,        &
                            ctrl_data%ene_info%pairlistdist, &
                            ctrl_data%ene_info%water_model,  &
                            ctrl_data%ens_info%ensemble,     &
                            ctrl_data%cons_info%rigid_bond,  &
                            boundary)


    ! set parameters for domain 
    !
    call setup_domain_pio(ctrl_data%cons_info, &
                          boundary, enefunc, constraints, domain)


    ! set parameters for communication
    !
    call setup_communicate(boundary, domain, comm)
    call setup_communicate_size(domain, comm)


    ! setup enefunc in each domain
    !
    call define_enefunc_pio(ctrl_data%ene_info, &
                            localres, constraints, domain, enefunc)

    call dealloc_localres(localres, LocalRestraint)


    ! set pairlist
    !
    call setup_pairlist(enefunc, domain, pairlist)


    ! set parameters for PME
    !
    if (ctrl_data%ene_info%electrostatic == ElectrostaticPME) then
      call setup_pme(domain, boundary, enefunc)
      call pme_pre  (domain, boundary)
    end if


    ! set parameters for dynamics
    !
    call setup_dynamics_pio(ctrl_data%dyn_info,   &
                            ctrl_data%bound_info, &
                            dynamics)


    ! set parameters for ensemble
    !
    call setup_ensemble(ctrl_data%ens_info, dynamics, ensemble)


    ! set parameters for constraints
    !
    call setup_constraints_pio(ctrl_data%cons_info, constraints)


    ! set output
    !
    call setup_output_md(ctrl_data%out_info, dynamics, output)

    !
    ! output deg of freedom
    !
    call update_num_deg_freedom('Reading from parallel I/O', 0,  &
                                domain%num_deg_freedom)

    ! change step number
    !
    if (dynamics%restart .and. dynamics%integrator == IntegratorLEAP) then
     dynvars%step = 1
    else
     dynvars%step = 0
    endif

    return

  end subroutine setup_spdyn_md_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_spdyn_min_pio
  !> @brief        setup variables and structures in minimization
  !! @authors      JJ
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   output      : information of output
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   pairlist    : information of nonbonded pairlist
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   minimize    : information of minimize
  !! @param[out]   constraints : information of constraints
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   domain      : information of each domain
  !! @param[out]   comm        : communicator for domain
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_spdyn_min_pio(ctrl_data, output, enefunc, pairlist,     &
                                 dynvars, minimize, constraints, boundary, &
                                 domain, comm)

    ! formal arguments
    type(s_ctrl_data),       intent(inout) :: ctrl_data
    type(s_output),          intent(inout) :: output
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_minimize),        intent(inout) :: minimize
    type(s_constraints),     intent(inout) :: constraints
    type(s_boundary),        intent(inout) :: boundary
    type(s_domain),          intent(inout) :: domain
    type(s_comm),            intent(inout) :: comm

    ! local variables
    type(s_localres)         :: localres
 
 
    ! read domain restart file (Parallel I/O)
    !
    call read_parallel_io_rst(ctrl_data, &
                              domain,    &
                              enefunc,   &
                              boundary)


    ! read input files
    !
    if (ctrl_data%inp_info%local_resfile /= '') &
    call input_localres(ctrl_data%inp_info%local_resfile, localres)


    ! set parameters for boundary condition
    !
    call setup_boundary_pio(ctrl_data%bound_info,            &
                            ctrl_data%ene_info%table,        &
                            ctrl_data%ene_info%pairlistdist, &
                            ctrl_data%ene_info%water_model,  &
                            ctrl_data%ens_info%ensemble,     &
                            ctrl_data%cons_info%rigid_bond,  &
                            boundary)


    ! set parameters for domain 
    !
    call setup_domain_pio(ctrl_data%cons_info, &
                          boundary, enefunc, constraints, domain)


    ! set parameters for communication
    !
    call setup_communicate(boundary, domain, comm)
    call setup_communicate_size(domain, comm)


    ! setup enefunc in each domain
    !
    call define_enefunc_pio(ctrl_data%ene_info, &
                            localres, constraints, domain, enefunc)

    call dealloc_localres(localres, LocalRestraint)


    ! set pairlist
    !
    call setup_pairlist(enefunc, domain, pairlist)


    ! set parameters for PME
    !
    if (ctrl_data%ene_info%electrostatic == ElectrostaticPME) then
      call setup_pme(domain, boundary, enefunc)
      call pme_pre  (domain, boundary)
    end if


    ! set parameters for minimize
    !
    call setup_minimize(ctrl_data%min_info, minimize)


    ! set output
    !
    call setup_output_min(ctrl_data%out_info, minimize, output)

    ! change step number
    !
    dynvars%step = 0

    return

  end subroutine setup_spdyn_min_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_parallel_io_rst
  !> @brief        read parallel I/O restart file
  !! @authors      JJ
  !! @param[in]    ctrl_data   : information of control parameters
  !! @param[out]   domain      : information of each domain
  !! @param[out]   enefunc     : information of energy function
  !! @param[out]   boundary    : information of boundary condition
  !! @param[out]   constraints : information of constraints
  !! @param[out]   dynvars     : information of dynamic variables
  !! @param[out]   dynamics    : information of molecular dynamics
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_parallel_io_rst(ctrl_data, domain, enefunc, boundary, &
                                  constraints, dynvars, dynamics)

    ! formal arguments
    type(s_ctrl_data),             intent(in)    :: ctrl_data
    type(s_domain),                intent(inout) :: domain
    type(s_enefunc),               intent(inout) :: enefunc
    type(s_boundary),              intent(inout) :: boundary
    type(s_constraints), optional, intent(inout) :: constraints
    type(s_dynvars),     optional, intent(inout) :: dynvars
    type(s_dynamics),    optional, intent(inout) :: dynamics


    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Parallel_Io_Rst> '
      write(MsgOut,'(A)') ' '
    end if

    call pio_read_domain_rst( &
                          pio_get_ranked_filename(ctrl_data%inp_info%rstfile), &
                          boundary,    &
                          domain,      &
                          enefunc,     &
                          constraints, &
                          dynvars,     &
                          dynamics)

    if (.not. main_rank) &
      return

    write(MsgOut,'(A,A)') ' Parallel I/O file: ', &
                                             trim(ctrl_data%inp_info%rstfile)
    write(MsgOut,'(A,L)') '          Restart : ', pio_restart
    write(MsgOut,'(A)')   ' '

    call pio_check_compatible( &
                          ctrl_data%ene_info%pairlistdist, &
                          ctrl_data%ene_info%table,        &
                          ctrl_data%ene_info%water_model,  &
                          ctrl_data%cons_info%rigid_bond,  &
                          ctrl_data%cons_info%fast_water,  &
                          ctrl_data%cons_info%water_model, &
                          ctrl_data%ens_info%ensemble,     &
                          ctrl_data%bound_info%pbc_info%box_size_x, &
                          ctrl_data%bound_info%pbc_info%box_size_y, &
                          ctrl_data%bound_info%pbc_info%box_size_z, &
                          ctrl_data%bound_info%origin_x,   &
                          ctrl_data%bound_info%origin_y,   &
                          ctrl_data%bound_info%origin_z,   &
                          ctrl_data%bound_info%domain_x,   &
                          ctrl_data%bound_info%domain_y,   &
                          ctrl_data%bound_info%domain_z,   &
                          ctrl_data%sel_info%groups,       &
                          ctrl_data%res_info%function,     &
                          ctrl_data%res_info%constant,     &
                          ctrl_data%res_info%select_index)

    return

  end subroutine read_parallel_io_rst

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    save_parallel_io_t0
  !> @brief        save t0 status to parallel I/O system
  !! @authors      JJ
  !! @param[in]    ctrl_data : information of control parameters
  !! @param[in]    domain    : information of each domain
  !! @param[in]    boundary  : information of boundary condition
  !! @param[in]    md        : flag for MD or minimization
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine save_parallel_io_t0(ctrl_data, boundary, md)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: md


    pio_t0_info%input_topfile            = ctrl_data%inp_info%topfile
    pio_t0_info%input_parfile            = ctrl_data%inp_info%parfile
    pio_t0_info%energy_pairlistdist      = ctrl_data%ene_info%pairlistdist
    pio_t0_info%energy_table             = ctrl_data%ene_info%table
    pio_t0_info%energy_watermodel        = ctrl_data%ene_info%water_model
    pio_t0_info%boundary_boxsizex        = boundary%box_size_x
    pio_t0_info%boundary_boxsizey        = boundary%box_size_y
    pio_t0_info%boundary_boxsizez        = boundary%box_size_z
    pio_t0_info%boundary_originx         = boundary%origin_x
    pio_t0_info%boundary_originy         = boundary%origin_y
    pio_t0_info%boundary_originz         = boundary%origin_z
    pio_t0_info%boundary_domain_xyz      = boundary%num_domain(1)* &
                                           boundary%num_domain(2)* &
                                           boundary%num_domain(3)

    allocate(pio_t0_info%selection_group(size(ctrl_data%sel_info%groups)))
    pio_t0_info%selection_group(:)       = ctrl_data%sel_info%groups(:)

    allocate(pio_t0_info%restraint_func (size(ctrl_data%res_info%function)))
    allocate(pio_t0_info%restraint_const(size(ctrl_data%res_info%function)))
    allocate(pio_t0_info%restraint_index(size(ctrl_data%res_info%function)))
    pio_t0_info%restraint_func(:)        = ctrl_data%res_info%function(:)
    pio_t0_info%restraint_const(:)       = ctrl_data%res_info%constant(:)
    pio_t0_info%restraint_index(:)       = ctrl_data%res_info%select_index(:)

    if (md) then

      pio_t0_info%constraint_rigidbond   = ctrl_data%cons_info%rigid_bond
      pio_t0_info%constraint_fastwater   = ctrl_data%cons_info%fast_water
      pio_t0_info%constraint_watermodel  = ctrl_data%cons_info%water_model
      pio_t0_info%ensemble_type          = ctrl_data%ens_info%ensemble

    else

      pio_t0_info%constraint_rigidbond   = .false.
      pio_t0_info%constraint_fastwater   = .false.
      pio_t0_info%constraint_watermodel  = 'NONE'
      pio_t0_info%ensemble_type          = 0

      pio_constraints%num_bonds          = 0
      pio_constraints%connect            = 0
      pio_constraints%water_rHH          = 0.0_wp
      pio_constraints%water_rOH          = 0.0_wp
      pio_constraints%water_massO        = 0.0_wp
      pio_constraints%water_massH        = 0.0_wp

      pio_dynvars%thermostat_momentum    = 0.0_wp
      pio_dynvars%barostat_momentum(1:3) = 0.0_wp
      pio_dynamics%iseed                 = 0

    end if

    return

  end subroutine save_parallel_io_t0

end module sp_setup_spdyn_mod
