!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_update_domain_mod
!> @brief   library subroutine used for integrator
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_update_domain_mod

  use sp_domain_mod
  use sp_pairlist_mod
  use sp_enefunc_mod
  use sp_communicate_mod
  use sp_migration_mod
  use sp_minimize_str_mod
  use sp_dynamics_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: domain_interaction_update_md
  public  :: domain_interaction_update_min

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    domain_interaction_update_md
  !> @brief        update interactions with MD
  !! @authors      JJ
  !! @param[in]    istep       : step number
  !! @param[in]    dynamics    : dynamics information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary condition information
  !! @param[inout] constraints : constraints information
  !! @param[inout] comm        : communication information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine domain_interaction_update_md(istep, dynamics, domain, enefunc, &
                                          pairlist, boundary, constraints, comm)

    ! formal arguments
    integer,                 intent(in)    :: istep
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_boundary),        intent(inout) :: boundary
    type(s_constraints),     intent(inout) :: constraints
    type(s_comm),            intent(inout) :: comm

    ! local variable
    integer                  :: ncell, nb


    ncell = domain%num_cell_local
    nb    = domain%num_cell_boundary

    if (mod(istep,dynamics%nbupdate_period) == 0) then
      call timer(TimerIntegrator, TimerOn)

#ifdef DEBUG
      if (main_rank) then
        write(6,'(a,3f15.8)') &
             "Debuging > bsize_[x,y,z]", &
             boundary%box_size_x,boundary%box_size_y,boundary%box_size_z
        write(6,'(a,2i8)')    &
             "Debuging > ncell, nb",ncell,nb
      end if
#endif

      if (enefunc%table%water_table) then

        if (constraints%rigid_bond) then

          if (istep == dynamics%nbupdate_period) then

            call alloc_domain(domain, DomainPtlMove, ncell+nb, ncell, 1)
            call alloc_domain(domain, DomainWaterMove, ncell+nb, ncell, 1)

            call alloc_constraints(constraints, ConstraintsHGroupMove, &
                                   constraints%connect, ncell+nb)

            call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
            if (enefunc%restraint) &
            call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)

          end if

          call update_outgoing_atom (boundary, constraints, domain)
          call update_outgoing_HGr  (boundary, constraints, domain)
          call update_outgoing_water(boundary, domain)

          call communicate_constraints(domain, comm, constraints)

          call update_incoming_atom (constraints, domain)
          call update_incoming_HGr  (constraints, domain)
          call update_incoming_water(domain)

          call update_cell_size_constraints    (domain, comm, constraints)
          call update_cell_boundary_constraints(domain, comm, boundary, &
                                                constraints)

          call update_enefunc(.true., domain, comm, enefunc, constraints)

        else

          if (istep == dynamics%nbupdate_period) then

            call alloc_domain(domain, DomainPtlMove, ncell+nb, ncell, 1)
            call alloc_domain(domain, DomainWaterMove, ncell+nb, ncell, 1)

            call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
            if (enefunc%restraint) &
            call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)

          end if

          call update_outgoing_solute (boundary, domain)
          call update_outgoing_water  (boundary, domain)

          call communicate_solutewater(domain, comm)

          call update_incoming_solute (domain)
          call update_incoming_water  (domain)

          call update_cell_size_solute   (domain, comm)
          call update_cell_boundary_table(domain, comm, boundary)

          call update_enefunc(.true., domain, comm, enefunc)

        end if

        call timer(TimerIntegrator, TimerOff)
        call timer(TimerPairList, TimerOn)
        call update_pairlist_table(enefunc, domain, pairlist)
        call timer(TimerPairList, TimerOff)
        call timer(TimerIntegrator, TimerOn)

      else

        if (constraints%rigid_bond) then

          if (istep == dynamics%nbupdate_period) then

            call alloc_domain(domain, DomainPtlMove, ncell+nb, ncell, 1)
            call alloc_domain(domain, DomainWaterMove, ncell+nb, ncell, 1)

            call alloc_constraints(constraints, ConstraintsHGroupMove, &
                                   constraints%connect, ncell+nb)

            call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
            if (enefunc%restraint) &
            call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)

          end if

          call update_outgoing_atom (boundary, constraints, domain)
          call update_outgoing_HGr  (boundary, constraints, domain)
          call update_outgoing_water(boundary, domain)

          call communicate_constraints(domain, comm, constraints)

          call update_incoming_atom (constraints, domain)
          call update_incoming_HGr  (constraints, domain)
          call update_incoming_water(domain)

          call update_cell_size_constraints    (domain, comm, constraints)
          call update_cell_boundary_constraints(domain, comm, boundary, &
                                                constraints)

          call update_enefunc(.false., domain, comm, enefunc, constraints)

          call timer(TimerIntegrator, TimerOff)
          call timer(TimerPairList, TimerOn)
          call update_pairlist_pbc(enefunc, domain, pairlist)
          call timer(TimerPairList, TimerOff)
          call timer(TimerIntegrator, TimerOn)

        else

          if (istep == dynamics%nbupdate_period) then

            call alloc_domain(domain, DomainPtlMove, ncell+nb, ncell, 1)
            call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
            if (enefunc%restraint) &
                 call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)

          end if

          call update_outgoing_ptl(boundary, domain)
          call communicate_ptl    (domain, comm)
          call update_incoming_ptl(domain)

          call update_cell_size    (domain, comm)
          call update_cell_boundary(domain, comm, boundary)

          call update_enefunc(.false., domain, comm, enefunc)

          call timer(TimerIntegrator, TimerOff)
          call timer(TimerPairList, TimerOn)
          call update_pairlist_pbc(enefunc, domain, pairlist)
          call timer(TimerPairList, TimerOff)
          call timer(TimerIntegrator, TimerOn)

        end if

      end if

      call timer(TimerIntegrator, TimerOff)
    end if

    return

  end subroutine domain_interaction_update_md
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    domain_interaction_update_min
  !> @brief        update interactions with minimization
  !! @authors      JJ
  !! @param[in]    istep       : step number
  !! @param[in]    minimize    : minimize information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary condition information
  !! @param[inout] comm        : communication information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine domain_interaction_update_min(istep, minimize, domain, enefunc, &
                                           pairlist, boundary, comm)

    ! formal arguments
    integer,                 intent(in)    :: istep
    type(s_minimize),        intent(in)    :: minimize
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_boundary),        intent(inout) :: boundary
    type(s_comm),            intent(inout) :: comm

    ! local variable
    integer                  :: ncell, nb


    ncell = domain%num_cell_local
    nb    = domain%num_cell_boundary

    if (mod(istep,minimize%nbupdate_period) == 0) then
      call timer(TimerIntegrator, TimerOn)

      if (enefunc%table%water_table) then

        if (istep == minimize%nbupdate_period) then

          call alloc_domain(domain, DomainPtlMove,   ncell+nb, ncell, 1)
          call alloc_domain(domain, DomainWaterMove, ncell+nb, ncell, 1)

          call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
          if (enefunc%restraint) &
          call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)
        end if

        call update_outgoing_solute (boundary, domain)
        call update_outgoing_water  (boundary, domain)

        call communicate_solutewater(domain, comm)

        call update_incoming_solute (domain)
        call update_incoming_water  (domain)

        call update_cell_size_solute   (domain, comm)
        call update_cell_boundary_table(domain, comm, boundary)

        call update_enefunc(.true., domain, comm, enefunc)

        call timer(TimerIntegrator, TimerOff)
        call timer(TimerPairList, TimerOn)
        call update_pairlist_table(enefunc, domain, pairlist)
        call timer(TimerPairList, TimerOff)
        call timer(TimerIntegrator, TimerOn)

      else

        if (istep == minimize%nbupdate_period) then

          call alloc_domain(domain, DomainPtlMove, ncell+nb, ncell, 1)
          call alloc_enefunc(enefunc, EneFuncBondCell, ncell, ncell+nb)
          if (enefunc%restraint) &
          call alloc_enefunc(enefunc, EneFuncRestCell, ncell, ncell+nb)
        end if

        call update_outgoing_ptl(boundary, domain)

        call communicate_ptl(domain, comm)

        call update_incoming_ptl(domain)

        call update_cell_size    (domain, comm)
        call update_cell_boundary(domain, comm, boundary)

        call update_enefunc(.false., domain, comm, enefunc)

        call timer(TimerIntegrator, TimerOff)
        call timer(TimerPairList, TimerOn)
        call update_pairlist_pbc(enefunc, domain, pairlist)
        call timer(TimerPairList, TimerOff)
        call timer(TimerIntegrator, TimerOn)

      end if

      call timer(TimerIntegrator, TimerOff)

    end if

    return

  end subroutine domain_interaction_update_min

end module sp_update_domain_mod
