!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_enefunc_charmm_mod
!> @brief   define potential energy functions
!! @authors Jaewoon Jung (JJ), Takeshi Imai (TI), Chigusa Kobayashi (CK), 
!!          Takaharu Mori (TM), Yuji Sugita (YS)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_enefunc_charmm_mod

  use sp_enefunc_restraints_mod
  use sp_enefunc_table_mod
  use sp_communicate_mod
  use sp_energy_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_enefunc_str_mod
  use sp_energy_str_mod
  use sp_domain_str_mod
  use dihedral_libs_mod
  use molecules_str_mod
  use fileio_par_mod
  use fileio_localres_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: define_enefunc_charmm
  public  :: define_enefunc_charmm_pio
  public  :: update_enefunc_charmm
  private :: setup_enefunc_bond
  private :: setup_enefunc_bond_constraint
  private :: setup_enefunc_angl
  private :: setup_enefunc_angl_constraint
  private :: setup_enefunc_dihe
  private :: setup_enefunc_impr
  private :: setup_enefunc_cmap
  private :: setup_enefunc_nonb
  private :: setup_enefunc_localres_bond
  private :: setup_enefunc_localres_angle
  private :: setup_enefunc_localres_dihed
  private :: setup_enefunc_bond_pio
  private :: setup_enefunc_angl_pio
  private :: setup_enefunc_dihe_pio
  private :: setup_enefunc_impr_pio
  private :: setup_enefunc_cmap_pio
  private :: setup_enefunc_nonb_pio
  private :: update_outgoing_enefunc_bond
  private :: update_incoming_enefunc_bond
  private :: update_outgoing_enefunc_angl
  private :: update_incoming_enefunc_angl
  private :: update_outgoing_enefunc_dihe
  private :: update_incoming_enefunc_dihe
  private :: update_outgoing_enefunc_impr
  private :: update_incoming_enefunc_impr
  private :: update_outgoing_enefunc_cmap
  private :: update_incoming_enefunc_cmap
  private :: update_outgoing_enefunc_restraint
  private :: update_incoming_enefunc_restraint
  private :: count_nonb_excl
  private :: count_nonb_excl_rest
  private :: count_nonb_excl_solute
  private :: count_nonb_excl_solute_rest
  private :: count_nonb_excl_constraint
  private :: count_nonb_excl_constraint_rest
  private :: pack_array_4to3
  private :: pack_array_3to2

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc_charmm
  !> @brief        a driver subroutine for defining potential energy functions
  !! @authors      YS, TI, JJ, CK
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    localres    : local restraint information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    constraints : constraints information
  !! @param[in]    restraints  : restraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc_charmm(ene_info, par, localres, molecule, &
                                   constraints, restraints, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info 
    type(s_par),             intent(in)    :: par
    type(s_localres),        intent(in)    :: localres
    type(s_molecule),        intent(in)    :: molecule
    type(s_constraints),     intent(in)    :: constraints
    type(s_restraints),      intent(in)    :: restraints
    type(s_domain),          intent(inout) :: domain 
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: ncel, ncelb


    ! base
    !
    ncel  = domain%num_cell_local
    ncelb = domain%num_cell_local + domain%num_cell_boundary

    call alloc_enefunc(enefunc, EneFuncBase, ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncBond, ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncAngl, ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncDihe, ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncImpr, ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncBondCell, ncel, ncelb)


    if (ene_info%table .and. ene_info%water_model(1:4) == 'TIP3') then
      enefunc%table%water_table = .true.
    else
      enefunc%table%water_table = .false.
    end if


    if (.not. constraints%rigid_bond) then

      ! bond
      !
      call setup_enefunc_bond(par, molecule, domain, enefunc)

      ! angle
      !
      call setup_enefunc_angl(par, molecule, domain, enefunc)

    else

      ! bond
      !
      call setup_enefunc_bond_constraint( &
                              par, molecule, domain, constraints, enefunc)

      ! angle
      !
      call setup_enefunc_angl_constraint(par, molecule, domain, enefunc)

    end if

    ! dihedral
    !
    call setup_enefunc_dihe(par, molecule, domain, enefunc)

    ! improper
    !
    call setup_enefunc_impr(par, molecule, domain, enefunc)

    ! cmap
    !
    call setup_enefunc_cmap(ene_info, par, molecule, domain, enefunc)

    ! nonbonded
    !
    call setup_enefunc_nonb(ene_info, par, constraints, domain, enefunc)

    ! lookup table
    !
    if (ene_info%table) &
    call setup_enefunc_table(ene_info, enefunc)

    ! restraint
    !
    call setup_enefunc_restraints(molecule, restraints, domain, enefunc)

    call setup_enefunc_localres_bond (localres, domain, enefunc)
    call setup_enefunc_localres_angle(localres, domain, enefunc)
    call setup_enefunc_localres_dihed(localres, domain, enefunc)

    ! write summary of energy function
    !
    if (main_rank) then
      write(MsgOut,'(A)') &
           'Define_Enefunc_Charmm> Number of Interactions in Each Term'
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  bond_ene        = ', enefunc%num_bond_all, &
           '  angle_ene       = ', enefunc%num_angl_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  torsion_ene     = ', enefunc%num_dihe_all, &
           '  improper_ene    = ', enefunc%num_impr_all
      write(MsgOut,'(A20,I10)')                          &
           '  cmap_ene        = ', enefunc%num_cmap_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  nb_exclusions   = ', enefunc%num_excl_all, &
           '  nb14_calc       = ', enefunc%num_nb14_all
      write(MsgOut,'(A)') ' '
    end if

    return

  end subroutine define_enefunc_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc_charmm_pio
  !> @brief        a driver subroutine for defining potential energy functions
  !! @authors      YS, TI, JJ, CK
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    localres    : local restraint information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc_charmm_pio(ene_info, localres, constraints, &
                                       domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info 
    type(s_localres),        intent(in)    :: localres
    type(s_constraints),     intent(in)    :: constraints
    type(s_domain),          intent(inout) :: domain 
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: ncel, ncelb


    ! base
    !
    ncel  = domain%num_cell_local
    ncelb = domain%num_cell_local + domain%num_cell_boundary

    call alloc_enefunc  (enefunc, EneFuncBondCell, ncel, ncelb)


    if (ene_info%table .and. ene_info%water_model(1:4) == 'TIP3') then
      enefunc%table%water_table = .true.
    else
      enefunc%table%water_table = .false.
    end if


    ! bond
    !
    call setup_enefunc_bond_pio(domain, enefunc)

    ! angle
    !
    call setup_enefunc_angl_pio(domain, enefunc)

    ! dihedral
    !
    call setup_enefunc_dihe_pio(domain, enefunc)

    ! improper
    !
    call setup_enefunc_impr_pio(domain, enefunc)

    ! cmap
    !
    call setup_enefunc_cmap_pio(ene_info, domain, enefunc)

    ! nonbonded
    !
    call setup_enefunc_nonb_pio(ene_info, constraints, domain, enefunc)

    ! lookup table
    !
    if (ene_info%table) &
    call setup_enefunc_table(ene_info, enefunc)

    ! restraint
    !
    call setup_enefunc_localres_bond (localres, domain, enefunc)
    call setup_enefunc_localres_angle(localres, domain, enefunc)
    call setup_enefunc_localres_dihed(localres, domain, enefunc)

    ! write summary of energy function
    !
    if (main_rank) then
      write(MsgOut,'(A)') &
           'Define_Enefunc_Charmm_Pio> Number of Interactions in Each Term'
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  bond_ene        = ', enefunc%num_bond_all, &
           '  angle_ene       = ', enefunc%num_angl_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  torsion_ene     = ', enefunc%num_dihe_all, &
           '  improper_ene    = ', enefunc%num_impr_all
      write(MsgOut,'(A20,I10)')                          &
           '  cmap_ene        = ', enefunc%num_cmap_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  nb_exclusions   = ', enefunc%num_excl_all, &
           '  nb14_calc       = ', enefunc%num_nb14_all
      write(MsgOut,'(A)') ' '
    end if

    return

  end subroutine define_enefunc_charmm_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_enefunc_charmm
  !> @brief        a driver subroutine for updating potential energy functions
  !! @authors      JJ
  !! @param[in]    table       : flag for table or not
  !! @param[inout] domain      : domain information
  !! @param[inout] comm        : communication information
  !! @param[inout] enefunc     : energy potential functions information
  !! @param[inout] constraints : constraints information [optional]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_enefunc_charmm(table, domain, comm, enefunc, constraints)

    ! formal arguments
    logical,                       intent(in)    :: table
    type(s_domain),                intent(inout) :: domain
    type(s_comm),                  intent(inout) :: comm 
    type(s_enefunc),               intent(inout) :: enefunc
    type(s_constraints), optional, intent(inout) :: constraints

    ! local variables
    logical                        :: first


    first = .false.

    call timer(TimerUpdate , TimerOn)

    ! bond
    !
    call update_outgoing_enefunc_bond(domain, enefunc)

    ! angle
    !
    call update_outgoing_enefunc_angl(domain, enefunc)

    ! dihedral
    !
    call update_outgoing_enefunc_dihe(domain, enefunc)

    ! improper
    !
    call update_outgoing_enefunc_impr(domain, enefunc)

    ! cmap
    !
    call update_outgoing_enefunc_cmap(domain, enefunc)

    ! restraint
    !
    call update_outgoing_enefunc_restraint(domain, enefunc)

    ! send the necessary bonding informations to other cells
    !
    if (enefunc%restraint) then

      call communicate_rest(domain, comm, enefunc)

    else

      call communicate_bond(domain, comm, enefunc)

    end if

    ! bond
    !
    call update_incoming_enefunc_bond(domain, enefunc)

    ! angle
    !
    call update_incoming_enefunc_angl(domain, enefunc)

    ! dihedral
    !
    call update_incoming_enefunc_dihe(domain, enefunc)

    ! improper
    !
    call update_incoming_enefunc_impr(domain, enefunc)

    ! cmap
    !
    call update_incoming_enefunc_cmap(domain, enefunc)

    ! restraint
    !
    call update_incoming_enefunc_restraint(domain, enefunc)

    ! nonbonded
    !
    call timer(TimerUpdate , TimerOff)

    if (present(constraints)) then

      if (enefunc%local_restraint) then
        call count_nonb_excl_constraint_rest(first, table, constraints, &
                                             domain, enefunc)

      else
        call count_nonb_excl_constraint(first, table, constraints, &
                                             domain, enefunc)

      end if

    else

      if (enefunc%local_restraint) then
        if (table) then
          call count_nonb_excl_solute_rest(first, domain, enefunc)

        else
          call count_nonb_excl_rest(first, domain, enefunc)

        end if
      else
        if (table) then
          call count_nonb_excl_solute(first, domain, enefunc)

        else
          call count_nonb_excl(first, domain, enefunc)

        end if
      end if

    endif

    return

  end subroutine update_enefunc_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond
  !> @brief        define BOND term for each cell in potential energy function
  !! @authors      YS, JJ, TM
  !! @param[in]    par      : CHARMM PAR information
  !! @param[in]    molecule : molecule information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond(par, molecule, domain, enefunc)

    ! formal arguments
    type(s_par),             intent(in)    :: par
    type(s_molecule),        intent(in)    :: molecule
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, icel_local
    integer                  :: icel1, icel2
    integer                  :: nbond, nbond_p, i1, i2, found
    character(6)             :: ci1, ci2

    real(wp),        pointer :: force(:,:), dist(:,:)
    integer,         pointer :: bond(:), list(:,:,:)
    integer,         pointer :: ncel, cell_pair(:,:)
    integer,         pointer :: id_g2l(:,:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    bond      => enefunc%num_bond
    list      => enefunc%bond_list
    force     => enefunc%bond_force_const
    dist      => enefunc%bond_dist_min

    nbond     = molecule%num_bonds
    nbond_p   = par%num_bonds

    do i = 1, nbond

      ci1 = molecule%atom_cls_name(molecule%bond_list(1,i))
      ci2 = molecule%atom_cls_name(molecule%bond_list(2,i))
      i1  = molecule%bond_list(1,i)
      i2  = molecule%bond_list(2,i)

      icel1 = id_g2l(1,i1)
      icel2 = id_g2l(1,i2)

      ! Check if it is in my domain
      !
      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local > 0 .and. icel_local <= ncel) then

          do j = 1, nbond_p
            if ((ci1 == par%bond_atom_cls(1,j) .and.  &
                 ci2 == par%bond_atom_cls(2,j)) .or.  &
                (ci1 == par%bond_atom_cls(2,j) .and.  &
                 ci2 == par%bond_atom_cls(1,j))) then

              bond (icel_local) = bond(icel_local) + 1
              list (1,bond(icel_local),icel_local) = i1
              list (2,bond(icel_local),icel_local) = i2
              force(bond(icel_local),icel_local)   = par%bond_force_const(j) 
              dist (bond(icel_local),icel_local)   = par%bond_dist_min(j)
              exit

            end if
          end do

          if (j == nbond_p + 1) &
            write(MsgOut,*) &
              'Setup_Enefunc_Bond> not found BOND: [', &
              ci1, ']-[', ci2, '] in parameter file. (ERROR)'

        end if

      end if

    end do
   
    found = 0 
    do i = 1, ncel
      found = found + bond(i)
      if (bond(i) > MaxBond) &
        call error_msg('Setup_Enefunc_Bond> Too many bonds.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif

    if (enefunc%num_bond_all /= nbond) &
      call error_msg('Setup_Enefunc_Bond> Some bond paremeters are missing.')

    return
    
  end subroutine setup_enefunc_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_constraint
  !> @brief        define BOND term between heavy atoms
  !! @authors      JJ
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    domain      : domain information
  !! @param[in]    constraints : constraints information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_constraint(par, molecule, domain, &
                                           constraints, enefunc)

    ! formal arguments
    type(s_par),                 intent(in)    :: par
    type(s_molecule),            intent(in)    :: molecule
    type(s_domain),      target, intent(in)    :: domain
    type(s_constraints), target, intent(in)    :: constraints
    type(s_enefunc),     target, intent(inout) :: enefunc

    ! local variable
    integer                      :: i, j, icel_local
    integer                      :: icel1, icel2
    integer                      :: nbond, nbond_p, i1, i2, found
    integer                      :: len_ri1, len_ri2
    integer                      :: wat_bonds, tmp_mole_no, mole_no, wat_found
    character(6)                 :: ci1, ci2
    character(4)                 :: atom1, atom2, ri1, ri2

    real(wp),            pointer :: force(:,:), dist(:,:)
    integer,             pointer :: bond(:), list(:,:,:), num_water
    integer,             pointer :: ncel, cell_pair(:,:), id_g2l(:,:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    bond      => enefunc%num_bond
    list      => enefunc%bond_list
    force     => enefunc%bond_force_const
    dist      => enefunc%bond_dist_min
    num_water => enefunc%table%num_water

    nbond     = molecule%num_bonds
    nbond_p   = par%num_bonds

    do i = 1, nbond

      i1    = molecule%bond_list(1,i)
      i2    = molecule%bond_list(2,i)
      atom1 = molecule%atom_name(i1)
      atom2 = molecule%atom_name(i2)

      if (atom1(1:1) /= 'H' .and. atom2(1:1) /= 'H') then

        ci1 = molecule%atom_cls_name(molecule%bond_list(1,i))
        ci2 = molecule%atom_cls_name(molecule%bond_list(2,i))

        icel1 = id_g2l(1,i1)
        icel2 = id_g2l(1,i2)

        ! Check if it is in my domain
        !
        if (icel1 /= 0 .and. icel2 /= 0) then

          icel_local = cell_pair(icel1,icel2)

          if (icel_local > 0 .and. icel_local <= ncel) then

            do j = 1, nbond_p
              if ((ci1 == par%bond_atom_cls(1, j) .and.  &
                   ci2 == par%bond_atom_cls(2, j)) .or.  &
                  (ci1 == par%bond_atom_cls(2, j) .and.  &
                   ci2 == par%bond_atom_cls(1, j))) then

                bond (icel_local) = bond(icel_local) + 1
                list (1,bond(icel_local),icel_local) = i1
                list (2,bond(icel_local),icel_local) = i2
                force(bond(icel_local),icel_local) = par%bond_force_const(j) 
                dist (bond(icel_local),icel_local) = par%bond_dist_min(j)
                exit

              end if
            end do

            if (j == nbond_p + 1) &
              write(MsgOut,*) &
                'Setup_Enefunc_Bond_Constraint> not found BOND: [', &
                ci1, ']-[', ci2, '] in parameter file. (ERROR)'

          end if

        end if

      end if

    end do


    ! for water molecule
    !
    wat_bonds = 0
    wat_found = 0

    if (constraints%fast_water) then

      tmp_mole_no = -1

      do i = 1, nbond

        ri1 = molecule%residue_name(molecule%bond_list(1,i))
        ri2 = molecule%residue_name(molecule%bond_list(2,i))
        len_ri1 = len_trim(ri1)
        len_ri2 = len_trim(ri2)

        if (ri1(1:len_ri1) == constraints%water_model(1:len_ri1) .and. &
            ri2(1:len_ri2) == constraints%water_model(1:len_ri2)) then

          wat_bonds=wat_bonds+1
          mole_no = molecule%molecule_no(molecule%bond_list(1,i))

          if (mole_no /= tmp_mole_no) then
            wat_found = wat_found +1
            tmp_mole_no = mole_no
          end if

        end if

      end do

      if (wat_found /= num_water) &
        call error_msg( &
          'Setup_Enefunc_Bond_Constraint> # of water is incorrect')

    end if

    found = 0 
    do i = 1, ncel
      found = found + bond(i)
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif

    if (constraints%fast_water) then

      if (enefunc%num_bond_all /= (nbond-constraints%num_bonds-wat_bonds)) then
        call error_msg( &
          'Setup_Enefunc_Bond_Constraint> Some bond paremeters are missing.')
      end if

    else

      if (enefunc%num_bond_all /= (nbond-constraints%num_bonds-3*num_water))then
        call error_msg( &
          'Setup_Enefunc_Bond_Constraint> Some bond paremeters are missing.')
      end if

    end if

    return
    
  end subroutine setup_enefunc_bond_constraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl
  !> @brief        define ANGLE term for each cell in potential energy function
  !! @authors      YS, JJ, TM
  !! @param[in]    par      : CHARMM PAR information
  !! @param[in]    molecule : molecule information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl(par, molecule, domain, enefunc)

    ! formal arguments
    type(s_par),     target, intent(in)    :: par
    type(s_molecule),        intent(in)    :: molecule
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j, icel_local
    integer                  :: icel1, icel2
    integer                  :: nangl, nangl_p, found
    integer                  :: list(3)
    character(6)             :: ci1, ci2, ci3

    real(wp),        pointer :: force(:,:), theta(:,:)
    real(wp),        pointer :: ubforce(:,:), ubrmin(:,:)
    integer,         pointer :: angle(:), alist(:,:,:)
    integer,         pointer :: ncel, cell_pair(:,:), id_g2l(:,:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    angle     => enefunc%num_angle
    alist     => enefunc%angle_list
    force     => enefunc%angle_force_const
    theta     => enefunc%angle_theta_min
    ubforce   => enefunc%urey_force_const
    ubrmin    => enefunc%urey_rmin

    nangl     = molecule%num_angles
    nangl_p   = par%num_angles

    do i = 1, nangl

      ci1 = molecule%atom_cls_name(molecule%angl_list(1,i))
      ci2 = molecule%atom_cls_name(molecule%angl_list(2,i))
      ci3 = molecule%atom_cls_name(molecule%angl_list(3,i))

      list(1) = molecule%angl_list(1,i)
      list(2) = molecule%angl_list(2,i)
      list(3) = molecule%angl_list(3,i)

      icel1 = id_g2l(1,list(1))
      icel2 = id_g2l(1,list(3))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          do j = 1, nangl_p
            if ((ci1 == par%angl_atom_cls(1,j) .and. &
                 ci2 == par%angl_atom_cls(2,j) .and. &
                 ci3 == par%angl_atom_cls(3,j)) .or. &
                (ci1 == par%angl_atom_cls(3,j) .and. &
                 ci2 == par%angl_atom_cls(2,j) .and. &
                 ci3 == par%angl_atom_cls(1,j))) then

              angle(icel_local) = angle(icel_local) + 1
              alist(1:3,angle(icel_local),icel_local) = list(1:3)

              force(angle(icel_local),icel_local)   = &
                   par%angl_force_const(j)
              theta(angle(icel_local),icel_local)   = &
                   par%angl_theta_min(j)*RAD
              ubforce(angle(icel_local),icel_local) = &
                   par%angl_ub_force_const(j)
              ubrmin(angle(icel_local),icel_local)  = &
                   par%angl_ub_rmin(j)
              exit

            end if
          end do

          if (j == nangl_p + 1) &
            write(MsgOut,*) &
              'Setup_Enefunc_Angl> not found ANGL: [',&
              ci1, ']-[', ci2, ']-[', ci3, '] in parameter file. (ERROR)'

        end if

      end if

    end do

    found = 0
    do i = 1, ncel
      found = found + angle(i)
      if (angle(i) > MaxAngle) &
        call error_msg('Setup_Enefunc_Angl> Too many angles.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_angl_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_angl_all = found
#endif

    if (enefunc%num_angl_all /= nangl) &
      call error_msg('Setup_Enefunc_Angl> Some angle paremeters are missing.')

    return

  end subroutine setup_enefunc_angl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl_constraint
  !> @brief        define ANGLE term for each cell in potential energy function
  !                with SETTLE constraint
  !! @authors      JJ
  !! @param[in]    par      : CHARMM PAR information
  !! @param[in]    molecule : molecule information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl_constraint(par, molecule, domain, enefunc)

    ! formal arguments
    type(s_par),     target, intent(in)    :: par
    type(s_molecule),        intent(in)    :: molecule
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j, icel_local
    integer                  :: icel1, icel2
    integer                  :: nangl, nangl_p, found
    integer                  :: list(3)
    character(6)             :: ci1, ci2, ci3
    character(4)             :: ri1, ri2, ri3

    real(wp),        pointer :: force(:,:), theta(:,:)
    real(wp),        pointer :: ubforce(:,:), ubrmin(:,:)
    integer,         pointer :: angle(:), alist(:,:,:), num_water
    integer,         pointer :: ncel, cell_pair(:,:), id_g2l(:,:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    angle     => enefunc%num_angle
    alist     => enefunc%angle_list
    force     => enefunc%angle_force_const
    theta     => enefunc%angle_theta_min
    ubforce   => enefunc%urey_force_const
    ubrmin    => enefunc%urey_rmin
    num_water => enefunc%table%num_water

    nangl     = molecule%num_angles
    nangl_p   = par%num_angles

    do i = 1, nangl

      ci1 = molecule%atom_cls_name(molecule%angl_list(1,i))
      ci2 = molecule%atom_cls_name(molecule%angl_list(2,i))
      ci3 = molecule%atom_cls_name(molecule%angl_list(3,i))
      ri1 = molecule%residue_name (molecule%angl_list(1,i))
      ri2 = molecule%residue_name (molecule%angl_list(2,i))
      ri3 = molecule%residue_name (molecule%angl_list(3,i))

      if (ri1(1:4) /= 'TIP3' .and. ri2(1:4) /= 'TIP3' .and. &
          ri3(1:4) /= 'TIP3') then

        list(1) = molecule%angl_list(1,i)
        list(2) = molecule%angl_list(2,i)
        list(3) = molecule%angl_list(3,i)

        icel1 = id_g2l(1,list(1))
        icel2 = id_g2l(1,list(3))

        if (icel1 /= 0 .and. icel2 /= 0) then

          icel_local = cell_pair(icel1,icel2)
  
          if (icel_local >= 1 .and. icel_local <= ncel) then

            do j = 1, nangl_p
              if ((ci1 == par%angl_atom_cls(1,j) .and. &
                   ci2 == par%angl_atom_cls(2,j) .and. &
                   ci3 == par%angl_atom_cls(3,j)) .or. &
                  (ci1 == par%angl_atom_cls(3,j) .and. &
                   ci2 == par%angl_atom_cls(2,j) .and. &
                   ci3 == par%angl_atom_cls(1,j))) then

                angle(icel_local) = angle(icel_local) + 1
                alist(1:3,angle(icel_local),icel_local) = list(1:3)

                force(angle(icel_local),icel_local)   = &
                     par%angl_force_const(j)
                theta(angle(icel_local),icel_local)   = &
                     par%angl_theta_min(j)*RAD
                ubforce(angle(icel_local),icel_local) = &
                     par%angl_ub_force_const(j)
                ubrmin(angle(icel_local),icel_local)  = &
                     par%angl_ub_rmin(j)
                exit

              end if
            end do

            if (j == nangl_p + 1) &
              write(MsgOut,*) &
                'Setup_Enefunc_Angl_Constraint> not found ANGL: [', &
                ci1, ']-[', ci2, ']-[', ci3, '] in parameter file. (ERROR)'

          end if

        end if

      end if

    end do

    found = 0
    do i = 1, ncel
      found = found + angle(i)
      if (angle(i) > MaxAngle) &
        call error_msg('Setup_Enefunc_Angl_Constraint> Too many angles.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_angl_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_angl_all = found
#endif

    if (enefunc%num_angl_all /= (nangl - num_water)) &
      call error_msg( &
        'Setup_Enefunc_Angl_Constraint> Some angle paremeters are missing.')

    return

  end subroutine setup_enefunc_angl_constraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dihe
  !> @brief        define DIHEDRAL term in potential energy function
  !! @authors      YS, JJ, TM
  !! @param[in]    par      : CHARMM PAR information
  !! @param[in]    molecule : molecule information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_dihe(par, molecule, domain, enefunc)

    ! formal arguments
    type(s_par),      target, intent(in)    :: par
    type(s_molecule), target, intent(in)    :: molecule
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: ndihe, ndihe_p
    integer                   :: i, j, icel_local
    integer                   :: icel1, icel2
    integer                   :: found, nw_found
    integer                   :: list(4)
    character(6)              :: ci1, ci2, ci3, ci4

    real(wp),         pointer :: force(:,:), phase(:,:)
    integer,          pointer :: dihedral(:), dlist(:,:,:), period(:,:)
    integer,          pointer :: ncel, cell_pair(:,:), id_g2l(:,:)
    logical,      allocatable :: no_wild(:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    dihedral  => enefunc%num_dihedral
    dlist     => enefunc%dihe_list 
    force     => enefunc%dihe_force_const
    period    => enefunc%dihe_periodicity
    phase     => enefunc%dihe_phase
   
    ndihe     = molecule%num_dihedrals
    ndihe_p   = par%num_dihedrals

    ! check usage of wild card
    !
    allocate(no_wild(ndihe_p))

    do i = 1, ndihe_p

      if ((par%dihe_atom_cls(1,i) /= WildCard) .and. &
          (par%dihe_atom_cls(4,i) /= WildCard)) then
        ! A-B-C-D type
        no_wild(i) = .true.
      else
        ! X-B-C-D type
        no_wild(i) = .false.
      end if

    end do

    ! find number of interactions
    !
    do i = 1, ndihe

      ci1 = molecule%atom_cls_name(molecule%dihe_list(1,i))
      ci2 = molecule%atom_cls_name(molecule%dihe_list(2,i))
      ci3 = molecule%atom_cls_name(molecule%dihe_list(3,i))
      ci4 = molecule%atom_cls_name(molecule%dihe_list(4,i))

      list(1) = molecule%dihe_list(1,i)
      list(2) = molecule%dihe_list(2,i)
      list(3) = molecule%dihe_list(3,i)
      list(4) = molecule%dihe_list(4,i)

      icel1 = id_g2l(1,list(1))
      icel2 = id_g2l(1,list(4))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          nw_found = 0
          do j = 1, ndihe_p
            if (no_wild(j)) then
              if (((ci1 == par%dihe_atom_cls(1,j)) .and. &
                   (ci2 == par%dihe_atom_cls(2,j)) .and. &
                   (ci3 == par%dihe_atom_cls(3,j)) .and. &
                   (ci4 == par%dihe_atom_cls(4,j))) .or. &
                  ((ci1 == par%dihe_atom_cls(4,j)) .and. &
                   (ci2 == par%dihe_atom_cls(3,j)) .and. &
                   (ci3 == par%dihe_atom_cls(2,j)) .and. &
                   (ci4 == par%dihe_atom_cls(1,j)))) then

                nw_found = nw_found + 1
                dihedral(icel_local) = dihedral(icel_local) + 1 
                dlist(1:4,dihedral(icel_local),icel_local) = list(1:4)

                force (dihedral(icel_local),icel_local) = &
                     par%dihe_force_const(j)
                period(dihedral(icel_local),icel_local) = &
                     par%dihe_periodicity(j)
                phase (dihedral(icel_local),icel_local) = &
                     par%dihe_phase(j) * RAD

              end if
            end if
          end do

          if (nw_found == 0) then
            do j = 1, ndihe_p
              if (.not.no_wild(j)) then
                if (((ci2 == par%dihe_atom_cls(2,j)) .and. &
                     (ci3 == par%dihe_atom_cls(3,j))) .or. &
                    ((ci2 == par%dihe_atom_cls(3,j)) .and. &
                     (ci3 == par%dihe_atom_cls(2,j)))) then

                  dihedral(icel_local) = dihedral(icel_local) + 1
                  dlist(1:4,dihedral(icel_local),icel_local) = list(1:4)

                  force (dihedral(icel_local),icel_local) = &
                       par%dihe_force_const(j)
                  period(dihedral(icel_local),icel_local) = &
                       par%dihe_periodicity(j)
                  phase (dihedral(icel_local),icel_local) = &
                       par%dihe_phase(j) * RAD

                end if
              end if
            end do
          end if

        end if

      end if

    end do

    deallocate(no_wild)

    found = 0
    do i = 1, ncel
      found = found + dihedral(i)
      if (dihedral(i) > MaxDihe) &
        call error_msg('Setup_Enefunc_Dihe> Too many dihedral angles.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_dihe_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_dihe_all = found
#endif

    if (enefunc%num_dihe_all < ndihe) &
      call error_msg( &
         'Setup_Enefunc_Dihe> Some dihedral paremeters are missing.')

    return

  end subroutine setup_enefunc_dihe

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_impr
  !> @brief        define IMPROPER term in potential energy function
  !! @authors      YS,JJ
  !! @param[in]    par      : CHARMM PAR information
  !! @param[in]    molecule : molecule information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_impr(par, molecule, domain, enefunc)

    ! formal variables
    type(s_par),      target, intent(in)    :: par
    type(s_molecule), target, intent(in)    :: molecule
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: nimpr, nimpr_p
    integer                   :: i, j, icel_local
    integer                   :: icel1, icel2
    integer                   :: found
    integer                   :: list(4)
    character(6)              :: ci1, ci2, ci3, ci4

    real(wp),         pointer :: force(:,:), phase(:,:)
    integer,          pointer :: improper(:), ilist(:,:,:)
    integer,          pointer :: ncel, cell_pair(:,:), id_g2l(:,:)
    integer,      allocatable :: wc_type(:)
    logical,      allocatable :: no_wild(:)


    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair
    id_g2l    => domain%id_g2l

    improper  => enefunc%num_improper
    ilist     => enefunc%impr_list
    force     => enefunc%impr_force_const
    phase     => enefunc%impr_phase

    nimpr     = molecule%num_impropers
    nimpr_p   = par%num_impropers

    ! check usage of wild card
    !
    allocate(wc_type(nimpr_p), no_wild(nimpr))

    do i = 1, nimpr_p

      if ((par%impr_atom_cls(1,i) /= WildCard) .and. &
          (par%impr_atom_cls(2,i) /= WildCard) .and. &
          (par%impr_atom_cls(3,i) /= WildCard) .and. &
          (par%impr_atom_cls(4,i) /= WildCard)) then

        ! A-B-C-D type
        wc_type(i) = 0

      else if ((par%impr_atom_cls(1,i) == WildCard) .and. &
               (par%impr_atom_cls(2,i) /= WildCard) .and. &
               (par%impr_atom_cls(3,i) /= WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then

        ! X-B-C-D type
        wc_type(i) = 1

      else if ((par%impr_atom_cls(1,i) == WildCard) .and. &
               (par%impr_atom_cls(2,i) == WildCard) .and. &
               (par%impr_atom_cls(3,i) /= WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then

        ! X-X-C-D type
        wc_type(i) = 2

      else if ((par%impr_atom_cls(1,i) /= WildCard) .and. &
               (par%impr_atom_cls(2,i) == WildCard) .and. &
               (par%impr_atom_cls(3,i) == WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then

        ! A-X-X-D type
        wc_type(i) = 3

      else
        call error_msg('Setup_Enefunc_Impr> Undefined Wild Card')

      end if

    end do

    ! setup parameters
    !
    do i = 1, nimpr

      no_wild(i) = .false.

      ci1 = molecule%atom_cls_name(molecule%impr_list(1,i))
      ci2 = molecule%atom_cls_name(molecule%impr_list(2,i))
      ci3 = molecule%atom_cls_name(molecule%impr_list(3,i))
      ci4 = molecule%atom_cls_name(molecule%impr_list(4,i))
      list(1:4) = molecule%impr_list(1:4,i)

      icel1 = id_g2l(1,list(1))
      icel2 = id_g2l(1,list(4))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          ! A-B-C-D type
          !
          do j = 1, nimpr_p
            if (wc_type(j) == 0) then
              if (((ci1 == par%impr_atom_cls(1,j)) .and. &
                   (ci2 == par%impr_atom_cls(2,j)) .and. &
                   (ci3 == par%impr_atom_cls(3,j)) .and. &
                   (ci4 == par%impr_atom_cls(4,j))) .or. &
                  ((ci1 == par%impr_atom_cls(4,j)) .and. &
                   (ci2 == par%impr_atom_cls(3,j)) .and. &
                   (ci3 == par%impr_atom_cls(2,j)) .and. &
                   (ci4 == par%impr_atom_cls(1,j)))) then

                improper(icel_local) = improper(icel_local) + 1
                ilist(1:4,improper(icel_local),icel_local) = list(1:4)

                force(improper(icel_local),icel_local) = &
                     par%impr_force_const(j)
                phase(improper(icel_local),icel_local) = &
                     par%impr_phase(j) * RAD
                no_wild(i) = .true.
                exit

              end if
            end if
          end do

          ! X-B-C-D type
          !
          if (.not.no_wild(i)) then
            do j = 1, nimpr_p
              if (wc_type(j) == 1) then
                if (((ci2 == par%impr_atom_cls(2,j)) .and. &
                     (ci3 == par%impr_atom_cls(3,j)) .and. &
                     (ci4 == par%impr_atom_cls(4,j))) .or. &
                    ((ci2 == par%impr_atom_cls(4,j)) .and. &
                     (ci3 == par%impr_atom_cls(3,j)) .and. &
                     (ci4 == par%impr_atom_cls(2,j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local),icel_local) = list(1:4)

                  force(improper(icel_local),icel_local) = &
                       par%impr_force_const(j)
                  phase(improper(icel_local),icel_local) = &
                       par%impr_phase(j) * RAD
                  no_wild(i) = .true.
                  exit

                end if
              end if
            end do
          end if

          ! X-X-C-D type
          !
          if (.not.no_wild(i)) then
            do j = 1, nimpr_p
              if (wc_type(j) == 2) then
                if (((ci3 == par%impr_atom_cls(3,j)) .and. &
                     (ci4 == par%impr_atom_cls(4,j))) .or. &
                    ((ci3 == par%impr_atom_cls(4,j)) .and. &
                     (ci4 == par%impr_atom_cls(3,j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local),icel_local) = list(1:4)

                  force(improper(icel_local),icel_local) = &
                       par%impr_force_const(j)
                  phase(improper(icel_local),icel_local) = &
                       par%impr_phase(j) * RAD
                  no_wild(i) = .true.
                  exit

                end if
              end if
            end do
          end if
 
          ! A-X-X-D type
          !
          if (.not.no_wild(i)) then
            do j = 1, nimpr_p
              if (wc_type(j) == 3) then
                if (((ci1 == par%impr_atom_cls(1,j)) .and. &
                     (ci4 == par%impr_atom_cls(4,j))) .or. &
                    ((ci1 == par%impr_atom_cls(4,j)) .and. &
                     (ci4 == par%impr_atom_cls(1,j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local),icel_local) = list(1:4)

                  force(improper(icel_local),icel_local) = &
                       par%impr_force_const(j)
                  phase(improper(icel_local),icel_local) = &
                       par%impr_phase(j) * RAD
                  no_wild(i) = .true.
                  exit

                end if
              end if
            end do
          end if
  
          if (.not.no_wild(i)) &
            write(MsgOut,*) &
              'Setup_Enefunc_Impr> Unknown IMPR type. [', &
              ci1, ']-[', ci2, ']-[', ci3, ']-[', ci4, '] (ERROR)'

        end if
      end if

    end do

    deallocate(wc_type, no_wild)

    found = 0
    do i = 1, ncel
      found = found + improper(i)
      if (improper(i) > MaxImpr) &
        call error_msg('Setup_Enefunc_Impr> Too many improper dihedral angles')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_impr_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_impr_all = found
#endif

    if (enefunc%num_impr_all < nimpr) &
      call error_msg( &
        'Setup_Enefunc_Impr> Some improper paremeters are missing.')

    return

  end subroutine setup_enefunc_impr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_cmap
  !> @brief        define cmap term in potential energy function with DD
  !! @authors      TY, TM
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    domain      : domain information
  !! @param[inout] enefunc     : energy potential functions informationn
  !!
  !! @note       In str "par", following variables have been defined:
  !!   cmap_atom_cls(8,imap) (char) : 8 atom classes (4 for phi and 4 for psi)
  !!   cmap_resolution(imap) (int ) : = 24 (for all imap) = 360degree/15degree
  !!   cmap_data(i,j,imap)   (real) : Ecmap for grid points. i and j are the
  !!                                  1st (psi?) and the 2nd (phi?) grid IDs.
  !!                                  1<=i,j<=24.
  !!   Where imap is ID of cmap type (1 <= imap <= 6).
  !!
  !! @note       TY determined to use natural spline (periodic = .false.)
  !!             because it is similar to the CHARMM way.
  !!             However, I notice that force will be more accurately continuous
  !!             at phi (or psi) = +-180 degree if periodic spline was used.
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_cmap(ene_info, par, molecule, domain, enefunc)

    ! formal variables
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_par),             intent(in)    :: par
    type(s_molecule),        intent(in)    :: molecule
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j, k, l, ityp
    integer                  :: ncmap_p, found, ngrid0
    integer                  :: list(2), icel1, icel2, icel_local
    integer                  :: flag_cmap_type, alloc_stat, dealloc_stat
    character(6)             :: ci1, ci2, ci3, ci4, ci5, ci6, ci7, ci8
    logical                  :: periodic

    integer,         pointer :: ncel, cell_pair(:,:), id_g2l(:,:)

    real(wp),    allocatable :: c_ij(:,:,:,:) ! cmap coeffs


    ncel            => domain%num_cell_local
    cell_pair       => domain%cell_pair
    id_g2l          => domain%id_g2l


    ! If 'periodic' is .true.,
    !   then cubic spline with periodic (in dihedral-angle space) boundary
    !   will be applied.
    ! If 'periodic' is .false.,
    !   then natural cubic spline without periodic boudnary
    !   will be applied to expanded cross-term maps.
    !   This is similar with CHARMM's source code.
    !
    periodic = ene_info%cmap_pspline
    ncmap_p  = par%num_cmaps

    ! begin
    !
    ngrid0 = 0
    do i = 1, ncmap_p
      ngrid0 = max(ngrid0, par%cmap_resolution(i))
    end do

    call alloc_enefunc(enefunc, EneFuncCmap, ncel, ngrid0, ncmap_p)

    alloc_stat = 0
    allocate(c_ij(4,4,ngrid0,ngrid0), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    do i = 1, ncmap_p
      enefunc%cmap_resolution(i) = par%cmap_resolution(i)
    end do

    ! derive cmap coefficients by bicubic interpolation
    !
    do ityp = 1, ncmap_p

      if (periodic) then
        call derive_cmap_coefficients_p(ityp, par, c_ij)
      else
        call derive_cmap_coefficients_np(ityp, par, c_ij)
      end if

      do l = 1, ngrid0
        do k = 1, ngrid0 
          do j = 1, 4
            do i = 1, 4
              enefunc%cmap_coef(i,j,k,l,ityp) = c_ij(i,j,k,l)
            end do
          end do
        end do
      end do
    end do

    enefunc%num_cmap(1:ncel) = 0
    do i = 1, molecule%num_cmaps

      list(1) = molecule%cmap_list(1,i)
      list(2) = molecule%cmap_list(8,i)

      icel1 = id_g2l(1,list(1))
      icel2 = id_g2l(1,list(2))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          ! ci* will be atom-type strings
          !
          ci1 = molecule%atom_cls_name(molecule%cmap_list(1,i))
          ci2 = molecule%atom_cls_name(molecule%cmap_list(2,i))
          ci3 = molecule%atom_cls_name(molecule%cmap_list(3,i))
          ci4 = molecule%atom_cls_name(molecule%cmap_list(4,i))
          ci5 = molecule%atom_cls_name(molecule%cmap_list(5,i))
          ci6 = molecule%atom_cls_name(molecule%cmap_list(6,i))
          ci7 = molecule%atom_cls_name(molecule%cmap_list(7,i))
          ci8 = molecule%atom_cls_name(molecule%cmap_list(8,i))
          flag_cmap_type = -1

          ! assign cmap type ID to each (psi,phi) pair
          !
          do j = 1, ncmap_p
            if (ci1 == par%cmap_atom_cls(1, j) .and. &
                ci2 == par%cmap_atom_cls(2, j) .and.   &
                ci3 == par%cmap_atom_cls(3, j) .and.   &
                ci4 == par%cmap_atom_cls(4, j) .and.   &
                ci5 == par%cmap_atom_cls(5, j) .and.   &
                ci6 == par%cmap_atom_cls(6, j) .and.   &
                ci7 == par%cmap_atom_cls(7, j) .and.   &
                ci8 == par%cmap_atom_cls(8, j)   ) then

              enefunc%num_cmap(icel_local) = enefunc%num_cmap(icel_local) + 1
              enefunc%cmap_list(1:8,enefunc%num_cmap(icel_local),icel_local) &
                   = molecule%cmap_list(1:8,i)
              enefunc%cmap_type(enefunc%num_cmap(icel_local),icel_local) = j
              flag_cmap_type = j
              exit

            end if
          end do

          ! if not found, print detail about the error.
          !

          if (flag_cmap_type <= 0) then
            write(MsgOut,*) 'Setup_Enefunc_Cmap> not found CMAP: '
            write(MsgOut,*) ' [',ci1,']-[',ci2,']-[',ci3,']-[',ci4,']-'
            write(MsgOut,*) ' [',ci5,']-[',ci6,']-[',ci7,']-[',ci8,'] '
            write(MsgOut,*) '  in parameter file. (ERROR)'
          end if

        end if
      end if

    end do

    deallocate(c_ij, stat=dealloc_stat)
    if (dealloc_stat /= 0) &
      call error_msg_dealloc

    ! write summary
    !
    if (main_rank) then
      if (periodic) then
        write(MsgOut,'(A)') &
    'Setup_Enefunc_Cmap> Periodic-boundary spline is used to derive cmap coefs.'
        write(MsgOut,'(A)') ''
      else
        write(MsgOut,'(A)') &
            'Setup_Enefunc_Cmap> Natural spline is used to derive cmap coefs.'
        write(MsgOut,'(A)') ''
      end if
    end if

    ! stop if parameter is missing
    !
    found = 0 
    do i = 1, ncel
      found = found + enefunc%num_cmap(i)
      if (enefunc%num_cmap(i) > MaxCmap) &
        call error_msg('Setup_Enefunc_Cmap> Too many cmaps.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_cmap_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_cmap_all = found
#endif

    if (enefunc%num_cmap_all /=  molecule%num_cmaps) &
      call error_msg('Setup_Enefunc_Cmap> Some cmap parameters are missing.')

    return

  end subroutine setup_enefunc_cmap

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_nonb
  !> @brief        define NON-BOND term in potential energy function
  !! @authors      YS, JJ, TI
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_nonb(ene_info, par, constraints, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_par),             intent(in)    :: par
    type(s_constraints),     intent(in)    :: constraints
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: eps14, rmin14, eps, rmin
    integer                  :: i, j
    integer                  :: nonb_p, ncel
    logical                  :: first


    first  = .true.
    nonb_p = par%num_atom_cls
    ncel   = domain%num_cell_local 

    ! set cutoff distance etc.
    !
    enefunc%switchdist   = ene_info%switchdist
    enefunc%cutoffdist   = ene_info%cutoffdist
    enefunc%pairlistdist = ene_info%pairlistdist
    enefunc%dielec_const = ene_info%dielec_const
    enefunc%force_switch = ene_info%vdw_force_switch
    enefunc%num_atom_cls = nonb_p

    if (ene_info%electrostatic == ElectrostaticPME) then
      enefunc%pme_use      = .true.
      enefunc%pme_alpha    = ene_info%pme_alpha
      enefunc%pme_ngrid_x  = ene_info%pme_ngrid_x
      enefunc%pme_ngrid_y  = ene_info%pme_ngrid_y
      enefunc%pme_ngrid_z  = ene_info%pme_ngrid_z
      enefunc%pme_nspline  = ene_info%pme_nspline

    else
      enefunc%pme_use      = .false.

    end if 

    ! set lennard-jones parameters
    !
    call alloc_enefunc(enefunc, EneFuncNbon, nonb_p)

    do i = 1, nonb_p

      lj_coef(1,i) = par%nonb_eps(i)
      lj_coef(2,i) = par%nonb_rmin(i)

      do j = 1, nonb_p

        ! combination rule
        !
        eps14  = sqrt(par%nonb_eps_14(i) * par%nonb_eps_14(j))
        rmin14 = par%nonb_rmin_14(i) + par%nonb_rmin_14(j)
        eps    = sqrt(par%nonb_eps(i) * par%nonb_eps(j))
        rmin   = par%nonb_rmin(i) + par%nonb_rmin(j)

        ! set parameters
        !
        enefunc%nb14_lj12(i,j) = eps14 * (rmin14 ** 12)
        enefunc%nb14_lj6(i,j)  = 2.0_wp * eps14 * (rmin14 ** 6)
        enefunc%nonb_lj12(i,j) = eps * (rmin ** 12)
        enefunc%nonb_lj6(i,j)  = 2.0_wp * eps * (rmin ** 6)

      end do
    end do

    call alloc_enefunc(enefunc, EneFuncNonb,     ncel, maxcell1)
    call alloc_enefunc(enefunc, EneFuncNonbList, ncel, maxcell1)

    if (enefunc%table%water_table) then

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(first, .true., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl_solute(first, domain, enefunc)

      end if

    else

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(first, .false., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl(first, domain, enefunc)

      end if

    end if

    return

  end subroutine setup_enefunc_nonb

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_localres_bond
  !> @brief        define local restraint for each cell in 
  !                potential energy function
  !! @authors      CK
  !! @param[in]    localres : local restraint information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_localres_bond(localres, domain, enefunc)

    ! formal arguments
    type(s_localres), target, intent(in)    :: localres
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variable
    integer                   :: i, ig, k, ix, iatm, icel_local, i1, i2
    integer                   :: irest, icel1, icel2, found, oldbonds

    real(wp),         pointer :: const(:),ref(:)
    real(wp),         pointer :: force(:,:), dist(:,:)
    integer,          pointer :: id_g2l(:,:), ncell
    integer,          pointer :: num_funcs
    integer,          pointer :: index_atoms(:,:)
    integer,          pointer :: func(:)
    integer,          pointer :: bond(:), list(:,:,:), cell_pair(:,:)
    integer,          pointer :: bkind(:,:)


    num_funcs   => localres%num_funcs
    func        => localres%func
    index_atoms => localres%index_atoms
    const       => localres%const
    ref         => localres%ref

    ncell       => domain%num_cell_local
    cell_pair   => domain%cell_pair
    id_g2l      => domain%id_g2l

    bond        => enefunc%num_bond
    list        => enefunc%bond_list
    force       => enefunc%bond_force_const
    dist        => enefunc%bond_dist_min
    bkind       => enefunc%bond_kind


    if (num_funcs > 0) then
      enefunc%local_restraint = .true.
    else
      return
    end if

    ! position restraint
    !
    irest = 0
    do i = 1, num_funcs

      if (localres%func(i) == LocalResBonds) then

        i1 = index_atoms(1,i)
        i2 = index_atoms(2,i)

        icel1 = domain%id_g2l(1,i1)
        icel2 = domain%id_g2l(1,i2)
        irest = irest+1

        if (icel1 /= 0 .and. icel2 /= 0) then
          icel_local = cell_pair(icel1,icel2)

          if (icel_local > 0 .and. icel_local <= ncell) then
            bond (icel_local) = bond(icel_local) + 1
            list (1:2,bond(icel_local),icel_local) = index_atoms(1:2,i)
            force(bond(icel_local),icel_local)     = 0.5_wp * const(i)
            dist (bond(icel_local),icel_local)     = ref(i)
            bkind(bond(icel_local),icel_local)     = 1
          end if
        end if

      end if

    end do

    found = 0
    oldbonds = enefunc%num_bond_all

    do i = 1,ncell
      found = found + bond(i)
      if (bond(i) > MaxBond) &
        call error_msg('Setup_Enefunc_Localres_Bond> Too many bonds.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif

    if (enefunc%num_bond_all-oldbonds /= irest) &
      call error_msg( &
           'Setup_Enefunc_Localres_Bond> Some Bond Paremeters are missing.')

    return

  end subroutine setup_enefunc_localres_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_localres_angle
  !> @brief        define local restraint for each cell in 
  !                potential energy function
  !! @authors      CK
  !! @param[in]    localres : local restraint information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_localres_angle(localres, domain, enefunc)

    ! formal arguments
    type(s_localres), target, intent(in)    :: localres
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variable
    integer                   :: i, ig, k, ix, iatm, icel_local
    integer                   :: i1, i2, i3
    integer                   :: irest, icel1, icel2, found, oldangle

    real(wp),         pointer :: const(:),ref(:)
    real(wp),         pointer :: ubforce(:,:), ubrmin(:,:)
    real(wp),         pointer :: force(:,:), theta(:,:)
    integer,          pointer :: id_g2l(:,:), ncell
    integer,          pointer :: num_funcs
    integer,          pointer :: index_atoms(:,:)
    integer,          pointer :: func(:)
    integer,          pointer :: angle(:), list(:,:,:), cell_pair(:,:)
    integer,          pointer :: akind(:,:)


    num_funcs   => localres%num_funcs
    func        => localres%func
    index_atoms => localres%index_atoms
    const       => localres%const
    ref         => localres%ref

    ncell       => domain%num_cell_local
    cell_pair   => domain%cell_pair
    id_g2l      => domain%id_g2l

    angle       => enefunc%num_angle
    list        => enefunc%angle_list
    force       => enefunc%angle_force_const
    theta       => enefunc%angle_theta_min
    ubforce     => enefunc%urey_force_const
    ubrmin      => enefunc%urey_rmin
    akind       => enefunc%angle_kind

    if (num_funcs > 0) then
      enefunc%local_restraint = .true.
    else
      return
    end if

    ! position restraint
    !
    irest = 0
    do i = 1, num_funcs

      if (localres%func(i) == LocalResAngles) then

        i1 = index_atoms(1,i)
        i2 = index_atoms(2,i)
        i3 = index_atoms(3,i)

        icel1 = domain%id_g2l(1,i1)
        icel2 = domain%id_g2l(1,i3)
        irest = irest+1

        if (icel1 /= 0 .and. icel2 /= 0) then

          icel_local = cell_pair(icel1,icel2)

          if (icel_local > 0 .and. icel_local <= ncell) then
            angle  (icel_local) = angle(icel_local) + 1
            list   (1:3,angle(icel_local),icel_local) = index_atoms(1:3,i)
            force  (angle(icel_local),icel_local)     = 0.5_wp * const(i)
            theta  (angle(icel_local),icel_local)     = ref(i)*RAD
            ubforce(angle(icel_local),icel_local)     = 0.0_wp
            ubrmin (angle(icel_local),icel_local)     = 0.0_wp
            akind  (angle(icel_local),icel_local)     = 1
          end if
        end if

      end if

    end do

    found = 0
    oldangle = enefunc%num_angl_all

    do i = 1,ncell
      found = found + angle(i)
      if (angle(i) > MaxAngle) &
        call error_msg('Setup_Enefunc_Localres_Angle> Too many angles.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_angl_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_angl_all = found
#endif

    if (enefunc%num_angl_all-oldangle /= irest) &
      call error_msg( &
           'Setup_Enefunc_Localres_Angle> Some Angle Paremeters are missing.')

    return

  end subroutine setup_enefunc_localres_angle

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_localres_dihed
  !> @brief        define local restraint for each cell in 
  !                potential energy function
  !! @authors      CK
  !! @param[in]    localres : local restraint information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_localres_dihed(localres, domain, enefunc)

    ! formal arguments
    type(s_localres), target, intent(in)    :: localres
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variable
    integer                   :: i, ig, k, ix, iatm, icel_local 
    integer                   :: i1, i2, i3, i4
    integer                   :: irest, icel1, icel2, found, olddihed

    real(wp),         pointer :: const(:),ref(:)
    real(wp),         pointer :: force(:,:), theta(:,:)
    integer,          pointer :: id_g2l(:,:), ncell
    integer,          pointer :: num_funcs
    integer,          pointer :: index_atoms(:,:)
    integer,          pointer :: func(:)
    integer,          pointer :: dihed(:), list(:,:,:), cell_pair(:,:)
    integer,          pointer :: dkind(:,:)


    num_funcs   => localres%num_funcs
    func        => localres%func
    index_atoms => localres%index_atoms
    const       => localres%const
    ref         => localres%ref

    ncell       => domain%num_cell_local
    cell_pair   => domain%cell_pair
    id_g2l      => domain%id_g2l

    dihed       => enefunc%num_dihedral
    list        => enefunc%dihe_list
    force       => enefunc%dihe_force_const
    theta       => enefunc%dihe_phase
    dkind       => enefunc%dihe_kind

    if (num_funcs > 0) then
      enefunc%local_restraint = .true.
    else
      return
    end if

    ! position restraint
    !
    irest = 0
    do i = 1, num_funcs

      if (localres%func(i) == LocalResDihedrals) then

        i1 = index_atoms(1,i)
        i2 = index_atoms(2,i)
        i3 = index_atoms(3,i)
        i4 = index_atoms(4,i)

        icel1 = domain%id_g2l(1,i1)
        icel2 = domain%id_g2l(1,i4)
        irest = irest+1

        if (icel1 /= 0 .and. icel2 /= 0) then
          icel_local = cell_pair(icel1,icel2)

          if (icel_local > 0 .and. icel_local <= ncell) then
            dihed(icel_local) = dihed(icel_local) + 1
            list (1:4,dihed(icel_local),icel_local) = index_atoms(1:4,i)
            force(dihed(icel_local),icel_local)     = const(i)
            theta(dihed(icel_local),icel_local)     = ref(i)*RAD
            dkind(dihed(icel_local),icel_local)     = 1
          end if
        end if

      end if

    end do

    found = 0
    olddihed = enefunc%num_dihe_all

    do i = 1,ncell
      found = found + dihed(i)
      if (dihed(i) > maxdihe) &
        call error_msg( &
             'Setup_Enefunc_Localres_Dihed> Too many dihedral angles.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_dihe_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_dihe_all = found
#endif

    if (enefunc%num_dihe_all-olddihed /= irest) &
      call error_msg( &
           'Setup_Enefunc_Localres_Dihed> Some Dihed Paremeters are missing.')

    return

  end subroutine setup_enefunc_localres_dihed

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_pio
  !> @brief        define BOND term for each cell in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, found
    integer,         pointer :: bond(:)


    bond      => enefunc%num_bond

    found = 0 
    do i = 1, domain%num_cell_local
      found = found + bond(i)
      if (bond(i) > MaxBond) &
        call error_msg('Setup_Enefunc_Bond_Pio> Too many bonds.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif

    return
    
  end subroutine setup_enefunc_bond_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl_pio
  !> @brief        define ANGLE term for each cell in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, found
    integer,         pointer :: angle(:)


    angle     => enefunc%num_angle

    found = 0
    do i = 1, domain%num_cell_local
      found = found + angle(i)
      if (angle(i) > MaxAngle) &
        call error_msg('Setup_Enefunc_Angl_Pio> Too many angles.') 
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_angl_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_angl_all = found
#endif

    return

  end subroutine setup_enefunc_angl_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dihe_pio
  !> @brief        define DIHEDRAL term in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_dihe_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: i, found
    integer,          pointer :: dihedral(:)


    dihedral  => enefunc%num_dihedral
   
    found = 0
    do i = 1, domain%num_cell_local
      found = found + dihedral(i)
      if (dihedral(i) > MaxDihe) &
        call error_msg('Setup_Enefunc_Dihe_Pio> Too many dihedral angles.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_dihe_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_dihe_all = found
#endif

    return

  end subroutine setup_enefunc_dihe_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_impr_pio
  !> @brief        define IMPROPER term in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_impr_pio(domain, enefunc)

    ! formal variables
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: i, found
    integer,          pointer :: improper(:)


    improper  => enefunc%num_improper

    found = 0
    do i = 1, domain%num_cell_local
      found = found + improper(i)
      if (improper(i) > MaxImpr) &
        call error_msg( &
            'Setup_Enefunc_Impr_Pio> Too many improper dihedral angles')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_impr_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_impr_all = found
#endif

    return

  end subroutine setup_enefunc_impr_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_nonb_pio
  !> @brief        define NON-BOND term in potential energy function
  !! @authors      NT
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_nonb_pio(ene_info, constraints, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_constraints),     intent(in)    :: constraints
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: ncel
    logical                  :: first


    first  = .true.
    ncel   = domain%num_cell_local 

    ! set cutoff distance etc.
    !
    enefunc%switchdist   = ene_info%switchdist
    enefunc%cutoffdist   = ene_info%cutoffdist
    enefunc%pairlistdist = ene_info%pairlistdist
    enefunc%dielec_const = ene_info%dielec_const
    enefunc%force_switch = ene_info%vdw_force_switch

    if (ene_info%electrostatic == ElectrostaticPME) then
      enefunc%pme_use      = .true.
      enefunc%pme_alpha    = ene_info%pme_alpha
      enefunc%pme_ngrid_x  = ene_info%pme_ngrid_x
      enefunc%pme_ngrid_y  = ene_info%pme_ngrid_y
      enefunc%pme_ngrid_z  = ene_info%pme_ngrid_z
      enefunc%pme_nspline  = ene_info%pme_nspline

    else
      enefunc%pme_use      = .false.

    end if 

    call alloc_enefunc(enefunc, EneFuncNonb,     ncel, maxcell)
    call alloc_enefunc(enefunc, EneFuncNonbList, ncel, maxcell)

    if (enefunc%table%water_table) then

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(first, .true., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl_solute(first, domain, enefunc)

      end if

    else

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(first, .false., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl(first, domain, enefunc)

      end if

    end if

    return

  end subroutine setup_enefunc_nonb_pio


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_cmap_pio
  !> @brief        define cmap term in potential energy function with DD
  !! @authors      NT
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : energy potential functions informationn
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_cmap_pio(ene_info, domain, enefunc)

    ! formal variables
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, found


    found = 0 
    do i = 1, domain%num_cell_local
      found = found + enefunc%num_cmap(i)
      if (enefunc%num_cmap(i) > MaxCmap) &
        call error_msg('Setup_Enefunc_Cmap_Pio> Too many cmaps.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_cmap_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_cmap_all = found
#endif

    return

  end subroutine setup_enefunc_cmap_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_bond
  !> @brief        update BOND term for each cell in potential energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_bond(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, ix, icel_local
    integer                  :: i1, i2, icel1, icel2

    real(wp),        pointer :: fc(:,:), r0(:,:)
    real(wp),        pointer :: buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_pair(:,:), id_g2l(:,:)
    integer,         pointer :: nbond(:), bondlist(:,:,:), bondkind(:,:)
    integer,         pointer :: bond_exit(:), index(:,:)
    integer,         pointer :: bond_add(:), buf_int(:,:,:)


    ncel_local  => domain%num_cell_local
    nboundary   => domain%num_cell_boundary
    cell_pair   => domain%cell_pair
    id_g2l      => domain%id_g2l

    nbond       => enefunc%num_bond
    bondlist    => enefunc%bond_list
    bondkind    => enefunc%bond_kind
    fc          => enefunc%bond_force_const
    r0          => enefunc%bond_dist_min
    bond_exit   => enefunc%bond_exit
    index       => enefunc%bond_exit_index
    bond_add    => enefunc%bond_add
    buf_int     => enefunc%buf_bond_integer
    buf_real    => enefunc%buf_bond_real

    bond_exit(1:ncel_local) = 0
    bond_add(1:ncel_local+nboundary) = 0

    do i = 1, ncel_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bondlist(1,ix,i))
        i1    = id_g2l(2,bondlist(1,ix,i))
        icel2 = id_g2l(1,bondlist(2,ix,i))
        i2    = id_g2l(2,bondlist(2,ix,i))

        icel_local = cell_pair(icel1,icel2)

        if (icel_local /= i) then

          bond_exit(i) = bond_exit(i) + 1
          index(bond_exit(i),i) = ix

          bond_add(icel_local) = bond_add(icel_local) + 1
          buf_real(1,bond_add(icel_local),icel_local) = r0(ix,i)
          buf_real(2,bond_add(icel_local),icel_local) = fc(ix,i)
          buf_int (1,bond_add(icel_local),icel_local) = bondlist(1,ix,i)
          buf_int (2,bond_add(icel_local),icel_local) = bondlist(2,ix,i)
          buf_int (3,bond_add(icel_local),icel_local) = bondkind(ix,i)

        end if

      end do
    end do

    return
    
  end subroutine update_outgoing_enefunc_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_bond
  !> @brief        update BOND term for each cell in potential energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_bond(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx, found 
    logical                  :: insert

    real(wp),        pointer :: buf_real(:,:,:), bonddist(:,:), bondforce(:,:)
    integer,         pointer :: ncel_local, nboundary, cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: bond_add(:), bond_exit(:), bond_exit_index(:,:)
    integer,         pointer :: buf_int(:,:,:), nbond(:)
    integer,         pointer :: bondlist(:,:,:), bondkind(:,:)


    ncel_local      => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    cell_g2l        => domain%cell_g2l
    id_g2l          => domain%id_g2l
 
    bond_add        => enefunc%bond_add
    bond_exit       => enefunc%bond_exit
    bond_exit_index => enefunc%bond_exit_index
    buf_int         => enefunc%buf_bond_integer
    buf_real        => enefunc%buf_bond_real
    bondlist        => enefunc%bond_list
    bondforce       => enefunc%bond_force_const
    bonddist        => enefunc%bond_dist_min
    nbond           => enefunc%num_bond
    bondkind        => enefunc%bond_kind

    found = 0 
    do i = 1, ncel_local

#ifdef DEBUG
      if (bond_add(i)+nbond(i)-bond_exit(i) > MaxBond) then
        call error_msg( &
            'Debug: update_incoming_enefunc_bond> bond is exceed MaxBond')
      end if
#endif 

      if (bond_add(i) > BondMove) &
        call error_msg( &
        'Debug: update_incoming_enefunc_bond> bond migration is exceed maximum')

      if (bond_add(i) >= bond_exit(i)) then
          
        do k = 1, bond_exit(i)
          bondlist (1,bond_exit_index(k,i),i) = buf_int (1,k,i)
          bondlist (2,bond_exit_index(k,i),i) = buf_int (2,k,i)
          bondkind(bond_exit_index(k,i),i)    = buf_int (3,k,i)
          bonddist (bond_exit_index(k,i),i)   = buf_real(1,k,i)
          bondforce(bond_exit_index(k,i),i)   = buf_real(2,k,i)
        end do

        do k = bond_exit(i)+1, bond_add(i)
          ix = k + nbond(i) - bond_exit(i)
          bondlist (1,ix,i) = buf_int(1,k,i)
          bondlist (2,ix,i) = buf_int(2,k,i)
          bondkind (ix,i)   = buf_int(3,k,i)
          bonddist (ix,i)   = buf_real(1,k,i)
          bondforce(ix,i)   = buf_real(2,k,i)
        end do

        nbond(i) = nbond(i) + bond_add(i) - bond_exit(i)

      else

        do k = 1, bond_add(i)
          bondlist (1,bond_exit_index(k,i),i) = buf_int (1,k,i)
          bondlist (2,bond_exit_index(k,i),i) = buf_int (2,k,i)
          bondkind(bond_exit_index(k,i),i)    = buf_int (3,k,i)
          bonddist (bond_exit_index(k,i),i)   = buf_real(1,k,i)
          bondforce(bond_exit_index(k,i),i)   = buf_real(2,k,i)
        end do

        j  = 0
        ix = nbond(i)
        k  = bond_add(i) + 1

        do while (j < (bond_exit(i)-bond_add(i)))

          insert = .true.
          do kx = k, bond_exit(i)
            if (ix == bond_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = bond_exit_index(k,i)
            bondlist (1,kx,i) = bondlist(1,ix,i)
            bondlist (2,kx,i) = bondlist(2,ix,i)
            bondkind (kx,i)   = bondkind(ix,i)
            bonddist (kx,i)   = bonddist(ix,i)
            bondforce(kx,i)   = bondforce(ix,i)
            j = j + 1
            k = k + 1
          end if
          ix = ix - 1

        end do

        nbond(i) = nbond(i) + bond_add(i) - bond_exit(i)

      end if

    end do

    return
   
  end subroutine update_incoming_enefunc_bond
 
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_angl
  !> @brief        update ANGLE term for each cell in potential energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_angl(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: icel1, icel2, icel3, i1, i2, i3
    integer                  :: i, ix, icel_local

    real(wp),        pointer :: fc(:,:), theta0(:,:), fc_ub(:,:), r0_ub(:,:)
    real(wp),        pointer :: coord(:,:,:), buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_g2l(:), cell_g2b(:)
    integer,         pointer :: cell_pair(:,:), id_g2l(:,:)
    integer,         pointer :: nangle(:), anglelist(:,:,:), anglekind(:,:)
    integer,         pointer :: angle_add(:), angle_exit(:)
    integer,         pointer :: angle_exit_index(:,:), buf_int(:,:,:)


    ncel_local       => domain%num_cell_local
    nboundary        => domain%num_cell_boundary
    cell_g2l         => domain%cell_g2l
    cell_g2b         => domain%cell_g2b
    cell_pair        => domain%cell_pair
    id_g2l           => domain%id_g2l
    coord            => domain%coord

    nangle           => enefunc%num_angle
    anglelist        => enefunc%angle_list
    anglekind        => enefunc%angle_kind
    fc               => enefunc%angle_force_const
    theta0           => enefunc%angle_theta_min
    fc_ub            => enefunc%urey_force_const
    r0_ub            => enefunc%urey_rmin
    angle_add        => enefunc%angle_add
    angle_exit       => enefunc%angle_exit
    angle_exit_index => enefunc%angle_exit_index
    buf_int          => enefunc%buf_angle_integer
    buf_real         => enefunc%buf_angle_real

    angle_add (1:ncel_local+nboundary) = 0 
    angle_exit(1:ncel_local) = 0 

    do i = 1, ncel_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,anglelist(1,ix,i))
        i1    = id_g2l(2,anglelist(1,ix,i))
        icel2 = id_g2l(1,anglelist(2,ix,i))
        i2    = id_g2l(2,anglelist(2,ix,i))
        icel3 = id_g2l(1,anglelist(3,ix,i))
        i3    = id_g2l(2,anglelist(3,ix,i))

        icel_local = cell_pair(icel1,icel3)

        if (icel_local /= i) then

          angle_exit(i) = angle_exit(i) + 1
          angle_exit_index(angle_exit(i),i) = ix
          angle_add(icel_local) = angle_add(icel_local) + 1
          buf_real(1,angle_add(icel_local),icel_local) = theta0(ix,i)
          buf_real(2,angle_add(icel_local),icel_local) = fc(ix,i)
          buf_real(3,angle_add(icel_local),icel_local) = r0_ub(ix,i)
          buf_real(4,angle_add(icel_local),icel_local) = fc_ub(ix,i)
          buf_int (1,angle_add(icel_local),icel_local) = anglelist(1,ix,i)
          buf_int (2,angle_add(icel_local),icel_local) = anglelist(2,ix,i)
          buf_int (3,angle_add(icel_local),icel_local) = anglelist(3,ix,i)
          buf_int (4,angle_add(icel_local),icel_local) = anglekind(ix,i)

        end if

      end  do
    end do

    return

  end subroutine update_outgoing_enefunc_angl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_angl
  !> @brief        update ANGLE term for each cell in potential energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_angl(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx
    logical                  :: insert

    real(wp),        pointer :: buf_real(:,:,:)
    real(wp),        pointer :: angletheta(:,:), angleforce(:,:)
    real(wp),        pointer :: ubrmin(:,:), ubforce(:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: angle_add(:), angle_exit(:)  
    integer,         pointer :: angle_exit_index(:,:)
    integer,         pointer :: anglekind(:,:)
    integer,         pointer :: buf_int(:,:,:), nangle(:), anglelist(:,:,:)


    ncel_local       => domain%num_cell_local
    nboundary        => domain%num_cell_boundary
    cell_g2l         => domain%cell_g2l
    id_g2l           => domain%id_g2l

    nangle           => enefunc%num_angle
    angle_add        => enefunc%angle_add
    angle_exit       => enefunc%angle_exit
    angle_exit_index => enefunc%angle_exit_index
    buf_int          => enefunc%buf_angle_integer
    buf_real         => enefunc%buf_angle_real
    anglelist        => enefunc%angle_list
    anglekind        => enefunc%angle_kind
    angleforce       => enefunc%angle_force_const
    angletheta       => enefunc%angle_theta_min
    ubforce          => enefunc%urey_force_const
    ubrmin           => enefunc%urey_rmin

    do i = 1, ncel_local

#ifdef DEBUG
      if (angle_add(i)+nangle(i)-angle_exit(i) > MaxAngle) &
        call error_msg( &
            'Debug: update_incoming_enefunc_angle> angle is exceed MaxAngle')
#endif 

      if (angle_add(i) > AngleMove) &
        call error_msg( &
     'Debug: update_incoming_enefunc_angle> angle migration is exceed maximum')

      if (angle_add(i) >= angle_exit(i)) then
          
        do k = 1, angle_exit(i)
          anglelist (1,angle_exit_index(k,i),i) = buf_int (1,k,i)
          anglelist (2,angle_exit_index(k,i),i) = buf_int (2,k,i)
          anglelist (3,angle_exit_index(k,i),i) = buf_int (3,k,i)
          anglekind (angle_exit_index(k,i),i)   = buf_int (4,k,i)
          angletheta(angle_exit_index(k,i),i)   = buf_real(1,k,i)
          angleforce(angle_exit_index(k,i),i)   = buf_real(2,k,i)
          ubrmin    (angle_exit_index(k,i),i)   = buf_real(3,k,i)
          ubforce   (angle_exit_index(k,i),i)   = buf_real(4,k,i)
        end do

        do k = angle_exit(i)+1, angle_add(i)
          ix = k + nangle(i) - angle_exit(i)
          anglelist (1,ix,i) = buf_int(1,k,i)
          anglelist (2,ix,i) = buf_int(2,k,i)
          anglelist (3,ix,i) = buf_int(3,k,i)
          anglekind (ix,i)   = buf_int(4,k,i)
          angletheta(ix,i)   = buf_real(1,k,i)
          angleforce(ix,i)   = buf_real(2,k,i)
          ubrmin    (ix,i)   = buf_real(3,k,i)
          ubforce   (ix,i)   = buf_real(4,k,i)
        end do
        nangle(i) = nangle(i) + angle_add(i) - angle_exit(i)

      else
        do k = 1, angle_add(i)
          anglelist (1,angle_exit_index(k,i),i) = buf_int (1,k,i)
          anglelist (2,angle_exit_index(k,i),i) = buf_int (2,k,i)
          anglelist (3,angle_exit_index(k,i),i) = buf_int (3,k,i)
          anglekind (angle_exit_index(k,i),i)   = buf_int (4,k,i)
          angletheta(angle_exit_index(k,i),i)   = buf_real(1,k,i)
          angleforce(angle_exit_index(k,i),i)   = buf_real(2,k,i)
          ubrmin    (angle_exit_index(k,i),i)   = buf_real(3,k,i)
          ubforce   (angle_exit_index(k,i),i)   = buf_real(4,k,i)
        end do

        j  = 0
        ix = nangle(i)
        k  = angle_add(i) + 1

        do while (j < (angle_exit(i)-angle_add(i)))

          insert = .true.

          do kx = k, angle_exit(i)
            if (ix == angle_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = angle_exit_index(k,i)
            anglelist (1,kx,i) = anglelist(1,ix,i) 
            anglelist (2,kx,i) = anglelist(2,ix,i) 
            anglelist (3,kx,i) = anglelist(3,ix,i) 
            anglekind (kx,i)   = anglekind(ix,i) 
            angletheta(kx,i)   = angletheta(ix,i) 
            angleforce(kx,i)   = angleforce(ix,i) 
            ubrmin    (kx,i)   = ubrmin(ix,i) 
            ubforce   (kx,i)   = ubforce(ix,i) 
            j = j + 1
            k = k + 1

          end if
          ix = ix - 1

        end do

        nangle(i) = nangle(i) + angle_add(i) - angle_exit(i)

      end if

    end do

    return
   
  end subroutine update_incoming_enefunc_angl 

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_dihe
  !> @brief        update DIHEDRAL term for each cell in potential energy
  !!               function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_dihe(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: icel1, icel2, icel3, icel4, i1, i2, i3, i4
    integer                  :: i, ix, icel_local

    real(wp),        pointer :: fc(:,:), phase(:,:)
    real(wp),        pointer :: coord(:,:,:), buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_g2l(:), cell_g2b(:)
    integer,         pointer :: cell_pair(:,:), id_g2l(:,:)
    integer,         pointer :: dihed_add(:), dihed_exit(:)
    integer,         pointer :: dihed_exit_index(:,:)
    integer,         pointer :: ndihe(:), dihelist(:,:,:), dihekind(:,:)
    integer,         pointer :: nperiod(:,:), buf_int(:,:,:)


    ncel_local       => domain%num_cell_local
    nboundary        => domain%num_cell_boundary
    cell_g2l         => domain%cell_g2l
    cell_g2b         => domain%cell_g2b
    cell_pair        => domain%cell_pair
    id_g2l           => domain%id_g2l
    coord            => domain%coord

    ndihe            => enefunc%num_dihedral
    dihelist         => enefunc%dihe_list
    dihekind         => enefunc%dihe_kind
    dihed_add        => enefunc%dihed_add
    dihed_exit       => enefunc%dihed_exit
    dihed_exit_index => enefunc%dihed_exit_index
    fc               => enefunc%dihe_force_const
    nperiod          => enefunc%dihe_periodicity
    phase            => enefunc%dihe_phase
    buf_int          => enefunc%buf_dihed_integer
    buf_real         => enefunc%buf_dihed_real

    dihed_add (1:ncel_local+nboundary) = 0 
    dihed_exit(1:ncel_local) = 0 


    do i = 1, ncel_local
      do ix = 1, ndihe(i)

        icel1 = id_g2l(1,dihelist(1,ix,i))
        i1    = id_g2l(2,dihelist(1,ix,i))
        icel2 = id_g2l(1,dihelist(2,ix,i))
        i2    = id_g2l(2,dihelist(2,ix,i))
        icel3 = id_g2l(1,dihelist(3,ix,i))
        i3    = id_g2l(2,dihelist(3,ix,i))
        icel4 = id_g2l(1,dihelist(4,ix,i))
        i4    = id_g2l(2,dihelist(4,ix,i))

        icel_local = cell_pair(icel1,icel4)

        if (icel_local /= i) then

          dihed_exit(i) = dihed_exit(i) + 1
          dihed_exit_index(dihed_exit(i),i) = ix
          dihed_add(icel_local) = dihed_add(icel_local) + 1
          buf_real(1,dihed_add(icel_local),icel_local) = phase(ix,i)
          buf_real(2,dihed_add(icel_local),icel_local) = fc(ix,i)
          buf_int (1,dihed_add(icel_local),icel_local) = dihelist(1,ix,i)
          buf_int (2,dihed_add(icel_local),icel_local) = dihelist(2,ix,i)
          buf_int (3,dihed_add(icel_local),icel_local) = dihelist(3,ix,i)
          buf_int (4,dihed_add(icel_local),icel_local) = dihelist(4,ix,i)
          buf_int (5,dihed_add(icel_local),icel_local) = nperiod(ix,i)
          buf_int (6,dihed_add(icel_local),icel_local) = dihekind(ix,i)

        end if

      end  do
    end do

    return

  end subroutine update_outgoing_enefunc_dihe

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_dihe
  !> @brief        update DIHEDRAL term for each cell in potential energy 
  !!               function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_dihe(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx, ndihe
    logical                  :: insert

    real(wp),        pointer :: dihephase(:,:), diheforce(:,:), buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary, cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: dihed_add(:), dihed_exit(:)
    integer,         pointer :: dihed_exit_index(:,:), buf_int(:,:,:)
    integer,         pointer :: ndihedral(:), dihelist(:,:,:), diheperio(:,:)
    integer,         pointer :: dihekind(:,:)


    ncel_local       => domain%num_cell_local
    nboundary        => domain%num_cell_boundary
    cell_g2l         => domain%cell_g2l
    id_g2l           => domain%id_g2l

    ndihedral        => enefunc%num_dihedral
    dihelist         => enefunc%dihe_list
    dihekind         => enefunc%dihe_kind
    diheforce        => enefunc%dihe_force_const
    diheperio        => enefunc%dihe_periodicity
    dihephase        => enefunc%dihe_phase
    dihed_add        => enefunc%dihed_add
    dihed_exit       => enefunc%dihed_exit
    dihed_exit_index => enefunc%dihed_exit_index
    buf_int          => enefunc%buf_dihed_integer
    buf_real         => enefunc%buf_dihed_real

    ndihe = enefunc%num_dihe_all

    do i = 1, ncel_local

#ifdef DEBUG
      if (dihed_add(i)+ndihedral(i)-dihed_exit(i) > MaxDihe) &
        call error_msg( &
            'Debug: update_incoming_enefunc_dihed> dihed is exceed MaxDihe')
#endif 

      if (dihed_add(i) > DiheMove) &
        call error_msg( &
    'Debug: update_incoming_enefunc_dihed> dihed migration is exceed maximum')

      if (dihed_add(i) >= dihed_exit(i)) then
          
        do k = 1, dihed_exit(i)
          dihelist (1,dihed_exit_index(k,i),i) = buf_int (1,k,i)
          dihelist (2,dihed_exit_index(k,i),i) = buf_int (2,k,i)
          dihelist (3,dihed_exit_index(k,i),i) = buf_int (3,k,i)
          dihelist (4,dihed_exit_index(k,i),i) = buf_int (4,k,i)
          diheperio(dihed_exit_index(k,i),i)   = buf_int (5,k,i)
          dihekind (dihed_exit_index(k,i),i)   = buf_int (6,k,i)
          dihephase(dihed_exit_index(k,i),i)   = buf_real(1,k,i)
          diheforce(dihed_exit_index(k,i),i)   = buf_real(2,k,i)
        end do

        do k = dihed_exit(i)+1, dihed_add(i)
          ix = k + ndihedral(i) - dihed_exit(i)
          dihelist (1,ix,i) = buf_int (1,k,i)
          dihelist (2,ix,i) = buf_int (2,k,i)
          dihelist (3,ix,i) = buf_int (3,k,i)
          dihelist (4,ix,i) = buf_int (4,k,i)
          diheperio(ix,i)   = buf_int (5,k,i)
          dihekind (ix,i)   = buf_int (6,k,i)
          dihephase(ix,i)   = buf_real(1,k,i)
          diheforce(ix,i)   = buf_real(2,k,i)
        end do

        ndihedral(i) = ndihedral(i) + dihed_add(i) - dihed_exit(i)

      else
        do k = 1, dihed_add(i)
          dihelist (1,dihed_exit_index(k,i),i) = buf_int (1,k,i)
          dihelist (2,dihed_exit_index(k,i),i) = buf_int (2,k,i)
          dihelist (3,dihed_exit_index(k,i),i) = buf_int (3,k,i)
          dihelist (4,dihed_exit_index(k,i),i) = buf_int (4,k,i)
          diheperio(dihed_exit_index(k,i),i)   = buf_int (5,k,i)
          dihekind (dihed_exit_index(k,i),i)   = buf_int (6,k,i)
          dihephase(dihed_exit_index(k,i),i)   = buf_real(1,k,i)
          diheforce(dihed_exit_index(k,i),i)   = buf_real(2,k,i)
        end do

        j  = 0
        ix = ndihedral(i)
        k  = dihed_add(i) + 1

        do while (j < (dihed_exit(i)-dihed_add(i)))

          insert = .true.
          do kx = k, dihed_exit(i)
            if (ix == dihed_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = dihed_exit_index(k,i)
            dihelist (1,kx,i) = dihelist (1,ix,i)
            dihelist (2,kx,i) = dihelist (2,ix,i)
            dihelist (3,kx,i) = dihelist (3,ix,i)
            dihelist (4,kx,i) = dihelist (4,ix,i)
            diheperio(kx,i)   = diheperio(ix,i)
            dihekind (kx,i)   = dihekind (ix,i)
            dihephase(kx,i)   = dihephase(ix,i)
            diheforce(kx,i)   = diheforce(ix,i)
            j = j + 1
            k = k + 1
          end if

          ix = ix - 1

        end do
        ndihedral(i) = ndihedral(i) + dihed_add(i) - dihed_exit(i)

      end if

    end do

    return
   
  end subroutine update_incoming_enefunc_dihe
 
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_impr
  !> @brief        update IMPROPER DIHEDRAL term for each cell in potential 
  !                energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_impr(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: icel1, icel2, icel3, icel4, i1, i2, i3, i4
    integer                  :: i, ix, icel_local

    real(wp),        pointer :: fc(:,:), phase(:,:)
    real(wp),        pointer :: coord(:,:,:), buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_g2l(:), cell_g2b(:)
    integer,         pointer :: cell_pair(:,:), id_g2l(:,:)
    integer,         pointer :: impr_add(:), impr_exit(:), impr_exit_index(:,:)
    integer,         pointer :: buf_int(:,:,:), nimproper(:), imprlist(:,:,:)


    ncel_local      => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    cell_g2l        => domain%cell_g2l
    cell_g2b        => domain%cell_g2b
    cell_pair       => domain%cell_pair
    id_g2l          => domain%id_g2l
    coord           => domain%coord

    nimproper       => enefunc%num_improper
    imprlist        => enefunc%impr_list
    impr_add        => enefunc%impr_add
    impr_exit       => enefunc%impr_exit
    impr_exit_index => enefunc%impr_exit_index
    fc              => enefunc%impr_force_const
    phase           => enefunc%impr_phase
    buf_int         => enefunc%buf_impr_integer
    buf_real        => enefunc%buf_impr_real

    impr_exit(1:ncel_local) = 0 
    impr_add (1:ncel_local+nboundary) = 0 

    do i = 1, ncel_local
      do ix = 1, nimproper(i)

        icel1 = id_g2l(1,imprlist(1,ix,i))
        i1    = id_g2l(2,imprlist(1,ix,i))
        icel2 = id_g2l(1,imprlist(2,ix,i))
        i2    = id_g2l(2,imprlist(2,ix,i))
        icel3 = id_g2l(1,imprlist(3,ix,i))
        i3    = id_g2l(2,imprlist(3,ix,i))
        icel4 = id_g2l(1,imprlist(4,ix,i))
        i4    = id_g2l(2,imprlist(4,ix,i))

        icel_local = cell_pair(icel1,icel2)

        if (icel_local /= i) then

          impr_exit(i) = impr_exit(i) + 1
          impr_exit_index(impr_exit(i),i) = ix
          impr_add(icel_local) = impr_add(icel_local) + 1
          buf_real(1,impr_add(icel_local),icel_local) = phase(ix,i)
          buf_real(2,impr_add(icel_local),icel_local) = fc(ix,i)
          buf_int (1,impr_add(icel_local),icel_local) = imprlist(1,ix,i)
          buf_int (2,impr_add(icel_local),icel_local) = imprlist(2,ix,i)
          buf_int (3,impr_add(icel_local),icel_local) = imprlist(3,ix,i)
          buf_int (4,impr_add(icel_local),icel_local) = imprlist(4,ix,i)

        end if

      end  do
    end do

    return

  end subroutine update_outgoing_enefunc_impr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_impr
  !> @brief        update IMPROPER DIHEDRAL term for each cell in potential 
  !!               energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_impr(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx
    logical                  :: insert

    real(wp),        pointer :: buf_real(:,:,:)
    real(wp),        pointer :: impr_force(:,:), impr_phase(:,:)
    integer,         pointer :: ncel_local, nboundary, cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: impr_add(:), impr_exit(:), impr_exit_index(:,:)
    integer,         pointer :: buf_int(:,:,:), nimproper(:), imprlist(:,:,:)


    ncel_local       => domain%num_cell_local
    nboundary        => domain%num_cell_boundary
    cell_g2l         => domain%cell_g2l
    id_g2l           => domain%id_g2l

    nimproper        => enefunc%num_improper
    imprlist         => enefunc%impr_list
    impr_add         => enefunc%impr_add
    impr_exit        => enefunc%impr_exit
    impr_exit_index  => enefunc%impr_exit_index
    impr_force       => enefunc%impr_force_const
    impr_phase       => enefunc%impr_phase
    buf_int          => enefunc%buf_impr_integer
    buf_real         => enefunc%buf_impr_real

    do i = 1, ncel_local
#ifdef DEBUG
      if (impr_add(i)+nimproper(i)-impr_exit(i) > MaxImpr) &
        call error_msg( &
            'Debug: update_incoming_enefunc_impr> improper is exceed MaxImpr')
#endif 

      if (impr_add(i) > ImprMove) &
        call error_msg( &
   'Debug: update_incoming_enefunc_impr> improper migration is exceed maximum')

      if (impr_add(i) >= impr_exit(i)) then
          
        do k = 1, impr_exit(i)
          imprlist(1,impr_exit_index(k,i),i) = buf_int (1,k,i)       
          imprlist(2,impr_exit_index(k,i),i) = buf_int (2,k,i)       
          imprlist(3,impr_exit_index(k,i),i) = buf_int (3,k,i)       
          imprlist(4,impr_exit_index(k,i),i) = buf_int (4,k,i)       
          impr_phase(impr_exit_index(k,i),i) = buf_real(1,k,i)
          impr_force(impr_exit_index(k,i),i) = buf_real(2,k,i)
        end do

        do k = impr_exit(i)+1, impr_add(i)
          ix = k + nimproper(i) - impr_exit(i)
          imprlist(1,ix,i) = buf_int (1,k,i)
          imprlist(2,ix,i) = buf_int (2,k,i)
          imprlist(3,ix,i) = buf_int (3,k,i)
          imprlist(4,ix,i) = buf_int (4,k,i)
          impr_phase(ix,i) = buf_real(1,k,i)
          impr_force(ix,i) = buf_real(2,k,i)
        end do

        nimproper(i) = nimproper(i) + impr_add(i) - impr_exit(i)

      else
        do k = 1, impr_add(i)
          imprlist(1,impr_exit_index(k,i),i) = buf_int (1,k,i)
          imprlist(2,impr_exit_index(k,i),i) = buf_int (2,k,i)
          imprlist(3,impr_exit_index(k,i),i) = buf_int (3,k,i)
          imprlist(4,impr_exit_index(k,i),i) = buf_int (4,k,i)
          impr_phase(impr_exit_index(k,i),i) = buf_real(1,k,i)
          impr_force(impr_exit_index(k,i),i) = buf_real(2,k,i)
        end do

        j  = 0
        ix = nimproper(i)
        k  = impr_add(i) + 1

        do while (j < (impr_exit(i)-impr_add(i)))

          insert = .true.
          do kx = k, impr_exit(i)
            if (ix == impr_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then
            kx = impr_exit_index(k,i)
            imprlist(1,kx,i) = imprlist(1,ix,i)
            imprlist(2,kx,i) = imprlist(2,ix,i)
            imprlist(3,kx,i) = imprlist(3,ix,i)
            imprlist(4,kx,i) = imprlist(4,ix,i)
            impr_phase(kx,i) = impr_phase(ix,i)
            impr_force(kx,i) = impr_force(ix,i)
            j = j + 1
            k = k + 1
          end if
          ix = ix - 1
        end do

        nimproper(i) = nimproper(i) + impr_add(i) - impr_exit(i)

      end if
    end do

    return
   
  end subroutine update_incoming_enefunc_impr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_cmap
  !> @brief        update CMAP term for each cell in potential 
  !                energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_cmap(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: icel1, icel2
    integer                  :: i, ix, k, icel_local

    real(wp),        pointer :: coord(:,:,:)
    integer,         pointer :: ncel_local, nboundary
    integer,         pointer :: cell_g2l(:), cell_g2b(:)
    integer,         pointer :: cell_pair(:,:), id_g2l(:,:)
    integer,         pointer :: cmap_exit(:), cmap_add(:), cmap_exit_index(:,:)
    integer,         pointer :: ncmap(:), cmaplist(:,:,:), cmaptype(:,:)
    integer,         pointer :: buf_int(:,:,:)


    ncel_local      => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    cell_g2l        => domain%cell_g2l
    cell_g2b        => domain%cell_g2b
    cell_pair       => domain%cell_pair
    id_g2l          => domain%id_g2l
    coord           => domain%coord

    ncmap           => enefunc%num_cmap
    cmaplist        => enefunc%cmap_list
    cmaptype        => enefunc%cmap_type
    cmap_add        => enefunc%cmap_add
    cmap_exit       => enefunc%cmap_exit
    cmap_exit_index => enefunc%cmap_exit_index
    buf_int         => enefunc%buf_cmap_integer

    cmap_exit(1:ncel_local) = 0 
    cmap_add (1:ncel_local+nboundary) = 0 

    do i = 1, ncel_local
      do ix = 1, ncmap(i)

        icel1 = id_g2l(1,cmaplist(1,ix,i))
        icel2 = id_g2l(1,cmaplist(8,ix,i))
        icel_local = cell_pair(icel1,icel2)

        if (icel_local /= i) then

          cmap_exit(i) = cmap_exit(i) + 1
          cmap_exit_index(cmap_exit(i),i) = ix
          cmap_add(icel_local) = cmap_add(icel_local) + 1
          do k = 1, 8
            buf_int(k,cmap_add(icel_local),icel_local) =cmaplist(k,ix,i)
          end do
          buf_int(9,cmap_add(icel_local),icel_local) = cmaptype(ix,i)

        end if

      end  do
    end do

    return

  end subroutine update_outgoing_enefunc_cmap

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_cmap
  !> @brief        update CMAP term for each cell in potential energy function
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_cmap(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx, ia
    logical                  :: insert

    integer,         pointer :: ncel_local, nboundary, cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: cmap_add(:), cmap_exit(:), cmap_exit_index(:,:)
    integer,         pointer :: buf_int(:,:,:), ncmap(:), cmap_list(:,:,:)
    integer,         pointer :: cmap_type(:,:)


    ncel_local      => domain%num_cell_local
    nboundary       => domain%num_cell_boundary
    cell_g2l        => domain%cell_g2l
    id_g2l          => domain%id_g2l

    ncmap           => enefunc%num_cmap    
    cmap_list       => enefunc%cmap_list
    cmap_type       => enefunc%cmap_type
    cmap_add        => enefunc%cmap_add
    cmap_exit       => enefunc%cmap_exit
    cmap_exit_index => enefunc%cmap_exit_index
    buf_int         => enefunc%buf_cmap_integer


    do i = 1, ncel_local

#ifdef DEBUG
      if (cmap_add(i)+ncmap(i)-cmap_exit(i) > MaxCmap) &
        call error_msg( &
          'Debug: update_incoming_enefunc_cmap> cmap is exceed MaxCmap')
#endif 

      if (cmap_add(i) > CmapMove) &
        call error_msg( &
      'Debug: update_incoming_enefunc_cmap> cmap migration is exceed maximum')

      if (cmap_add(i) >= cmap_exit(i)) then
          
        do k = 1, cmap_exit(i)
          do ia = 1, 8
            cmap_list(ia,cmap_exit_index(k,i),i) = buf_int(ia,k,i)       
          end do
          cmap_type(cmap_exit_index(k,i),i) = buf_int(9,k,i)       
        end do

        do k = cmap_exit(i)+1, cmap_add(i)

          ix = k + ncmap(i) - cmap_exit(i)
          do ia = 1, 8
            cmap_list(ia,ix,i) = buf_int(ia,k,i)
          end do
          cmap_type(ix,i) = buf_int(9,k,i)

        end do

        ncmap(i) = ncmap(i) + cmap_add(i) - cmap_exit(i)

      else

        do k = 1, cmap_add(i)
          do ia = 1, 8
            cmap_list(ia,cmap_exit_index(k,i),i) = buf_int(ia,k,i)
          end do
          cmap_type(cmap_exit_index(k,i),i)   = buf_int(9,k,i)
        end do

        j  = 0
        ix = ncmap(i)
        k  = cmap_add(i) + 1

        do while (j < (cmap_exit(i)-cmap_add(i)))

          insert = .true.
          do kx = k, cmap_exit(i)
            if (ix == cmap_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then

            kx = cmap_exit_index(k,i)
            do ia = 1, 8
              cmap_list(ia,kx,i) = cmap_list(ia,ix,i)
            end do

            cmap_type(kx,i) = cmap_type(ix,i)
            j = j + 1
            k = k + 1

          end if

          ix = ix - 1

        end do

        ncmap(i) = ncmap(i) + cmap_add(i) - cmap_exit(i)

      end if

    end do

    return
   
  end subroutine update_incoming_enefunc_cmap

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_outgoing_enefunc_restraint
  !> @brief        update restraint term for each cell 
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_outgoing_enefunc_restraint(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, ix, icel_local

    real(wp),        pointer :: force(:,:,:), coord(:,:,:)
    real(wp),        pointer :: buf_real(:,:,:)
    integer,         pointer :: ncel_local, nboundary, id_g2l(:,:)
    integer,         pointer :: nrestraint(:), restraint_atom(:,:)
    integer,         pointer :: restraint_exit(:), index(:,:)
    integer,         pointer :: restraint_add(:), buf_int(:,:)


    if (.not. enefunc%restraint) &
      return

    ncel_local     => domain%num_cell_local
    nboundary      => domain%num_cell_boundary
    id_g2l         => domain%id_g2l

    nrestraint     => enefunc%num_restraint
    restraint_atom => enefunc%restraint_atom
    force          => enefunc%restraint_force
    coord          => enefunc%restraint_coord
    restraint_exit => enefunc%restraint_exit
    index          => enefunc%restraint_exit_index
    restraint_add  => enefunc%restraint_add
    buf_int        => enefunc%buf_restraint_integer
    buf_real       => enefunc%buf_restraint_real

    restraint_exit(1:ncel_local) = 0
    restraint_add (1:ncel_local+nboundary) = 0

    do i = 1, ncel_local
      do ix = 1, nrestraint(i)

        icel_local = id_g2l(1,restraint_atom(ix,i))

        if (icel_local /= i) then

          restraint_exit(i) = restraint_exit(i) + 1
          index(restraint_exit(i),i) = ix

          restraint_add(icel_local) = restraint_add(icel_local) + 1
          buf_real(1:4,restraint_add(icel_local),icel_local) = force(1:4,ix,i)
          buf_real(5:7,restraint_add(icel_local),icel_local) = coord(1:3,ix,i)
          buf_int (restraint_add(icel_local),icel_local) = restraint_atom(ix,i)

        end if

      end do
    end do

    return

  end subroutine update_outgoing_enefunc_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_incoming_enefunc_restraint
  !> @brief        update incoing restraint term for each cell 
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_incoming_enefunc_restraint(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ix, kx, found
    logical                  :: insert

    real(wp),        pointer :: buf_real(:,:,:), force(:,:,:), coord(:,:,:)
    integer,         pointer :: ncel_local, nboundary, cell_g2l(:), id_g2l(:,:)
    integer,         pointer :: restraint_add(:), restraint_exit(:)
    integer,         pointer :: restraint_exit_index(:,:)
    integer,         pointer :: buf_int(:,:), nrestraint(:)
    integer,         pointer :: restraint_atom(:,:)


    if (.not. enefunc%restraint) &
      return

    ncel_local           => domain%num_cell_local
    nboundary            => domain%num_cell_boundary
    cell_g2l             => domain%cell_g2l
    id_g2l               => domain%id_g2l

    restraint_add        => enefunc%restraint_add
    restraint_exit       => enefunc%restraint_exit
    restraint_exit_index => enefunc%restraint_exit_index
    buf_int              => enefunc%buf_restraint_integer
    buf_real             => enefunc%buf_restraint_real
    restraint_atom       => enefunc%restraint_atom
    force                => enefunc%restraint_force
    coord                => enefunc%restraint_coord
    nrestraint           => enefunc%num_restraint

    found = 0
    do i = 1, ncel_local

#ifdef DEBUG
      if (restraint_add(i)+nrestraint(i)-restraint_exit(i) > MaxRest) &
        call error_msg( &
          'Debug: update_incoming_enefunc_rest> restraint is exceed MaxRest')
#endif 

      if (restraint_add(i) > RestMove) &
        call error_msg( &
  'Debug: update_incoming_enefunc_rest> restraint migration is exceed maximum')

      if (restraint_add(i) >= restraint_exit(i)) then

        do k = 1, restraint_exit(i)
          restraint_atom(restraint_exit_index(k,i),i) = buf_int(k,i)
          force(1:4,restraint_exit_index(k,i),i) = buf_real(1:4,k,i)
          coord(1:3,restraint_exit_index(k,i),i) = buf_real(5:7,k,i)
        end do

        do k = restraint_exit(i)+1, restraint_add(i)
          ix = k + nrestraint(i) - restraint_exit(i)
          restraint_atom(ix,i) = buf_int(k,i)
          force(1:4,ix,i) = buf_real(1:4,k,i)
          coord(1:3,ix,i) = buf_real(5:7,k,i)
        end do

        nrestraint(i) = nrestraint(i) + restraint_add(i) - restraint_exit(i)

      else

        do k = 1, restraint_add(i)
          restraint_atom(restraint_exit_index(k,i),i) = buf_int(k,i)
          force(1:4,restraint_exit_index(k,i),i)  = buf_real(1:4,k,i)
          coord(1:3,restraint_exit_index(k,i),i)  = buf_real(5:7,k,i)
        end do

        j  = 0
        ix = nrestraint(i)
        k  = restraint_add(i) + 1

        do while (j < (restraint_exit(i)-restraint_add(i)))

          insert = .true.
          do kx = k, restraint_exit(i)
            if (ix == restraint_exit_index(kx,i)) then
              insert = .false.
              j = j + 1
              exit
            end if
          end do

          if (insert) then

            kx = restraint_exit_index(k,i)
            restraint_atom(kx,i) = restraint_atom(ix,i)
            force(1:4,kx,i) = force(1:4,ix,i)
            coord(1:3,kx,i) = coord(1:3,ix,i)

            j = j + 1
            k = k + 1

          end if

          ix = ix - 1

        end do

        nrestraint(i) = nrestraint(i) + restraint_add(i) - restraint_exit(i)

      end if
    end do

    return

  end subroutine update_incoming_enefunc_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl
  !> @brief        exclude 1-2, 1-3 interactions
  !! @authors      JJ    
  !! @param[in]    first   : flag for first call or not
  !! @param[inout] domain  : structure of domain
  !! @param[inout] enefunc : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl(first, domain, enefunc)

    ! formal arguments
    logical,                 intent(in)    :: first
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, i, ix, k, i1, i2
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: num_excl, num_nb14
    integer                  :: found1, found2
    logical                  :: duplicate

    integer,         pointer :: ncell_local, max_atom, natom(:), id_g2l(:,:)
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: nbond(:), bondlist(:,:,:)
    integer,         pointer :: nangle(:), anglelist(:,:,:)
    integer,         pointer :: ndihedral(:), dihelist(:,:,:), pairlist(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)

 
    ncell_local     => domain%num_cell_local 
    max_atom        => domain%max_num_atom
    natom           => domain%num_atom
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    pairlist        => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    bondlist        => enefunc%bond_list
    nangle          => enefunc%num_angle
    anglelist       => enefunc%angle_list
    ndihedral       => enefunc%num_dihedral
    dihelist        => enefunc%dihe_list
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1

    ncell = domain%num_cell_local + domain%num_cell_boundary

    max_atom = 0
    do i = 1, ncell
      max_atom = max(max_atom,natom(i))
    end do 

    ! initialization
    !
    num_nonb_excl  (1:max_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:max_atom,1:ncell_local) = 0
    num_nb14_calc  (1:max_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:max_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                = 0
    num_excl_total1(1:ncell_local)            = 0
    num_nb14_total (1:maxcell)                = 0
    num_nb14_total1(1:ncell_local)            = 0

    ! exclude 1-2 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bondlist(1,ix,i))
        i1    = id_g2l(2,bondlist(1,ix,i))
        icel2 = id_g2l(1,bondlist(2,ix,i))
        i2    = id_g2l(2,bondlist(2,ix,i))

        if (icel1 < icel2) then
          num_excl = num_nonb_excl(i1,pairlist(icel2,icel1)) + 1
          num_nonb_excl(i1,pairlist(icel2,icel1)) = num_excl
          nonb_excl_list(num_excl,i1,pairlist(icel2,icel1)) = i2
          num_excl_total(pairlist(icel2,icel1)) = &
               num_excl_total(pairlist(icel2,icel1)) + 1

        else if (icel1 > icel2) then
          num_excl = num_nonb_excl(i2,pairlist(icel1,icel2)) + 1
          num_nonb_excl(i2,pairlist(icel1,icel2)) = num_excl
          nonb_excl_list(num_excl,i2,pairlist(icel1,icel2)) = i1
          num_excl_total(pairlist(icel1,icel2)) = &
               num_excl_total(pairlist(icel1,icel2)) + 1

        else if (i1 < i2) then
          num_excl = num_nonb_excl1(i1,i) + 1
          num_nonb_excl1(i1,i) = num_excl
          nonb_excl_list1(num_excl,i1,i) = i2
          num_excl_total1(i) = num_excl_total1(i) + 1

        else if (i1 > i2) then
          num_excl = num_nonb_excl1(i2,i) + 1
          num_nonb_excl1(i2,i) = num_excl
          nonb_excl_list1(num_excl,i2,i) = i1
          num_excl_total1(i) = num_excl_total1(i) + 1

        end if
      end do
    end do
             
    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,anglelist(1,ix,i))
        i1    = id_g2l(2,anglelist(1,ix,i))
        icel2 = id_g2l(1,anglelist(3,ix,i))
        i2    = id_g2l(2,anglelist(3,ix,i))

        if (icel1 < icel2) then

          num_excl = num_nonb_excl(i1,pairlist(icel2,icel1))
          duplicate = .false.

          do k = 1, num_excl
            if (i2 == nonb_excl_list(k,i1,pairlist(icel2,icel1))) &
              duplicate = .true.
          end do

          if (.not. duplicate) then
            num_excl = num_excl + 1
            num_nonb_excl(i1,pairlist(icel2,icel1)) = num_excl
            nonb_excl_list(num_excl,i1,pairlist(icel2,icel1)) = i2
            num_excl_total(pairlist(icel2,icel1)) = &
                 num_excl_total(pairlist(icel2,icel1)) + 1
          end if

        else if (icel1 > icel2) then

          num_excl = num_nonb_excl(i2,pairlist(icel1,icel2))
          duplicate = .false.

          do k = 1, num_excl
            if (i1 == nonb_excl_list(k,i2,pairlist(icel1,icel2))) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_excl = num_excl + 1
            num_nonb_excl(i2,pairlist(icel1,icel2)) = num_excl
            nonb_excl_list(num_excl,i2,pairlist(icel1,icel2)) = i1
            num_excl_total(pairlist(icel1,icel2)) = &
                 num_excl_total(pairlist(icel1,icel2)) + 1
          end if

        else if (i1 < i2) then

          num_excl = num_nonb_excl1(i1,i)
          duplicate = .false.

          do k = 1, num_excl
            if (i2 == nonb_excl_list1(k,i1,i)) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_excl = num_excl + 1
            num_nonb_excl1(i1,i) = num_excl
            nonb_excl_list1(num_excl,i1,i) = i2
            num_excl_total1(i) = num_excl_total1(i) + 1
          end if

        else if (i1 > i2) then

          num_excl = num_nonb_excl1(i2,i)
          duplicate = .false.

          do k = 1, num_excl
            if (i1 == nonb_excl_list1(k,i2,i)) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_excl = num_excl + 1
            num_nonb_excl1(i2,i) = num_excl
            nonb_excl_list1(num_excl,i2,i) = i1
            num_excl_total1(i) = num_excl_total1(i) + 1
          end if

        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihelist(1,ix,i))
        i1    = id_g2l(2,dihelist(1,ix,i))
        icel2 = id_g2l(1,dihelist(4,ix,i))
        i2    = id_g2l(2,dihelist(4,ix,i))

        if (icel1 < icel2) then

          num_nb14 = num_nb14_calc(i1,pairlist(icel2,icel1))
          duplicate = .false.

          do k = 1, num_nonb_excl(i1,pairlist(icel2,icel1))
            if (i2 == nonb_excl_list(k,i1,pairlist(icel2,icel1))) &
              duplicate = .true.
          end do

          do k = 1, num_nb14
            if (i2 == nb14_calc_list(k,i1,pairlist(icel2,icel1))) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            num_nb14_calc(i1,pairlist(icel2,icel1)) = num_nb14
            nb14_calc_list(num_nb14,i1,pairlist(icel2,icel1)) = i2
            num_nb14_total(pairlist(icel2,icel1)) = &
                 num_nb14_total(pairlist(icel2,icel1)) + 1
          end if

        else if (icel1 > icel2) then

          num_nb14 = num_nb14_calc(i2,pairlist(icel1,icel2))
          duplicate = .false.

          do k = 1, num_nonb_excl(i2,pairlist(icel1,icel2))
            if (i1 == nonb_excl_list(k,i2,pairlist(icel1,icel2))) &
              duplicate = .true.
          end do

          do k = 1, num_nb14
            if (i1 == nb14_calc_list(k,i2,pairlist(icel1,icel2))) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            num_nb14_calc(i2,pairlist(icel1,icel2)) = num_nb14
            nb14_calc_list(num_nb14,i2,pairlist(icel1,icel2)) = i1
            num_nb14_total(pairlist(icel1,icel2)) = &
                 num_nb14_total(pairlist(icel1,icel2)) + 1
          end if

        else if (i1 < i2) then

          num_nb14 = num_nb14_calc1(i1,i)
          duplicate = .false.

          do k = 1, num_nb14
            if (i2 == nb14_calc_list1(k,i1,i)) &
              duplicate = .true.
          end do

          do k = 1, num_nonb_excl1(i1,i)
            if (i2 == nonb_excl_list1(k,i1,i)) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            num_nb14_calc1(i1,i) = num_nb14
            nb14_calc_list1(num_nb14,i1,i) = i2
            num_nb14_total1(i) = num_nb14_total1(i) + 1
          end if

        else if (i1 > i2) then

          num_nb14 = num_nb14_calc1(i2,i)
          duplicate = .false.

          do k = 1, num_nb14
            if (i1 == nb14_calc_list1(k,i2,i)) &
              duplicate = .true.
          end do

          do k = 1, num_nonb_excl1(i2,i)
            if (i1 == nonb_excl_list1(k,i2,i)) &
              duplicate = .true.
          end do

          if (.not.duplicate) then
            num_nb14 = num_nb14 + 1
            num_nb14_calc1(i2,i) = num_nb14
            nb14_calc_list1(num_nb14,i2,i) = i1
            num_nb14_total1(i) = num_nb14_total1(i) + 1
          end if

        end if

      end do
    end do

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel)>0.and.cell_pair(jcel,icel)<=ncell_local) then
            found1 = found1 + num_excl_total(pairlist(jcel,icel))
            found2 = found2 + num_nb14_total(pairlist(jcel,icel))
          end if 
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do
#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(natom, domain%cell_pairlist1, num_nonb_excl, &
                     nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(natom, domain%cell_pairlist1, num_nb14_calc, &
                     nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(natom, num_nonb_excl1, ncell_local,          &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(natom, num_nb14_calc1, ncell_local,          &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_rest
  !> @brief        exclude 1-2, 1-3 interactions w/o local restraint
  !! @authors      JJ, CK 
  !! @param[in]    first   : flag for first call or not
  !! @param[inout] domain  : structure of domain
  !! @param[inout] enefunc : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_rest(first, domain, enefunc)

    ! formal arguments
    logical,                 intent(in)    :: first
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, i, ix, k, i1, i2
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: num_excl, num_nb14
    integer                  :: found1, found2
    integer                  :: fkind
    logical                  :: duplicate

    integer,         pointer :: ncell_local, max_atom, natom(:), id_g2l(:,:)
    integer,         pointer :: cell_pair(:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: nbond(:), bondlist(:,:,:)
    integer,         pointer :: nangle(:), anglelist(:,:,:)
    integer,         pointer :: ndihedral(:), dihelist(:,:,:), pairlist(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)

 
    ncell_local     => domain%num_cell_local 
    max_atom        => domain%max_num_atom
    natom           => domain%num_atom
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    pairlist        => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    bondlist        => enefunc%bond_list
    nangle          => enefunc%num_angle
    anglelist       => enefunc%angle_list
    ndihedral       => enefunc%num_dihedral
    dihelist        => enefunc%dihe_list
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1

    ncell = domain%num_cell_local + domain%num_cell_boundary

    max_atom = 0
    do i = 1, ncell
      max_atom = max(max_atom,natom(i))
    end do 

    ! initialization
    !
    num_nonb_excl  (1:max_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:max_atom,1:ncell_local) = 0
    num_nb14_calc  (1:max_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:max_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                = 0
    num_excl_total1(1:ncell_local)            = 0
    num_nb14_total (1:maxcell)                = 0
    num_nb14_total1(1:ncell_local)            = 0

    ! exclude 1-2 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bondlist(1,ix,i))
        i1    = id_g2l(2,bondlist(1,ix,i))
        icel2 = id_g2l(1,bondlist(2,ix,i))
        i2    = id_g2l(2,bondlist(2,ix,i))
        fkind = enefunc%bond_kind(ix,i)

        if (fkind == 0) then

          if (icel1 < icel2) then
            num_excl = num_nonb_excl(i1,pairlist(icel2,icel1)) + 1
            num_nonb_excl(i1,pairlist(icel2,icel1)) = num_excl
            nonb_excl_list(num_excl,i1,pairlist(icel2,icel1)) = i2
            num_excl_total(pairlist(icel2,icel1)) = &
                 num_excl_total(pairlist(icel2,icel1)) + 1
         
          else if (icel1 > icel2) then
            num_excl = num_nonb_excl(i2,pairlist(icel1,icel2)) + 1
            num_nonb_excl(i2,pairlist(icel1,icel2)) = num_excl
            nonb_excl_list(num_excl,i2,pairlist(icel1,icel2)) = i1
            num_excl_total(pairlist(icel1,icel2)) = &
                 num_excl_total(pairlist(icel1,icel2)) + 1
         
          else if (i1 < i2) then
            num_excl = num_nonb_excl1(i1,i) + 1
            num_nonb_excl1(i1,i) = num_excl
            nonb_excl_list1(num_excl,i1,i) = i2
            num_excl_total1(i) = num_excl_total1(i) + 1
         
          else if (i1 > i2) then
            num_excl = num_nonb_excl1(i2,i) + 1
            num_nonb_excl1(i2,i) = num_excl
            nonb_excl_list1(num_excl,i2,i) = i1
            num_excl_total1(i) = num_excl_total1(i) + 1
         
          end if

        end if
      end do
    end do
             
    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,anglelist(1,ix,i))
        i1    = id_g2l(2,anglelist(1,ix,i))
        icel2 = id_g2l(1,anglelist(3,ix,i))
        i2    = id_g2l(2,anglelist(3,ix,i))
        fkind = enefunc%angle_kind(ix,i)

        if (fkind == 0) then

          if (icel1 < icel2) then
         
            num_excl = num_nonb_excl(i1,pairlist(icel2,icel1))
            duplicate = .false.
         
            do k = 1, num_excl
              if (i2 == nonb_excl_list(k,i1,pairlist(icel2,icel1))) &
                duplicate = .true.
            end do
         
            if (.not. duplicate) then
              num_excl = num_excl + 1
              num_nonb_excl(i1,pairlist(icel2,icel1)) = num_excl
              nonb_excl_list(num_excl,i1,pairlist(icel2,icel1)) = i2
              num_excl_total(pairlist(icel2,icel1)) = &
                   num_excl_total(pairlist(icel2,icel1)) + 1
            end if
         
          else if (icel1 > icel2) then
         
            num_excl = num_nonb_excl(i2,pairlist(icel1,icel2))
            duplicate = .false.
         
            do k = 1, num_excl
              if (i1 == nonb_excl_list(k,i2,pairlist(icel1,icel2))) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_excl = num_excl + 1
              num_nonb_excl(i2,pairlist(icel1,icel2)) = num_excl
              nonb_excl_list(num_excl,i2,pairlist(icel1,icel2)) = i1
              num_excl_total(pairlist(icel1,icel2)) = &
                   num_excl_total(pairlist(icel1,icel2)) + 1
            end if
         
          else if (i1 < i2) then
         
            num_excl = num_nonb_excl1(i1,i)
            duplicate = .false.
         
            do k = 1, num_excl
              if (i2 == nonb_excl_list1(k,i1,i)) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_excl = num_excl + 1
              num_nonb_excl1(i1,i) = num_excl
              nonb_excl_list1(num_excl,i1,i) = i2
              num_excl_total1(i) = num_excl_total1(i) + 1
            end if
         
          else if (i1 > i2) then
         
            num_excl = num_nonb_excl1(i2,i)
            duplicate = .false.
         
            do k = 1, num_excl
              if (i1 == nonb_excl_list1(k,i2,i)) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_excl = num_excl + 1
              num_nonb_excl1(i2,i) = num_excl
              nonb_excl_list1(num_excl,i2,i) = i1
              num_excl_total1(i) = num_excl_total1(i) + 1
            end if
         
          end if

        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihelist(1,ix,i))
        i1    = id_g2l(2,dihelist(1,ix,i))
        icel2 = id_g2l(1,dihelist(4,ix,i))
        i2    = id_g2l(2,dihelist(4,ix,i))
        fkind = enefunc%dihe_kind(ix,i)

        if (fkind == 0) then

          if (icel1 < icel2) then
         
            num_nb14 = num_nb14_calc(i1,pairlist(icel2,icel1))
            duplicate = .false.
         
            do k = 1, num_nonb_excl(i1,pairlist(icel2,icel1))
              if (i2 == nonb_excl_list(k,i1,pairlist(icel2,icel1))) &
                duplicate = .true.
            end do
         
            do k = 1, num_nb14
              if (i2 == nb14_calc_list(k,i1,pairlist(icel2,icel1))) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_nb14 = num_nb14 + 1
              num_nb14_calc(i1,pairlist(icel2,icel1)) = num_nb14
              nb14_calc_list(num_nb14,i1,pairlist(icel2,icel1)) = i2
              num_nb14_total(pairlist(icel2,icel1)) = &
                   num_nb14_total(pairlist(icel2,icel1)) + 1
            end if
         
          else if (icel1 > icel2) then
         
            num_nb14 = num_nb14_calc(i2,pairlist(icel1,icel2))
            duplicate = .false.
         
            do k = 1, num_nonb_excl(i2,pairlist(icel1,icel2))
              if (i1 == nonb_excl_list(k,i2,pairlist(icel1,icel2))) &
                duplicate = .true.
            end do
         
            do k = 1, num_nb14
              if (i1 == nb14_calc_list(k,i2,pairlist(icel1,icel2))) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_nb14 = num_nb14 + 1
              num_nb14_calc(i2,pairlist(icel1,icel2)) = num_nb14
              nb14_calc_list(num_nb14,i2,pairlist(icel1,icel2)) = i1
              num_nb14_total(pairlist(icel1,icel2)) = &
                   num_nb14_total(pairlist(icel1,icel2)) + 1
            end if
         
          else if (i1 < i2) then
         
            num_nb14 = num_nb14_calc1(i1,i)
            duplicate = .false.
         
            do k = 1, num_nb14
              if (i2 == nb14_calc_list1(k,i1,i)) &
                duplicate = .true.
            end do
         
            do k = 1, num_nonb_excl1(i1,i)
              if (i2 == nonb_excl_list1(k,i1,i)) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_nb14 = num_nb14 + 1
              num_nb14_calc1(i1,i) = num_nb14
              nb14_calc_list1(num_nb14,i1,i) = i2
              num_nb14_total1(i) = num_nb14_total1(i) + 1
            end if
         
          else if (i1 > i2) then
         
            num_nb14 = num_nb14_calc1(i2,i)
            duplicate = .false.
         
            do k = 1, num_nb14
              if (i1 == nb14_calc_list1(k,i2,i)) &
                duplicate = .true.
            end do
         
            do k = 1, num_nonb_excl1(i2,i)
              if (i1 == nonb_excl_list1(k,i2,i)) &
                duplicate = .true.
            end do
         
            if (.not.duplicate) then
              num_nb14 = num_nb14 + 1
              num_nb14_calc1(i2,i) = num_nb14
              nb14_calc_list1(num_nb14,i2,i) = i1
              num_nb14_total1(i) = num_nb14_total1(i) + 1
            end if
         
          end if

        end if

      end do
    end do

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel)>0.and.cell_pair(jcel,icel)<=ncell_local) then
            found1 = found1 + num_excl_total(pairlist(jcel,icel))
            found2 = found2 + num_nb14_total(pairlist(jcel,icel))
          end if 
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do
#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(natom, domain%cell_pairlist1, num_nonb_excl, &
                     nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(natom, domain%cell_pairlist1, num_nb14_calc, &
                     nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(natom, num_nonb_excl1, ncell_local,          &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(natom, num_nb14_calc1, ncell_local,          &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl_rest

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_solute
  !> @brief        exclude 1-2, 1-3 interactions
  !! @authors      JJ    
  !! @param[in]    first   : flag for first call or not
  !! @param[inout] domain  : structure of domain
  !! @param[inout] enefunc : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_solute(first, domain, enefunc)

    ! formal arguments
    logical,                 intent(in)    :: first
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, ncell_local, i, ix, k, i1, i2
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: num_excl, num_nb14, id, omp_get_thread_num
    integer                  :: found1, found2
    logical                  :: duplicate

    integer,         pointer :: nsolute(:), id_g2l(:,:)
    integer,         pointer :: nbond(:), bond_list(:,:,:)
    integer,         pointer :: nangle(:), angl_list(:,:,:)
    integer,         pointer :: ndihedral(:), dihe_list(:,:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: cell_pair(:,:), cell_pairlist2(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)


    nsolute         => domain%num_solute
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    cell_pairlist2  => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    nangle          => enefunc%num_angle
    ndihedral       => enefunc%num_dihedral
    bond_list       => enefunc%bond_list
    angl_list       => enefunc%angle_list
    dihe_list       => enefunc%dihe_list
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1

    ncell_local = domain%num_cell_local
    ncell       = domain%num_cell_local + domain%num_cell_boundary

    domain%max_num_atom = 0
    do i = 1, ncell
      domain%max_num_atom = max(domain%max_num_atom,domain%num_atom(i))
    end do

    ! initialization
    !
    num_nonb_excl  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_nb14_calc  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                           = 0
    num_excl_total1(1:ncell_local)                       = 0
    num_nb14_total (1:maxcell)                           = 0
    num_nb14_total1(1:ncell_local)                       = 0

    ! exclude 1-2 interaction
    !
    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, icel1, icel2, icel, i1, i2, num_excl, duplicate,  &
    !$omp          k, num_nb14) 
    !        
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bond_list(1,ix,i))
        icel2 = id_g2l(1,bond_list(2,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel,nthread) == id) then

            i1 = id_g2l(2,bond_list(1,ix,i))
            i2 = id_g2l(2,bond_list(2,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (icel1 < icel2) then
                num_excl = num_nonb_excl(i1,icel) + 1
                num_nonb_excl(i1,icel) = num_excl
                nonb_excl_list(num_excl,i1,icel) = i2
                num_excl_total(icel) = num_excl_total(icel) + 1

              else if (icel1 > icel2) then
                num_excl = num_nonb_excl(i2,icel) + 1
                num_nonb_excl(i2,icel) = num_excl
                nonb_excl_list(num_excl,i2,icel) = i1
                num_excl_total(icel) = num_excl_total(icel) + 1

              end if
            end if

          end if

        else

          if (mod(i,nthread) == id) then

            i1    = id_g2l(2,bond_list(1,ix,i))
            i2    = id_g2l(2,bond_list(2,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (i1 < i2) then

                num_excl = num_nonb_excl1(i1,i) + 1
                num_nonb_excl1(i1,i) = num_excl
                nonb_excl_list1(num_excl,i1,i) = i2
                num_excl_total1(i) = num_excl_total1(i) + 1

              else if (i1 > i2) then

                num_excl = num_nonb_excl1(i2,i) + 1
                num_nonb_excl1(i2,i) = num_excl
                nonb_excl_list1(num_excl,i2,i) = i1
                num_excl_total1(i) = num_excl_total1(i) + 1

              end if
            end if

          end if

        end if

      end do
    end do

    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,angl_list(1,ix,i))
        icel2 = id_g2l(1,angl_list(3,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel,nthread) == id) then

            i1 = id_g2l(2,angl_list(1,ix,i))
            i2 = id_g2l(2,angl_list(3,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (icel1 < icel2) then

                num_excl = num_nonb_excl(i1,icel)
                duplicate = .false.

                do k = 1, num_excl
                  if (i2 == nonb_excl_list(k,i1,icel)) &
                    duplicate = .true.
                end do

                if (.not. duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl(i1,icel) = num_excl
                  nonb_excl_list(num_excl,i1,icel) = i2
                  num_excl_total(icel) =  num_excl_total(icel) + 1
                end if

              else if (icel1 > icel2) then

                num_excl = num_nonb_excl(i2,icel)
                duplicate = .false.

                do k = 1, num_excl
                  if (i1 == nonb_excl_list(k,i2,icel)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl(i2,icel) = num_excl
                  nonb_excl_list(num_excl,i2,icel) = i1
                  num_excl_total(icel) = num_excl_total(icel) + 1
                end if

              end if
            end if

          end if

        else

          if (mod(i,nthread) == id) then

            i1    = id_g2l(2,angl_list(1,ix,i))
            i2    = id_g2l(2,angl_list(3,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (i1 < i2) then

                num_excl = num_nonb_excl1(i1,i)
                duplicate = .false.

                do k = 1, num_excl
                  if (i2 == nonb_excl_list1(k,i1,i)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl1(i1,i) = num_excl
                  nonb_excl_list1(num_excl,i1,i) = i2
                  num_excl_total1(i) = num_excl_total1(i) + 1
                end if

              else if (i1 > i2) then

                num_excl = num_nonb_excl1(i2,i)
                duplicate = .false.

                do k = 1, num_excl
                  if (i1 == nonb_excl_list1(k,i2,i)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl1(i2,i) = num_excl
                  nonb_excl_list1(num_excl,i2,i) = i1
                  num_excl_total1(i) = num_excl_total1(i) + 1
                end if

              end if
            end if

          end if
        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihe_list(1,ix,i))
        icel2 = id_g2l(1,dihe_list(4,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel,nthread) == id) then

            i1 = id_g2l(2,dihe_list(1,ix,i))
            i2 = id_g2l(2,dihe_list(4,ix,i))

            if (icel1 < icel2) then

              num_nb14 = num_nb14_calc(i1,icel)
              duplicate = .false.

              do k = 1, num_nonb_excl(i1,icel)
                if (i2 == nonb_excl_list(k,i1,icel)) &
                  duplicate = .true.
              end do

              do k = 1, num_nb14
                if (i2 == nb14_calc_list(k,i1,icel)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc(i1,icel) = num_nb14
                nb14_calc_list(num_nb14,i1,icel) = i2
                num_nb14_total(icel) = num_nb14_total(icel) + 1
              end if

            else if (icel1 > icel2) then

              num_nb14 = num_nb14_calc(i2,icel)
              duplicate = .false.

              do k = 1, num_nonb_excl(i2,icel)
                if (i1 == nonb_excl_list(k,i2,icel)) &
                  duplicate = .true.
              end do

              do k = 1, num_nb14
                if (i1 == nb14_calc_list(k,i2,icel)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc(i2,icel) = num_nb14
                nb14_calc_list(num_nb14,i2,icel) = i1
                num_nb14_total(icel) = num_nb14_total(icel) + 1
              end if

            end if
          end if

        else

          if (mod(i,nthread) == id) then

            i1    = id_g2l(2,dihe_list(1,ix,i))
            i2    = id_g2l(2,dihe_list(4,ix,i))

            if (i1 < i2) then

              num_nb14 = num_nb14_calc1(i1,i)
              duplicate = .false.

              do k = 1, num_nb14
                if (i2 == nb14_calc_list1(k,i1,i)) &
                  duplicate = .true.
              end do

              do k = 1, num_nonb_excl1(i1,i)
                if (i2 == nonb_excl_list1(k,i1,i)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc1(i1,i) = num_nb14
                nb14_calc_list1(num_nb14,i1,i) = i2
                num_nb14_total1(i) = num_nb14_total1(i) + 1
              end if

            else if (i1 > i2) then

              num_nb14 = num_nb14_calc1(i2,i)
              duplicate = .false.

              do k = 1, num_nb14
                if (i1 == nb14_calc_list1(k,i2,i)) &
                  duplicate = .true.
              end do

              do k = 1, num_nonb_excl1(i2,i)
                if (i1 == nonb_excl_list1(k,i2,i)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc1(i2,i) = num_nb14
                nb14_calc_list1(num_nb14,i2,i) = i1
                num_nb14_total1(i) = num_nb14_total1(i) + 1
              end if

            end if
          end if

        end if

      end do
    end do

    !$omp end parallel

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel) > 0 .and. &
              cell_pair(jcel,icel) <= ncell_local) then
            found1 = found1 + num_excl_total(cell_pairlist2(jcel,icel))
            found2 = found2 + num_nb14_total(cell_pairlist2(jcel,icel))
          end if
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do

#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1,       &
                     num_nonb_excl, nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1,       &
                     num_nb14_calc, nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(domain%num_atom, num_nonb_excl1, ncell_local, &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(domain%num_atom, num_nb14_calc1, ncell_local, &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl_solute

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_solute_rest
  !> @brief        exclude 1-2, 1-3 interactions w/o local restraint
  !! @authors      JJ    
  !! @param[in]    first   : flag for first call or not
  !! @param[inout] domain  : structure of domain
  !! @param[inout] enefunc : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_solute_rest(first, domain, enefunc)

    ! formal arguments
    logical,                 intent(in)    :: first
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, ncell_local, i, ix, k, i1, i2
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: num_excl, num_nb14, id, omp_get_thread_num
    integer                  :: found1, found2
    integer                  :: fkind
    logical                  :: duplicate

    integer,         pointer :: nsolute(:), id_g2l(:,:)
    integer,         pointer :: nbond(:), bond_list(:,:,:)
    integer,         pointer :: nangle(:), angl_list(:,:,:)
    integer,         pointer :: ndihedral(:), dihe_list(:,:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: cell_pair(:,:), cell_pairlist2(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)


    nsolute         => domain%num_solute
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    cell_pairlist2  => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    nangle          => enefunc%num_angle
    ndihedral       => enefunc%num_dihedral
    bond_list       => enefunc%bond_list
    angl_list       => enefunc%angle_list
    dihe_list       => enefunc%dihe_list
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1

    ncell_local = domain%num_cell_local
    ncell       = domain%num_cell_local + domain%num_cell_boundary

    domain%max_num_atom = 0
    do i = 1, ncell
      domain%max_num_atom = max(domain%max_num_atom,domain%num_atom(i))
    end do

    ! initialization
    !
    num_nonb_excl  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_nb14_calc  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                           = 0
    num_excl_total1(1:ncell_local)                       = 0
    num_nb14_total (1:maxcell)                           = 0
    num_nb14_total1(1:ncell_local)                       = 0

    ! exclude 1-2 interaction
    !
    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, icel1, icel2, icel, i1, i2, num_excl, duplicate,  &
    !$omp          k, num_nb14, fkind) 
    !        
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bond_list(1,ix,i))
        icel2 = id_g2l(1,bond_list(2,ix,i))
        fkind = enefunc%bond_kind(ix,i)

        if (fkind == 0) then

          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel,nthread) == id) then
         
              i1 = id_g2l(2,bond_list(1,ix,i))
              i2 = id_g2l(2,bond_list(2,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (icel1 < icel2) then
                  num_excl = num_nonb_excl(i1,icel) + 1
                  num_nonb_excl(i1,icel) = num_excl
                  nonb_excl_list(num_excl,i1,icel) = i2
                  num_excl_total(icel) = num_excl_total(icel) + 1
         
                else if (icel1 > icel2) then
                  num_excl = num_nonb_excl(i2,icel) + 1
                  num_nonb_excl(i2,icel) = num_excl
                  nonb_excl_list(num_excl,i2,icel) = i1
                  num_excl_total(icel) = num_excl_total(icel) + 1
         
                end if
              end if
         
            end if
         
          else
         
            if (mod(i,nthread) == id) then
         
              i1    = id_g2l(2,bond_list(1,ix,i))
              i2    = id_g2l(2,bond_list(2,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (i1 < i2) then
         
                  num_excl = num_nonb_excl1(i1,i) + 1
                  num_nonb_excl1(i1,i) = num_excl
                  nonb_excl_list1(num_excl,i1,i) = i2
                  num_excl_total1(i) = num_excl_total1(i) + 1
         
                else if (i1 > i2) then
         
                  num_excl = num_nonb_excl1(i2,i) + 1
                  num_nonb_excl1(i2,i) = num_excl
                  nonb_excl_list1(num_excl,i2,i) = i1
                  num_excl_total1(i) = num_excl_total1(i) + 1
         
                end if
              end if
         
            end if
         
          end if

        end if

      end do
    end do

    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,angl_list(1,ix,i))
        icel2 = id_g2l(1,angl_list(3,ix,i))
        fkind = enefunc%angle_kind(ix,i)

        if (fkind == 0) then

          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel,nthread) == id) then
         
              i1 = id_g2l(2,angl_list(1,ix,i))
              i2 = id_g2l(2,angl_list(3,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (icel1 < icel2) then
         
                  num_excl = num_nonb_excl(i1,icel)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i2 == nonb_excl_list(k,i1,icel)) &
                      duplicate = .true.
                  end do
         
                  if (.not. duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl(i1,icel) = num_excl
                    nonb_excl_list(num_excl,i1,icel) = i2
                    num_excl_total(icel) =  num_excl_total(icel) + 1
                  end if
         
                else if (icel1 > icel2) then
         
                  num_excl = num_nonb_excl(i2,icel)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i1 == nonb_excl_list(k,i2,icel)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl(i2,icel) = num_excl
                    nonb_excl_list(num_excl,i2,icel) = i1
                    num_excl_total(icel) = num_excl_total(icel) + 1
                  end if
         
                end if
              end if
         
            end if
         
          else
         
            if (mod(i,nthread) == id) then
         
              i1    = id_g2l(2,angl_list(1,ix,i))
              i2    = id_g2l(2,angl_list(3,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (i1 < i2) then
         
                  num_excl = num_nonb_excl1(i1,i)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i2 == nonb_excl_list1(k,i1,i)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl1(i1,i) = num_excl
                    nonb_excl_list1(num_excl,i1,i) = i2
                    num_excl_total1(i) = num_excl_total1(i) + 1
                  end if
         
                else if (i1 > i2) then
         
                  num_excl = num_nonb_excl1(i2,i)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i1 == nonb_excl_list1(k,i2,i)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl1(i2,i) = num_excl
                    nonb_excl_list1(num_excl,i2,i) = i1
                    num_excl_total1(i) = num_excl_total1(i) + 1
                  end if
         
                end if
              end if
         
            end if
          end if

        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihe_list(1,ix,i))
        icel2 = id_g2l(1,dihe_list(4,ix,i))
        fkind = enefunc%dihe_kind(ix,i)

        if (fkind == 0) then
          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel,nthread) == id) then
         
              i1 = id_g2l(2,dihe_list(1,ix,i))
              i2 = id_g2l(2,dihe_list(4,ix,i))
         
              if (icel1 < icel2) then
         
                num_nb14 = num_nb14_calc(i1,icel)
                duplicate = .false.
         
                do k = 1, num_nonb_excl(i1,icel)
                  if (i2 == nonb_excl_list(k,i1,icel)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nb14
                  if (i2 == nb14_calc_list(k,i1,icel)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc(i1,icel) = num_nb14
                  nb14_calc_list(num_nb14,i1,icel) = i2
                  num_nb14_total(icel) = num_nb14_total(icel) + 1
                end if
         
              else if (icel1 > icel2) then
         
                num_nb14 = num_nb14_calc(i2,icel)
                duplicate = .false.
         
                do k = 1, num_nonb_excl(i2,icel)
                  if (i1 == nonb_excl_list(k,i2,icel)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nb14
                  if (i1 == nb14_calc_list(k,i2,icel)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc(i2,icel) = num_nb14
                  nb14_calc_list(num_nb14,i2,icel) = i1
                  num_nb14_total(icel) = num_nb14_total(icel) + 1
                end if
         
              end if
            end if
         
          else
         
            if (mod(i,nthread) == id) then
         
              i1    = id_g2l(2,dihe_list(1,ix,i))
              i2    = id_g2l(2,dihe_list(4,ix,i))
         
              if (i1 < i2) then
         
                num_nb14 = num_nb14_calc1(i1,i)
                duplicate = .false.
         
                do k = 1, num_nb14
                  if (i2 == nb14_calc_list1(k,i1,i)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nonb_excl1(i1,i)
                  if (i2 == nonb_excl_list1(k,i1,i)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc1(i1,i) = num_nb14
                  nb14_calc_list1(num_nb14,i1,i) = i2
                  num_nb14_total1(i) = num_nb14_total1(i) + 1
                end if
         
              else if (i1 > i2) then
         
                num_nb14 = num_nb14_calc1(i2,i)
                duplicate = .false.
         
                do k = 1, num_nb14
                  if (i1 == nb14_calc_list1(k,i2,i)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nonb_excl1(i2,i)
                  if (i1 == nonb_excl_list1(k,i2,i)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc1(i2,i) = num_nb14
                  nb14_calc_list1(num_nb14,i2,i) = i1
                  num_nb14_total1(i) = num_nb14_total1(i) + 1
                end if
         
              end if
            end if
         
          end if

        end if

      end do
    end do

    !$omp end parallel

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel) > 0 .and. &
              cell_pair(jcel,icel) <= ncell_local) then
            found1 = found1 + num_excl_total(cell_pairlist2(jcel,icel))
            found2 = found2 + num_nb14_total(cell_pairlist2(jcel,icel))
          end if
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do

#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1,       &
                     num_nonb_excl, nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1,       &
                     num_nb14_calc, nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(domain%num_atom, num_nonb_excl1, ncell_local, &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(domain%num_atom, num_nb14_calc1, ncell_local, &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl_solute_rest

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_constraint
  !> @brief        exclude 1-2, 1-3 interactions and constraints
  !! @authors      JJ    
  !! @param[in]    first       : flag for first call or not
  !! @param[in]    water_table : flag for water table
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : structure of domain
  !! @param[inout] enefunc     : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_constraint(first, water_table, constraints, &
                                        domain, enefunc)

    ! formal arguments
    logical,                     intent(in)    :: first
    logical,                     intent(in)    :: water_table
    type(s_constraints), target, intent(in)    :: constraints
    type(s_domain),      target, intent(inout) :: domain
    type(s_enefunc),     target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, ncell_local, i, ix, k, i1, i2, i3
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: ic, j, ih
    integer                  :: num_excl, num_nb14, id, omp_get_thread_num
    integer                  :: found1, found2
    logical                  :: duplicate

    integer,         pointer :: nsolute(:), nwater(:), water_list(:,:,:)
    integer,         pointer :: id_g2l(:,:)
    integer,         pointer :: nbond(:), bond_list(:,:,:)
    integer,         pointer :: nangle(:), angl_list(:,:,:)
    integer,         pointer :: ndihedral(:), dihe_list(:,:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: cell_pair(:,:), cell_pairlist2(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)
    integer,         pointer :: HGr_local(:,:), HGr_bond_list(:,:,:,:)


    HGr_local       => constraints%HGr_local
    HGr_bond_list   => constraints%HGr_bond_list 

    nsolute         => domain%num_solute
    nwater          => domain%num_water
    water_list      => domain%water_list
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    cell_pairlist2  => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    nangle          => enefunc%num_angle
    ndihedral       => enefunc%num_dihedral
    bond_list       => enefunc%bond_list
    angl_list       => enefunc%angle_list
    dihe_list       => enefunc%dihe_list
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1

    ncell_local = domain%num_cell_local
    ncell       = domain%num_cell_local + domain%num_cell_boundary

    domain%max_num_atom = 0
    do i = 1, ncell
      domain%max_num_atom = max(domain%max_num_atom,domain%num_atom(i))
    end do

    ! initialization
    !
    num_nonb_excl  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_nb14_calc  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                           = 0
    num_excl_total1(1:ncell_local)                       = 0
    num_nb14_total (1:maxcell)                           = 0
    num_nb14_total1(1:ncell_local)                       = 0

    ! exclude 1-2 interaction
    !
    !$omp parallel default(shared)                                     &
    !$omp private(id, i, ix, icel1, icel2, icel, i1, i2, i3, num_excl, &
    !$omp         duplicate, k, num_nb14, ic, j, ih) 
    !        
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bond_list(1,ix,i))
        icel2 = id_g2l(1,bond_list(2,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel-1,nthread) == id) then

            i1 = id_g2l(2,bond_list(1,ix,i))
            i2 = id_g2l(2,bond_list(2,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (icel1 < icel2) then
                num_excl = num_nonb_excl(i1,icel) + 1
                num_nonb_excl(i1,icel) = num_excl
                nonb_excl_list(num_excl,i1,icel) = i2
                num_excl_total(icel) = num_excl_total(icel) + 1

              else if (icel1 > icel2) then
                num_excl = num_nonb_excl(i2,icel) + 1
                num_nonb_excl(i2,icel) = num_excl
                nonb_excl_list(num_excl,i2,icel) = i1
                num_excl_total(icel) = num_excl_total(icel) + 1

              end if
            end if
          end if

        else

          if (mod(i-1,nthread) == id) then

            i1 = id_g2l(2,bond_list(1,ix,i))
            i2 = id_g2l(2,bond_list(2,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (i1 < i2) then
                num_excl = num_nonb_excl1(i1,i) + 1
                num_nonb_excl1(i1,i) = num_excl
                nonb_excl_list1(num_excl,i1,i) = i2
                num_excl_total1(i) = num_excl_total1(i) + 1

              else if (i1 > i2) then
                num_excl = num_nonb_excl1(i2,i) + 1
                num_nonb_excl1(i2,i) = num_excl
                nonb_excl_list1(num_excl,i2,i) = i1
                num_excl_total1(i) = num_excl_total1(i) + 1

              end if
            end if

          end if

        end if

      end do
    end do

    ! exclude constraint
    !
    do icel = id+1, ncell_local, nthread
      do ic = 1, constraints%connect
        do j = 1, HGr_local(ic,icel)

          i1 = HGr_bond_list(1,j,ic,icel)
          do ih = 1, ic
            i2 = HGr_bond_list(ih+1,j,ic,icel)
            num_excl = num_nonb_excl1(i1,icel) + 1
            num_nonb_excl1(i1,icel) = num_excl
            nonb_excl_list1(num_excl,i1,icel) = i2
            num_excl_total1(icel) = num_excl_total1(icel) + 1

          end do

        end do
      end do
    end do

    ! exclude water
    !
    if (.not. water_table) then
      do icel = id+1, ncell_local, nthread
        do ic = 1, nwater(icel)

          i1 = water_list(1,ic,icel)
          i2 = water_list(2,ic,icel)
          i3 = water_list(3,ic,icel)

          num_nonb_excl1(i1,icel) = 2
          num_nonb_excl1(i2,icel) = 1
          num_nonb_excl1(i3,icel) = 0
          nonb_excl_list1(1,i1,icel) = i2
          nonb_excl_list1(2,i1,icel) = i3
          nonb_excl_list1(1,i2,icel) = i3
          num_excl_total1(icel) = num_excl_total1(icel) + 3

        end do
      end do
    end if

    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,angl_list(1,ix,i))
        icel2 = id_g2l(1,angl_list(3,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel-1,nthread) == id) then

            i1 = id_g2l(2,angl_list(1,ix,i))
            i2 = id_g2l(2,angl_list(3,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (icel1 < icel2) then

                num_excl = num_nonb_excl(i1,icel)
                duplicate = .false.

                do k = 1, num_excl
                  if (i2 == nonb_excl_list(k,i1,icel)) &
                    duplicate = .true.
                end do

                if (.not. duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl(i1,icel) = num_excl
                  nonb_excl_list(num_excl,i1,icel) = i2
                  num_excl_total(icel) =  num_excl_total(icel) + 1
                end if

              else if (icel1 > icel2) then

                num_excl = num_nonb_excl(i2,icel)
                duplicate = .false.

                do k = 1, num_excl
                  if (i1 == nonb_excl_list(k,i2,icel)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl(i2,icel) = num_excl
                  nonb_excl_list(num_excl,i2,icel) = i1
                  num_excl_total(icel) = num_excl_total(icel) + 1
                end if

              end if
            end if
          end if

        else

          if (mod(i-1,nthread) == id) then

            i1 = id_g2l(2,angl_list(1,ix,i))
            i2 = id_g2l(2,angl_list(3,ix,i))

            if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
              if (i1 < i2) then

                num_excl = num_nonb_excl1(i1,i)
                duplicate = .false.

                do k = 1, num_excl
                  if (i2 == nonb_excl_list1(k,i1,i)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl1(i1,i) = num_excl
                  nonb_excl_list1(num_excl,i1,i) = i2
                  num_excl_total1(i) = num_excl_total1(i) + 1
                end if

              else if (i1 > i2) then

                num_excl = num_nonb_excl1(i2,i)
                duplicate = .false.

                do k = 1, num_excl
                  if (i1 == nonb_excl_list1(k,i2,i)) &
                    duplicate = .true.
                end do

                if (.not.duplicate) then
                  num_excl = num_excl + 1
                  num_nonb_excl1(i2,i) = num_excl
                  nonb_excl_list1(num_excl,i2,i) = i1
                  num_excl_total1(i) = num_excl_total1(i) + 1
                end if

              end if
            end if

          end if
        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihe_list(1,ix,i))
        icel2 = id_g2l(1,dihe_list(4,ix,i))

        if (icel1 /= icel2) then

          icel  = cell_pairlist2(icel1,icel2)
          if (mod(icel-1,nthread) == id) then

            i1 = id_g2l(2,dihe_list(1,ix,i))
            i2 = id_g2l(2,dihe_list(4,ix,i))

            if (icel1 < icel2) then

              num_nb14 = num_nb14_calc(i1,icel)
              duplicate = .false.

              do k = 1, num_nonb_excl(i1,icel)
                if (i2 == nonb_excl_list(k,i1,icel)) &
                  duplicate = .true.
              end do

              do k = 1, num_nb14
                if (i2 == nb14_calc_list(k,i1,icel)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc(i1,icel) = num_nb14
                nb14_calc_list(num_nb14,i1,icel) = i2
                num_nb14_total(icel) = num_nb14_total(icel) + 1
              end if

            else if (icel1 > icel2) then

              num_nb14 = num_nb14_calc(i2,icel)
              duplicate = .false.

              do k = 1, num_nonb_excl(i2,icel)
                if (i1 == nonb_excl_list(k,i2,icel)) &
                  duplicate = .true.
              end do

              do k = 1, num_nb14
                if (i1 == nb14_calc_list(k,i2,icel)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc(i2,icel) = num_nb14
                nb14_calc_list(num_nb14,i2,icel) = i1
                num_nb14_total(icel) = num_nb14_total(icel) + 1
              end if

            end if
          end if

        else

          if (mod(i-1,nthread) == id) then

            i1 = id_g2l(2,dihe_list(1,ix,i))
            i2 = id_g2l(2,dihe_list(4,ix,i))

            if (i1 < i2) then

              num_nb14 = num_nb14_calc1(i1,i)
              duplicate = .false.

              do k = 1, num_nb14
                if (i2 == nb14_calc_list1(k,i1,i)) &
                  duplicate = .true.
              end do

              do k = 1, num_nonb_excl1(i1,i)
                if (i2 == nonb_excl_list1(k,i1,i)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc1(i1,i) = num_nb14
                nb14_calc_list1(num_nb14,i1,i) = i2
                num_nb14_total1(i) = num_nb14_total1(i) + 1
              end if

            else if (i1 > i2) then

              num_nb14 = num_nb14_calc1(i2,i)
              duplicate = .false.

              do k = 1, num_nb14
                if (i1 == nb14_calc_list1(k,i2,i)) &
                  duplicate = .true.
              end do

              do k = 1, num_nonb_excl1(i2,i)
                if (i1 == nonb_excl_list1(k,i2,i)) &
                  duplicate = .true.
              end do

              if (.not.duplicate) then
                num_nb14 = num_nb14 + 1
                num_nb14_calc1(i2,i) = num_nb14
                nb14_calc_list1(num_nb14,i2,i) = i1
                num_nb14_total1(i) = num_nb14_total1(i) + 1
              end if

            end if

          end if

        end if

      end do
    end do

    !$omp end parallel

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel) > 0 .and. &
              cell_pair(jcel,icel) <= ncell_local) then
            found1 = found1 + num_excl_total(cell_pairlist2(jcel,icel))
            found2 = found2 + num_nb14_total(cell_pairlist2(jcel,icel))
          end if
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do

#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1, &
                     num_nonb_excl, nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1, &
                     num_nb14_calc, nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(domain%num_atom, num_nonb_excl1, ncell_local, &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(domain%num_atom, num_nb14_calc1, ncell_local, &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl_constraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    count_nonb_excl_constraint_rest
  !> @brief        exclude 1-2, 1-3 interactions and constraints w/o local
  !                restraint
  !! @authors      JJ    
  !! @param[in]    first       : flag for first call or not
  !! @param[in]    water_table : flag for water table
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : structure of domain
  !! @param[inout] enefunc     : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine count_nonb_excl_constraint_rest(first, water_table, constraints, &
                                             domain, enefunc)

    ! formal arguments
    logical,                     intent(in)    :: first
    logical,                     intent(in)    :: water_table
    type(s_constraints), target, intent(in)    :: constraints
    type(s_domain),      target, intent(inout) :: domain
    type(s_enefunc),     target, intent(inout) :: enefunc

    ! local variables
    integer                  :: ncell, ncell_local, i, ix, k, i1, i2, i3
    integer                  :: icel, icel1, icel2, jcel
    integer                  :: ic, j, ih
    integer                  :: num_excl, num_nb14, id, omp_get_thread_num
    integer                  :: found1, found2
    integer                  :: fkind
    logical                  :: duplicate

    integer,         pointer :: nsolute(:), nwater(:), water_list(:,:,:)
    integer,         pointer :: id_g2l(:,:)
    integer,         pointer :: nbond(:), bond_list(:,:,:)
    integer,         pointer :: nangle(:), angl_list(:,:,:)
    integer,         pointer :: ndihedral(:), dihe_list(:,:,:)
    integer,         pointer :: num_nonb_excl(:,:), num_nonb_excl1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), num_nb14_calc1(:,:)
    integer,         pointer :: cell_pair(:,:), cell_pairlist2(:,:)
    integer,         pointer :: nonb_excl_list(:,:,:), nonb_excl_list1(:,:,:)
    integer,         pointer :: nb14_calc_list(:,:,:), nb14_calc_list1(:,:,:)
    integer,         pointer :: num_excl_total(:), num_nb14_total(:)
    integer,         pointer :: num_excl_total1(:), num_nb14_total1(:)
    integer,         pointer :: HGr_local(:,:), HGr_bond_list(:,:,:,:)


    HGr_local       => constraints%HGr_local
    HGr_bond_list   => constraints%HGr_bond_list 

    nsolute         => domain%num_solute
    nwater          => domain%num_water
    water_list      => domain%water_list
    id_g2l          => domain%id_g2l
    cell_pair       => domain%cell_pair
    cell_pairlist2  => domain%cell_pairlist2

    nbond           => enefunc%num_bond
    nangle          => enefunc%num_angle
    ndihedral       => enefunc%num_dihedral
    bond_list       => enefunc%bond_list
    angl_list       => enefunc%angle_list
    dihe_list       => enefunc%dihe_list
    num_nonb_excl   => enefunc%num_nonb_excl
    num_nonb_excl1  => enefunc%num_nonb_excl1
    num_nb14_calc   => enefunc%num_nb14_calc
    num_nb14_calc1  => enefunc%num_nb14_calc1
    nonb_excl_list  => enefunc%nonb_list
    nonb_excl_list1 => enefunc%nonb_list1
    nb14_calc_list  => enefunc%nb14_list
    nb14_calc_list1 => enefunc%nb14_list1
    num_excl_total  => enefunc%num_excl_total
    num_excl_total1 => enefunc%num_excl_total1
    num_nb14_total  => enefunc%num_nb14_total
    num_nb14_total1 => enefunc%num_nb14_total1

    ncell_local = domain%num_cell_local
    ncell       = domain%num_cell_local + domain%num_cell_boundary

    domain%max_num_atom = 0
    do i = 1, ncell
      domain%max_num_atom = max(domain%max_num_atom,domain%num_atom(i))
    end do

    ! initialization
    !
    num_nonb_excl  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nonb_excl1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_nb14_calc  (1:domain%max_num_atom,1:maxcell)     = 0
    num_nb14_calc1 (1:domain%max_num_atom,1:ncell_local) = 0
    num_excl_total (1:maxcell)                           = 0
    num_excl_total1(1:ncell_local)                       = 0
    num_nb14_total (1:maxcell)                           = 0
    num_nb14_total1(1:ncell_local)                       = 0

    ! exclude 1-2 interaction
    !
    !$omp parallel default(shared)                                     &
    !$omp private(id, i, ix, icel1, icel2, icel, i1, i2, i3, num_excl, &
    !$omp         duplicate, k, num_nb14, ic, j, ih, fkind) 
    !        
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    do i = 1, ncell_local
      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bond_list(1,ix,i))
        icel2 = id_g2l(1,bond_list(2,ix,i))
        fkind = enefunc%bond_kind(ix,i)

        if (fkind == 0) then
          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel-1,nthread) == id) then
         
              i1 = id_g2l(2,bond_list(1,ix,i))
              i2 = id_g2l(2,bond_list(2,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (icel1 < icel2) then
                  num_excl = num_nonb_excl(i1,icel) + 1
                  num_nonb_excl(i1,icel) = num_excl
                  nonb_excl_list(num_excl,i1,icel) = i2
                  num_excl_total(icel) = num_excl_total(icel) + 1
         
                else if (icel1 > icel2) then
                  num_excl = num_nonb_excl(i2,icel) + 1
                  num_nonb_excl(i2,icel) = num_excl
                  nonb_excl_list(num_excl,i2,icel) = i1
                  num_excl_total(icel) = num_excl_total(icel) + 1
         
                end if
              end if
            end if
         
          else
         
            if (mod(i-1,nthread) == id) then
         
              i1 = id_g2l(2,bond_list(1,ix,i))
              i2 = id_g2l(2,bond_list(2,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (i1 < i2) then
                  num_excl = num_nonb_excl1(i1,i) + 1
                  num_nonb_excl1(i1,i) = num_excl
                  nonb_excl_list1(num_excl,i1,i) = i2
                  num_excl_total1(i) = num_excl_total1(i) + 1
         
                else if (i1 > i2) then
                  num_excl = num_nonb_excl1(i2,i) + 1
                  num_nonb_excl1(i2,i) = num_excl
                  nonb_excl_list1(num_excl,i2,i) = i1
                  num_excl_total1(i) = num_excl_total1(i) + 1
         
                end if
              end if
         
            end if
         
          end if

        end if

      end do
    end do

    ! exclude constraint
    !
    do icel = id+1, ncell_local, nthread
      do ic = 1, constraints%connect
        do j = 1, HGr_local(ic,icel)

          i1 = HGr_bond_list(1,j,ic,icel)
          do ih = 1, ic
            i2 = HGr_bond_list(ih+1,j,ic,icel)
            num_excl = num_nonb_excl1(i1,icel) + 1
            num_nonb_excl1(i1,icel) = num_excl
            nonb_excl_list1(num_excl,i1,icel) = i2
            num_excl_total1(icel) = num_excl_total1(icel) + 1

          end do

        end do
      end do
    end do

    ! exclude water
    !
    if (.not. water_table) then
      do icel = id+1, ncell_local, nthread
        do ic = 1, nwater(icel)

          i1 = water_list(1,ic,icel)
          i2 = water_list(2,ic,icel)
          i3 = water_list(3,ic,icel)

          num_nonb_excl1(i1,icel) = 2
          num_nonb_excl1(i2,icel) = 1
          num_nonb_excl1(i3,icel) = 0
          nonb_excl_list1(1,i1,icel) = i2
          nonb_excl_list1(2,i1,icel) = i3
          nonb_excl_list1(1,i2,icel) = i3
          num_excl_total1(icel) = num_excl_total1(icel) + 3

        end do
      end do
    end if

    ! exclude 1-3 interaction
    !
    do i = 1, ncell_local
      do ix = 1, nangle(i)

        icel1 = id_g2l(1,angl_list(1,ix,i))
        icel2 = id_g2l(1,angl_list(3,ix,i))
        fkind = enefunc%angle_kind(ix,i)

        if (fkind == 0) then

          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel-1,nthread) == id) then
         
              i1 = id_g2l(2,angl_list(1,ix,i))
              i2 = id_g2l(2,angl_list(3,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (icel1 < icel2) then
         
                  num_excl = num_nonb_excl(i1,icel)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i2 == nonb_excl_list(k,i1,icel)) &
                      duplicate = .true.
                  end do
         
                  if (.not. duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl(i1,icel) = num_excl
                    nonb_excl_list(num_excl,i1,icel) = i2
                    num_excl_total(icel) =  num_excl_total(icel) + 1
                  end if
         
                else if (icel1 > icel2) then
         
                  num_excl = num_nonb_excl(i2,icel)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i1 == nonb_excl_list(k,i2,icel)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl(i2,icel) = num_excl
                    nonb_excl_list(num_excl,i2,icel) = i1
                    num_excl_total(icel) = num_excl_total(icel) + 1
                  end if
         
                end if
              end if
            end if
         
          else
         
            if (mod(i-1,nthread) == id) then
         
              i1 = id_g2l(2,angl_list(1,ix,i))
              i2 = id_g2l(2,angl_list(3,ix,i))
         
              if (i1 <= nsolute(icel1) .and. i2 <= nsolute(icel2)) then
                if (i1 < i2) then
         
                  num_excl = num_nonb_excl1(i1,i)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i2 == nonb_excl_list1(k,i1,i)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl1(i1,i) = num_excl
                    nonb_excl_list1(num_excl,i1,i) = i2
                    num_excl_total1(i) = num_excl_total1(i) + 1
                  end if
         
                else if (i1 > i2) then
         
                  num_excl = num_nonb_excl1(i2,i)
                  duplicate = .false.
         
                  do k = 1, num_excl
                    if (i1 == nonb_excl_list1(k,i2,i)) &
                      duplicate = .true.
                  end do
         
                  if (.not.duplicate) then
                    num_excl = num_excl + 1
                    num_nonb_excl1(i2,i) = num_excl
                    nonb_excl_list1(num_excl,i2,i) = i1
                    num_excl_total1(i) = num_excl_total1(i) + 1
                  end if
         
                end if
              end if
         
            end if
          end if

        end if

      end do
    end do

    ! count 1-4 interaction
    !
    do i = 1, ncell_local
      do ix = 1, ndihedral(i)

        icel1 = id_g2l(1,dihe_list(1,ix,i))
        icel2 = id_g2l(1,dihe_list(4,ix,i))
        fkind = enefunc%dihe_kind(ix,i)
        
        if (fkind == 0) then

          if (icel1 /= icel2) then
         
            icel  = cell_pairlist2(icel1,icel2)
            if (mod(icel-1,nthread) == id) then
         
              i1 = id_g2l(2,dihe_list(1,ix,i))
              i2 = id_g2l(2,dihe_list(4,ix,i))
         
              if (icel1 < icel2) then
         
                num_nb14 = num_nb14_calc(i1,icel)
                duplicate = .false.
         
                do k = 1, num_nonb_excl(i1,icel)
                  if (i2 == nonb_excl_list(k,i1,icel)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nb14
                  if (i2 == nb14_calc_list(k,i1,icel)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc(i1,icel) = num_nb14
                  nb14_calc_list(num_nb14,i1,icel) = i2
                  num_nb14_total(icel) = num_nb14_total(icel) + 1
                end if
         
              else if (icel1 > icel2) then
         
                num_nb14 = num_nb14_calc(i2,icel)
                duplicate = .false.
         
                do k = 1, num_nonb_excl(i2,icel)
                  if (i1 == nonb_excl_list(k,i2,icel)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nb14
                  if (i1 == nb14_calc_list(k,i2,icel)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc(i2,icel) = num_nb14
                  nb14_calc_list(num_nb14,i2,icel) = i1
                  num_nb14_total(icel) = num_nb14_total(icel) + 1
                end if
         
              end if
            end if
         
          else
         
            if (mod(i-1,nthread) == id) then
         
              i1 = id_g2l(2,dihe_list(1,ix,i))
              i2 = id_g2l(2,dihe_list(4,ix,i))
         
              if (i1 < i2) then
         
                num_nb14 = num_nb14_calc1(i1,i)
                duplicate = .false.
         
                do k = 1, num_nb14
                  if (i2 == nb14_calc_list1(k,i1,i)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nonb_excl1(i1,i)
                  if (i2 == nonb_excl_list1(k,i1,i)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc1(i1,i) = num_nb14
                  nb14_calc_list1(num_nb14,i1,i) = i2
                  num_nb14_total1(i) = num_nb14_total1(i) + 1
                end if
         
              else if (i1 > i2) then
         
                num_nb14 = num_nb14_calc1(i2,i)
                duplicate = .false.
         
                do k = 1, num_nb14
                  if (i1 == nb14_calc_list1(k,i2,i)) &
                    duplicate = .true.
                end do
         
                do k = 1, num_nonb_excl1(i2,i)
                  if (i1 == nonb_excl_list1(k,i2,i)) &
                    duplicate = .true.
                end do
         
                if (.not.duplicate) then
                  num_nb14 = num_nb14 + 1
                  num_nb14_calc1(i2,i) = num_nb14
                  nb14_calc_list1(num_nb14,i2,i) = i1
                  num_nb14_total1(i) = num_nb14_total1(i) + 1
                end if
         
              end if
         
            end if
         
          end if
        end if

      end do
    end do

    !$omp end parallel

    ! Check the total number of exclusion list
    !
    if (first) then

      found1 = 0
      found2 = 0

      do icel = 1, ncell-1
        do jcel = icel+1, ncell
          if (cell_pair(jcel,icel) > 0 .and. &
              cell_pair(jcel,icel) <= ncell_local) then
            found1 = found1 + num_excl_total(cell_pairlist2(jcel,icel))
            found2 = found2 + num_nb14_total(cell_pairlist2(jcel,icel))
          end if
        end do
      end do

      do icel = 1, ncell_local
        found1 = found1 + num_excl_total1(icel)
        found2 = found2 + num_nb14_total1(icel)
      end do

#ifdef MPI
      call mpi_reduce(found1, enefunc%num_excl_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
      call mpi_reduce(found2, enefunc%num_nb14_all, 1, mpi_integer, mpi_sum, &
                      0, mpi_comm_country, ierror)
#else
    enefunc%num_excl_all = found1
    enefunc%num_nb14_all = found2
#endif
    end if

    ! pack into small dimension
    !
    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1, &
                     num_nonb_excl, nonb_excl_list, enefunc%nonb_excl_list)

    call pack_array_4to3(domain%num_atom, domain%cell_pairlist1, &
                     num_nb14_calc, nb14_calc_list, enefunc%nb14_calc_list)

    call pack_array_3to2(domain%num_atom, num_nonb_excl1, ncell_local, &
                     nonb_excl_list1, enefunc%nonb_excl_list1)

    call pack_array_3to2(domain%num_atom, num_nb14_calc1, ncell_local, &
                     nb14_calc_list1, enefunc%nb14_calc_list1)

    return

  end subroutine count_nonb_excl_constraint_rest

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pack_array_4to3
  !> @brief        pack 4D-array into 3D-array
  !! @authors      YS, JJ
  !! @param[in]    num_ele  : 
  !! @param[in]    num_list :
  !! @param[in]    list_1d  :
  !! @param[out]   list_3d  :
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pack_array_4to3(num_ele, cell_list, num_list, list_3d, list_2d)

    ! formal arguments
    integer,                 intent(in)    :: num_ele(:)
    integer,                 intent(in)    :: cell_list(:,:)
    integer,                 intent(in)    :: num_list(:,:)
    integer,                 intent(in)    :: list_3d(:,:,:)
    integer,                 intent(inout) :: list_2d(:,:)

    ! local variables
    integer                  :: i, j, k, icel, jcel
    integer                  :: id, omp_get_thread_num


    !$omp parallel default(shared) private(id, jcel, icel, k, i, j)
    !

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do jcel = id+1, maxcell, nthread
      icel = cell_list(1,jcel)
      k = 0
      do i = 1, num_ele(icel)
        do j = 1, num_list(i,jcel)
          k = k + 1
          list_2d(k,jcel) = list_3d(j,i,jcel)
        end do
      end do
    end do

    !$omp end parallel 

    return

  end subroutine pack_array_4to3

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pack_array_3to2
  !> @brief        pack 3D-array into 2D-array
  !! @authors      YS
  !! @param[in]    num_ele  : 
  !! @param[in]    num_list :
  !! @param[in]    list_1d  :
  !! @param[out]   list_2d  :
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pack_array_3to2(num_ele, num_list, ncell_local, list_2d, list_1d)

    ! formal arguments
    integer,                 intent(in)    :: num_ele(:)
    integer,                 intent(in)    :: num_list(:,:)
    integer,                 intent(in)    :: ncell_local
    integer,                 intent(in)    :: list_2d(:,:,:)
    integer,                 intent(inout) :: list_1d(:,:)

    ! local variables
    integer                  :: i, j, k, icel
    integer                  :: id, omp_get_thread_num


    !$omp parallel default(shared) private(id, icel, k, i, j)
    !

#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do icel = id+1, ncell_local, nthread
      k = 0
      do i = 1, num_ele(icel)
        do j = 1, num_list(i,icel)
          k = k + 1
          list_1d(k,icel) = list_2d(j,i,icel)
        end do
      end do
    end do

    !$omp end parallel 

    return

  end subroutine pack_array_3to2

end module sp_enefunc_charmm_mod
